;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)

(depends-on %module/ nisp)

(depends-on :at-run-time %optop/ types)

(deffunc causation-view - Internalized-causation-defn
	                (proc - Causation-defn
			 view-finder - (Fun Internalized-causation-defn
					    (Internalized-durative-defn))
			 dom - Domain)
   (let ((intdef (causation-defn-projectable proc dom)))
      (cond ((is Internalized-durative-defn intdef)
	     (funcall view-finder intdef))
	    ((is Internalized-causation-defn intdef)
	     intdef)
	    (t
	     (signal-problem setup-process-complex
		"Can't get appropriate view of "
		intdef)))))

(defproc causation-defn-projectable - Internalized-causation-defn
	                (cd - Causation-defn dom - Domain)
   (let ((intform (!_internalized cd)))
      (cond ((not intform)
	     (!= intform (causation-defn-internalize cd dom))
	     (!= (!_internalized cd)
		 intform)))
      intform))

(defopfunc causation-defn-internalize - Internalized-causation-defn
	                 (cdefn - Causation-defn dom - Domain)
                         (ignore dom)
   (signal-problem causation-defn-projectable
      "Don't know how to internalize " cdefn))

(defmethfunc causation-defn-internalize - Internalized-action-defn
	        (adefn - Action-defn dom - Domain)
   (multi-let (((idefn _ linkspots)
		(causation-defn-basic-internalize adefn dom)))
      (opt::prelink linkspots dom)
      (be Internalized-action-defn idefn)))

(defmethfunc causation-defn-internalize - Internalized-method-defn
	           (mdefn - Method-defn dom - Domain)
   (multi-let (((cint _ linkspots)
		(causation-defn-basic-internalize mdefn dom))
	       (boundvars
		(<# (\\ (b) (!_varname b))
			(!_varbdgs mdefn))))
      (multi-let (((mint m-linkspots)
		   (internalize (!_maint mdefn) boundvars ':consequent))
		  ((exp-int e-linkspots)
		   (internalize (!_expansion mdefn)
				boundvars ':consequent)))
	 (opt::prelink (nconc linkspots m-linkspots e-linkspots) dom)
	 (with cint
	    (make-inst Internalized-method-defn
	       :function !>function
	       :term !>term
	       :condition !>condition
	       :type-filter !>type-filter
	       :effect !>effect
	       :expansion exp-int
	       :maint mint)))))

(defmethfunc causation-defn-internalize - Internalized-process-defn
	            (pdefn - Process-defn dom - Domain)
   (multi-let (((cint _ linkspots)
		(causation-defn-basic-internalize pdefn dom))
	       (boundvars
                  (<# (\\ (b) (!_varname b))
                      (!_varbdgs pdefn))))
     (let ((int-conds (process-condition-classify (!_condition cint) dom)))
	(multi-let (((int-a a-linkspots)
		     (cond ((!_start-effect pdefn)
			    (internalize (!_start-effect pdefn)
					 boundvars ':consequent))
			   (t (values false !()))))
		    ((int-o o-linkspots)
		     (cond ((!_stop-effect pdefn)
			    (internalize (!_stop-effect pdefn)
					 boundvars ':consequent))
			   (t (values false !())))))
	  (opt::prelink (nconc linkspots a-linkspots o-linkspots) dom)
	  (with cint
	     (make-inst Internalized-process-defn
		:function !>function
		:term !>term
		:condition !>condition
		:type-filter !>type-filter
		:sorted-condition int-conds
		:start-effect int-a
		:effect !>effect
		:stop-effect int-o))))))

(defmethfunc causation-defn-internalize - Internalized-durative-defn
	             (dadefn - Durative-action-defn dom - Domain)
   (let ((boundvars
	        ;;(append '(D= D< D>) ...)
            (<# (\\ (b) (!_varname b))
                (!_varbdgs dadefn)))
	 (typefilt (opt::qvarbdgs-constraints (!_varbdgs dadefn)
					      (domain-place-env dom)))
	 (da-term (!_term dadefn)))
      (multi-let (((intc c-linkspots)
		   (internalize (!_condition dadefn) boundvars ':antecedent))
		  ((inte e-linkspots)
		   (internalize (!_effect dadefn) boundvars ':consequent))
		  ((intd d-linkspots)
		   (internalize (!_duration dadefn) boundvars ':consequent)))
	 (opt::prelink (nconc c-linkspots e-linkspots d-linkspots)
		       dom)
	 (multi-let ((oa-cond (extract-foxlong-modals intc '(over all) false))
		     ((start-eff stop-eff start-conds overall-conds)
		      (durative-effect-split inte da-term))
		     ((dur-= dur-< dur->)
		      (duration-ineqs-classify intd)))
	    (let-fun ()
	       (make-inst Internalized-durative-defn
		  :process-view (internalize-process-view)
		  :start-view (internalize-start-view)
		  :stop-view (cond ((null dur-=) (internalize-stop-view))
				   (t false)))

	  :where

 (:def internalize-process-view - Internalized-process-defn
			   ()
;;;;    (trace-around intern-proc-view
;;;;       (:> " (intern-proc-view: " ")")
    (let ((ae-cond (extract-foxlong-modals intc '(at end) false))
	  (elapcomp `((fl-v (elapsed (dur-happening ,da-term)))
		      ,(cond ((null dur-=) '?D<)
			     (t '?D=)))))
       (let ((condition
		(n-conn-join
		   'and
		   (cond ((and (null dur-=) (null dur-<))
			  '(and))
			 (t
			  `(=< ,@elapcomp)))
		   `(dur-in-progress ,da-term ?D= ?D> ?D<)
		   '(and (opt:is opt:Number ?D=)
		         (opt:is opt:Number ?D>)
		         (opt:is opt:Number ?D<))
		    oa-cond)))
;;;;	  (out (:to *error-output*)
;;;;	     "Condition = " condition :% :%)	       
	  (make-inst Internalized-process-defn
	     :function 'dur-happening ;;;(car da-term)
	     :term `(dur-happening ,da-term)
	     :condition condition
	     :type-filter typefilt
	     :sorted-condition (process-condition-classify condition dom)
	     :start-effect '(and)
	     :effect (</ (\\ (trace-eff ovc)
			    (conn-join 'and 
				       trace-eff
				       `(when (not ,ovc)
					   (not (dur-trace
						       throughout
						       ,da-term
						       ,ovc)))))
			 '(and)
			 overall-conds)
	     :stop-effect `(and (when ,(cond ((null dur-=)
					      (n-conn-join 'and
						  `(dur-in-progress
						     ,da-term ?D= ?D> ?D<)
						  oa-cond
						  ae-cond))
					     (t
					      (conn-join
					         'and oa-cond ae-cond)))
				      ,stop-eff)
				(not (dur-in-progress
				         ,da-term ?D= ?D> ?D<))))))
;;;;       (:< (val &rest _) "intern-proc-view: " val))
    )

 (:def internalize-start-view - Internalized-action-defn
			 ()
    (multi-let ((as-cond (extract-foxlong-modals intc '(at start) true)))
       (let ((dur-min-max 
		`((eval ,(cond ((null dur-<) 99999999.0)
			       (t `(min ,@dur-<)))
			?D<)
		  (eval ,(cond ((null dur->) -99999999.0)
			       (t `(max ,@dur->)))
			?D>)
		  (opt:is opt:Number ?D=)
		  (opt:is opt:Number ?D>)
		  (opt:is opt:Number ?D<)))
	     (effect
		(conn-join
		   'and
		   `(dur-in-progress ,da-term ?D= ?D> ?D<)
		   (conn-join 'and
			      start-eff
			      `(and ,@(<# (\\ (sc - Sexp)
					     `(when ,sc
						 (dur-trace
						    initial
						    ,da-term
						    ,sc)))
					  start-conds)
				    ,@(<# (\\ (tc - Sexp)
					     `(when ,tc
						 (dur-trace
						    throughout
						    ,da-term
						    ,tc)))
					  overall-conds))))))
	  (cond ((null dur-=)
		 (make-inst Internalized-action-defn
		    :function 'dur-start
		    :term `(dur-start ,da-term)
		    :condition (n-conn-join 'and
				  as-cond
				  oa-cond
				  `(and ,@dur-min-max
					(=< ?D> ?D<)
					(= ?D= ?D<)))
		    :type-filter typefilt
		    :effect effect))
		(t
		 (make-inst Internalized-action-defn
		    :function 'dur-start
		    :term `(dur-start ,da-term)
		    :condition (n-conn-join 'and
				  as-cond
				  oa-cond
				  `(and ,@(<# (\\ (e)
						 `(eval ,e ?D=))
					      dur-=)
					,@dur-min-max
					(=< ?D> ?D=)
					(=< ?D= ?D<)))
		    :type-filter typefilt
		    :effect effect))))))

 (:def internalize-stop-view - Internalized-action-defn ()
    (make-inst Internalized-action-defn
       :function 'dur-stop
       :term `(dur-stop ,da-term)
       :condition `(and (dur-in-progress ,da-term ?D= ?D> ?D<)
			(>= (elapsed (dur-happening ,da-term))
			    ?D>)
			(=< (elapsed (dur-happening ,da-term))
			    ?D<)
			(opt:is opt:Number ?D=)
			(opt:is opt:Number ?D>)
			(opt:is opt:Number ?D<))
       :type-filter typefilt
       :effect `(not (dur-in-progress ,da-term ?D= ?D> ?D<)))))))))
      
(deffunc durative-effect-split - (Multvals start-effect stop-effect
					   - Sexp
					   at-start-conds overall-conds
					   - (Lst Sexp))
                         (eff da-term - Sexp)
   (let ((at-start-whens !())
	 (over-all-whens !())
	 - (Lst Sexp))
      (letrec ()
	 (let ((start-effect (walk-collecting-start eff))
	       (stop-effect (walk-collecting-stop eff)))
	    ;; As a side effect, at-start-whens and over-all-whens
	    ;; are now set to "at start"s and "over all"s found in
	    ;; 'when' antecedents
	    (values start-effect stop-effect
		    at-start-whens over-all-whens))
       :where

 (walk-collecting-start - Sexp (eff - Sexp)
;;;;    (trace-around coll-start
;;;;       (:> " (coll-start: " eff ")")
    (match-cond eff
       ?( (and ?@conjl)
	 (</ (\\ (fmla c)
		(conn-join 'and fmla (walk-collecting-start c)))
	     '(and)
	     conjl))
       ?( (when ?ante ?subeff)
	 (let ((ca (when-convert ante)))
	    `(when ,ca ,(walk-collecting-start subeff))))
       ?( (at start ?e)
	 e)
       (t '(and)))
;;;;       (:< (val &rest _) "coll-start: " val))
    )

 (walk-collecting-derivs - Sexp (eff - Sexp)
    (match-cond eff
       ?( (and ?@conjl)
	 (</ (\\ (fmla c)
		(conn-join 'and fmla (walk-collecting-derivs c)))
	     '(and)
	     conjl))
       ?( (when ?ante ?subeff)
	 (let ((ca (when-convert ante)))
	    `(when ,ca ,(walk-collecting-derivs subeff))))
       ?( (?(:& ?inc-or-dec ?(:\| increase decrease))
	   ?f ?(:\| (* \#t ?d)
		   (* ?d \#t)))
	 `(derivative ,f ,(cond ((eq inc-or-dec 'decrease) `(- ,d))
				(t d))))
       ?(t '(and))))

 (walk-collecting-stop - Sexp (eff - Sexp)
;;;;    (trace-around coll-stop
;;;;       (:> " (coll-stop: " eff ")")
    (match-cond eff
       ?( (and ?@conjl)
	 (</ (\\ (fmla c)
		(conn-join 'and fmla (walk-collecting-stop c)))
	     '(and)
	     conjl))
       ?( (when ?ante ?subeff)
	 (let ((ca (when-convert ante)))
	    `(when ,ca ,(walk-collecting-stop subeff))))
       ?( (at start ?e)
	 '(and))
       ?( (at end ?e)
	 e)
       ?( (?(:\| increase decrease)
	   ?_ ?(:\| (* \#t ?_)
		   (* ?_ \#t)))
	 '(and))
       (t eff))
;;;;       (:< (val &rest _) "coll-stop: " val))
    )
	 
 ;; Convert a 'when' antecedent to form that doesn't depend on
 ;; action at a (temporal) distance
 (when-convert - Sexp (a - Sexp)
    (match-cond a
       ?( (and ?@conjl)
	 `(and ,@(<# when-convert conjl)))
       ?( (at ?(:& ?which ?(:\| start end))
	      ?p)
	 (cond ((eq which 'start)
		(!= at-start-whens
		    (adjoin= #'equal p *-*))
		`(dur-trace initial ,da-term ,p))
	       (t p)))
       ?( (over all ?p)
	 (!= over-all-whens (adjoin p *-*))
	 `(dur-trace throughout ,da-term ,p))
       (t a)))
	     
;;;; (adjoin-non-variant - (Lst Sexp) (p - Sexp l - (Lst Sexp))
;;;;    (adjoin= (\\ (p1 p2)
;;;;		(variants p1 p2 dummy-id1* dummy-id2* (empty-env)))
;;;;	     p l))
 
)))

(deffunc extract-foxlong-modals - Sexp (exp - Sexp time-spec - Sexp
					assume - Boolean)
   (match-cond exp
      ?( (and ?@conjl)
	(</ (\\ (r c)
	       (conn-join 'and r (extract-foxlong-modals c time-spec assume)))
	    '(and) conjl))
      ?( (?,@time-spec ?p)
	p)
      ?( ?(:\| (at start ?_) (at end ?_) (over all ?_))
	;; If any other modality occurs, ignore it
	'(and))
      (assume
       ;; If time-spec is the default, take exp
       exp)
      (t '(and))))
	     
(deffunc duration-ineqs-classify - (Multvals d= d< d> - (Lst Sexp))
	                     (exp - Sexp)
   (match-cond exp
      ?( (and ?@conjl)
	(repeat :for ((c :in conjl)
		      cd= cd< cd>
		      :collectors d= d< d>)
	   (!= < cd= cd< cd> >
	       (duration-ineqs-classify c))
	:nconc (:into d= cd=)
	:nconc (:into d< cd<)
	:nconc (:into d> cd>)
	:result (values d= d< d>)))
      ?( (= duration ?d)
	(values (list d) !() !()))
      ?( (?(:\| < =<) duration ?d)
	(values !() (list exp) !()))
      ?( (?(:\| > >=) duration ?d)
        (values !() !() (list exp)))
      (t
       (signal-problem duration-ineqs-classify
	  "Unintelligible duration constraint: " exp))))

(deffunc causation-defn-basic-internalize
                        - (Mlv Internalized-causation-defn
			       (Lst Symbol)  ;; the boundvars
			       (Lst Linkspot))
              (cdefn - Causation-defn dom - Domain)
              ;;(ignore dom)
   (let ((boundvars (<# (\\ (b) (!_varname b))
			(!_varbdgs cdefn)))
	 - (Lst Symbol)
	 (condition (!_condition cdefn))
	 (effect (!_effect cdefn)) - (~ Typed-exp)
	 (linkspots '())
	 ls  - (Lst Linkspot)
	 intcond inteff - Sexp
	 intfilt - (Lst Sexp))
     (!= < intcond ls >
	 (cond (condition
		(internalize 
		   condition
		   boundvars
		   ':antecedent))
	       (t
		(values false !()))))
     (!= linkspots ls)
     (!= < inteff ls >
	 (cond (effect
		(internalize effect boundvars ':consequent))
	       (t
		(values false !()))))
     (cond ((matchq (exists ?@_) intcond)
	    (dbg-save condition intcond effect)
	    (breakpoint causation-defn-basic-internalize
	       "intcond = " intcond
	       :% " effect = " effect)))
     (!= linkspots (nconc ls *-*))
     (!= intfilt (opt::qvarbdgs-constraints (!_varbdgs cdefn)
					    (domain-place-env dom)))
     ;;(prelink linkspots dom)
     (values
	(make-inst Internalized-causation-defn
	    :function (!_function cdefn)
	    :term (!_term cdefn)
	    :condition intcond
	    :type-filter intfilt
	    :effect inteff)
	boundvars
	linkspots)))

;;; For debugging
(deffunc sit-path - (Lst Situation) (sit - Situation sit-space - Situation-space)
   (let ((start (find-init-situation sit)))
      (let-fun ((find-path - (Lst Situation) (acts - (Lst Action))
		   ;; Given list 'acts', find sequence of sits
		   ;; from 'start' following act list in reverse.
		   (repeat
		    :until (or (null acts)
			       (not (matchq (transition ?_)
					    (!_act (car acts)))))
		      (!= acts (cdr *-*)))
		   (cond ((null acts) (list start))
			 (t
			  (let ((prev (find-path (cdr acts)))
				(act (match-cond (!_act (car acts))
					(:? (wait ?_)
					   'wait)
					(t (car acts)))))
			     (let ((prior (car prev)))
			        (repeat :for ((s :in (!_succs prior sit-space)))
				 :result (progn
					    (dbg-save prior)
					   (signal-problem sit-path
					    "Can't find sit for action "
					    (car acts)
					    " after " prior))
				 :until (eq (!_act s) act)
				 :result (cons (!_sit s) prev))))))))
	 (find-path (!_path sit)))))

;;;; Result is 
;;;; disjunction of 
;;;;  (conjunction of (discrete conjunction)
;;;;                  and (disjunction of (continuous conjunction)))
(deffunc process-condition-classify - (Lst (Lstructure ()
					      discrete
					      - (Lst Sexp)
					      continuous
					      - (Lst (Lst (Lrcd Ineq-rel
								Sexp Sexp)))))
	           (condition - Sexp dom - Domain)
   (letrec ()
      (combine-discretes
	 (<# (\\ (conjunction)
		(classify-discrete-vs-continuous
		   conjunction dom))
	     (disjunctive-normal-form-no-/=~ condition true)))
    :where
      (combine-discretes - (Lst (Lstructure ()
				   discrete - (Lst Sexp)
				   continuous - (Lst (Lst (Lrcd Ineq-rel
								Sexp Sexp)))))
                         (dnf-version - (Lst (Lstructure ()
						discrete - (Lst Sexp)
						continuous
						   - (Lst (Lrcd Ineq-rel
								Sexp Sexp)))))
;;;;	 (trace-around discs-comb
;;;;	    (:> " (discs-comb: "  dnf-version ")")
	 (repeat :for ((d :in dnf-version) - (Lstructure ()
						discrete - (Lst Sexp)
						continuous
						   - (Lst (Lrcd Ineq-rel
								Sexp Sexp)))
		       (res !()) - (Lst (Lstructure ()
					   discrete - (Lst Sexp)
					   continuous
					      - (Lst (Lst (Lrcd Ineq-rel
								Sexp Sexp))))))
;;;;	    (out "Processing discrete = " (!_discrete d)
;;;;		 :% " continuous = " (!_continuous d) :%)
	    (let ((p (assoc= #'equal (!_discrete d) res)))
	       (cond ((not p)
		      (!= p (tuple (!_discrete d) !()))
		      (!= res (cons p *-*))))
	       (!= (cadr p) (cons (!_continuous d) *-*)))
;;;;	    (out "Now res = " res :%)
	  :result res)
;;;;	    (:< (val &rest _) "discs-comb: " val))
	 )))


;;; Turn all occurrences of /=~ into disjunctions of < and >.
;;; While you're at it, flush all not's.
(deffunc disjunctive-normal-form-no-/=~ - (Lst (Lst Sexp))
			     (e - Sexp polarity - Boolean)
   (letrec ()
      (let ((ordinary (disjunctive-normal-form e polarity)))
	 (<! expand-/=~ ordinary))
    :where
      (expand-/=~ - (Lst (Lst Sexp))
		 (conjuncts - (Lst Sexp))
	 (cond ((null conjuncts) (list conjuncts))
	       (t
		(let ((firstres
		          (match-cond (car conjuncts)
			      ?( (not (?rel ?@args))
				(cond ((eq rel '=~)
				       (/=~-flush args))
				      ((is Ineq-rel rel)
				       (list `((,(ineq-rel-negation rel)
						,@args))))
				      (t (list (list (car conjuncts))))))
			      ?( (/=~ ?@args) 
				(/=~-flush args))
			      (t
				(list `(,(car conjuncts))))))
		      (remres (expand-/=~ (cdr conjuncts)))
		      - (Lst (Lst Sexp)))
		   ;; 'firstres' is a DNF expansion of the first
		   ;; conjunct.  'remres' is a DNF expansion of the
		   ;; rest of them
		   (repeat :for ((fr :in firstres))
		    :nconc
		      (repeat :for ((res :in remres))
		       :collect `(,@fr ,@res)))))))

      (/=~-flush - (Lst (Lst Sexp)) (args - (Lst Sexp))
	 (list `((< ,@args))
	       `((> ,@args))))))

(deffunc ineq-rel-negation - Ineq-rel (r - Ineq-rel)
	 (alist-entry
	    r
	    (be (Lst (Lrcd Ineq-rel Ineq-rel))
		'((> =<) (>= <) (< >=) (=< >) (<= >)
		  (/=~ =~) (=~ /=~)))
	    (signal-problem disjunctive-normal-form-no-/=~
	       "Bogus Ineq-rel " r)))

;;; Actually, "continuous" means "continuous and involving fluents."
(deffunc classify-discrete-vs-continuous - (Lstructure ()
					      discrete - (Lst Sexp)
					      continuous
					        - (Lst (Lrcd Ineq-rel
							     Sexp Sexp)))
	                          (atomic-formulas - (Lst Sexp) dom - Domain)
   (let-fun ((is-discrete - Boolean
		      (fmla - Sexp)
	       (cond ((memq (car fmla) ineq-rels*)
		      (not (exists (a :in (cdr fmla))
			      (is-fluent-term a dom))))
		     ((eq (car fmla) 'not)
		      (cond ((is-discrete (cadr fmla))
			     true)
			    (t
			     (err-out "Warning: classifying " fmla
				      " as continuous")
			     false)))
		     (t true))))
      (multi-let (((discrete - (Lst Sexp)
		    continuous - (Lst (Lrcd Ineq-rel Sexp Sexp)))
		   (be * (classify atomic-formulas #'is-discrete))))
	 (tuple discrete continuous))))

(deffunc durative-process-view - Internalized-process-defn
	                       (d - Internalized-durative-defn)
   (!_process-view d))

(deffunc durative-start-view - Internalized-action-defn
	                     (d - Internalized-durative-defn)
   (!_start-view d))

(deffunc durative-stop-view - Internalized-action-defn
	                     (d - Internalized-durative-defn)
   (!_stop-view d))



(repeat :for ((p :in '(> >= < =< =~ /=~)))
   (!= bad-reduction-preds*
       (adjoin p *-* :test #'eq)))

(deffunc plausibly-reducible - Boolean (pred - CSymbol dom - Domain)
   (and (not (memq pred bad-reduction-preds*))
	(not (pred-is-rigid pred dom))))

(deffunc pred-is-rigid - Boolean (pred - CSymbol dom - Domain)
   (or (memq pred '(opt:is opt::goal-call))
       (let ((pred-defn (find-domain-bdg-val pred dom)))
	  (and pred-defn
	       (is Predicate pred-defn)
	       (opt::Predicate-rigid pred-defn)))))

(deffunc is-fluent-term - Boolean (a - Sexp dom - Domain)
   (cond ((atom a) false)
	 ((eq (strip-car a) 'fl-v)
	  true)
	 (t
	  (exists (b :in (cdr a))
	     (is-fluent-term b dom)))))

(deffunc lit-pred - CSymbol (lit - Sexp)
   (multi-let (((_ pat)
		(lit-analyze lit)))
       (let ((pred (car pat)))
	  (opt::strip-sym-with-type pred)
	  (be CSymbol pred))))
