;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)

(depends-on %module/ nisp)

(eval-when (:load-toplevel)
   ;; We eval here because it's the expansion of the 'opt-load' macro
   ;; and the macros in the file that defines the domain.
   ;; (Bad design choices, which should be fixed -- someday.)
   (eval '(opt-load %optop/ convoys)))

(depends-on :at-run-time %optop/ gcdbg)

(defnisptype Convoy-test-res
    (Structure num - Fixnum
	       score - Number
	       length - Fixnum
	       run-time - Run-time-spec
	       maxmatches num-goal-nodes missteps - Integer
	       soln - (~ Plan-prob-soln)))

(defproc convoy-test-and-show - Void (i - Fixnum reuse-rmgraphs - Boolean)
      (convoy-test-res-show (convoy-test i reuse-rmgraphs false)
			    reuse-rmgraphs))

(defproc convoy-test-series - Void (i j - Fixnum reuse-rmgraphs - Boolean)
   (repeat :for ((num = i :to j))
      (out (:to *query-io*) num 1)
      (convoy-test-res-show (convoy-test num reuse-rmgraphs true)
			    reuse-rmgraphs)))

(specdecl situation-copy-scheme* - (Const :copy-all :incremental))

(defproc convoy-test-res-show - Void (ctr - Convoy-test-res reuse-rmgraphs - Boolean)
   (with ctr
      (let ((steps (reverse (!_acts (!_plan !>soln)))))
	 (out "Situation copy scheme: " situation-copy-scheme*
	      " // reuse-rmgraphs* = " reuse-rmgraphs
	      :% !>num 1 "Score: " (:f "~,0f" !>score)
	      1 " Length: " !>length
	      1 "Search: " (:f "~,1f" (/ (->float !>missteps)
					 (+ !>missteps !>length)))
	      1 "Time: " !>run-time
	      :% 3 "Goal nodes: " !>num-goal-nodes
	      1 "Matches: " !>maxmatches
	      :% 3 "Time/goalnode: " (:f "~,1f" (/ (->float (!_run-time !>run-time))
						   !>num-goal-nodes))
		 1 "Time/maxmatch: " (:f "~,1f" (/ (->float (!_run-time !>run-time))
						   !>maxmatches))
		 1 "Goal nodes/plan step: "
		    (:f "~,0f" (/ (->float !>num-goal-nodes) !>length))
	      :% "Solution:" 
	      :% (:e (repeat :for ((s :in steps))
			(:o 6 s :%)))
	      :% "Actions + time: " (:i> 6) :%
	         (:e (plan-print (!_plan !>soln)))
	      :%))))
      
(specdecl 
;;;;          last-tr* - (Lstructure ()
;;;;			soln
;;;;			maxmatches goal-nodes-num run-time real-time numtried
;;;;			mode)
	  (profile-type* false) - (~ (Const :time :space)))

(specdecl nisptype::bool-type* - Opt-type)

(defproc convoy-test - Convoy-test-res
	             (num - Fixnum copy destroy-after - Boolean)
   (bind ((reuse-rmgraphs* copy)
	  (plan-dbg-amt* 0)
;;;;	  (plan-dbg* false)
;;;;	  (large-scale-plan-dbg* false)
	  )
      (multi-let (((probname prob-stmt)
		   (convoy-prob num)))
	 (eval prob-stmt)
	 (excl:gc t)
	 (multi-let (((solns - (Lst Plan-prob-soln)
		       numtried goal-nodes-num maxmatches - Integer
		       run-time - Run-time-spec
		       _ - Integer
		       _ - (Const :best-first-search
				  :hill-climbing/random-restarts))
		      (cond (profile-type*
			     (search-with-profiling probname))
			    (t
			     (search-for-plan probname)))))
	    (cond (destroy-after
		   (opt::set-global-opt-sym probname false nisptype::bool-type*)))
	    (letrec ()
	       (cond ((null solns)
		      (make Convoy-test-res
			    num -100000 100000
			    run-time
			    maxmatches goal-nodes-num (missteps) false))
		     (t
		      (make Convoy-test-res
			    num
			    (!_score (car solns))
			    (!_length (!_plan (car solns)))
			    run-time
			    maxmatches goal-nodes-num
			    (missteps)
			    (car solns))))
	     :where
		(missteps - Integer ()
		   (out "numtried = " numtried " length = "
			   (!_length (!_plan (car solns))) :%)
		   (let ((r (- numtried
			       (!_length (!_plan (car solns)))
			       1)))
		      (out "missteps = " r :%)
		      r))
		)))))

(defun search-with-profiling (probname)
   (prof:with-profiling (:type profile-type*)
     (search-for-plan probname)))

(specdecl search-for-plan
	  - (Fun (Multvals solns - (Lst Plan-prob-soln)
			   numtried goal-nodes-num maxmatches - Integer
			   run-time - Run-time-spec
			   real-time - Integer
			   mode - (Const :best-first-search
					 :hill-climbing/random-restarts))
		 (Obj . Obj)))

(deffunc convoy-prob - (Mlv Symbol Sexp) (n - Integer)
   (let ((convoys
	    (repeat :for ((i = 1 :to (* 2 n)))
	      :collect (build-symbol convoy- (< i))))
	 (sources
	    (repeat :for ((i = 1 :to (* 2 n)))
	     :collect (build-symbol source- (< i))))
	 (dests
	    (repeat :for ((i = 1 :to (* 2 n)))
	     :collect (build-symbol dest- (< i))))
	 (probname (build-symbol convoys- (< n))))
      (values probname
	      `(define (problem ,probname)
		  (:domain convoys)
		  (:objects
		      ,@convoys
		      - Convoy
		      ,@sources ,@dests mid-s mid-d
		      - Point)

		  (:facts
		      ,@(repeat :for ((cv :in convoys))
			 :append `((= (base-speed ,cv)
				      ,(/ 10.0 3600))))  ; 10 km/hr
		      ,@(repeat :for ((src :in sources))
			 :append `((= (geo-dist ,src mid-s) 10.0)))
		      (= (geo-dist mid-s mid-d) 10.0)
		      ,@(repeat :for ((dst :in dests))
			 :append `((= (geo-dist mid-d ,dst) 10.0)))
		      ,@(repeat :for ((src :in (nthtail n sources))
				      (dst :in (nthtail n dests)))
			 :append `((= (geo-dist ,src ,dst) 30.0))))
		  (:init
		      ,@(repeat :for ((cv :in convoys)
				      (src :in sources))
			 :append `((at ,cv ,src)
				   (current-value (dist ,cv ,src) 0.0)
				   (current-value (dist ,cv mid-s) 10.0)))
		      ,@(repeat :for ((src :in sources))
			 :append `((connected ,src mid-s)
				   (current-value
				       (traffic ,src mid-s)
				       0)))
		      (connected mid-s mid-d)
		      (current-value (traffic mid-s mid-d) 0)
		      ,@(repeat :for ((dst :in dests))
			 :append `((connected mid-d ,dst)
				   (current-value
				      (traffic mid-d ,dst)
				      0)))
		      ,@(repeat :for ((cv :in (nthtail n convoys))
				      (src :in (nthtail n sources))
				      (dst :in (nthtail n dests)))
			 :append `((connected ,src ,dst)
				   (current-value (dist ,cv ,dst) 30.0)
				   (current-value
				       (traffic ,src ,dst)
				       0))))
		  (:goal
		     (and 
		      ,@(repeat :for ((cv :in convoys)
				      (dst :in dests))
			 :append `((at ,cv ,dst)))))
		  (:metric minimize (total-time))))))
