;-*- Mode: Common-lisp; Package: nisp; Readtable: ytools; -*-
(in-package :nisp)

(depends-on %module/ nisp)

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(almost-copy copy-with-changes)))

;;; (copy-with-changes object [- type]
;;;     (field1 val1
;;;      field2 val2
;;;        ...))
;;; creates a copy of 'object' with the given fields changed.  
;;; There is an implicit (with object ...) wrapped around the
;;; the computations of the vals. 

(datafun decl-compl copy-with-changes
   (defun :^ (exp dest-type)
      (multi-let (((copied type new-slot-vals)
		   (let ((old-type-decl
			     (cond ((is-type-desig (caddr exp))
				    (out (:to *error-output*)
					 "Old type declaration in " exp :%)
				    (caddr exp))
				   (t false))))
		      (let ((exp1 (cond (old-type-decl
					 `(,(car exp) ,(cadr exp) ,@(cdddr exp)))
					(t exp))))
			 (let ((type-decl (memq '- exp1)))
			    (let ((exp2
				     (cond (type-decl
					    `(,@(ldiff exp type-decl)
						,@(cddr type-decl)))
					   (t exp1))))
			       ;;;;(err-out "Ready to 'match-let' " exp2)
			       (match-let (?_ ?copied ?new-slot-vals ?@unch)
					   exp2
				  (cond ((not (null unch))
					 (out (:to *error-output*)
					    "Explicit list of unchanged slots in "
					    :% exp
					    :% "  -- ignoring" :%)))
				  (values copied
					  (cond (type-decl
						 (designated-type (cadr type-decl)))
						(old-type-decl
						 (designated-type old-type-decl))
						(t dest-type))
					  new-slot-vals))))))))
;;;;	 (err-out "copied = " copied " type = " type
;;;;		 " new-slot-vals = " new-slot-vals)
	 (let ((obdc (decl-compile copied type)))
	    (let ((classtype (Dclcmp-typ obdc))
		  (classexp (Dclcmp-exp obdc)))
	       (let ((class-initargs (class-assemble-initargs classtype)))
		  ;; 'class-initargs' is a list of pairs (keyword slot)
		  (let ((changed-args
			   (bind ((implicit-ob* `(*dc ,obdc)))
			      (repeat :for ((kl = new-slot-vals
						:then (cddr kl)))
			       :until (null kl)
			       :within
				 (let ((slot (alref class-initargs
						    (first kl))))
				    ;;;;(err-out "Compiling " kl)
				    (cond (slot
					   (:continue
					    :nconc (allocate-new kl slot)))
					  (t
					   (signal-problem copy-with-changes
					      "Illegal initarg " (first kl)
					      " for type " classtype))))
			       :where
			          (:def allocate-new (kl slot)
				     (list (first kl)
					   (decl-compile-exp
					      (cond ((*-*-occurs-in
						        (second kl))
						    `(let ((*-* !>,(first kl)))
							(second kl)))
						    (t
						     (second kl)))
					      (type-slot-fun
						 classtype
						 slot 'type true)))))))
			(kept-args
			   (repeat :for ((initpair :in class-initargs))
			    :when (not (repeat :for ((kl = new-slot-vals
							 :then (cddr kl)))
					:until (null kl)
					:result false
					:until (eq (first kl) (car initpair))
					:result true))
			    :nconc (list (first initpair)
					 (match-let (?ob ?_ ?_ ?acc)
						    (must-type-slot
						       classexp classtype
						       (second initpair)
						       'acc)
					    `(,acc ,ob))))))
		     (type-trans
			`(make-inst ,(type-feature classtype 'class-name)
				    ,@changed-args
				    ,@kept-args)
			classtype
			dest-type))))))))
	
(defmacro almost-copy (ob &rest stuff)
   `(decl () (copy-with-changes ,ob ,@stuff)))

(defmacro copy-with-changes (ob &rest stuff)
   `(decl () (copy-with-changes ,ob ,@stuff)))

(deffunc *-*-occurs-in - Boolean (exp - Sexp)
		(cond ((atom exp) (eq exp '*-*))
		      ((memq (car exp) '(quote !=)) false)
		      (t
		       (some #'*-*-occurs-in exp))))