;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)

(depends-on %module/ nisp)

(defproc l-node-see - Void (ln - L-node)
   (with ln
     (out
      "L-node " (!_propo !>occ) "@" (!_length !>pp) " ^" !>height
      :% 1 !>status
      :% " Effort-specs: "
      :% (:q ((null !>effort-specs) "none")
	     (t
	      (:e (repeat :for ((es :in !>effort-specs)
				(i = 1 :by 1))
		     (:o i 1 es :%)))))
      :%
      (:q ((not (memq !>status '(*unchecked *true)))
	   1 (:pp-nl :linear)
	   "Instances: "
	   (:pp-nl :linear)
	   !>instances
	   1 (:pp-nl :linear)
	   (len !>regressions)
	   " regressions"))
      :%)))

(defproc effort-spec-see - Void (es - Effort-spec height - Fixnum)
   (let-fun ()
      (es-walk es '() height)
    :where
      (es-walk - Void (es - Effort-spec
		       ancestors - (Lst Effort-spec) 
                       height - Fixnum)
         (let ((cycle (one-effort-spec-see es ancestors)))
            (cond ((and (not cycle) (> height 0))
		   (out-indent t 2
                      (!= height (- height 1))
		      (!= ancestors (cons es *-*))
		      (repeat :for ((sub :in (!_children es))
				   (i = 1 :by 1))
			 (red-walk i sub ancestors height)))))))

      (red-walk - Void (i - Integer sub - Reduction
			ancestors - (Lst Effort-spec)
			height - Fixnum)
	 (out (:to *query-io*)
	      i "> "
	      (:q ((!_effective sub) "*"))
	      (:q ((!_action sub)
		   (!_action sub))
		  (t "In previous plan prefix ..."))
	       :% 3 (:q ((eq (!_instances sub) '*unknown)
			"-- instances unknown")
		       ((null (!_instances sub))
			"-- no instances")
		       (t
			"Instances: " (!_instances sub)))
	       :% 3 "Linkenv: " (!_linkenv sub) :%)
	 
	 (out-indent t 2
	    (repeat :for ((sub-es :in (!_preconds sub)))
	       (es-walk sub-es ancestors height))))))

(defproc one-effort-spec-see - Boolean (es - Effort-spec 
                                        ancestors - (Lst Effort-spec))
   (let ((cycle (memq es ancestors)))
      (out (:to *query-io*)
	   (!_height es) ;;;;(:q ((!_in-progress es) "?"))
	   ": "
	   (!_conjuncts es) 
	   (:q (cycle " \\\\"))
	   :% " Effort = " (!_effort es)

	   " [id = " (!_varid es) "] " 
	   (:q ((null (!_constraints es))
	       :% " No constraints")
	      (t
	       :% " Constraints: " (!_constraints es) :%))
	   (:q ((has-qvars (!_literal es))
	       (:q ((null (!_instances es))
		   1 "-- No instances")
		  (t
		   :% 3 "Instances: " (!_instances es)))))
	   :%
           (:q ((eq (!_actions es) '*in-progress)
	       3 "Actions in progress")
              ((null (!_actions es))
	       (:q ((not (= (!_effort es) 0))
		   3 "No actions")))
              (t
	       3  "Actions: " (!_actions es)))
	   :%
	   (:q ((and (not (null ancestors))
		    (not (nisp::exists (p in (!_parents es))
                            (eq (car ancestors) (!_reducee p)))))
	       "**PARENT MISSING**" :%)
	      ((null (!_parents es))
	       "TOP" :%))
	   (:q ((nisp::exists (p in (!_parents es))
		  (not (memq (!_reducee p) ancestors)))
	       "Other parents" :%)) 
;(<? (\\ (P) (NOT (MEMQ P ANCESTORS)))
;				     (!_PARENTS ES))
;	       :%)
)
      cycle))

(specdecl esl* - (Lst Effort-spec)
          pp* - Plan-prefix
          et* - Exp_index
	  es* - Effort-spec
	  es-trail* - (Lst Effort-spec)
	  ln* - L-node)

(defproc last-effort-specs - Integer ()
    (!= pp* (caar (!_contents last-effort-tab*)))
    (!= et* (cadar (!_contents last-effort-tab*)))
    (!= esl* (opt::exp_index-contents et* false))
    (!= es-trail* '())
    (!= ln* (lookup-l-node 'top pp* last-effort-tab*))
    (!= es* (list-elt (!_effort-specs ln*) 0))
    (len esl*))

(specdecl (display-effort-zeroes* true) - Boolean)

(specdecl (rgmark* 0) - Integer)

(defproc display-effective-subgraph - Void (es - Effort-spec)
   (!= rgmark* (+ *-* 1))
   (let-fun ()
      (disp-sub es 0)
    :where
      (disp-sub - Void (es - Effort-spec indent - Fixnum)
         (one-effort-spec-flash es indent)
         (repeat :for ((ch :in (!_children es)))
          :result nil
          :until (!_effective ch)
          :result (let ((indent (+ indent 3)))
                    (out (:_ indent) "Action " (!_action ch) :%)
                    (repeat :for ((sub-es :in (!_preconds ch))
				   - Effort-spec)
                       (cond ((or display-effort-zeroes*
                                  (> (!_effort sub-es) 0))
                              (cond ((and (= (!_mark sub-es) rgmark*)
                                          (> (!_effort sub-es) 0))
                                     (out (:to *query-io*)
                                          (:_ indent)
                                          (!_literal sub-es)
					  " already displayed"
                                          :%))
                                    (t
                                     (!= (!_mark sub-es) rgmark*)
                                     (disp-sub sub-es indent)))))))))))

(defproc one-effort-spec-flash - Void (es - Effort-spec indent - Fixnum)
   (let ()
      (out (:to *query-io*)
	   (:_ indent)
	   (!_height es) ;;;;(:q ((!_in-progress es) "?"))
	      ": "
	   (!_literal es)
           (:q ((> (!_effort es) 0)
	       " Effort = " (!_effort es))
              (t " True"))
	   ;(:q (CYCLE " \\\\"))
           (:q ((> (!_effort es) 0)
	       :%
	       (:q ((eq (!_actions es) '*in-progress)
		   "Actions in progress ")
		  ((eq (!_actions es) '*unknown)
		   "Actions unknown ")
		  ((null (!_actions es)) "No actions")
		  (t  "Actions: " (!_actions es)))
	       (:q ((not (atom (!_actions es)))
		       ;; (NOT (ATOM (!_CRITICAL ES)))
		   :% (:_ (+ indent 1)))
		  (t 1))))
	       ;(:q ((EQ (!_CRITICAL ES) '*UNKNOWN)
;		   "Criticals unknown")
	;	  (T "Critical: " (!_CRITICAL ES)))))
	   :%)))

(defproc next-es - Void (pat)
   (cond ((is-Number pat)
	  (!= pat `(1 ,pat))))
   (cond ((is-list-of pat #'is-Number)
	  (repeat :for ((okay true))
	   :until (or (null pat) (not okay))
	     (cond ((and (is Integer (car pat))
			 (> (car pat) 0)
			 (=< (car pat) (len (!_children es*))))
		    (let ((red (elt (!_children es*) (- (car pat) 1)))
			  - Reduction)
		       (!= pat (cdr pat))
		       (cond ((and (not (null pat))
				   (is Integer (car pat))
				   (> (car pat) 0)
				   (=< (car pat) (len (!_preconds red))))
			      (!= es* (elt (!_preconds red)
					   (- (car pat) 1)))
			      (!= pat (cdr pat)))
			     (t
			      (out "Bogus: " pat :%)
			      (!= okay false)))))
		   (t
		    (out "Bogus: " pat :%)
		    (!= okay false)))))
	 ((eq pat 'back)
	  (cond ((not (null es-trail*))
		 (!= es* (car es-trail*))
		 (!= es-trail* (cdr *-*)))))
	 (t
	  (!= es* (lookup-effort-spec pat pp* last-effort-tab*))))
   (cond ((or (null es-trail*)
	      (not (eq (car es-trail*) es*)))
	  (!= es-trail* (cons es* *-*))))
   (effort-spec-see es* 1))

(defproc next-l-node - Void (pat - Sexp)
   (let ((ln (lookup-l-node pat pp* last-effort-tab*)))
      (cond (ln
	     (!= ln* ln)
	     (l-node-see ln))
	    (t "?"))))

(defproc reg-meth-see - Void (rm - Regress-method)
   (with rm
      (out !>change
	   (:q ((not !>good) "[nogood]"))
	   (:q (!>zone " for sits reachable from " !>zone))
	   :%
	   !>precond-id ": " !>bdgs
	   :% "Action: " !>action
	   :% "Precond: " !>precond
	   :%)))