;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)

;;; From lisp listener, connects to Demo_home.class, the main Java program
;;; for running the demo.

(defconstant +demo-port+ 7778)

(defvar server-address* nil)

(defvar last-socket*)
(defvar demo-socket*)

(defvar demo-dbg* false)

(defclass socket-stream (acl-socket:socket-stream-internet-active)
  ())

(defun simply-connect (&optional (addr server-address*))
  (!= demo-socket*
      (acl-socket:make-socket
         :connect ':active
	 :remote-host addr
	 :remote-port +demo-port+)))

(defun test (x)
   (bind ((*error-output*
	     (make-two-way-stream
		*terminal-io*
		(make-broadcast-stream
		   *terminal-io*
		   demo-socket*))))
      (dbg-out plan-dbg*
	 x)))

(defun connect-to-demo (&optional (server-host false))
      (cond (server-host
	     (cond ((is-String server-host)
		    (!= server-address* server-host))
		   (t
		    (signal-problem connect-to-demo
		       "Illegal server-host " server-host))))
	    (server-address*
	     (!= server-host server-address*))
	    (t 
	     (out (:to *query-io*) "Name or IP address of server machine: ")
	     (setq server-host
                   (string-trim " \t" (in (:from *query-io*) :linestring)))
	     (cond ((string= server-host "")
		    (setq server-host server-address*))
		   (t
		    (setq server-address* server-host)))))
      (block connect
	 (bind ((demo-socket*
			  (acl-socket:make-socket
				:connect ':active
				:remote-host server-host
				:remote-port +demo-port+)))
	    (!= last-socket* demo-socket*)
	    (let-fun ()
	       (unwind-protect
		  (bind ((tpl:*print*
			       (\\ (x &optional (srm *standard-output*))
				  (try-send-to-socket-stream " " x)
				  (cond ((eq srm 't)
					 (!= srm *standard-output*)))
				  (print x srm)))
			 (tpl:*read*
			    (\\ (&optional (isrm *terminal-io*) (eof-error-p t)
					   (eof-val nil) (recursive-p nil))
			       (let ((r (read isrm eof-error-p eof-val recursive-p)))
				  (try-send-to-socket-stream "> " r)
				  r)))
			 (*error-output*
			    (make-two-way-stream
			       *error-output*
			       (make-broadcast-stream
				  *terminal-io*
				  demo-socket*))))
		     (simple-repl))
		 (progn
		    (format t "Abort!~%")
		    (cond (demo-socket*
			   (close demo-socket*)))))
	     :where
	       (try-send-to-socket-stream (prefix x)
		  (handler-case
		     (progn (dbg-out demo-dbg*
			       "Sending to socket: " (:a prefix) x :%)
			    (out (:to demo-socket*) (:a prefix) x :%)
			    (force-output demo-socket*))
		   (error ()
		      (out (:to *terminal-io*)
			   "Can't write to socket stream; bailing" :%)
		      (return-from connect false))))))))

(defun simple-repl ()
   (let-fun ()
      (repeat :for ((r = (prompt-for-input) :then :again))
       :until (eq r ':pop)
	  (force-output *terminal-io*)    
	  (let ((vals (values->list (eval r))))
	     (repeat :for ((v :in vals))
		(funcall tpl:*print* v *terminal-io*)
		(out (:to *terminal-io*) :%)))
	  (force-output *terminal-io*))
    :where
      (prompt-for-input ()
	 (out (:to *terminal-io*) "> ")
	 (funcall tpl:*read* *terminal-io*))))

(defun xmit (&optional (term-string "---+++---"))
   (repeat :for (:collector lines)
    :within
      (let ((r (in (:from *terminal-io*) :linestring)))
	 (:continue
	  :until (string= r term-string)
	  :collect r))
    :result
       (repeat :for ((r :in lines))
	  (out (:to demo-socket*)
	     (:a r) :%)))
   (values))