;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: domain.lisp,v 1.32 2005/12/08 21:09:10 dvm Exp $

(depends-on %module/ nisp)

(depends-on :at-run-time %optop/ types cause-intern)

(deffunc find-domain - Domain (name - Symbol)
   (let ((dom (opt::try-domain-with-name name false)))
      (cond ((not dom)
             (signal-problem find-domain :fatal
                "Undefined domain " name)))
      (domain-prepare-for-planning dom)))

(defproc domain-prepare-for-planning - Domain (dom)
   (cond ((is Domain dom)
	  dom)
	 (t
	  (multiple-value-let (dset dget)
			      (opt::domain-name-procs dom)
	     (let ((value-of-name (funcall dget)))
		(cond ((is Domain value-of-name)
		       value-of-name)
		      (t
		       (let ((prepped-main-parent
				(cond ((domain-is-local dom)
				       (domain-prepare-for-planning
					  (car (!_(Opt-domain parents) dom))))
				      (t false))))
;;;;                         (dbg-save dom prepped-main-parent)
;;;;                         (breakpoint domain-prepare-for-planning
;;;;                             "Prepped " dom " -> " prepped-main-parent)
			  (let ((pd (opt::domain-indexify dom))
				- Indexed-domain)
			     (let ((udom
				     (with pd
					(make-inst Domain
					 :name !>name
					 :generation !>generation
					 :fill-time !>fill-time
					 :parents !>parents
					 :requirements !>requirements
					 :local-bdgs !>local-bdgs
					 :inherited-bdgs !>inherited-bdgs
					 ;;:AXIOMS !>AXIOMS
					 ;;:facts !>facts
					 :actions !>actions
					 :occasions !>occasions
					 :act-dif-tab false
					 :pred-causers false
					 :causation-effects false
					 :namespace !>namespace
					 :namespace-tab !>namespace-tab))))
				(cond (prepped-main-parent
				       (!= (!_parents udom)
					   (cons prepped-main-parent
						 (cdr *-*)))))
				(funcall dset udom)
				udom))))))))))

(defmethproc domain-make-current - Void (dom - Domain)
   (let ((needed-work false))
      (let ((susp1 false))   ;(SUSPICIOUS-ANCESTORS DOM)
	;;;;(opt::indexed-domain-make-current dom)
	(let ((susp2 (suspicious-ancestors dom)))
	   (cond ((or susp1 susp2 (domain-has-old-act-dif-tab dom))
		  (!= (!_act-dif-tab dom) false)
		  (!= (!_pred-causers dom) false)
		  ;;;;(!= (!_achievability dom) !())
		  (!= (!_causation-effects dom) false)
		  (!= (!_act-dif-time dom) (next-generation))
;;;;		  (cond ((or susp1 susp2)
;;;;			 (out (:to *query-io*)
;;;;			      "Domain " dom " suspicious "
;;;;			      (:q (susp1
;;;;				  (:q (susp2 "before and after")
;;;;				     (t "before")))
;;;;				 (t "after"))
;;;;			      " making it current" :%)
;;;;			 (dbg-save :run-loud dom)
;;;;			 (breakpoint Domain/domain-make-current
;;;;			    "What's going on?")))
		  (cond (susp2
			 (reset-ancestors dom)
			 (cond ((suspicious-ancestors dom)
				(error-break domain-make-current :fatal
				   "Horribly suspicious ancestors: " dom)))))))
	   (let ((adom (domain-for-causer-cache dom)))
	      (cond ((and (not (eq adom dom))
			  (domain-has-old-act-dif-tab adom))
		     (!= (!_act-dif-tab adom) false)
		     (!= (!_pred-causers adom) false)
		     (!= (!_causation-effects adom) false)
		     (!= (!_act-dif-time adom) (next-generation))
		     (!= (!_infinite-objects adom) ':compute)
		     (!= needed-work true)))
	      (cond ((call-next-method dom)
		     (!= needed-work true)))
	      needed-work)))))

(deffunc domain-has-old-act-dif-tab - Boolean (domx - Domain)
                    (let ((g (!_act-dif-time domx)))
                      (exists (a in (!_ancestors domx))
                         (> (!_generation a) g))))

(defproc domain-destroy - Void (thname - Symbol)
   (let ((dom (be * (get-global-opt-sym thname))) - (~ Opt-domain))
      (cond (dom
             (out (:to *query-io*) "Destroying " dom :%)
             (opt::map-global-opt-symbols
                (\\ (sym val)
                   (cond ((and (is Plan-problem val)
	                       (exists (a in (!_(Domain ancestors)
                                                (!_(Plan-problem domain) val)))
                                   (or (eq dom a)
                                       (eq (!_name dom) (!_name a)))))
                          (out (:to *query-io*) " flushing " sym)
                          (opt::flush-global-opt-sym sym)))))
             (out (:to *query-io*) :%)
             (opt::flush-global-opt-sym thname)))))

(deffunc suspicious-ancestors - Boolean (dom - Domain)
   (null (!_ancestors dom)))

(deffunc domain-has-infinite-obj-terms - Boolean (dom - Domain)
   (let-fun ()
      (let ((res (!_infinite-objects dom)))
	 (cond ((eq res ':compute)
		(!= res
		    (or (domain-declares-requirement dom ':numbers)
			(block search-for-function
			   (walk-domain-bdgs dom
			      (\\ (bdg - Domain-bdg _ - Opt-domain)
				 (let ((v (!_val bdg)))
				    (cond ((and (is Functional-constant v)
						(troublesome v))
;;;;					   (err-out
;;;;					      "Domain " dom :% " has function " v)
					   (return-from search-for-function
					      v))))))
			   false)))
		(!= (!_infinite-objects dom) res)))
	 res)
    :where
       (:def troublesome (fc - Functional-constant)
	   (and (not (memq (!_name fc) '(not step-value)))
		(not (is Predicate fc))
		(not (is Causation-function fc))
		(not (!_callable fc))
		(not (!_type-inverter fc))))))

#|
NOT USED
(deffunc inferrable - (Lst Sexp) (pat - Sexp dom - Domain)
   (multi-let (((_ pat)
		(lit-analyze pat)))
      (let ((pred (car pat)))
	 (opt::strip-sym-with-type pred)
	 (symbol-inferrability pred dom))))

NOT USED
(deffunc symbol-inferrability - (Lst Sexp) (pred - Symbol dom - Domain)
   (let ((p (alref (!_inferrability dom) pred ':unknown)))
      (cond ((eq p ':unknown)
	     (let ((infl (compute-inferrability pred dom)))
	        (!= (alref (!_inferrability dom) pred)
		    infl)
		infl))
	    (t p))))

NOT USED
(deffunc compute-inferrability - (Lst Fact) (pred - Symbol dom - Domain)
   (let ((rule-groups (all-rule-groups dom)))
      (repeat :for ((rg :in rule-groups) - Rule-group)
       :nconc
	 (repeat :for ((rule :in (!_rules rg)))
          :when (is Fact rule)
	  :within
	     (let-fun ((search-for-concluder - Boolean (e - Sexp)
			  (match-cond e
			     (:? (<- (?,pred ?@_) ?_)
				true)
			     (:? (-> ?_ ?c)
				(search-for-concluder c))
			     (t false))))
		(let ((exp (opt:fact-internalize rule)))
		   (cond ((search-for-concluder exp)
			  (:continue
			   :collect exp)))))))))
|#

(specdecl (achieve-break* false) - Boolean
	  (ach-id* (new-Varid)) - Varid)

(deffunc achievable - Boolean
	     (pat - Sexp dom - Domain)
   (or (not (null (direct-causers pat dom)))
       (not (null (causal-impliers pat dom)))))

(deffunc direct-causers - (Lst Causation-defn)
	      (pat - Sexp dom - Domain)
   (!= dom (domain-for-causer-cache *-*))
   (nodupq 
      (nconc
	 (symbol-direct-influencers (lit-pred pat) dom)
	 (let ((fluterms (<! collect-fluent-terms (cdr pat))))
	    (nconc (<! (\\ (ft) (elapsed-causers ft dom))
		       fluterms)
		   (repeat :for ((ft :in fluterms))
		    :append (symbol-direct-influencers
			       (be Symbol
				   (cond ((consp ft) (car ft))
					 (t ft)))
				dom)))))))

(deffunc collect-fluent-terms - (Lst Sexp) (pat - Sexp)
   (cond ((atom pat) !())
	 ((eq (strip-car pat) 'fl-v)
	  (cond ((consp (cadr pat))
		 (let ((x (strip-car (cadr pat))))
		    (cond ((is-symbolish x)
			   (be Sexp (list (cons x (cdr (cadr pat))))))
			  (t !(Sexp)))))
		((is Sym-with-type (cadr pat))
		 (list (!_(Sym-with-type actual) (cadr pat))))
		(t (list (cadr pat)))))
	 (t
	  (<! collect-fluent-terms (cdr pat)))))

(deffunc elapsed-causers - (Lst Causation-defn) (fluterm - Sexp dom - Domain)
   (match-cond fluterm
      ?( (elapsed (?proc ?@args))
	(decl (proc - CSymbol)
	   (list (name-causation-defn proc 'dur-happening args dom))))
      (t !())))

;;;;(deffunc function-symbols - (Lst Symbol) (x - Sexp)
;;;;   (cond ((is Symbol x) (list x))
;;;;	 ((atom x) '())
;;;;	 (t
;;;;	  (<! function-symbols x))))

(deffunc causal-impliers - (Lst Occasion)
	      (pat - Sexp dom - Domain)
   (symbol-causal-impliers (lit-pred pat) dom))

(defproc symbol-causal-impliers - (Lst Occasion)
	                (pred - CSymbol dom - Domain)
   (!= dom (domain-for-causer-cache *-*))
   (memoize-val 
      (repeat :for ((rg :in (all-rule-groups dom)) - Rule-group)
       :nconc
	 (repeat :for ((rule :in (!_rules rg)))
	  :when (is Fact rule)
	  :within
	    (decl (rule - Fact)
	       (let ((occ (uniquify-occasion (fact-internalize rule)
					     dom)))
		 (cond ((memq pred (rule-links-causer-to-preds
				       occ dom))
			(:continue
			 :collect occ)))))))
      :store-as
	 (second (val-or-initialize
		   (href (place-pred-causers dom)
			 pred)
		   :init (be (Lrcd (Const :unknown) (Const :unknown))
			     (tuple ':unknown ':unknown))))
      :missing-if ':unknown))

;;; Not currently used; would be slightly more efficient
;;; because it uses index to search.  But it's less general,
;;; and the results are cached anyway.
(defproc fetch-causal-impliers - (Lst Occasion)
	                   (pat - Sexp dom - Domain)
   (!= dom (domain-for-causer-cache *-*))
   (memoize-val
      (let ((rules (opt::sit-assertion-fetch
		      `(<- ,pat ?_) ach-id* (empty-env)
		      true
		      (domain-place-fact-sit
				 dom false))))
	 (repeat :for ((trip :in rules))
	  :when (not (null (rule-links-causer-to-preds
				(second trip) dom)))
	  :collect (second trip)))
      :store-as
	 (second
	    (let ((pred (lit-pred pat)))
	       (val-or-initialize
		      (href (place-pred-causers dom)
			    pred)
		      :init (be (Lrcd (Const :unknown) (Const :unknown))
				(tuple ':unknown ':unknown)))))
      :missing-if ':unknown))

(deffunc domain-for-causer-cache - Domain (dom - Domain)
   ;; Local domains cannot have their own action definitions, so
   ;; achievability info should be stored in the main parent.
   ;; EXCEPTION: If domain requires timed initial literals,
   ;; you don't know exactly what predicates are achievable until
   ;; you see the inits.
   (cond ((and (domain-is-local dom)
	       (not (domain-declares-requirement
		         dom
		         ':timed-initial-literals)))
	  (be Domain (car (!_(Domain parents) dom))))
	 (t dom)))

;;; All the Causations that can change the truth value of atomic
;;; formulas with this pred
(defproc symbol-direct-influencers - (Lst Causation-defn)
	           (pred - CSymbol dom - Domain)
   (memoize-val
      (compute-direct-influencers pred dom)
      :store-as
	 (first (val-or-initialize
		    (href (place-pred-causers dom) pred)
		    :init (be (Lrcd (Const :unknown) (Const :unknown))
			      (tuple ':unknown ':unknown))))
      :missing-if ':unknown))

;;; 'pred' is either a predicate or a fluent function.
(deffunc compute-direct-influencers - (Lst Causation-defn)
	                       (pred - CSymbol dom - Domain)
   (cond ((pred-is-rigid pred dom) '())
	 (t
	  (repeat :for ((a :in (!_ancestors dom)))
;;;;             (!= a* a)
	   :nconc (repeat :for ((vb :in (domain-get-local-bdgs a)))
		   :within
		   (cond ((is Domain-bdg vb)
			  (let ((v (!_val vb)))
			     (cond ((is-simple-causation-function v)
				    (decl (v - Causation-function)
				       (cond ((memq pred
						    (find-causation-effects
						       v dom))
					      (:continue
					       :collect
						  (!_defn v)))))))))))))))

(specdecl (link-id* (new-Varid)) - Varid)

;;; If 'rule' has indirect links from its antecedent literals to
;;; Causations, then return a list of all the preds such that
;;; the rule can be used to derive atomic formulas involving those
;;; preds.  If it doesn't, return !().
(deffunc rule-links-causer-to-preds - (Lst CSymbol)
	                  (rule - Occasion dom - Domain)
   (memoize-val
      (let ((fsit (domain-place-fact-sit dom false)))
	 (let-fun ()
	    (cond ((search-for-pred-path rule !())
		   (rule-affected-preds
		      (!_propo rule) true link-id* (empty-env)
		      '(and) dom))
		  (t !()))

	  :where

	     (:def search-for-pred-path - Boolean
		              (rule - Occasion
			       pred-path - (Lst Back-chainer))
		(repeat :for ((a :in (opt::formula-antecedent-atoms
					(!_propo rule))))
		 :result false
		 :within
		   (let ((next-pred (car a)))
		      (opt::strip-sym-with-type next-pred)
		      (:continue
		       :until (memq next-pred pred-path)
		       :result false
		       :until (not (null (symbol-direct-influencers
					  next-pred dom)))
		       :result true
		       :within
			 (let ((pushed-path
				  (cons next-pred pred-path))
			       (bcl (opt::backchain-fetch
				       a dummy-id* (empty-env)
				       false fsit)))
			    (:continue
			     :until (exists (feeder :in bcl)
				      (search-for-pred-path
					 (third feeder)
					 pushed-path))
			     :result true))))))))
      :store-as
	 (href (place-causation-effects dom)
	       rule
	       ':uncomputed)
      :missing-if ':uncomputed))

(specdecl (effect-id* (new-Varid)) - Varid)

;;;;(deffunc find-rule-effects - (Lst CSymbol)
;;;;	                (rule - Fact dom - Domain)
;;;;   (let ((ce-tab (place-causation-effects dom)))
;;;;      (let ((p (href ce-tab rule ':uncomputed)))
;;;;	 (cond ((eq p ':uncomputed)
;;;;		(let ((rule (uniquify-occasion-with-vars
;;;;			        (fact-internalize rule)
;;;;				dom)))
;;;;		   (let ((preds (rule-affected-preds
;;;;				   (!_propo rule) true
;;;;				    effect-id* (empty-env) '(and) dom)))
;;;;		      (!= preds (nodupq *-*))
;;;;		      (!= (href ce-tab rule) preds)
;;;;		      preds)))
;;;;	       (t p)))))

(deffunc rule-affected-preds - (Lst CSymbol) (rule - Sexp polarity - Boolean
					      id - Varid env - Bdgenv
					      condition - Sexp dom - Domain)
   (cond ((is Qvar rule)
	  (let ((dnf-pre (disjunctive-normal-form condition true)))
	     (repeat :for ((disjunct :in dnf-pre))
	      :nconc
		(let ((rl (reduce-to-rebind
				 disjunct rule id id env dom)))
		   (repeat :for ((trip :in rl))
		      ;; trip is (new conjuncts, new qvar val, new env)
		    :nconc (rule-affected-preds
			      (second trip) polarity id (third trip)
			      `(and ,@(first trip) ,condition) dom))))))
	 ((atom rule) '())
	 (t
	  (let ((fctn (car rule)))
	     (opt::strip-sym-with-type fctn)
	     (decl (fctn - CSymbol)
		(let-fun ()
		   (case fctn
		     ((and or)
		      (cond ((boole-eq polarity (eq fctn 'and))
			     (repeat :for ((conjunct :in (cdr rule)))
			      :nconc
				 (rule-affected-preds conjunct polarity id env
						      condition dom)))
			    (t
			     (repeat :for ((disjunct :in (cdr rule) :tail dtail))
			      :nconc
				 (rule-affected-preds
				    disjunct polarity id env
				    `(and ,condition
					  (let ((sibs (append (ldiff (cdr rule) dtail)
							      (cdr disjunct))))
					     (cond (polarity
						    (<# literal-neg sibs))
						   (t sibs))))
				    dom)))))
		      ((if -> >-)
		       (impl-affected-preds (cadr rule) (caddr rule)))
		      ((<-)
		       (impl-affected-preds (caddr rule) (cadr rule)))
		      ((not)
		       (rule-affected-preds (cadr rule) (not polarity)
					    id env condition dom))
		      ((forall exists exists!)
		       (rule-affected-preds
			  (caddr rule) polarity id env condition dom))
		      (t
		       (cond (polarity
			      (list fctn)
;;;;			      (cond ((exists (d :in (disjunctive-normal-form
;;;;						       condition true))
;;;;					(exists (c :in d)
;;;;					   (achievable c dom)))
;;;;				     (list fctn))
;;;;				    (t !()))
			      )
			     (t !()))))
	         :where
	           (:def impl-affected-preds - (Lst CSymbol) (ante conse - (Lst Sexp))
		      (cond (polarity
			     (rule-affected-preds conse true id env
						  `(and ,ante ,condition) dom))
			    (t
			     (nconc (rule-affected-preds ante true id env
							 condition dom)
				    (rule-affected-preds conse false id env
							 condition dom)))))))))))

(deffunc is-simple-causation-function - Boolean (v - Obj)
   (and (is Causation-function v)
	(decl (v - Causation-function)
	   (and (let ((ftype (!_type v)))
		  (= (nity::type-find-feature ftype 'nity::level global-opt-env*)
		     0))
		(let ((defn (!_defn v)))
		   (or (not (is Action-defn defn))
		       (decl (defn - Action-defn)
			  (not (!_has-methods defn)))))))))

(deffunc find-causation-effects - (Lst CSymbol) (v - Causation-function
					        dom - Domain)
   (let ((vname (!_name v))
	 (ce-tab (place-causation-effects dom)))
      (let ((effl (href ce-tab vname ':uncomputed)))
;;;;	 (dbg-save effl v dom)
;;;;	 (breakpoint find-action-effects
;;;;	    "p = " effl)
	 (cond ((eq effl ':uncomputed)
		(let ((proj-version (causation-defn-projectable
				        (!_defn v) dom)))
;;;;		   (dbg-save proj-version)
;;;;		   (breakpoint find-action-effects
;;;;		      "proj-version = " proj-version)
		   (letrec ((find-effects (idefn-or-iview
					     - Internalized-causation-defn)
			       (union
				  (effect-affected-preds
				     (!_effect idefn-or-iview)
				     true (new-Varid) (empty-env)
				     'true 'true dom)
				  (cond ((is Internalized-process-defn
					     idefn-or-iview)
					 (union
					    (cond ((!_start-effect
						      idefn-or-iview)
						   (effect-affected-preds
						      (!_start-effect
							 idefn-or-iview)
						      true (new-Varid)
						      (empty-env)
						      'true 'true dom))
						  (t !()))
					    (cond ((!_stop-effect
						      idefn-or-iview)
						   (effect-affected-preds
						      (!_stop-effect
							 idefn-or-iview)
						      true (new-Varid)
						      (empty-env)
						      'true 'true dom))
						  (t !()))))))))
		      (let ((eff (cond ((is Internalized-durative-defn
					    proj-version)
					(nconc (find-effects
						  (!_process-view
						     proj-version))
					       (find-effects
						  (!_start-view
						     proj-version))
					       (cond ((!_stop-view
						          proj-version)
						      (find-effects
							 (!_stop-view
							     proj-version)))
						     (t !()))))
				       (t
					(decl (proj-version
						 - Internalized-causation-defn)
					   (find-effects
					      proj-version))))))
;;;;			(dbg-save vname dom eff)
;;;;			(breakpoint find-causation-effects
;;;;			   "About to cache " vname 1 eff)
			 (!= eff (nodupq eff))
			 (!= (href ce-tab vname)
			     eff)
			 eff))))
	       (t (be (Lst CSymbol) effl))))))

(deffunc effect-affected-preds - (Lst CSymbol)
			       (effect - Sexp
				allow-time-spec - Boolean
				eff-id - Varid env - Bdgenv
				precond secondary - Sexp dom - Domain)
;;;;   (dbg-save effect)
;;;;   (breakpoint effect-affected-preds
;;;;      "effect = " effect)
   (cond ((is Qvar effect)
	  ;; Assume that all useful reductions are known before planning
	  ;; starts.  (No action adds facts of form (<- (pred ...) ...).)
	  (reduced-precond-affected-preds
	     effect allow-time-spec eff-id env precond secondary dom))
	 ((atom effect) '())
	 (t
	  (let ((fctn (car effect)))
	     (opt::strip-sym-with-type fctn)
;;;;	     (dbg-save fctn)
;;;;	     (breakpoint effect-affected-preds
;;;;		"fctn = " fctn)
	     (decl (fctn - CSymbol)
		(case fctn
		   ((and)
		    (<! (\\ (e)
			   (effect-affected-preds
			      e allow-time-spec eff-id env precond secondary dom))
			(cdr effect)))
		   ((not)
		    (effect-affected-preds (cadr effect)
					   allow-time-spec eff-id env
					   precond secondary dom))
		   ((when if)
		    (effect-affected-preds (caddr effect)
					   allow-time-spec eff-id env
					   precond
					   `(and ,(if-when-precond
						     (car effect) (cadr effect))
						 ,secondary)
					   dom))
		   ((forall)
		    (effect-affected-preds (caddr effect)
					   allow-time-spec eff-id env
					   precond secondary dom))
		   ((at over)
		    (let (q)
		       (cond ((and allow-time-spec
				   (matchq (?@(:\| (at ?(:\| start end))
						  (over all)
					    ?q))
					   ;; Avoid strip-sym-with-type lossage:
					   `(,fctn ,(cdr effect))))
			      (effect-affected-preds
				  q false eff-id env
				  precond secondary dom))
			     (t
			      (cons fctn
				    (<! collect-fluent-funs (cdr effect)))))))
		   ((derivative deriv-comp)
		    (match-let (?q ?d) (cdr effect)
		       (nconc (be (Lst CSymbol)
				  (cond ((is-Pair q) (list (car q)))
					(t (list q))))
			      (collect-fluent-funs d))))
		   ((increase decrease assign change)
		    (let ((fl (cadr effect)))
		       (strip-sym-with-type fl)
		       (cond ((is-Pair fl)
			      (!= fl (car fl))
			      (strip-sym-with-type fl)))
		       (decl (fl - CSymbol)
			  (list fl))))
		   (t
		    (list fctn)
   ;;;;		 (cons fctn
   ;;;;		       (<! collect-fluent-funs (cdr effect)))
		   )))))))

;;; The fluent-funs are symbols like 'container-capacity',
;;; which occur in internalized formulas in form
;;;     (fl-v (container-capacity ...))
(deffunc collect-fluent-funs - (Lst Symbol) (pat - Sexp)
   (cond ((atom pat) !())
	 ((eq (strip-car pat) 'fl-v)
	  (let ((x (strip-car (cadr pat))))
	     (cond ((is CSymbol x) (list (!_sym x)))
		   (t !(Symbol)))))
	 (t
	  (<! collect-fluent-funs (cdr pat)))))

(defun if-when-precond (if-when precond)
   (cond ((eq if-when 'if)
	  `(after-action ,precond))
	 (t precond)))

(deffunc reduced-precond-affected-preds - (Lst CSymbol)
	                     (effect - Qvar allow-time-spec - Boolean
			      eff-id - Varid env - Bdgenv
			      precond secondary - Sexp dom - Domain)
   (let ((dnf-pre (disjunctive-normal-form
		     precond true))
	 (dnf-sec (disjunctive-normal-form
		     secondary true)))
      (nconc (<! (\\ (pconj)
		    (let ((rl (reduce-to-rebind
				 pconj effect eff-id eff-id
				 env dom)))
		       (<! (\\ (r)
			      (effect-affected-preds
			         (cadr r) allow-time-spec eff-id (caddr r)
				 `(and ,@(car r)) secondary dom))
			   rl)))
		 dnf-pre)
	     (<! (\\ (sconj)
		    (let ((rl (reduce-to-rebind
				 sconj effect eff-id eff-id
				 env dom)))
		       (<! (\\ (r)
			      (effect-affected-preds
			         (cadr r) allow-time-spec eff-id (caddr r)
				 precond `(and ,@(car r)) dom))
			   rl)))
		 dnf-sec))))

;; Find and apply one reduction of one goal in pre that affects the 
;; value of var.  Returns list of triples: new conjuncts, new value of
;; var, new Bdgenv.
;; This function is greedy; it stops as soon it finds a conjunct that
;; does the job.
(deffunc reduce-to-rebind - (Lst (Lrcd (Lst Sexp) Sexp Bdgenv))
			  (pre - (Lst Sexp)
			   var - Qvar
			   id-pre id-var - Varid bdgs - Bdgenv dom - Domain)
   (let ((fs (domain-place-fact-sit dom true)))
      (repeat :for ((p :in pre :tail pl-tail) - Sexp
		   (unreduced '()) - (Lst Sexp)
		   (sought '()) - (Lst (Lrcd (Lst Sexp) Sexp Bdgenv))
		   (found false) - Boolean)
	 (letrec ((subst-answer - (Lst (Lrcd (Lst Sexp) Sexp Bdgenv))
				(new-conjuncts - (Lst Sexp) e - Bdgenv)
		     ;;(out (tr subst-answer ("subst> " p 1 new-conjuncts 1 e)
		     (multiple-value-let (ans e)
					 (safe-varsubst var id-var e)
			(cond ((is Qvar ans)
			       '())
			      (t
			       (list (tuple `(,@(reverse unreduced)
					      ,@new-conjuncts
					      ,@(cdr pl-tail))
					    ans
					    e)))))
		     ;;("subst< " (car out-vals*))))
		     ))
	    (let ((al (deduce p id-pre bdgs fs dom)))
	       (cond ((null al)
		      (cond ((plausibly-reducible (be CSymbol (car p))
						  dom)
			     (let ((bl (find-reductions p id-pre bdgs fs dom)))
				(!= sought
				    (<! (\\ (q)
					   (subst-answer (car q) (cadr q)))
					bl))))))
		     (t
		      ;; Have to also deal with the case
		      ;; where goal is actually true in dom.
		      (!= sought
			  (<! (\\ (a)
				 (subst-answer '() a))
			      al))))
	       (cond ((null sought)
		      (!= unreduced (cons p *-*)))
		     (t
		      (!= found true)))))
       :until found
       :result sought)))

(deffunc name-causation-defn - Causation-defn
	           (sym special - CSymbol args - (Lst Sexp) dom - Domain)
   (strip-sym-with-type sym)
   (cond ((eq sym special)
	  (name-causation-defn
	     (strip-car (car args)) 'dummy (cdr (car args)) dom))
	 (t
	  (let ((symval
		   (find-domain-bdg-val sym dom)))
	     (cond ((and symval (is Causation-function symval))
		    (!_(Causation-function defn) symval))
		   (t
		    (signal-problem name-causation-defn
		       "Can't get causation definition associated with: "
		       `(,sym ,@args))))))))

(deffunc uniquify-like-occasion - Uterm (pat - Sexp context - Domain
					 createfn - (Fun Uterm (Sexp Domain)))
  (let ((index (opt::indexed-domain-place-occasion-index context)))
     (repeat :for ((cand :in (exp-fetch pat index true))
		   - Uterm)
      :result (let ((new-uterm (funcall createfn pat context)))
		  (exp-ob-index new-uterm index true)
		  new-uterm)
      :until (equal (!_term cand) pat)
      :result cand)))

(deffunc must-be-list - Boolean (l - Sexp kind)
   (cond ((and (atom l) (not (null l)))
	  (error-break must-be-list :fatal
	     "Illegal :" (d (string-upcase kind)) "S: " l)))
   true)

(deffunc occasion-neg - Occasion (occ - Sexp th - Domain)
   (let ((n (cond ((eq (car occ) 'not) (cadr occ))
                  (t `(not ,occ)))))
      (uniquify-occasion n th)))

(deffunc must-be-well-formed - (Lst Sexp) (l - Sexp)
         (cond ((well-formed-literal l)
		(list l))
	       (t
		(error-break occasion-neg
		   "Illegal: " l
		   (:novalue "I will ignore it"))
		'())))

(deffunc well-formed-literal - Boolean (fmla - Sexp)
   (and (consp fmla)
	(or (is CSymbol (car fmla))
	    (is Sym-with-type (car fmla)))
	(repeat :for ((r (cdr fmla) (cdr r)) - Sexp
		   (i = 0 :by 1) - Fixnum)
	 :until (null r)
	 :result true
	 :while (and (consp r)
		    (well-formed-term (car r)))
	 :result false)))

(deffunc well-formed-term - Boolean (a - Sexp)
  (or (is Qvar a) (is Number a) (is CSymbol a)
      (well-formed-literal a)))

(deffunc mapeltreduce2 - (Mlv Obj Obj) (fcn - (Fun (Mlv Obj Obj)
						   (Obj Obj (Lst Obj)) nil)
					i1 i2 - Obj
					l - (Lst Obj))
   (cond ((null l) (values i1 i2))
	 (t
	  (multiple-value-let (i1 i2)
			      (funcall fcn i1 i2 (car l))
	     (mapeltreduce2 fcn i1 i2 (cdr l))))))

(deffunc seq-uniquify - (Lst Action) (l - (Lst Sexp) d - Domain)
   (<# (\\ (a) (uniquify-action a d))
       l))

(deffunc domain-is-local - Boolean (d - Opt-domain)
   (matchq (domain ?_) (!_name d)))

(deffunc uterm-exp->sexp - Sexp (ut - Uterm-exp)
   (cond ((is Uterm ut)
	  (!_term ut))
	 ((atom ut)
	  (be Sexp ut))
	 (t
	  (<# uterm-exp->sexp
	      (be (Lst Uterm) ut)))))

(deffunc place-causation-effects - (Htb (Lst CSymbol)) (dom - Domain)
   (val-or-initialize
       (!_causation-effects dom)
       :init (make-eq-hash-table)))

(deffunc place-pred-causers - (Htb (Lrcd (Alt (Lst Causation-defn)
					      (Const :unknown))
					 (Alt (Lst Occasion)
					      (Const :unknown))))
	      (dom - Domain)
   (val-or-initialize
       (!_pred-causers dom)
       :init (make-eq-hash-table)))


;;;;(defun sit-ini (name dom)
;;;;   (let ((sit (opt::find-domain-bdg-val name dom)))
;;;;      (cond (sit
;;;;	     (opt::situation-initialize sit))
;;;;	    (t "No such situation"))))
;;;;

(deffunc env-find-domain - (~ Domain) (env - Typenv)
   (repeat :for ((x :in (!_bdgs env)))
    :result false
    :until (is Domain x)
    :result (be Domain x)))

;; Conveniences
(defun dom-prep (name)
   (let ((dom (try-domain-with-name name false)))
      (cond (dom
	     (!= dom (domain-prepare-for-planning dom))
	     (domain-make-current dom)
	     dom)
	    (t '"No such domain"))))

(deffunc prob-prep - Plan-problem (x)
   (let ((pp
           (cond ((is Symbol x)
                  (let ((pp (one-value (get-global-opt-sym x))))
                     (cond ((or (is Plan-problem pp)
                                (is Uninitialized-plan-problem pp))
                            pp)
                           (t
                            (error-break coerce-to-plan-problem :fatal
                               "Not the name of a plan-problem: " x)))))
	         ((or (is Plan-problem x) (is Uninitialized-plan-problem x))
                  x)
                 (t
                  (error-break coerce-to-plan-problem :fatal
                     "Not a plan-problem: " x))))
         - Uninitialized-plan-problem)
      (let ((sp (!_sit pp)))
         (cond (sp
                (let ((dom (!_domain sp)))
                   (cond ((not (is Domain dom))
                          (!= dom (domain-prepare-for-planning dom))))
		   (domain-make-current dom)
;;;;                   (cond ((and (!_act-dif-tab dom)
;;;;		               (exists (a in (!_ancestors dom))
;;;;			          (> (!_generation a)
;;;;			             (!_act-dif-time dom))))
;;;;                          (!= (!_achievability dom) '())
;;;;                          (!= (!_act-dif-tab dom) false)))
		   (let ((pprob (opt::problem-initialize pp)))
		      ;;(fire-up-processes sp dom)
		      pprob)))
	       (t
		(signal-problem plan-problem-initialize
		   "Plan problem has no initial situation: "
		   pp))))))