;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: execute-nocmplr.lisp,v 1.1 2005/02/24 04:55:54 dvm Exp $

(depends-on %module/ nisp)

;;; Functions for executing and projecting action expansions
(depends-on :at-run-time %optop/ types projutil)

(depends-on :at-compile-time %optop/ almost)

;;; The code from forceproj.lisp has been turned inside out (oops, I mean 
;;; "refactored") and subsumed under execution in :forced mode.

;;; Questions: Which comes first when you're executing both an expansion
;;; and a reduction tree?  (The reduction tree, because it can achieve goals
;;; the expansion set up.)

(defnisptype Exec-state :forward)

;;; :find-open-subgoals is the mode used in 'make-or-edit-rm-graph' to find
;;; gaps that an rm-graph can fill.

(defnisptype Exec-mode (Const :real :projected :forced :one-step :zero-step))
;;; :one-step and :zero-step are not forced.  See below.

(defnisptype Exec-status (Const :finished :continuing :ready
				:blocked :stalled :failed))
;;; :failed -- Infeasible step attempted
;;; :continuing -- Okay so far, go on to next step (= Projection-result :okay)
;;; :ready -- In :zero-step mode, means thread reached point where the next
;;;               thing is to take an action.
;;; :blocked -- All threads blocked waiting for something to change 
;;; :stalled -- Exceeded resource allotment
;;; :finished -- Done, quiescent, hooray

(defnisptype Thread-cont (Fun (Mlv Exec-status Exec-state)
			      (Exec-state) t))

;;; These are just for debugging; the Obj's are just going to be printed.
(defnisptype Thread-descrip (Structure this-one - Obj stack - (Lst Obj)))

(defnisptype Thread (Structure continuation - Thread-cont
			       descrip - Thread-descrip))

(deffunc new-Thread (cont this-one stack)
   (make Thread cont (make Thread-descrip this-one stack)))

;;; The type of !_data depends on !_kind, which is either :open-subgoals,
;;; :link, or :ready, or ...?
;;; ':ready' means not really blocked, but don't want to proceed.
;;;          -> data = nil
;;; ':open-subgoals' -> data = (subgoal in dnf (Lst (Lst Sexp)), continuation)
;;;        continuation is function that takes list of answers to subgoal,
;;;        a list of Threads (= (!_threads blockage)),
;;;        and returns a list of Threads.  
(defnisptype Gate (nisp:Object kind - Symbol data - Obj))

(defnisptype Blockage (Structure gate - Gate threads - (Lst Thread)))

(defnisptype Link-state (Lst (Lrcd Plan-link Obj)))

;;; The fields of this datatype must never be set, because the
;;; same Exec-state can occur in different branches of the search space.
(defnisptype Exec-state (Structure mode - Exec-mode
				   forcings steps - Integer
				   elapsed - (Alt Time-int (Const *never))
				   ;; -- *never only if mode is :failed?
				   active - (Lst Thread)
				   sit - Situation
				   sit-space - Situation-space
				   hist - (Lst Exec-state)
				    ;; -- In non-:real modes, the situations
				    ;; and execution states
				    ;;; leading up to this one,
				    ;; in rev chrono order.
				   links - Link-state
				   env - (Lst (Lrcd Plan-link Obj))
				   blocked - (Lst Blockage))
;;;;  (start-sit - Situation (es - Exec-state)
;;;;     (car (car (!_hist es))))
  (past-sits - (Lst Situation) (es - Exec-state)
     (<# !_sit (!_hist es))))

;;;;Have to fix opt-eval so it expects an env arg. <- huh?

;;;;How do you evaluate during real execution?
;;;;E.g., need some notion of "the current actual situation."

(defproc continue-execution - (Multvals status - Exec-status
					new-state - Exec-state
					rem-cycles - Integer)
			    (state - Exec-state
			     max-time - Time-int max-steps - Integer 
			     cycles - Integer)
   (repeat :for ((status ':continuing) - Exec-status)
    :until (or (not (eq status ':continuing))
	       (= cycles 0)
	       (> (!_steps state) max-steps)
	       (> (be Time-int (!_elapsed state)) max-time))
    :result (values (cond ((eq status ':continuing)
			   ':stalled)
			  ;; -- This is the *only* way to return a :stalled
			  ;; status.  Threads never return it.
			  (t status))
		    state cycles)
      (!= state (check-subgoal-blockages state))
      (!= < status state >
	  (funcall (!_continuation (car (!_active state)))
		   (almost-copy state
		      (:active (cdr !>active)
		       :hist (cons state !>hist)))))
      (!= cycles (- *-* 1))))

;;; This version does a nondeterministic simulation just one step (situation
;;; change) ahead.  It returns all the possible actions that could be executed
;;; next.
(defproc execute-one-step - (Lst (Lrcd (~ Action) Exec-status Exec-state))
	                  (state - Exec-state)
   (!= state (check-subgoal-blockages state))
   (repeat :for (;;;;(initsit (!_sit state)) - Situation
		 ;; Managed as a stack --
		 (expansion-queue !()) - (Lst Exec-state)
		 :collector results - (Lrcd (~ Action) Exec-status Exec-state))
    :nconc
      (repeat :for ((th :in (!_active state)) - Thread
		    new-status - Exec-status new-state - Exec-state)
	 (!= < new-status new-state >
	     (funcall (!_continuation th)
		      (almost-copy state
			 (:active (remove1q th (!_active state))
			  :hist (cons state !>hist)))))
       :within
         (let ((act (cond ((eq (!_sit new-state)
			       (!_sit state))
			   false)
			  (t (between-action (!_sit state)
					     (!_sit new-state)
					     (!_sit-space state)))))
	       - (~ (Alt Action (Const wait transition))))
	    (cond (act
		   (let ((e (assq act (!_elements results))))
		      (cond (e
			     (cond ((exec-status-better new-status (cadr e))
				    (!= (cadr e) new-status)
				    (!= (caddr e) new-state))))
			    (t
			     (:continue
			         :collect (tuple act new-status new-state))))))
		  ((eq new-status ':finished)
		   (:continue
		       :collect (tuple false ':finished new-state)))
		  ((eq new-state ':continuing)
		   (push new-state expansion-queue)))))
      :until (null expansion-queue)
         (!= state (off-list expansion-queue))
      :result results))

;;; Execute as far as possible without actually doing anything.  This will
;;; expose all the open subgoals and other blockages.
(deffunc exec-state-normalize - Exec-state (es0 - Exec-state)
   (let ()
      (repeat :for ((es (check-subgoal-blockages es0)) - Exec-state
		    ;;;; status - Exec-status
		    )
       :until (null (!_active es))
	 (!= < _ es >
	     (funcall (!_continuation (car (!_active es)))
		      (almost-copy es
			 (:active (cdr !>active)
			  :hist (cons es !>hist)
			  :mode ':zero-step))))
       :result (multi-let (((ready others - (Lst Blockage))

			    (classify (!_blocked es)
				      (\\ (b - Blockage)
					 (eq (!_kind (!_gate b))
					     ':ready)))))
		  (almost-copy es 
		     (:active (append (<$ !_threads ready)
				      (!_active es))
		      :blocked others))))))

(deffunc exec-status-better - Boolean (stat1 stat2 - Exec-status)
   (memq stat1
	 (cdr (memq stat2 '(:failed :blocked :continuing :finished)))))

(deffunc between-action - Action
	      (sit1 sit2 - Situation sit-space - Situation-space)
   (repeat :for ((ss :in (!_succs sit1 sit-space)))
    :result (progn
	       (dbg-save :run-loud sit1 sit2)
	       (signal-problem between-action
		  "Can't find action leading from " sit1 " to " sit2))
    :until (and (eq (!_sit ss) sit2)
		(is Action (!_act ss)))
    :result (be Action (!_act ss))))

(specdecl (check-id* (new-Varid)) - Varid)

(deffunc check-subgoal-blockages - Exec-state (state - Exec-state)
   (let ((sit (!_sit state)))
      (repeat :for ((b :in (!_blocked state)) - Blockage
		    (dom (!_domain (find-init-situation sit))) - Domain
		    :collector keep - Blockage free - Thread)
       :within
	 (cond ((eq (!_kind (!_gate b)) ':open-subgoals)
		(decl ((d (!_data (!_gate b)))
		       - (Lrcd Sexp (Fun (Lst Thread) ((Lst Bdgenv) (Lst Thread)))))
		   (let ((al (deduce `(or ,@(<# (\\ (c) `(and ,@c))
						(car d)))
				     check-id* (empty-env) sit dom)))
		      (cond ((null al)
			     (:continue
			      :collect (:into keep b)))
			    (t
			     ;; Notify the thread that 
			     (:continue
			      :nconc (:into free
					    (funcall (cadr d)
						     al (!_threads b)))))))))
	       (t
		(:continue
		 :collect (:into keep b))))
       :result (cond ((null free)
		      state)
		     (t
		      (almost-copy state
			 (:active (nconc free !>active)
			  :blocked keep)))))))


(specdecl (plan-id* (new-Varid)) - Varid)

(defnisptype Projection-result (Alt (Const :failed)
				    ;; Looking for open subgoals,
				    ;; but they're not here --
				    (Const :ready)
				    (Lstructure &flag :okay
				       sit - Situation
				       elapsed - Time-int
				       numsteps - Integer)
				    (Lstructure &flag :forced
				       num-forcings - Integer
				       sit - Situation
				       elapsed - Time-int
				       numsteps - Integer)
				    (Lstructure &flag :open-subgoals
				       dnf - (Lst (Lst Sexp)))))

(deffunc cached-thread-result - (Mlv (~ Exec-status) Exec-state Action)
	                      (es - Exec-state act - Sexp newth - Thread dom - Domain)
   (let ((uact (uniquify-action act dom)))
      (let ((pr (recover-cached-transition
		    (!_mode es) (!_sit es) uact (!_sit-space es))))
	 (cond (pr
		(multi-let (((status new-es)
			     (projected-thread-result pr es newth)))
		   (values status new-es uact)))
	       (t
		(values false es uact))))))

;;; If result is false, that means there is no useful cached transition.
(deffunc recover-cached-transition - (~ Projection-result)
				   (mode - Exec-mode
				    sit - Situation act - Action
				    sit-space - Situation-space)
   (let ((sss (find-old-projection sit act sit-space))
	 - (~ Sitsuccspec))
      (cond (sss
	     (with sss
		(cond ((null !>open-subgoals)
		       (cond (!>sit
			      (cond ((= !>forcings 0)
				     (tuple (atomconst :okay)
					    !>sit
					    (be Time-int !>elapsed)
					    !>steps))
				    ((eq mode ':forced)
				     (tuple (atomconst :forced)
					    !>forcings
					     !>sit
					     (be Time-int !>elapsed)
					     !>steps))
				    (t ':failed)))
			     (t ':failed)))
		      (t
		       (cond (!>sit
			      (cond ((= !>forcings 0)
				     (signal-problem recover-cached-transition
					"Unforced execution of " act
					" in spite of open subgoals "
					:% " [to wit -- "
					!>open-subgoals "]"))
				    ((eq mode ':projected)
				     ':failed)
				    (t
				     (tuple (atomconst :forced)
					    !>forcings
					    !>sit
					    (be Time-int !>elapsed)
					    !>steps))))
			     (t false))))))
	    (t false))))

(deffunc projected-thread-result - (Mlv Exec-status Exec-state)
                             (pr - Projection-result es - Exec-state newth - Thread)
   (with es
      (match-cond pr
	 (:? :failed
	    (values ':failed es))
	 (:? ?(:\| (?(:& :okay ?stat)
		    ?new-sit ?elapsed ?numsteps)
		   (?(:& :forced ?stat)
		    ?num-forcings ?new-sit ?elapsed ?numsteps))
	    (values ':continuing
	      (make-inst Exec-state
		 :mode !>mode
		 :forcings (cond ((eq stat ':forced)
				  (+ !>forcings num-forcings))
				 (t !>forcings))
		 :steps (+ !>steps numsteps)
		 :elapsed (be Time-int
			      (+ (be Time-int !>elapsed)
				 elapsed))
		 :active (cons newth !>active)
		 :sit (be Situation new-sit)
		 :sit-space !>sit-space
		 :hist (cons es !>hist)
		 :links !>links
		 :env !>env
		 :blocked !>blocked)))
	 (:? ?(:\| :ready
		   (:open-subgoals ?dnf))
	    (values (cond ((null !>active) ':blocked)
			  (t ':continuing))
	       (make-inst Exec-state
		  :mode !>mode :forcings !>forcings 
		  :steps !>steps :elapsed !>elapsed
		  :active !>active
		  :sit !>sit
		  :sit-space !>sit-space
		  :hist (cons es !>hist)
		  :links !>links
		  :env !>env
		  :blocked 
		      (cons (make Blockage
			       (make Gate ':open-subgoals dnf)
			       (list newth))
			    !>blocked))))
	 (:? :failed
	    (values ':failed es))
	 (t
	  (signal-problem projected-thread-result
	     "Unintelligible projection result " pr)))))
   
;;; THIS SHOULD BE CALLED BY force-project

(defproc forced-action-debug - Void (act - Action
				     goal - (~ Sexp) goal-achieved - Boolean
				     forced - Boolean
				     steps - Integer
				     elapsed - (Alt Time-int (Const *never)))
   (out (:to *error-output*)
	    (:q (forced  
		 "Action " act :% " infeasible" :% " FORCING" :%))
	    (:q (goal
		 (:q (goal-achieved
		      "Okay, goal now satisfied")
		     (t
		      "Goal not satisfied"))))
	    1
	    (:q ((not (eql elapsed 0.0))
		 "Elapsed = " elapsed 1))
	    (:q ((not (eql steps 0))
		 "Steps = " steps 1))
	    :%))