;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: fluregress-exc.lisp,v 1.1 2006/01/28 14:30:46 dvm Exp $

(depends-on %module/ nisp)

(depends-on :at-run-time %optop/ types domain cause-intern)

(defnisptype Embryo-regress-method
    (Structure ()
       change - (Const :add :del)
       inductive - Boolean
       ;; If non-false, the regress method requires you to
       ;; wait until !_quant becomes zero (starting negative).
       ;; This makes no sense unless the sign of the change
       ;; is the same as the sign of the occurrence of the
       ;; changee in the quantity to be increased --
       quant - (~ Sexp)
       id - Varid bdgs - Bdgenv
       precond secondary - Sexp
       view - (~ (Alt Occasion Internalized-causation-defn))
       act-sort - (~ (Const *wait *watch *act)))
;;; Important: The quant's freevars, if any, are wrt lit-id*.
;;; The precond and secondary are wrt effect-id*
;;; 'finish-regress-method' must translate these to new Varid's.
   (conser (change - (Const :add :del)
	    inductive - Boolean
	    quant - (~ Sexp)
	    id - Varid bdgs - Bdgenv
	    precond secondary - Sexp)
      (make ^^ change inductive quant id bdgs precond secondary
	       false false)))

;;;;(specdecl (rs-id* (new-Varid)) - Varid)

;;;;(specdecl (test1* #'car) - (Fun Sexp (Sexp))
;;;;	  test2* - Symbol)

(specdecl (lit-id* (new-Varid)) - Varid)

;; Always use same ids internally to link !_literal of Regress-spec
;; to !_precond of Regress-method

(deffunc fluent-regression-methods - (Lst Regress-method)
	                (polarity - Boolean lit - Sexp th - Domain)
   (cond ((memq (car lit) ineq-rels*)
	  (multi-let (((ineq quant)
		       (inequality-normal-form polarity lit)))
	     (nconc (<! (\\ (erm)
			   (finish-regress-method
			       erm `(,ineq ,quant 0) th))
			(elapsed-regression-methods ineq quant th))
		    (repeat :for ((cdefn :in (direct-causers lit th)))
		     :when (is Causation-defn cdefn)
		     :nconc
		        (decl (cdefn - Causation-defn)
			   (<! finish-it
			       (causation-defn-regress
				   cdefn
				   (\\ - (Lst Embryo-regress-method)
				       (idefn - Internalized-causation-defn
					effect - Sexp)
				     (ineq-regress-unify
					ineq quant effect
					lit-id* effect-id* (empty-env)
					(!_condition idefn)
					'(and) th))
				   th)))
		     :where
			(finish-it (erm - Embryo-regress-method)
			   (finish-regress-method
			        erm
				`(,ineq ,quant 0) th))))))
	 ;; Eventually we will have the ability to handle things like
	 ;; (null (contents bucket17))
	 (t !())))

(deffunc inequality-normal-form - (Multvals ineq - (Const > >= =~ /=~)
					    lhs - Sexp)
                                (polarity - Boolean
				 e - Sexp)
   (let (r lhs rhs)
      (cond ((matchq (?r ?lhs ?rhs) e)
	     (cond ((not polarity)
		    (!= r (cadr (assq r '((> =<) (>= <) (< >=) (=< >)
					  (=~ /=~) (/=~ =~)))))))
	     (cond ((memq r '(=< <))
		    (!= lhs (prog1 rhs (!= rhs lhs)))
		    (!= r (cond ((eq r '=<) '>=)
				(t '>)))))
	     (cond ((not (eql rhs '0))
		    (!= lhs `(- ,lhs ,rhs))))
	     (values r lhs))
	    (t
	     (error-break inequality-normal-form :fatal
		"Ill-formed inequality " e)))))

;;;;(specdecl (elproc-id* (new-Varid))
;;;;	  (elquant-id* (new-Varid))
;;;;	  - Varid)

(deffunc elapsed-regression-methods - (Lst Embryo-regress-method)
	                 (ineq - (Const > >= =~ /=~) quant - Sexp
			  dom - Domain)
			 (ignore ineq)
   (repeat :for ((env :in (term-occurrences '(elapsed ?proc)
					    quant effect-id* lit-id*
					    true (empty-env))))
    :nconc
       (multi-let (((intel coeff _)
		    (occurrence-analyze '(elapsed ?proc) quant
					effect-id* lit-id* env)))
	  (cond (intel
		 (multi-let (((ok elcar elcdr _)
			      (opt::expression-car-cdr
				  '?proc effect-id* env false)))
		    (cond (ok
			   (process-elapser-methods elcar elcdr quant coeff
						    env dom))
			  (t 
			   (signal-problem elapsed-regression-methods
			      "Illegal term as arg in "
			      (safe-varsubst '?proc effect-id* env)
			      (:continue "I will ignore it"))
			   !()))))
		(t !())))))

(deffunc process-elapser-methods - (Lst Embryo-regress-method)
	              (elcar - Symbol elcdr quant coeff - Sexp
		       env - Bdgenv dom - Domain)
   (multi-let (((same-case opp-case)
		(same-sign-analysis coeff '(fl-^ 1.0)))
	       (intproc
		(causation-view 
		     (name-causation-defn 'dur-happening elcar elcdr dom)
		     #'durative-process-view
		     dom)))
      (multi-let (((ok env)
		   (unify `(,elcar ,@elcdr) (!_term intproc)
			  effect-id* effect-id* env)))
	 (cond (ok
		(repeat :for ((erm :in (nconc (ineq-deriv-regress-method
						 ':add quant lit-id* env
						 (!_condition intproc)
						 '(and) same-case)
					      (ineq-deriv-regress-method
						 ':del quant lit-id* env
						 (!_condition intproc)
						 '(and) opp-case))))
		   (!= (!_view erm) intproc)
		 :collect erm))
	       (t
		(signal-problem process-elapser-methods
		   "Couldn't unify " `(,elcar ,@elcdr)
		   :% " and " (!_term intproc)))))))

(deffunc causation-defn-regress - (Lst Embryo-regress-method)
            (cdefn - Causation-defn
	     regressor - (Fun (Lst Embryo-regress-method)
			      (Internalized-causation-defn Sexp))
	     th - Domain)
   (let ((idef (causation-defn-projectable cdefn th)))
      (cond ((is Durative-action-defn cdefn)
	     (decl (idef - Internalized-durative-defn)
		(<!         ;;;; (\\ (view)
		            ;;;;    (let ((rid (new-Varid)))
			    ;;;;       (<# (\\ (erm)
			    ;;;;              (tuple rid erm))
		            ;;;;           (funcall regressor view rid))))
		    (\\ (view)
		       (cond (view
			      (causation-regressor-apply view regressor))
			     (t !())))
		    (list (!_process-view idef)
			  (!_start-view idef)
			  (!_stop-view idef)))))
	    (t
	     (causation-regressor-apply idef regressor)))))

(deffunc causation-regressor-apply - (Lst Embryo-regress-method)
                           (idefn - Internalized-causation-defn
			    regressor
			      - (Fun (Lst Embryo-regress-method)
				     (Internalized-causation-defn
				      Sexp)))
   (let-fun ()
      (cond ((is Internalized-process-defn idefn)
	     (with idefn
		(nconc
		   (add-view (funcall regressor idefn !>start-effect)
			     '*watch)
		   (add-view (funcall regressor idefn !>effect)
			     '*wait)
		   (add-view (funcall regressor idefn !>stop-effect)
			     '*wait))))
	    (t
	     (add-view (funcall regressor idefn (!_effect idefn))
		       '*act)))
    :where
       (add-view - (Lst Embryo-regress-method)
		 (erml - (Lst Embryo-regress-method)
		  act-sort - (Const *wait *watch *act))
;;;;	  (cond ((and (not (null erml))
;;;;		      (eq act-sort '*watch))
;;;;		 (dbg-save idefn erml)
;;;;		 (breakpoint causation-regressor-apply
;;;;		    "erml = " erml)))
	  (repeat :for ((erm :in erml))
             (!= (!_view erm) idefn)
	     (!= (!_act-sort erm) act-sort))
	  erml)))

;;; We're looking for ways to make 'quant' increase toward 0 (:add) or
;;; decrease toward 0 (:del).
(deffunc ineq-regress-unify - (Lst Embryo-regress-method)
                       (ineq - (Const > >= =~ /=~)
		        quant - Sexp
			effects - Sexp
			quant-id eff-id - Varid env - Bdgenv
                        precond secondary - Sexp dom - Domain)
   (cond ((or (is Qvar effects) (atom effects))
	  (leaf-ineq-regress-unify ineq quant true effects
				   quant-id eff-id env
				   precond secondary dom))
	 (t
	  (case (car effects)
	     (and (repeat :for (e :in (cdr effects))
		   :nconc 
		     (ineq-regress-unify
				  ineq quant e
				  quant-id eff-id env
				  precond secondary dom)))
	     ((when if)
	      (ineq-regress-unify
	         ineq quant (caddr effects)
		 quant-id eff-id env
		 precond `(and ,(if-when-precond
				   (car effects) (cadr effects))
			       ,secondary)
		 dom))
	     (forall
	      (ineq-regress-unify
	         ineq quant (caddr effects)
		 quant-id eff-id env precond secondary dom))
	     (not
	      (leaf-ineq-regress-unify
	         ineq quant false (cadr effects)
		 quant-id eff-id env precond secondary dom))
	     (t
	      (leaf-ineq-regress-unify
	         ineq quant true effects
		 quant-id eff-id env precond secondary dom))))))

(deffunc leaf-ineq-regress-unify - (Lst Embryo-regress-method)
                            (ineq - (Const > >= =~ /=~)
			     quant - Sexp
                             leaf-polarity - Boolean leaf - Sexp
			     quant-id eff-id - Varid env - Bdgenv
                             precond secondary - Sexp dom - Domain)
   (multi-let (((leaf env)
		(safe-varsubst leaf eff-id env)))
      (cond ((is Qvar leaf)
	     (cond (leaf-polarity
		    (ineq-try-reduce-precond
		       ineq quant leaf quant-id eff-id env
		       precond secondary dom))
		   (t
		    (out (:to *error-output*)
			 "Can't handle negated variable effect: " leaf :%)
		    '())))
	    ((atom leaf) '())
	    (leaf-polarity
	     (fluent-regressions ineq quant leaf
				 quant-id eff-id env
				 precond secondary dom))
	    (t
;;;;	     (out (to *error-output*)
;;;;		"Warning: Useless leaf in causation definition: "
;;;;		`(not ,leaf) :%)
	     !(Embryo-regress-method)))))

;; For each conjunctive path through the preconds, 
;; find a precond that can be reduced in fact-sit of dom, thus binding
;; leaf (a Qvar).
(deffunc ineq-try-reduce-precond - (Lst Embryo-regress-method)
                            (ineq - (Const > >= =~ /=~)
			     quant - Sexp
			     leaf - Qvar
			     quant-id eff-id - Varid env - Bdgenv
			     precond secondary - Sexp dom - Domain)
   (let ((dnf-pre (disjunctive-normal-form
		     precond true))
	 (dnf-sec (disjunctive-normal-form
		     secondary true)))
      (nconc (<! (\\ (pconj)
		    (let ((rl (reduce-to-rebind
				 pconj leaf eff-id eff-id
				 env dom)))
		       (<! (\\ (r)
			      (ineq-regress-unify 
				 ineq quant (cadr r)
				 quant-id eff-id (caddr r)
				 `(and ,@(car r)) secondary dom))
			   rl)))
		 dnf-pre)
	     (<! (\\ (sconj)
		    (let ((rl (reduce-to-rebind
				 sconj leaf eff-id eff-id
				 env dom)))
		       (<! (\\ (r)
			      (ineq-regress-unify 
				 ineq quant (cadr r)
				 quant-id eff-id (caddr r)
				 precond `(and ,@(car r)) dom))
			   rl)))
		 dnf-sec))))

(deffunc fluent-regressions - (Lst Embryo-regress-method)
                            (ineq - (Const > >= =~ /=~)
			     quant - Sexp
                             leaf - Sexp
			     quant-id eff-id - Varid env - Bdgenv
			     primary secondary - Sexp
			     dom - Domain)
   (multi-let (((leaf-fn leaf-args env)
		(opt::e-car-cdr leaf eff-id env)))
;;;;      (let ((fmla `(fluent-regress-handler
;;;;		      ,ineq ,leaf-fn ?fn-name)))
;;;;	 (dbg-save fmla dom)
;;;;	 (breakpoint fluent-regressions
;;;;	    "Ready to infer " fmla) ...)

	 (<! (\\ (ans - (Lst Varbdg))
		(let ((fn-name (get-var-val-1 'fn-name ans)))
;;;;		   (out "Fluent regressor: " fn-name :%)
		   (funcall (symbol-function fn-name)
			    ineq quant leaf-fn leaf-args
			    quant-id eff-id env
			    primary secondary dom)))
	     ;;; We deduce it because in some domains there are
	     ;;; cases besides <, >, et al.
	     (deduce `(fluent-regress-handler
			     ,ineq ,leaf-fn ?fn-name)
		     dummy-id* !() 
		     (domain-find-init-sit dom)
		     dom))))

;;; Assert facts like (fluent-regress-handler < assign ineq-assign-regress)
;;; The freevars in 'primary' and 'secondary' args to these handlers
;;; are wrt Varid 'flu-id' (same as leaf-fn and leaf-args).

;;; Important: These handlers intentionally do not return exactly the
;;; right answer, i.e., a sufficient condition for the Causation to
;;; cause the inequality to become true.  That's because they don't
;;; bother regressing the actual inequality through the 'assign',
;;; 'derivative', or whatever.  Instead, they just mark the resulting
;;; Embryo-regress-method as inductive.

(deffunc ineq-inc-dec-regress - (Lst Embryo-regress-method)
	                  (ineq - (Const > >= =~ /=~)
			   quant - Sexp
			   leaf-fn - (Const increase decrease)
			   leaf-args - (Lst Sexp)
			   quant-id flu-id - Varid env - Bdgenv
			   primary secondary - Sexp dom - Domain)
   (match-let (?flu ?delta) leaf-args
      (ineq-assign-regress
         ineq quant leaf-fn
	 `(,flu (,(cond ((eq leaf-fn 'decrease) '-)
			 (t '+))
		 (fl-v ,flu)
		 ,delta))
	 quant-id flu-id env primary secondary dom)))

(deffunc ineq-assign-regress - (Lst Embryo-regress-method)
	                  (ineq - (Const > >= =~ /=~)
			   quant - Sexp
			   _ - Symbol leaf-args - (Lst Sexp)
			   quant-id flu-id - Varid env - Bdgenv
			   primary secondary - Sexp dom - Domain)
   (match-let (?flu ?newval) leaf-args
      (multi-let (((leaf-intel leaf-coeff leaf-k)
		   (occurrence-analyze flu newval flu-id flu-id env)))
	 (cond ((and leaf-intel
		     (is Number leaf-coeff)
		     (or (not (= leaf-coeff 0))
			 (not (and (is Number leaf-k)
				   (= leaf-k 0)))))
		(repeat :for ((env :in (term-occurrences
					   flu quant flu-id quant-id
					   true env)))
		 :nconc (cond ((= leaf-coeff 0)
			       (direc-assign-fluent-regress
				      ineq quant flu newval 'to-constant
				      quant-id flu-id env
				      primary secondary dom))
			      (t
			       (multi-let (((intel coeff _)
					    (occurrence-analyze
						flu quant
						flu-id quant-id env)))
				 ;; direction is the direction this leaf will
				 ;; push quant.
				 (let ((direction
					  (cond (intel
						 (quant-flu-direc-analyze
						    coeff leaf-coeff leaf-k))
						(t false)))
				       - (Const up down no-op to-constant
						messy nil))
				    (cond ((or (not direction)
					       (eq direction 'no-op))
					   !())
					  (t
					   (decl (direction
						    - (Const up down
							     to-constant
							     messy))
					      (direc-assign-fluent-regress
						   ineq quant flu newval
						   direction
						   quant-id flu-id env
						   primary secondary
						   dom))))))))))
	       (t !())))))

;;; coeff is from a single occurrence of the changing fluent
;;; in 'quant'.  Return the net impact this change will have on 'quant',
;;; ignoring any other occurrences of the fluent in 'quant'.
(deffunc quant-flu-direc-analyze - (Const up down no-op messy)
	                         (coeff - Sexp
				  leaf-coeff - Number leaf-k - Sexp)
   (cond ((is Number coeff)
	  (cond ((= coeff 0)
		 (signal-problem ineq-assign-regress
		    "fluent missing in quant"))
		;; At this point coeff and leaf-coeff are both
		;; non-zero numbers
		((= leaf-coeff 1)
		 (cond ((is Number leaf-k)
			(cond ((= leaf-k 0)
			       'no-op)
			      ((same-sign leaf-k coeff)
			       'up)
			      (t 'down)))
		       (t 'messy)))
		(t 'messy)))
	 (t 'messy)))

(deffunc direc-assign-fluent-regress - (Lst Embryo-regress-method)
	                 (ineq - (Const > >= =~ /=~)
			  lhs flu newval - Sexp
			  direction - (Const up down to-constant messy)
			  lhs-id flu-id - Varid env - Bdgenv
			  primary secondary - Sexp dom - Domain)
      (let ((certain-outcome
	       (cond ((eq direction 'to-constant)
		      (test-ineq ineq lhs flu newval lhs-id flu-id env dom))
		     (t false)))
	    - (Const true false nil))
	 (cond (certain-outcome
		(list (make Embryo-regress-method
			 (be (Const :add :del)
			     (cond ((eq certain-outcome
					'true)
				    ':add)
				   (t ':del)))
			 false lhs flu-id env
			 primary secondary)))
	       (t
		(decl (direction - (Const up down messy))
		   (nconc (fluent-regress-add
			     ineq lhs
			     direction flu-id env
			     primary secondary)
			  (fluent-regress-del
			     ineq lhs 
			     direction flu-id env
			     primary secondary)))))))

;;; Test whether the inequality `(,ineq ,lhs 0) is guaranteed to
;;; be true when 'newval' is substituted for 'flu' in 'lhs',
;;; *in any situation* in the space associated with 'dom'. 
(deffunc test-ineq - (Const true false nil) (ineq - (Const > >= =~ /=~)
					     lhs - Sexp
					     flu newval - Sexp
					     lhs-id flu-id - Varid
					     env - Bdgenv
					     dom - Domain)
   (multi-let (((new-lhs _)
		(fluent-subst newval flu lhs flu-id lhs-id env)))
      (cond ((has-qvars new-lhs) false)
	    (t
	     (multi-let (((_ expval)
			  (opt-eval new-lhs
				    (domain-find-init-sit dom)
				    !()
				    dom)))
		(cond ((case ineq
			  (>
			   (> expval 0))
			  (>=
			   (>= expval 0))
			  (=
			   (= expval 0))
			  (/=
			   (not (= expval 0))))
		       'true)
		      (t 'false)))))))

;;; 'newnorm' is new value of lhs of (> fluent 0)
(deffunc fluent-regress-add - (Lst Embryo-regress-method)
                            (ineq - (Const > >= =~ /=~)
			     quant - Sexp
			     direction - (Const up down messy)
			     precond-id - Varid env - Bdgenv
			     primary secondary - Sexp)
   (cond ((and (eq direction 'down)
	       (memq ineq '(> >=)))
	  ; This rule can never cause the inequality to become true.
	  '())
	 (t
	  (list (make Embryo-regress-method
		      ':add true quant precond-id env 
		      primary secondary)))))

(deffunc fluent-regress-del - (Lst Embryo-regress-method)
                            (ineq - (Const > >= =~ /=~)
			     quant - Sexp
			     direction - (Const up down messy)
			     precond-id - Varid env - Bdgenv
			     primary secondary - Sexp)
   (cond ((and (eq direction 'up)
	       (memq ineq '(> >=)))
	  ; This rule can never cause the inequality to become false
	  '())
	 (t
	  (list (make Embryo-regress-method
		      ':del true quant precond-id env
		      primary secondary)))))

(deffunc ineq-derivative-regress - (Lst Embryo-regress-method)
	                  (_ - (Const > >= =~ /=~)
			   quant - Sexp
			   _ - (Const derivative)
			   leaf-args - (Lst Sexp)
			   quant-id flu-id - Varid env - Bdgenv
			   primary secondary - Sexp _ - Domain)
   (match-let (?flu ?deriv) leaf-args
      (repeat :for ((env :in (term-occurrences flu quant flu-id quant-id
					       true env)))
       :nconc (multi-let (((intel coeff _)
			   (occurrence-analyze flu quant flu-id quant-id
					       env)))
		 (cond (intel
			(multi-let (((same-case opp-case)
				     (same-sign-analysis coeff `(fl-^ ,deriv))))
			   (nconc (ineq-deriv-regress-method
				     ':add quant flu-id env
				     primary secondary same-case)
				  (ineq-deriv-regress-method
				     ':del quant flu-id env
				     primary secondary opp-case))))
		       (t !()))))))

;;; Returns < same-case, opp-case >,  
(deffunc same-sign-analysis - (Multvals same opp
					- (Alt (Const :triv-true :triv-false)
					       Sexp))
                         (coeff deriv - Sexp)
      (cond ((eq (strip-car deriv)
		 'fl-^)
	     (let ((dval (cadr deriv)))
		(cond ((and (is Number dval)
			    (is Number coeff))
		       (cond ((= dval 0)
			      (values ':triv-false ':triv-false))
			     ((eq (< dval 0) (< coeff 0))
			      (values ':triv-true ':triv-false))
			     (t
			      (values ':triv-false ':triv-true))))
		      (t
		       (values `(same-sign ,coeff ,dval)
			       `(opp-sign ,coeff ,dval))))))
	    (t
	     (values `(same-sign ,coeff (fl-v ,deriv))
		     `(opp-sign ,coeff (fl-v ,deriv))))))

(deffunc ineq-deriv-regress-method - (Lst Embryo-regress-method)
                        (change - (Const :add :del)
			 quant - Sexp flu-id - Varid env - Bdgenv
			 primary secondary - Sexp
			 sign-anal - (Alt (Const :triv-true :triv-false)
					  Sexp))
   (cond ((eq sign-anal ':triv-false)
	  !())
	 (t
	  (list (make Embryo-regress-method
		   change true quant flu-id env
		   primary
		   (cond ((eq sign-anal ':triv-true)
			  secondary)
			 (t
			  `(and ,secondary ,sign-anal))))))))

;;; Substitute newval for flu in lhs.  Because ids are different,
;;; we have to be careful in the usual way.
(deffunc fluent-subst - (Multvals new-lhs - Sexp new-env - Bdgenv)
                      (newval flu lhs - Sexp 
		       flu-id lhs-id - Varid env - Bdgenv)
   (cond ((atom lhs)
	  (values lhs env))
	 (t
	  (let ((fcn (car lhs)))
	     (strip-sym-with-type fcn)
	     (cond ((and (eq fcn 'fl-v)
			 (variants flu (cadr lhs) flu-id lhs-id env))
		    (opt::varid-convert newval flu-id lhs-id env))
		   (t
		    (repeat :for ((a :in (cdr lhs)) new-a - Sexp
				  :collectors res - Sexp)
		       (!= < new-a env >
			   (fluent-subst newval flu a flu-id lhs-id env))
		     :collect new-a
		     :result (values `(,(car lhs) ,@res) env))))))))

;;; Try to see lhs as (+ (* coeff flu) offset); if so,
;;; 'intelligible' = true.
;;; If not 'intelligible', then coeff and offset are meaningless.
;;; If coeff and offset are numbers, then the thing is intelligible
;;; indeed.  If they contain qvars, they are wrt lhs-id

(deffunc occurrence-analyze - (Multvals 
				  intelligible - Boolean
				  coeff offset - Sexp)
                      (flu lhs - Sexp
		       flu-id lhs-id - Varid env - Bdgenv)
   (cond ((and (eq (strip-car lhs) 'fl-v)
	       (variants flu (cadr lhs) flu-id lhs-id env))
	  (values true 1 0))
	 ((atom lhs)
	  (values true 0 lhs))
	 ((memq (car lhs) '(+ -))
	  (repeat :for ((x :in (cdr lhs)) - Sexp
			(add-fcn-name '+
				 :then (be (Const + -) (car lhs)))
			- (Const + -)
			(add-fcn #'+
				 :then (cond ((eq (car lhs) '-) #'-)
					     (t #'+)))
			- (Fun Number Number)
			(coeff 0) (offset 0)
			- Sexp
			(all-intel true) - Boolean)
	      (multi-let (((xintel xcoeff xoffset)
			   (occurrence-analyze
			      flu x flu-id lhs-id env)))
		 (cond (xintel
			(!= coeff (maybe-add *-* xcoeff))
			(!= offset (maybe-add *-* xoffset)))
		       (t
			(!= all-intel false))))
	    :result (cond (all-intel
			   (values true coeff offset))
			  (t
			   (values false nil nil)))
	    :where
	       (maybe-add (e1 e2)
		  (cond ((is-Number e1)
			 (cond ((is Number e2)
				(funcall add-fcn e1 e2))
			       ((= e1 0)
				(cond ((eq add-fcn-name '+) e2)
				      (t `(- ,e2))))
			       (t `(,add-fcn-name ,e1 ,e2))))
			((is-Number e2)
			 (cond ((= e2 0)
				e1)
			       (t
				`(,add-fcn-name ,e1 ,e2))))
			(t
			 `(,add-fcn-name ,e1 ,e2))))))
	 ((memq (car lhs) '(* /))
	  (repeat :for ((x :in (cdr lhs)) - Sexp
		        xintel - Boolean xcoeff xoffset - Sexp
			(hairy-coeff false) - (~ Sexp)
			hc-offset - Sexp
			(prod 1) - Number
			(mult-fcn-name '*
				 :then (be (Const * /) (car lhs)))
			- (Const * /)
			(mult-fcn #'*
				 :then (cond ((eq (car lhs) '/) #'/)
					     (t #'*)))
			- (Fun Number (Number Number))
			:collectors

			arg-analyses
			  - (Lstructure () intellible - Boolean
					   coeff offset - Sexp))
	   :result (cond (hairy-coeff
			  (values true
				  (maybe-mult prod hairy-coeff)
				  (maybe-mult prod hc-offset)))
			 (t
			  (values true 0 prod)))
	     (!= < xintel xcoeff xoffset >
		 (occurrence-analyze flu x flu-id lhs-id env))
	   :while (and xintel
		       (or (not hairy-coeff)
			   (eql xcoeff 0)))
	     (cond ((and (eql xcoeff 0) (is Number xoffset))
		    (!= prod (funcall mult-fcn *-* xoffset)))
		   (t
		    (!= hairy-coeff xcoeff)
		    (!= hc-offset xoffset)))
	   :else :result (values false nil nil)

	   :where
	      (maybe-mult - Sexp (prod - Number exp - Sexp)
		 (cond ((is Number exp)
			(funcall mult-fcn prod exp))
		       (t
			`(,mult-fcn-name ,prod ,exp))))))
	 (t
	  (repeat :for ((a :in (cdr lhs))
			intel coeff)
	   :result (values true 0 lhs)
	     (!= < intel coeff _ >
		 (occurrence-analyze flu a flu-id lhs-id env))
	   :while (and intel (eql coeff 0))
	   :result (values false nil nil)))))

(deffunc ineq-opposite - (Const < =< /= =) (ineq - (Const > >= =~ /=~))
   (be * (cadr (assq ineq '((> <=) (>= <) (=~ /=~) (/=~ =~))))))

;;; For each subterm of 'expressions' that unifies with 'find', 
;;; return a unifying substitution.
;;; If 'must-be-definite' is true, then unifying substitution must
;;; eliminate variables in 'term'.  
(deffunc term-occurrences - (Lst Bdgenv)
			  (find expression - Sexp find-id exp-id - Varid
			   must-be-definite - Boolean
			   env - Bdgenv)
   (let-fun ((search-for-them - (Lst Bdgenv)
			     (exp - Sexp)
	       (multi-let (((ok unenv)
			    (unify find exp find-id exp-id env)))
		  (cond (ok
			 (cond ((and must-be-definite
				     (has-uvars find find-id unenv))
				!())
			       (t
				(list unenv))))
			((atom exp) !())
			(t
			 (repeat :for ((e :in exp))
			  :nconc (search-for-them e)))))))
      (search-for-them expression)))

(deffunc finish-regress-method - (Lst Regress-method)
                               (regression - Embryo-regress-method
				goal - Sexp
			          ;;;; cdefn - Causation-defn
				  ;;;; rid - Varid
				th - Domain)
   (!= regression
       (regression-flush-optimism *-* effect-id*))
   (let ((causer (!_view regression))
	 - (~ (Alt Occasion Internalized-causation-defn)))
      (let ((idefn (cond ((is Internalized-causation-defn causer)
			  causer)
			 (t false)))
	    - (~ Internalized-causation-defn))
	 (let-fun ()
	    (repeat :for ((goals :in (disjunctive-normal-form
					(relevant-preconditions)
					true))
			  - (Lst Sexp)
			  (act-sort (!_act-sort regression))
			  - (Const *wait *watch *act)
			  consistent - Boolean)
	       (cond ((exists (g :in goals)
			 (car-eq g 'exists))
		      (signal-problem finish-regress-method
			 "Existential in regress method: "
			 :% goals
			 (:proceed "I will trudge on"))))
	       ;; 'after-action's are only useful during projection
	       (!= goals
		   (remove-if (\\ (g) (matchq (after-action ?_) g))
			      *-*))
	       (!= < consistent goals >
		   (goal-list-simplify goals th))
	     :when consistent
	     :nconc
	       (let ((goal-id (new-Varid))
		     (precond-id (new-Varid)))
		  (multi-let (((goals env)
			       (varid-convert goals effect-id* precond-id
					      (!_bdgs regression))))
		     (cond ((forall (g :in goals)
			       (not (goal-trivially-false g th)))
			    (subst-and-convert act-sort goals
					       goal-id precond-id env))
			   (t !())))))
	  :where

 (:def subst-and-convert - (Lst Regress-method) 	  
                    (act-sort - (Const *wait *watch *act)
		     goals - (Lst Sexp)
		     goal-id precond-id - Varid env - Bdgenv)
;;;;    (trace-around subst-and-convert
;;;;       (:> "(subst-and-convert: " act-sort "/ <- " goals ")")
    (multi-let (((act-term env)
		 (cond (idefn
			(varid-convert (!_term idefn)
				       effect-id* precond-id env))
		       (t
			(values ':conclude env)))))
       (multi-let (((quant env)
		    (varid-convert (!_quant regression)
				   lit-id* goal-id env)))
	  (multi-let (((goal env)
		       (varid-convert
			  goal lit-id* goal-id env)))
	     (cond ((eq act-sort '*wait)
		    (cond (quant
			   (!= act-term
			       `(wait-non-negative
				   ,(one-value
				       (varid-convert quant
						      goal-id precond-id env))
				   ,act-term)))
			  (t
			   (!= act-term `(wait-while ,act-term)))))
		   ((eq act-sort '*watch)
		    (!= act-term 'watch)))
	     (multi-let (((_ precond)
			  (try-postpone-negs
			     (normalize-step-value-goals
				 goals act-term)
			     (cond ((eq (!_change regression)
					':del)
				    (raw-uvars act-term))
				   (t '())))))
		(cond ((act-uvar-check act-term precond precond-id env)
		       (list (make Regress-method
				(!_change regression)
				goal
				causer
				(!_inductive regression)
				quant
				goal-id precond-id env
				act-term precond)))
		      ((eq (!_change regression) ':add)
		       (err-out
			  "Rejected regression-method that"
			  " doesn't bind all action vars")
		       (dbg-save :run-loud
			  act-term precond precond-id goal-id env regression)
		       (breakpoint finish-regress-method
			  "act-term = " act-term)
		       !())
		      (t !()))))))
;;;;       (:< (val &rest _) "subst-and-convert: " val))
    )

 (:def relevant-preconditions - (Lst Sexp) ()
     (cond ((and idefn
		 (eq (!_change regression)
		     ':del))
	    (!_secondary regression))
	   (t
	    `(and ,(!_precond regression)
		  ,(!_secondary regression)
		  ,@(cond (idefn
			   (!_(Internalized-causation-defn
				 type-filter)
			      idefn))
			  (t !()))))))))))

(deffunc act-uvar-check - Boolean (act precond - Sexp id - Varid env - Bdgenv)
   (let ((auv (opt::all-uvars act id env))
	 (puv (opt::all-uvars precond id env))
	 - (Lst Expclo))
      (forall (ask :in auv)
	 (exists (psk :in puv)
	    (and (= (!_id psk) (!_id ask))
		 (eq (!_(Qvar sym) (!_skel psk))
		     (!_(Qvar sym) (!_skel ask))))))))

;;;;(defproc weird-elapsed-check - Void (erm - Embryo-regress-method)
;;;;   (let ((rq (!_quant erm))
;;;;	 (act (!_term (!_view erm))))
;;;;      (cond ((car-eq act 'dur-happening)
;;;;	     (let-fun ((find-weird-elap - Void (qe - Sexp)
;;;;			 (match-cond qe
;;;;			    ?( (elapsed (dur-happening ?dur))
;;;;			      (cond ((not (variants dur (cadr act)
;;;;						    lit-id* ...)))))
;;;; 


(specdecl nisptype:void-type* - Opt-type)

(deffunc goal-list-simplify - (Mlv Boolean (Lst Sexp))
	                    (goals - (Lst Sexp) dom - Domain)
   (repeat :for ((gl goals) - (Lst Sexp)
		 g0 td0 ob0 td1 - Sexp
		 redundant (consistent true) - Boolean
		 (domenv (domain-place-env dom)) - Typenv
		 :collectors good-goals - Sexp)
    :until (null gl)
    :result (values true good-goals)
       (!= g0 (car gl))
    :until (goal-trivially-false g0 dom)
    :result (values false nil)
       (!= gl (cdr gl))
       (!= redundant false)
       (cond ((matchq (dur-reach ?@_) g0)
	      ;; These occur only as subgoals of 
	      ;; '(wait-while (dur-happening <durative-action>))',
	      ;; and they contribute nothing, so flush them:
	      (!= redundant true))
	     ((matchq (is ?td0 ?ob0) g0)
	      ;;;;(out "Got: " g0 :%)
	      (repeat :for ((g1 :in gl)
			    (ty0 (designated-type td0 true domenv))
			    (redunds !(Sexp)))
	          ;;;;(out "Trying: " g1 :%)
		  (cond ((matchq (is ?td1 ?,ob0) g1)
;;;;			 (and (eq (car g1) 'is)
;;;;			      (equal (caddr g1) ob0))
;;;;			 (!= td1 (cadr g1))
			 ;;;;(out "Oh boy: " g1 :%)
			 (let ((ty1 (designated-type td1 true domenv)))
			    (multi-let (((expr int-ty _)
					 (opt::types-intersect
					     false true ty0 true ty1
					     !() domenv domenv !())))
			       (cond ((and expr (eq int-ty opt::void-type*))
				      (!= consistent false))
				     (t
				      (let ((sub01 
					       (opt::type-acceptable
						  false ty0 ty1 !()
						  domenv domenv !()))
					    (sub10
					       (opt::type-acceptable
						  false ty1 ty0 !()
						  domenv domenv !())))
					 (cond (sub01
;;;;						(out "Redundant: "
;;;;						     g1 :%)
						(!= redunds
						    (cons g1 *-*)))
					       (sub10
						(!= redundant true))))))))))
		:until redundant
		:result (!= gl (<? (\\ (g) (not (memq g redunds)))
				   *-*)))))
    :while consistent
    :result (values false nil)
    :when (not redundant)
    :collect g0))
						       

(deffunc goal-list-non-trivial - Boolean (goals th)
   (repeat :for ((g :in goals))
    :result (progn
	       (out (:to *error-output*)
		  "All goals okay: " goals :%)
	       true)
    :until (goal-trivially-false g th)
    :result (progn
	       (out (:to *error-output*)
		  "Bad goal: " g :%)
	       false)))

;; Returns false if find inconsistency
(deffunc regression-flush-optimism - (~ Embryo-regress-method)
	             (r - Embryo-regress-method
		      rid - Varid)  ; id for vars in (!_precond r).
   (cond ((and optimist-fn*
	       (exists (b in (!_bdgs r))
		  (is Optimist-var (!_varname b))))
	  ;;(out "Flushing optimism from" :% 3 r :%)
	  (multiple-value-let 
		       (opt-list env - (Lst Varbdg))
		       (classify (!_bdgs r)
				 (\\ (b - Varbdg)
				    (is Optimist-var (!_varname b))))
	     (cond ((null opt-list)
		    (signal-problem regression-flush-optimism
		       "Fumbled optimist bdg"
		       :fatal))
		   ((exists (b in opt-list)
			(not (is-step-value-mismatch b)))
		    (signal-problem regression-flush-optimism
		       "Some optimist matches don't involve "
		       "(step-value this-step): " :%
		       (<? neg is-step-value-mismatch opt-list)
		       :fatal)))
	     (let ((opt1val (!_val (car opt-list))))
		(let ((mismatch1 (cadr (!_skel opt1val)))
		      (id1 (!_id opt1val)))
		   (repeat :for ((ovb :in (cdr opt-list)) - Varbdg
				 (wishes (list (lrecord mismatch1 id1)))
				    - (Lst (Lrcd Sexp Varid))
				 optKval - Expclo
				 mismatchK - Sexp
				 idK - Varid)
		    :result (multi-let (((wl new-env)
					 (regression-wish-fulfillment
					      wishes rid env)))
			       (with r
				  (make Embryo-regress-method
				     !>change !>inductive !>quant
				     !>id new-env
				     (!_precond r)
				     `(and ,(!_secondary r)
					   ,@wl))))
		      (!= optKval (!_val ovb))
		      (!= mismatchK (cadr (!_skel optKval)))
		      (!= idK (!_id optKval))
		    :while (one-value (unify mismatchK mismatch1
					 idK id1 (!_bdgs r)))
		      (!= wishes (cons (lrecord mismatchK idK)
				       *-*))
		    :result (progn (dbg-out optimism-dbg*
				     "Found odd mismatch: " mismatchK
				     " vs. " mismatch1 :%)
				  false))))))
	 (t r)))

(deffunc is-step-value-mismatch - Boolean (b - Varbdg)
	 (is-step-value-occ (!_(Optimism-rcd opt-pat)
			       (car (!_skel (!_val b))))))

(deffunc regression-wish-fulfillment - (Mlv (Lst Sexp) Bdgenv)
	                             (mismatches - (Lst (Lrcd Sexp Varid))
				      rid - Varid
				      env - (Lst Varbdg))
   (repeat :for ((mis :in mismatches) - (Lrcd Sexp Varid)
		(res '()) - (Lst Sexp)
		pat - Sexp)
      (!= < pat env >
	  (varid-convert (car mis) (cadr mis) rid env))
      (!= res (cons `(= (step-value this-step) ,pat) *-*))
    :result (values res env)))

(specdecl (normvar-count* 0) - Integer)

;; For every goal that mentions (step-value this-step),
;; replace that term with V, and add 
;; (normal-step-value <this-step's act> V) to the goal list.
(deffunc normalize-step-value-goals - (Lst Sexp)
				    (goals - (Lst Sexp)
				     act - Sexp)
   (let-fun ()
      (cond ((<v step-value-occurs-in goals)
	     (let ((rev-gl '()) - (Lst Sexp)
		   normval - Sexp)
		(repeat :for ((gl goals (cdr gl)) - (Lst Sexp)
			     g - Sexp
			     (eqg false) - (~ Sexp))
		 :until (null gl)
		 :result (progn (!= normval 
				   (make Qvar
				      (build-symboid
					 normstepval-
					 (++ normvar-count*))
				      '()))
			       (!= rev-gl goals))
		   (!= g (car gl))
		   (match-cond g
		      ?( (= ?x ?y)
			(cond ((is-step-value-occ x)
			       (cond ((not (is-step-value-occ y))
				      ;; Very unlikely that they're both
				      ;; step-value-occs, but you never
				      ;; know
				      (!= eqg g)
				      (!= normval y))))
			      ((is-step-value-occ y)
			       (!= eqg g)
			       (!= normval x))
			      (t
			       (!= rev-gl (cons g *-*)))))
		      (t
		       (!= rev-gl (cons g *-*))))
		 :until eqg
		 :result (!= rev-gl
			    (nreconc *-* (cdr gl))))
		`((normal-step-value ,act ,normval)
		  ,@(<# (\\ (g) (replace-step-value g normval))
			rev-gl))))
	    (t
	     goals))))

(deffunc step-value-occurs-in - Boolean (x - Sexp)
    (cond ((is-step-value-occ x)
	   true)
	  ((or (atom x) (is Qvar x)) nil)
	  (t
	   (exists (y in x) (step-value-occurs-in y)))))

(deffunc replace-step-value - Sexp (x val - Sexp)
   (let-fun ((sv-replace - Sexp (x - Sexp)
	       (cond ((is-step-value-occ x)
		      val)
		     ((or (atom x) (is Qvar x))
		      x)
		     (t (<# sv-replace x)))))
      (sv-replace x)))

(defun is-step-value-occ (x)
    (let (sv)
       (and (matchq (?sv this-step)
		    x)
	    (or (eq sv 'step-value)
		(and (is Sym-with-type sv)
		     (eq (!_(Sym-with-type actual) sv) 'step-value))))))

(specdecl (absurd-id* (new-Varid)) - Varid)

(deffunc goal-trivially-false - Boolean (g - Sexp dom - Domain)
   (let ((fsit (domain-place-fact-sit dom true)))
      (let ((ansl (deduce `(absurd ,g) absurd-id* (empty-env) fsit dom)))
         (let ((res 
		  (exists (a :in ansl)
		     (opt::no-new-bdgs a (empty-env)))))
;;;;	    (cond (res
;;;;		   (out (:to *error-output*)
;;;;		      "Rejecting absurd goal " g :%)))
	    res))))

(specdecl nisptype::type-type* - Opt-type)

(deffunc is-type-name - Boolean (s - Symbol d - Domain)
   (let ((b (find-domain-bdg s d)))
      (cond (b
	     (eq (!_type b) nisptype::type-type*))
             ;(IS Opt-type (!_VAL B))
            (t 'nil))))

(deffunc domain-find-init-sit - Initial-situation (dom - Domain)
   (match-cond (!_name dom)
      ?( (domain ?s)
	(cond ((is Initial-situation s)
	       s)
	      (t
	       (signal-problem domain-find-init-sit
		  "Can't extract initial situation from domain named "
		  (!_name dom)))))
      (t
       (signal-problem domain-find-init-sit
	  "Domain does not correspond to initial situation: " dom))))

(decl ((fludom (try-domain-with-name 'opt::fluents false))
       (procdom (try-domain-with-name 'opt::processes false)))
   (cond ((and fludom procdom)
	  (!= fludom (domain-indexify *-*))
	  (!= procdom (domain-indexify *-*))
	  (domain-make-current fludom)
	  (domain-make-current procdom)
	  (let (;;;;(flu-fs (opt::domain-place-fact-sit fludom true))
		;;;;(proc-fs (opt::domain-place-fact-sit procdom true))
		)
	     (let-fun ()
		(add-facts ;;;;flu-fs
		   fludom
		   '((fluent-regress-handler <  assign
					        ineq-assign-regress)
		     (fluent-regress-handler >  assign
						ineq-assign-regress)
		     (fluent-regress-handler >= assign 
						ineq-assign-regress)
		     (fluent-regress-handler =< assign 
						ineq-assign-regress)
		     (fluent-regress-handler =~  assign 
						 ineq-assign-regress)
		     (fluent-regress-handler /=~ assign 
						 ineq-assign-regress)
		     
		     (fluent-regress-handler <  increase 
						ineq-inc-dec-regress)
		     (fluent-regress-handler >  increase 
						ineq-inc-dec-regress)
		     (fluent-regress-handler >= increase 
						ineq-inc-dec-regress)
		     (fluent-regress-handler =< increase
						ineq-inc-dec-regress)
		     (fluent-regress-handler =~  increase 
						 ineq-inc-dec-regress)
		     (fluent-regress-handler /=~ increase
						 ineq-inc-dec-regress)

		     (fluent-regress-handler <  decrease 
						ineq-inc-dec-regress)
		     (fluent-regress-handler >  decrease 
						ineq-inc-dec-regress)
		     (fluent-regress-handler >= decrease 
						ineq-inc-dec-regress)
		     (fluent-regress-handler =< decrease
						ineq-inc-dec-regress)
		     (fluent-regress-handler =~  decrease
						 ineq-inc-dec-regress)
		     (fluent-regress-handler /=~ decrease
						 ineq-inc-dec-regress)))
		(add-facts ;;;;proc-fs
		     procdom
		     '((fluent-regress-handler <  derivative
					          ineq-derivative-regress)
		       (fluent-regress-handler >  derivative
					          ineq-derivative-regress)
		       (fluent-regress-handler >= derivative
						  ineq-derivative-regress)
		       (fluent-regress-handler =< derivative
						  ineq-derivative-regress)
		       (fluent-regress-handler =~  derivative
						  ineq-derivative-regress)
		       (fluent-regress-handler /=~ derivative
						  ineq-derivative-regress)))
	      :where

	         (add-facts - Void (;;;;fact-sit - Situation
				    dom - Indexed-domain
				    assertions - (Lst Sexp))
		    (let ((frh-rg
			     (opt::domain-place-rule-group
				 dom 'fluent-regress-handlers)))
		       (cond ((opt::is-Rule-group frh-rg)
			      (!= (opt::Rule-group-rules frh-rg)
				  (<# uninternalize assertions)))
			     (t
			      (signal-problem fluregress
				 "Can't set up rule group for "
				 "fluent regress handlers"))))
		       
;;;;		    (with-thawed-situation fact-sit
;;;;		       (changing-fact-index fact-sit
;;;;			  (repeat :for ((a :in assertions))
;;;;			     (assertion-add a ground-id* !()
;;;;					    fact-sit dom))))
		    (!= (!_generation dom) (next-generation)))

		 (uninternalize - Obj (a - Sexp)
		    (opt::make-Fact
		       (make-inst opt::Unchecked-typed-exp
		          :source a
			  :type prop-type*)
		       false)))))
	 (t
	  (signal-problem fluregress
	     "Fluent and/or process domain inaccessible"))))
