;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: forceproj.lisp,v 1.19 2005/08/27 05:30:41 dvm Exp $

(depends-on %module/ nisp)

(depends-on :at-run-time %optop/ rmgraph discrtproj continproj types)

(depends-on :at-compile-time 
	                 %ytools/ multvalhacks)

(specdecl acts - (Lst Action))

;;; If skip-first is true, we're projecting the sequel of a feasible
;;; action, which has already been projected.
(deffunc reduction-tree-project - (Multvals num-steps - Integer
					    elapsed - Time-int
					    attained - Situation
					    penalty - Number)
	             (rt - Reduction-tree skip-first - Boolean 
		      steps-so-far - Integer
		      time-so-far max-time-left - Time-int
		      sits - (Lst Situation) sit-space - Situation-space dom - Domain)
   (dbg-out forceproj-dbg*
      "Projecting reduction tree")
   (let ((done-effort-specs !()) - (Lst Effort-spec)
	 (penalty 0) - Integer
	 (total-time time-so-far) - Float
	 (num-steps steps-so-far) - Integer
	 (sit (car sits))
	 ;;;;(start-sit (car sits))
	 (last-sit (car sits))
	 - Situation
;;;;	 (prev-sit (and (not (null (cdr sits)))
;;;;			(cadr sits)))
;;;;	 - (~ Situation)
	 (acts !()) - (Lst Action)
	 (past-sits (cdr sits)) ;;;;(start-sits sits)
	 - (Lst Situation))
     (let-fun (;; Returns list of actions done at tr or below
	       (walk-tree - (Lst Action) (tr - Reduction-tree
					  already-done - (Collector Action)
					  ;; -- list of actions to left
					  ;; of tr or a supernode.
					  skip-first - Boolean)
;;;;		  (trace-around walk-tree
;;;;		     (:> "(walk-tree: " (!_goal-es tr) ")")
;;;;		  (dbg-save tr)
		  (cond ((memq (!_goal-es tr) done-effort-specs)
			 !())
			((and skip-first
			      (null (!_preconds tr))
			      ;; waits always have to be projected here,
			      ;; because 'plan-prefix-grow' just projects
			      ;; the first 'wait' segment.
			      (not (matchq (?(:\| wait-non-negative
						 wait-while)
					     ?@_)
					   (!_term (!_achiever tr)))))
			 (list (!_achiever tr)))
			(t
			 (repeat :for ((prec :in (!_preconds tr))
				         - Reduction-tree
				       (sf = skip-first :then false)
				          - Boolean
				       :collectors sub-done - Action)
;;;;			    (out "Recursing on " prec :%)
			    (repeat :for ((sub
					     :in
					     (out-indent *standard-output* 3
						(walk-tree
						   prec already-done sf))))
			     :when (and (not (member= #'action-subsumed
						      sub
						      (!_elements
						          already-done)))
					(not (member= #'action-subsumed
						      sub
						      (!_elements sub-done))))
			        (one-collect sub-done sub)
				(one-collect already-done sub))
			  :result
			     (cond ((eq (!_term (!_achiever tr)) 'done)
				    sub-done)
				   ((member= #'action-subsumed
					     (!_achiever tr) sub-done)
				    (run-step tr)
				    sub-done)
				   (t
				    (let ((s (member= #'action-subsumed
						      (!_achiever tr)
						      (!_elements
						         already-done))))
				       (cond (s
					      (dbg-out forceproj-dbg*
						 "Action " (!_achiever tr)
						 :% " subsumed")
					      (cons (car s) sub-done))
					     (t
					      (run-step tr)
					      (cons (!_achiever tr)
						    sub-done)))))))))
;;;;		     (:< (val &rest _) "walk-tree: " val))
		  ))

	 (walk-tree rt (make (Collector Action)) skip-first)
	 (values num-steps total-time sit penalty)
       
       :where
       
 ;; Note that this is nonsymmetrical.  tr1 is subsumed by tr2 if
 ;; they're the same *or* they're both waits for the same process,
 ;; and tr2 waits for it to be done.
 (:def action-subsumed - Boolean (a1 a2 - Action)
;;;;    (trace-around action-subsumed
;;;;       (:> "(action-subsumed: " a1 :% 4 a2")")
    (or (eq a1 a2)
	(let ((t1 (!_term a1))
	      (t2 (!_term a2)))
	   (and (consp t1) (consp t2)
		(eq (car t1) 'wait-non-negative)
		(eq (car t2) 'wait-while)
		(equal (lastelt t1) (lastelt t2)))))
;;;;       (:< (val &rest _) "action-subsumed: " val))
    )

 (:def run-step - Void (tr - Reduction-tree)
    (let ((act (!_achiever tr))
	  (goal-es (!_goal-es tr)))
       (let ((immed-goal (!_literal goal-es))
	     (id (!_varid goal-es))
	     ;;;;(act1 (uniquify-action act dom))
	     )
	  (let ((now-instances (deduce immed-goal id !() sit dom)))
	     (dbg-out forceproj-dbg*
		"Contemplating " act  
		:% " Purpose: " immed-goal
		" Already? " (not (null now-instances)))
	     (cond ((eq (!_term act) 'watch)
		    ;; Do nothing, because the goal should follow from
		    ;; premises already achieved.
		    (cond ((null now-instances)
			   ;; Uh-oh
			   (!= penalty (+ *-* 1)))))
		   ((null now-instances)
		    (multi-let (((forced goal-achieved succ elapsed steps)
				 (force-project sit last-sit past-sits 
						act immed-goal id
						sit-space dom)))
		       (!= acts (cons act *-*))
		       (!= past-sits (cons sit *-*))
		       (cond (succ
			      (!= sit succ)))
		       (!= num-steps (+ *-* steps))
		       (!= total-time
			   (+ *-* (cond ((eq elapsed '*never) max-time-left)
					(t (be Time-int elapsed)))))
		       (cond (forced
			      (!= penalty (+ *-* 1))))
		       (cond ((not goal-achieved)
			      (!= penalty (+ *-* 1))))))
		   ((has-qvars immed-goal)  ;;;;(and (not skip-next) ...)
		    (multi-let (((succ elapsed steps)
				 (act-or-wait-project sit act sit-space dom)))
		       (cond ((and succ
				   (new-instances
				      immed-goal id sit now-instances))
			      (!= acts (cons act *-*))
			      (!= past-sits (cons sit *-*))
			      (!= sit succ)
			      (!= num-steps (+ *-* steps))
			      (!= total-time
				  (+ *-*
				     (cond ((eq elapsed '*never)
					    max-time-left)
					   (t (be Time-int elapsed)))))))))))))
    (dbg-out forceproj-dbg*
       "Total time = " (:f "~,2f" total-time) " total steps = " num-steps))

 (:def new-instances - Boolean (goal - Sexp id - Varid sit - Situation
			   now-instances - (Lst Bdgenv))
    (let ((new-ansl (deduce goal id !() sit dom)))
       (and (exists (a :in new-ansl)
	       (not (exists (now :in now-instances)
		       (one-value (bdgenvs-combine now a)))))
	    (forall (now :in now-instances)
	       (exists (a :in new-ansl)
		  (one-value (bdgenvs-combine now a)))))))

;;;; (pathological - Boolean (sit - Situation)
;;;;    (or (sit-patho sit)
;;;;	(sit-patho last-sit)
;;;;	(and prev-sit (sit-patho prev-sit))))

)))

;;;;(specdecl acts - (Lst Action))

;;; Returns < forced, new-sit, elapsed-time, num-steps >, 
;;; where 'forced' is true if either
;;; 'act' is infeasible, or it's feasible but fails to make 'goal' true.
(defproc force-project - (Multvals forced goal-achieved - Boolean
				   new-sit - (~ Situation)
				   elapsed - (Alt Time-int (Const *never))
				   num-steps - Integer)
	   (sit start-sit - Situation past-sits - (Lst Situation)
	    act - Action goal - Sexp id - Varid
	    sit-space - Situation-space dom - Domain)
   (let ((sss (find-old-projection sit act sit-space))
	 - (~ Sitsuccspec)
	 (act-term (!_term act)) eltact quant - Sexp
	 succ-sit - (~ Situation)
	 forced - Boolean
	 steps - Fixnum
	 adds dels al dl - (Lst Occasion)
	 elapsed - (Alt Time-int (Const *never))
	 (prev-sit (and (not (null past-sits))
			(car past-sits)))
	 - (~ Situation)
	 frac - Float)
      (letrec ()
	 (cond ((and sss (!_sit sss))
		(with sss
		   (!= succ-sit !>sit)
		   (!= forced !>forced)
		   (!= elapsed !>elapsed)
		   (!= frac !>frac)
		   (!= steps !>steps)))
	       (t
		(cond (sss (succ-erase sit act sit-space)))
		(multi-let (((procterm negterm)
			     (is-wait act)))
		   (cond (procterm
			  (cond ((process-find procterm sit)
				 (!= < succ-sit adds dels elapsed frac >
				     (process-project-and-record
					sit start-sit prev-sit procterm negterm
					sit-space dom))
				 (!= forced
				     (or (not succ-sit)
					 (let ((new-pc (!_processes succ-sit)))
					    (and new-pc
						 (not (!_kosher new-pc)))))))
				(t
				 (!= < succ-sit al dl >
				     (process-existence-force
					   procterm negterm sit past-sits
					   sit-space dom))
				 (cond ((not (process-find procterm succ-sit))
					(dbg-save :run-loud
					   procterm sit succ-sit dom)
					(signal-problem force-project
					   "Process cannot be forced to"
					   " exist: " procterm)))
				 (!= < succ-sit adds dels elapsed frac >
				     (process-project-and-record
					succ-sit start-sit prev-sit
					procterm negterm sit-space dom))
				 (!= < adds dels >
				     (add-del-combine al dl adds dels))
				 (!= steps 0)
				 (succ-record true sit elapsed 0
					      act succ-sit adds dels
					      frac !() sit-space)
				 (!= forced true))))
			  ((matchq (until-non-negative ?quant ?eltact)
				   act-term)
			   (!= elapsed 0.0)
			   (!= < forced succ-sit adds dels frac steps >
			       (until-project sit (uniquify-action eltact dom)
					      quant past-sits sit-space dom))
			   )
			  (t
			   (!= < forced succ-sit adds dels >
			       (discrete-action-force-project
				  sit act past-sits sit-space dom))
			   (!= elapsed 0.0)
			   (!= frac 0)
			   (!= steps 1))))))
      (let ((goal-achieved
		 (and succ-sit
		      (not (null (deduce goal id !() succ-sit dom))))))
	 (cond ((and forced
		     (not (eq elapsed '*never))
		     (> frac 0.0)
		     (< frac 1.0))
		(!= elapsed (be Time-int
				(/ (be Time-int elapsed)
				   frac)))))
	 (dbg-out forceproj-dbg*
		  (:q (forced  
		       "Action " act
		       :% " infeasible"
		       :% " FORCING" :%))
		  (:q (goal-achieved
		       "Okay, goal now satisfied")
		      (t
		       "Goal not satisfied"))
		  1
;;;;		  "Elapsed, steps = " elapsed ", " steps
		  (:q ((not (eql elapsed 0.0))
		       "Elapsed = " elapsed 1))
		  (:q ((not (eql steps 0))
		       "Steps = " steps 1))
		  :%)
	 (values forced
		 goal-achieved
		 succ-sit
		 elapsed
		 steps))
    :where
       )))

(defproc act-or-wait-project - (Mlv (~ Situation)
				    (Alt Time-int (Const *never))
				    Integer)
	                     (sit - Situation act - Action sit-space - Situation-space
			      dom - Domain)
   (let ((negterm false) (procterm false)- (~ Sexp))
      (matchq ?(:\| (wait-non-negative ?negterm ?procterm)
		   (wait-while ?procterm))
	      (!_term act))
      (cond (procterm
	     (multi-let (((final-sit adds dels elapsed frac)
			  (process-project
				sit !() procterm negterm sit-space dom)))
		(!= final-sit
		    (succ-record
		       false sit elapsed 0 act final-sit adds dels frac !() sit-space))
		(values final-sit elapsed 0)))
	    (t
	     (values (nth-value 0
			(fresh-act-project true sit act sit-space dom true))
		     0.0 1)))))

(defproc process-project-and-record - (Multvals succ - Situation
						adds dels - (Lst Occasion)
					        elapsed - (Alt Time-int
							       (Const *never))
						frac - Float)
	                            (sit start-sit - Situation
				     prev-sit - (~ Situation)
				     procterm negterm - Sexp
				     sit-space - Situation-space dom - Domain)
   (letrec ()
      (multi-let (((final-sit adds dels elapsed frac)
		   (process-project
			 sit (list start-sit prev-sit)
			 procterm negterm sit-space dom)))
	 (!= final-sit
	     (succ-record
		false sit elapsed 0
		(uniquify-action (wait-act procterm negterm)
				 dom)
		final-sit adds dels frac !() sit-space))
	 (values final-sit adds dels elapsed frac))
    :where
       ))



;;; This procedure does two rather different things, depending on whether
;;; 'negterm' is false or a fluent to be made non-negative.  
;;; In the latter case it is interested
;;; only in when that inequality becomes true.  In the former it
;;; is interested only in when the process denoted by 'procterm' finishes.
;;; So if the process ends unexpectedly, it's a failure if negterm is a
;;; term, a success if it's false.
;;; Returned value 'frac' is fraction of ground covered from starting value
;;; of negterm to 0.  It's zero if negterm=false.
;;; The 'others' arg is an obsolete debugging thing.
(deffunc process-project - (Multvals final-sit - (~ Situation)
				     adds deletes - (Lst Occasion)
				     elapsed - (Alt Time-int
						    (Const *never))
				     frac - Float)
	       (sit - Situation others - (Lst (~ Situation))
		procterm - Sexp
		negterm - (~ Sexp)
		sit-space - Situation-space dom - Domain)
               ;; -- negterm = false means that we want to project
	       ;; until the process stops
      (dbg-out forceproj-dbg*
	       "Projecting process " procterm)
      (repeat :for ((next-sit sit) - (~ Situation)
		    ;; -- never false except when error occurs
		    (termcond
		       (and negterm (be Uterm-ineq
					`(>= ,(uniquify-continuous-fluent-subterms
					       negterm dom)
					     0.0))))
		    - (~ Uterm-ineq)
		    process - (~ Process-instance)
		    (elapsed 0.0) - Time-int
		    (numwaits 0) - Fixnum
		    (timegrain (domain-temporal-grain-size dom))
		      - Float
		    delta-exasp - (~ Time-int)
		    delta-elap - (Alt Time-int (Const *never))
		    processes - Process-complex
		    trigs - (Lst Ineq-trigger)
		    (adds !()) (dels !()) al dl - (Lst Occasion)
;;;;		    save-sit - Situation
		    :collectors
		    exasperation - Time-int
		    
		    ;; -- Forecast times until desired trigger
		    ;; point is reached, in reverse chrono order.
		    )
         ;; until we reach goal or run out of patience
	 (!= processes (!_processes next-sit))
	 (!= process (process-find procterm next-sit))
       :while process
       :result (progn
		   (cond (negterm
			  (let ((frac (compute-frac)))
			     (cond ((ground-deduce `(=~ ,frac 1.0)
						   next-sit dom)
;;;;				    (cond ((pathological next-sit)
;;;;					   (dbg-save next-sit sit others)
;;;;					   (signal-problem process-project
;;;;					      "Pathology after ground-deduce")))
				    (dbg-out forceproj-dbg*
				       "Process ends, satisfying "
				       termcond)
				    (values next-sit adds dels elapsed frac))
				   (t
				    (dbg-out forceproj-dbg*
				       "Process ends without satisfying "
				       termcond)
;;;;				    (dbg-save sit next-sit frac negterm)
;;;;				    (breakpoint process-project
;;;;				       "frac = " frac)
				    (values false !() !()
					    (cond ((=< frac 0.0) '*never)
						  (t elapsed))
					    frac)))))
			 (t
			  (dbg-out forceproj-dbg*
			     "Process ends successfully")
			  (values next-sit adds dels elapsed 0))))
	 (!= trigs (cond (termcond
			  (be (Lst Ineq-trigger)
			      (<? is Ineq-trigger
				  (continuous-trigs
					   (list termcond)
					   (list '(made-it))
					   (!_model processes)
					   next-sit dom))))
			 (t
			  (repeat :for ((tr :in (!_triggers processes))
					:collectors stop-trigs)
			   :when (is-stopper tr)
			   :collect tr
			   :result (cond ((null stop-trigs)
					  ;; Make one up
					  (list (make Ineq-trigger
						   `(dummy)
						   (\\ (ti - Time-int)
						      (- ti
							 way-in-the-future*))
						   `(- *T* (way-in-the-future
							    = ,way-in-the-future*))
						   (list '(made-it)))))
					 (t stop-trigs))))))
       ;; This can happen only if negterm is non-false and has become
       ;; non-negative, i.e., we made it --
       :until (null trigs)
         (!= < delta-exasp _ _ >
	     (compute-trigger-time trigs timegrain))
       :while (and delta-exasp (> delta-exasp 0.0))
       :result
          (let ((frac (compute-frac)))
	     (dbg-out forceproj-dbg*
		"Process "
		(:q ((null trigs)
		     "ends happily")
		    (t "never ends"))
		:% "  accomplished fraction " frac " of " termcond)
	     (values next-sit adds dels elapsed frac))
       :collect (:into exasperation delta-exasp)
	  (dbg-out forceproj-dbg* "exasperation = " (!_elements exasperation) :%)
       :until (exasperated)
       :result
           (progn
	      (dbg-out forceproj-dbg*
		 "Process ends in exasperation")
	      (values false !() !() '*never 0))
	 (bind ((watch-sit* (cadr others)))
            (!= < next-sit al dl delta-elap >
		(wait-project true next-sit sit-space dom)))
	 (cond ((not next-sit)
		(signal-problem process-project
		   "Unexpected inability to project a 'wait'")))
	 (!= adds (union al (complementq *-* dl)))
	 (!= dels (union dl (complementq *-* al)))
	 (!= numwaits (+ *-* 1))
         (!= elapsed (+ *-* (be Time-int delta-elap)))

       :where

	  (:def is-stopper - Boolean (trig - Ineq-trigger)
	     (exists (eff :in (!_effects trig))
		(cond ((is Ineq-trigger eff)
		       (is-stopper eff))
		      (t
		       (decl (eff - Sexp)
			  (matchq (process-transition
				      stopping ?,procterm)
				  eff))))))

	  (:def exasperated - Boolean ()
	     (cond ((< numwaits 10)
		    false)
		   (t
		    (let ((exlts (!_elements exasperation)))
		       (let ((delta0 (car exlts)))
			  (let ((decreases
				   (cons delta0
					 (repeat :for
						 ((delta :in (cdr exlts))
						  (prev-delta delta0)
						    - Time-int)
					  :when (< delta prev-delta)
					  :collect delta
					     (!= prev-delta delta)))))
			     (< (* 3 (len decreases))
				numwaits)))))))

	  (:def compute-frac - Float ()
	     (prog1
			  (cond ((and negterm (> elapsed 0))
				 (let ((initval
					  (nth-value
					     1 (opt-eval negterm sit !() dom)))
				       (finalval
					  (nth-value
					     1 (opt-eval negterm next-sit
							 !() dom)))
				       - Number)
				   (/ (- finalval initval)
				      (- (float initval)))))
				(t 0))
	       ))

	  ))

;;;;(specdecl (adjustno* 0) - Fixnum)

;;; Project until the 'quant' is non-negative
(defproc until-project - (Multvals forced - Boolean
				   final-sit - (~ Situation)
				   adds deletes - (Lst Occasion)
				   frac - Float steps - Fixnum)
	       (sit - Situation act - Action quant - Sexp
		past-sits - (Lst Situation) sit-space - Situation-space dom - Domain)
   (let ((orig-val (nth-value 1 (opt-eval quant sit !() dom))))
      (cond ((not orig-val)
	     (signal-problem until-project
		"Can't evaluate " quant)))
      (cond ((>= orig-val 0)
	     (values false sit !() !() 1.0 0))
	    (t
	     (repeat :for ((val orig-val) - Sexp val-okay - Boolean
			   (steps = 1 :by 1) - Fixnum
			   (succ-sit sit) - Situation
			   (forced false) f1 - Boolean
			   (adds !()) (deletes !()) al dl - (Lst Occasion)
			   :collector vals - Number)
		(!= < f1 succ-sit al dl >
		    (discrete-action-force-project
		       succ-sit act past-sits sit-space dom))
		(cond (f1 (!= forced true)))
		(!= < adds deletes >
		    (add-del-combine adds deletes al dl))
		(!= < val-okay val > (opt-eval quant succ-sit !() dom))
		(cond ((not val-okay)
		       (signal-problem until-project
			  "Can't evaluate " quant)))
	      :until (>= val 0)
	      :result (values forced succ-sit adds deletes 1.0 steps)
	      :collect val
	      :until (nums-not-going-anywhere (!_elements vals))
	      :result (values true false !() !()
			      (/ (- val orig-val)
				 (- (float orig-val)))
			      steps))))))

(defproc discrete-action-force-project - (Multvals forced - Boolean
						   new-sit - Situation
						   adds deletes
						   - (Lst Occasion))
	                               (sit - Situation act - Action
					past-sits - (Lst Situation)
					sit-space - Situation-space dom - Domain)
   (cond ((eq (!_term act) 'watch)
	  ;; Control should never reach here, but if it does the following
	  ;; is the correct value to return --
	  (values false sit !() !()))
	 (t
	  (multi-let (((succ-sit adds dels _ _ )
		       (fresh-act-project true sit act sit-space dom true)))
;;;;	     (dbg-save succ-sit sit adds dels act sit-space dom)
;;;;	     (breakpoint discrete-action-force-project
;;;;		"succ-sit = " succ-sit)
	     (cond (succ-sit
		    (values false succ-sit adds dels))
		   (t
		    (succ-erase sit act sit-space)
		    (multi-let (((feas-sit al dl)
				 (act-feasibility-force sit past-sits
							act sit-space dom)))
		       (!= < succ-sit adds dels _ _ >
			   (fresh-act-project true feas-sit act sit-space dom true))
		       (cond ((not succ-sit)
			      (dbg-save :run-loud feas-sit sit past-sits
					act dom)
			      (signal-problem discrete-action-force-project
				 "Act cannot be forced to be"
				 " feasible: " act)))
		       (!= < adds dels >
			   (add-del-combine al dl adds dels))
		       (values true succ-sit adds dels))))))))

(deffunc nums-not-going-anywhere - Boolean (vals - (Lst Number))
   (let ((delta0 (car vals)) (numvals (len vals)))
      (and (> numvals 10)
	   (let ((decreases
		    (cons delta0
			  (repeat :for
				  ((delta :in (cdr vals))
				   (prev-delta delta0)
				     - Number)
			   :when (< delta prev-delta)
			   :collect delta
			      (!= prev-delta delta)))))
	      (>= (* 3 (len decreases))
		  numvals)))))

(defproc process-existence-force - (Multvals new-sit - Situation
					     adds dels - (Lst Occasion))
	                         (procterm - Sexp negterm - (~ Sexp)
				  sit - Situation past-sits - (Lst Situation)
				  sit-space - Situation-space dom - Domain)
   (let ((ipdefn
	    (causation-view
	       (name-causation-defn
		   (be CSymbol (car procterm))
		   'dur-happening (cdr procterm) dom)
	       #'durative-process-view
	       dom)))
      (let ((w-act (wait-act procterm negterm)))
	 (make-succ-sit-with-difs
	    ipdefn procterm w-act sit past-sits sit-space dom))))

(deffunc wait-act - Sexp (procterm negterm - Sexp)
   (cond (negterm
	  `(wait-non-negative ,negterm ,procterm))
	 (t `(wait-while ,procterm))))

(defproc act-feasibility-force - (Multvals new-sit - Situation
					   adds dels - (Lst Occasion))
	                       (sit - Situation past-sits - (Lst Situation)
				act - Action
				sit-space - Situation-space dom - Domain)
   (let ((actexp (!_act act)))
      (let ((actdef
	       (causation-view
		  (name-causation-defn
		     (be CSymbol (car actexp))
		     'dur-start (cdr actexp) dom)
		  #'durative-start-view
		  dom)))
	 (make-succ-sit-with-difs
	    actdef actexp actexp sit past-sits sit-space dom))))

;;; IN the next few functions, we are using dummy-id1* for pieces
;;; of a goal derived from a particular causation definition ('idefn') --

(defproc make-succ-sit-with-difs - (Multvals succ - Situation
					     adds dels - (Lst Occasion))
                                 (idefn - Internalized-causation-defn
				  unifexp actexp - Sexp
				  sit - Situation past-sits - (Lst Situation)
				  sit-space - Situation-space dom - Domain)
   (control-nest
      (let ((feas-act
	       (uniquify-action `(make-feasible ,actexp) dom)))
	 (let ((sss (find-old-projection sit feas-act sit-space)))
	    (cond (sss
		   (with sss
		      (values !>sit !>adds !>deletes)))
		  (t
		   (multi-let (((ok - Boolean env - Bdgenv)
				(unify (!_term idefn)  unifexp
				       dummy-id1*      dummy-id2*
				       (empty-env))))
		      (let ((new-occs !()) - (Lst Occasion))
			 (cond (ok
				(!= < new-occs env >
				    (find-assertable-differences
				       (!_condition idefn)
				       env sit past-sits dom))
				(multi-let (((adds deletes)
					     (find-fluent-differences
						 new-occs env sit dom)))
				   :then))
			       (t
				(signal-problem make-succ-sit-with-difs
				   "Term " unifexp :% "fails to unify with"
				   " defining term of " :% 1 idefn)))))))))
      :then
      (multi-let (((new-sit final-adds final-deletes)
		   (build-new-sit
		      true feas-act
		      adds deletes dummy-id1* env
		      sit sit-space dom)))
;;;;				(dbg-save feas-act new-occs env sit dom
;;;;					  new-sit adds deletes)
;;;;				(breakpoint make-succ-sit-with-difs
;;;;				   "Got new sit")
	 (values
	    (succ-record
	       true sit 0.0 1 feas-act new-sit
	       final-adds final-deletes
	       0.0 !() sit-space)
	    adds deletes))))




(specdecl (difs-fill-dbg* false) - Boolean)

(deffunc find-assertable-differences - (Mlv (Lst Occasion) Bdgenv)
	                    (p - Sexp env - Bdgenv
			     sit - Situation past-sits - (Lst Situation)
			     dom - Domain)
   (multi-let (((precond env)
		(safe-varsubst p dummy-id1* env)))
      (dbg-out difs-fill-dbg*
	       "Trying to force " precond " to be true")
      (let ((dp (disjunctive-normal-form precond true))
	    - (Lst (Lst Sexp))
	    result-env - Bdgenv new-occs - (Lst Occasion))
;;;;	 (dbg-save dp precond)
;;;;	 (breakpoint find-assertable-differences
;;;;	    "dp = " dp)
	 (repeat :for ((conj :in dp) - (Lst Sexp)
		       (found-difs false) - Boolean)
	  :result (progn
		     (dbg-save :run-loud dp sit past-sits dom env)
		     (signal-problem find-assertable-differences
			"Unable to complete " precond))
	    (let ((mml (maximal-matches
			  conj !()
			  dummy-id1* env
			  sit dom)))
	       ;; Look for a maximal match whose difs are assertable.
	       (repeat :for ((mm :in mml) - Maximal-match)
		  (cond ((forall (d :in (!_dif mm))
			    (goal-has-assertable-pred (!_goal d)))
			 (letrec ((elim ()
				   (!= < found-difs result-env new-occs >
				       (difs-elim-vars
						   (!_dif mm)
						   (!_constraints mm)
						   (!_subst mm)
						   sit past-sits dom))))
			    (cond (difs-fill-dbg*
				   (out (:to *error-output*)
				      "Trying difs "
				      (debug-varsubst
				          (!_dif mm) dummy-id1*(!_subst mm))
				      :% (:i> 3) :%
				      (:e (elim))))
				  (t (elim))))))
		:until found-difs))
	  :until found-difs)
	 (values new-occs result-env))))

(deffunc goal-has-assertable-pred - Boolean (goal - Sexp)
   (cond ((eq (car goal) 'not)
	  (goal-has-assertable-pred (cadr goal)))
	 (t
	  (not (memq (car goal) preds-forbidden-in-dif*)))))
;;; Not any more! --
;;;;	(not (memq (car goal) ineq-rels*))

(deffunc difs-elim-vars - (Multvals consis - Boolean
				    env - Bdgenv
				    new-occs - (Lst Occasion))
	                (difs - (Lst Constrained-goal)
			 constraints - (Lst Sexp)
			 env - Bdgenv
			 sit - Situation past-sits - (Lst Situation)
			 dom - Domain)
   (repeat :for ((dif :in difs) - Constrained-goal
		 (result-env env) - Bdgenv
		 (constraints constraints) - (Lst Sexp)
		 dif-filled - Boolean difexp goal-inst - Sexp
		 (unfilled-difs !()) (filled-difs !()) - (Lst Sexp))
      (!= < difexp result-env >
	  (safe-varsubst (!_goal dif) dummy-id1* result-env))
      (!= dif-filled (not (has-qvars difexp)))
      (cond (dif-filled
	     (!= goal-inst difexp)
	     (dbg-out difs-fill-dbg*
		      "Difference: " difexp)))
      (repeat :for ((past-sit :in past-sits) - Situation
		    (n = 1 :by 1) - Fixnum
		    al - (Lst Bdgenv))
       :until dif-filled
	 (!= al (deduce (!_goal dif) dummy-id1* result-env past-sit dom))
	 (repeat :for ((a :in al) new-env - Bdgenv
		       consis - Boolean)
	  :result nil
            (!= < goal-inst new-env >
		(safe-varsubst (!_goal dif) dummy-id1* a))
	    (!= < consis new-env >
		(constraints-satisfy
			     true (!_constraints dif) dummy-id1*
			     result-env sit dom))
	    (!= < goal-inst new-env >
		(safe-varsubst goal-inst dummy-id1* new-env))
	  :until (and consis (not (has-qvars goal-inst)))
	  :result (progn (dbg-out difs-fill-dbg*
			    "Found (" n " steps back) difference: "
			    goal-inst :%)
			 (!= dif-filled true))))
      (cond (dif-filled
	     (!= filled-difs (cons goal-inst *-*)))
	    (t
	     (!= unfilled-difs (cons (!_goal dif) *-*))
	     (!= constraints (append (!_constraints dif) *-*))))
    :result
       (multi-let (((consis result-env)
		    (constraints-satisfy
			true constraints dummy-id1*
			result-env sit dom)))
	  (cond (consis
		 (let ((new-occs-from-unfilled
			    (repeat :for ((d :in unfilled-difs) - Sexp
					  :collector instances - Occasion)
			       (!= < d result-env >
				   (safe-varsubst d dummy-id1* result-env))
			       (cond ((has-uvars d dummy-id1* result-env)
				      (signal-problem difs-elim-vars
					 "Constraining vars failed to"
					 " instantiate " d)))
			     :collect (uniquify-occasion d dom))))
		    (dbg-out difs-fill-dbg*
		       "Filled differences using constraints: "
		       (<# !_term new-occs-from-unfilled))
		    (values true
			    result-env
			    (nconc 
			       (<# (\\ (d)
				      (uniquify-occasion d dom))
				   filled-difs)
			       new-occs-from-unfilled))))
		(t
		 (dbg-out difs-fill-dbg*
		    "Attempt to fill differences founders due to"
		    " constraint inconsistency")
		 (values false nil nil))))))

(deffunc find-fluent-differences - (Multvals adds deletes - (Lst Occasion))
                       (diff-occs - (Lst Occasion) env - Bdgenv
			sit - Situation dom - Domain)
   (repeat :for ((occ :in diff-occs) - Occasion
		 (vartab !()) - (Lst (Lrcd Uterm Qvar))
		 :collectors other-adds other-dels - Occasion
			     flu-exps - Sexp)
    :within
      (let ((term (!_propo occ)))
	 (cond ((memq (strip-car term) ineq-rels*)
		(let-fun ((:def flu-trans - Sexp (e - Sexp)
			     (cond ((atom e) e)
				   (t
				    (let ((fn (strip-car e)))
				      (cond ((eq fn 'fl-v)
					     (fluent->qvar (cadr e)))
					    (t
					     `(,fn ,@(<# flu-trans (cdr e))))
				      ))))))
		   (:continue
		    :collect (:into flu-exps (flu-trans term)))
		 :where
		   (:def fluent->qvar - Qvar (term - Sexp)
		       ;; We do this roundabout thing because of
		       ;; nuisances such as the possible occurrence
		       ;; of Sym-with-types in the wretched term.
		       (let ((ut (uniquify-term term dom)))
			  (val-or-initialize
			     (alref vartab ut)
			   :init (let ((new-var (make-Qvar (gen-var 'flu) !())))
				    (one-collect
				        flu-exps
					`(opt:is opt:Number ,new-var))
				    new-var))))))
	       ((eq (car term) 'not)
		(:continue
		 :collect (:into other-dels
				 (uniquify-occasion (cadr term)
						    dom))))
	       (t
		(:continue
		 :collect (:into other-adds occ)))))
    :result
       (multi-let (((satisfiable sat-env)
		    (constraints-satisfy
		       true flu-exps dummy-id1* env sit dom)))
	  (cond (satisfiable
		 (repeat :for ((e :in vartab) - (Lrcd Uterm Qvar)
			       :collectors adds deletes - Occasion)
		  :within
		    (let ((bdg (uvar-lookup (Qvar-sym (second e))
					    dummy-id1* sat-env)))
		       (cond ((and bdg
				   (is-Number
				      (!_skel (!_val bdg))))
			      (multi-let (((old-val-occ new-val-occ)
					   (flu-change-occs
					      (first e)
					      (!_skel (!_val bdg)))))
				 (cond (old-val-occ
					(:continue
					 :collect (:into deletes
							 old-val-occ)
					 :collect (:into adds
							 new-val-occ))))))
			     (t
			      (signal-problem find-fluent-differences
				 "Constraint solution " sat-env
				 :% " has bogus or nonexistent value for "
				    (second e) "_[" dummy-id1* "]"))))
		  :result (values (append adds other-adds)
				  (append deletes other-dels))

		  :where
		     (:def flu-change-occs - (Multvals old new
						       - (~ Occasion))
			          (flutrm - Uterm new-val - Number)
			 (let ((fluent (be Fluexp (!_term flutrm)))
			       - Fluexp)
			    (multi-let (((okay old-val)
					 (fluent-current-value
					      fluent sit dom)))
			       (cond (okay
				      (cond ((= new-val old-val)
					     (values false false))
					    (t
					     (values (uniquify-occasion
							`(current-value
							    ,fluent ,old-val)
							dom)
						     (uniquify-occasion
						        `(current-value
							    ,fluent ,new-val)
							dom)))
				      ))
				     (t
				      (signal-problem flu-change-occs
					 "Fluent " fluent " has no value in "
					 sit))))))))
		(t
		 (signal-problem find-fluent-differences
		    "Constraints unsatisfiable: " flu-exps))))))


;;;;(deffunc act-is-a-wait - Boolean (act - Sexp)
;;;;   (or (eq act 'wait)
;;;;       (and (not (atom act))
;;;;	    (memq (car act)
;;;;		  '(wait-while wait-non-negative wait)))))

