;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: lisp-nisp-ty.lisp,v 1.11 2005/12/10 03:01:27 dvm Exp $

(depends-on %module/ nisp)

#+pre-chunk (depends-on %module/ nispobj)

(eval-when (:slurp-toplevel :compile-toplevel :load-toplevel :execute)
   (defvar lisp-class-package* nil))

;;; (lisp-class-as-nisp-type name [class-name]
;;;    [ option | slotspec ]* [:augtype -patches-])
;;; where option is one of
;;;   (:package <class-pkg>)  -- defaults to value of lisp-class-package*
;;;   (:expect-nisp-class-descriptor <boolean>) -- defaults to true
;;;       (If false, was not defined by Nisp's def-class)
;;;   (:include <types>)
;;;   (:initexp <term>) -- Provides term that variables of this type
;;;        will be bound to.  (The term is evaluated once. and stored
;;;        in a global variable.)  If this is not provided, it can
;;;        usually be produced automatically.
;;; and slotspec is one of
;;;   (:conser ...) -- passed directly to defnisptype (as 'conser')
;;;   ([| :access | :set | :both | :all ] <slotname> - <type> [<slotdefn>])
;;;      where first element defaults to ':all';
;;;            optional slotdefn defaults to class-name-slotname.
;;;        <slotname> may be "is".  If no explicit is, the macro
;;;        inserts slotspec (is - Boolean (x) (is-class-name x)) for you.
;;; class-name is a symbol giving the name of the class;
;;;     defaults to pkg::class-name
;;; Flag ':augtype' precedes ordinary patches to be added to the the type
;;;   definition
;;; Note that the macro can't figure out :include relations on its own.
;;; If class C2 :includes class C1, it won't know that type C2 includes type
;;; C1 after
;;;     (lisp-class-as-nisp-type C1 ...)  (lisp-class-as-nisp-type C2 ...)
;;; unless the second has an (:include C1) option.
(defmacro lisp-class-as-nisp-type (name &rest slots)
   (let ((realslots '())
         (other '())
	 (aug-patches '())
         (package lisp-class-package*)
         (is-def false)
         (conser '())
	 (require-nisp-class-descriptor true)
	 class-name-string
	 (class-name (cond ((null slots) false)
			   (t (car slots))))
         (initexp false))
      (cond (class-name
	     (cond ((is-Symbol class-name)
		    (!= class-name-string (symbol->string class-name))
		    (!= slots (cdr *-*)))
		   ((is-String class-name)
		    (out (:to *error-output*) 
			    "Warning -- explicit string " class-name
			    " used in lisp-class-as-nisp-type" :%)
		    (!= class-name-string class-name)
		    (!= slots (cdr *-*)))
		   (t
		    ;; class-name really is a slot
		    ;; The actual class-name is the same as name
		    (!= class-name-string (symbol->string name)))))
	    (t
	     (!= class-name-string (symbol->string name))))
      (repeat :for ((s :in slots :tail sl))
       :result nil
       :until (eq s ':augtype)
       :result (!= aug-patches (cdr sl))
         (cond ((or (atom s) (eq (car s) ':include))
                (!= other (cons s *-*)))
               ((eq (car s) ':package)
                (!= package (cadr s)))
               ((eq (car s) ':conser)
                (!= conser (list s)))
	       ((eq (car s) ':expect-nisp-class-descriptor)
		(!= require-nisp-class-descriptor
		    (memq (cadr s) '(t true))))
               ((eq (car s) ':initexp)
                (!= initexp s))
               (t
                (cond ((not (memq (car s) '(:all :access :set :both)))
                       (!= s (cons ':all *-*))))
                (cond ((eq (cadr s) 'is)
                       (!= is-def true)))
                (!= realslots (cons s *-*)))))
      (repeat :for (pkg)
;;;;	 (err-out "pkg = " pkg " lisp-class-package* = " lisp-class-package*
;;;;		  :%)
	 (!= pkg (cond ((typep package 'package) package)
		       (t (find-package package))))
       :until pkg
       :result (!= package pkg)
	 (dbg-save :run-loud pkg)
	 (signal-problem lisp-class-as-nisp-type
	    "Importing Lisp class " (:a class-name-string) :%
	    "lisp-class-package* must be set to a package designator"
	    " if lisp-class-as-nisp-type doesn't have an explicit"
	    " :package argument"
	    (:prompt-for "Name of package to set lisp-class-package* to"
			 nil)))
      (let ((importee (find-symbol class-name-string package))
	    (importee-descrip false)
	    (importee-kind ':structure))
	 (cond ((not importee)
		(signal-problem lisp-class-as-nisp-type
		   "Lisp-class name " (:a class-name-string) " not found in "
		   package))
	       ((progn (!= importee-descrip
			   (yt::get-ytools-class-descriptor importee))
		       (and require-nisp-class-descriptor
			    (not importee-descrip)))
		(signal-problem lisp-class-as-nisp-type
		   "Lisp-class name `" (:a class-name-string)
		   "' is not associated with a Nisp class descriptor"
		   (:novalue "I will work with it anyway"))))
	 (!= importee-kind
	     (cond (importee-descrip
		    (yt::ytd-medium importee-descrip))
		   (t ':structure)))
         (cond ((and (null conser)
                     (not initexp)
                     importee-descrip)
                (!= initexp
                    `(:initexp
                      (,(yt::ytd-key-conser importee-descrip))))))
	 (let ((type-def
	       `(defnisptype ,name
		   (,(cond ((eq importee-kind ':object) 'nisp::Object)
			   (t 'Structure))
		    (:already-defined ,(intern class-name-string package))
		    ,@other)
		   ,@conser
		   ,@(include-if (not is-def)
			`(is - Boolean
			     *integrable (x) (,(intern (concatenate 'string
							  (symbol->string
							     '#:is-)
							  class-name-string)
						       package)
					      x)))
		   ,@(repeat :for ((s :in realslots))
		      :nconc
			 (match-cond s
			    ?( (?mode ?slot - ?type ?@defn)
			      (cond ((memq (car defn) '(*integrable :integrable))
				     (list `(,mode ,slot - ,type
					     :integrable ,@(cdr defn))))
				    (t
				     (list `(,mode ,slot - ,type
					     :integrable (x - ,name)
					       (,(cond ((null defn)
							(intern
							    (concatenate 'string
							       class-name-string
							       "-"
							       (symbol->string slot))
							    package))
						       (t (car defn)))
						x))))))
			    (t
			     (signal-problem lisp-class-as-nisp-type
				"Importing Lisp class " (:a class-name-string) :%
				"Ill-formed slot specifier: " s
				(:continue "I will ignore it"))
			     '())))
		   ,@aug-patches)))
	    (cond ((or conser initexp)
		   (let ((new-type-var (gen-var 'new-type))
			 (class-const (build-symbol null "-" (:< name) "*")))
		      `(progn ,type-def
			      (eval-when (:slurp-toplevel)
				 (declaim (special ,class-const)))
			      (eval-when (:compile-toplevel :load-toplevel
					  :execute :slurp-toplevel)
				 #-pre-chunk (:slurp-filter :slurp-nisp-types)
				 (let ((,new-type-var (nisp::designated-type
							 ',name)))
				    (!= (nisp::type-feature
					   ,new-type-var 'ydecl::initexp)
					',class-const)))
			      (defparameter ,class-const
                                 ,(cond (initexp
                                         (cadr initexp))
                                        (t
                                         `(build-minimal-instance ,name)))))))
                  (t type-def))))))

;;; This is similar in content and purpose to 'initexp-const-defn' in
;;; ydecl/strtype.lisp.
(defmacro build-minimal-instance (typename)
   (let ((type (nisp::designated-type typename)))
      (let ((conser-atypes (nisp::type-slot-fun type 'conser 'nisp::atypes true)))
;;;;	 (trace-around bmo
;;;;             (:> "(bmo: " ',typename ")")
          `(make ,typename
		,@(<# ydecl::safe-type-initexp
		      conser-atypes))
;;;;             (:< (val &rest _) "bmo: " val))
          )))
