;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: maxmatch.lisp,v 1.18 2005/12/10 05:45:05 dvm Exp $

(depends-on %module/ nisp)

(depends-on (:at :run-time :slurp-time) %optop/ screamish)

(depends-on (:at :run-time)  %optop/ types domain)

(depends-on :at-run-time %nity/ simpleargs)

(defnisptype Constrained-goal
             (Structure goal - Sexp constraints - (Lst Sexp)
		(:handler
		    (print (cg - Constrained-goal srm - Stream)
		       (with cg
			  (out (:to srm)
			     "#{" 
			     (:q ((null !>constraints)
				  "Simple")
				 (t "Constrained"))
			     " goal " !>goal
			     (:q ((not (null !>constraints))
				  " + " !>constraints))
			     "}"))))))

(defnisptype Maximal-match (Structure subst - Bdgenv
				      ;;;;id - Varid
				      dif - (Lst Constrained-goal)
				      constraints - (Lst Sexp)
				      intsect - (Lst Sexp)))
;;; dif = subgoals that are not true.
;;; intsect = subgoals that are true in the current situation.
;;; We now have to include the intsect, because the set of subgoals out
;;; is not the same as the set we started with, if a reduction occurred.


(specdecl (maxmatch-dbg* false) (show-matches* false)
	  (maxmatch-count* false)
	  (break-on-unsatisfiable-constraints* true)
	  (track-maxmatch-size* true)
;;;;	  plan-dbg*
	  - Boolean
	  plan-dbg-amt* - Fixnum	  
          (maxmatch-excessive* 2000)
	  (maxmatch-explosion* 20)
          (completion-excessive* 50)
	  biggest-maxmatch*
	  - Integer
	  
	  unify-count* unify-success-count*
          (maximal-match-count* 0) (mm-unify-count* 0) - Integer
	  (unachievable-many-switch* 'postpone)
	  - (Const postpone fail warn-and-postpone warn-and-fail))
;;;;          conj* - (Lst Sexp) res* - (Lst (Lrcd Sexp Maximal-match))
;;;;          sit* - Situation dom* - Domain raw* - (Lst Maximal-match))

;;(specdecl alldif* vidx* subst* e* AVOID* dif* - obj)

(specdecl (call-count* 0) - Fixnum)

(defmacro bind-if (test bdgs &rest body)
   (let ((body-fun-name (gensym)))
      `(letrec ((,body-fun-name () ,@body))
	  (cond (,test
		 (bind ,bdgs (,body-fun-name)))
		(t (,body-fun-name))))))

(specdecl (preds-forbidden-in-dif* '(eval eval-test goal-call))
	  - (Lst Symbol)
	  (consider-match-reduction* true)
	  (many-allowed-in-dif* true)
	  (signal-maxmatch-excesses* true)
	  - Boolean)

(deffunc maximal-matches - (Lst Maximal-match) 
                         (conj constraints - (Lst Sexp)
			  vid - Varid env - Bdgenv
			  sit - Situation th - Domain)
;;;;   (!= call-count* (+ *-* 1))
;;;;   (out "maxmatch call-count = " call-count* :% :%)
;;;;   (bind-if (= call-count* 2)
;;;;      ((maxmatch-dbg* true)) ...)
   (!= maximal-match-count* (+ *-* 1))
;;;;   (!= conj* conj)
   (let (;;(VID (NEW-VARID))
         (uc unify-count*)
         (suc unify-success-count*)
	 (num-matches 0)
	 (num-failures 0)
	 (depth-tot 0))
      (let-fun ()
         (let ((matches (mm-avoid conj constraints '() '() '() env 0)))
	    (let ((ud (- unify-count* uc))
		  (sud (- unify-success-count* suc))
		  (fringe-size (+ num-matches num-failures)))
            (!= mm-unify-count* (+ *-* ud))
;;;;            (repeat :for ((mm :in matches) - Maximal-match
;;;;                         (res '()) - (Lst (Lrcd Sexp Maximal-match)))
;;;;               (multiple-value-let (xi env)
;;;;				   (safe-varsubst conj vid (!_subst mm))
;;;;				   (ignore env)
;;;;		  (!= res (cons (lrecord xi mm) *-*)))
;;;;             :result (progn ...))
	    (cond ((= fringe-size 0)
		   (signal-problem maximal-matches
		      "Zero matches and zero failures in max-match of "
		      conj))
		  ((= depth-tot 0)
		   (signal-problem maximal-matches
		      "Zero depth-tot in max-match of "
		      conj))
		  (t
		   (cond ((or maxmatch-dbg*
			      (and (> plan-dbg-amt* 1)
				   track-maxmatch-size*
				   (> fringe-size biggest-maxmatch*)))
			  (!= biggest-maxmatch* fringe-size)
			  (let ((av-depth (/ (float depth-tot) fringe-size)))
			     (err-out
				"Number of leaves in "
				(:q ((not maxmatch-dbg*)
				     "biggest max-match tree so far ")
				    (t
				     "max-match tree "))
				"= "
				fringe-size
				"   = " num-matches " successes + "
					num-failures " failures"
				(:q ((not (= num-matches (len matches)))
				     " [ (len matches) = " (len matches) "]"))
				:% "  Average depth = "
				   (:f "~5,1f" av-depth) " [conj len = "
				(len conj) "]"
				:% "  Branching factor = "
				(:f "~4,2f" (expt fringe-size (/ 1.0 av-depth)))
				:%))))))
	    (dbg-out maxmatch-dbg*
	       sud "/" ud " unifications in match of "
	       (:pp conj) :% "Matches: "
	       (:e (cond ((> num-matches 50)
			  (:o num-matches " maximal matches" :%))
			 ((null matches)
			  (:o "NONE" :%))
			 (t
			  (repeat :for ((r :in matches)
					(i = 1 :by 1))
			    (:o
			     :% 2 "Difference: "
			     (:pp (!_dif r))
			     :% 2 "Intersection: "
			     (:pp (!_intsect r))
			     ;;;;:% 2 "Substitution: "
			     ;;;;    (!_subst r)
			     :%
			     ))))))
	    (cond (maxmatch-dbg*
		   (breakpoint maximal-matches
		      "Number of matches: " num-matches)))
	    (cond ((and signal-maxmatch-excesses*
			(> ud maxmatch-excessive*))
	    ;;;;			      (dbg-save :run-loud conj sit th matches matches)
		   (signal-problem maximal-matches 
		      ud " UNIFICATIONS IN MAXIMAL MATCH OF "
		      conj
		      :proceed)))
	    matches))

       :where

 ;; The recursion from here on out is basically pure depth-first backtracking.
 ;; Each routine guesses a piece of the match and then calls someone else to
 ;; complete it.

 ;; The 'depth' argument to all of the subroutines is the number of choice
 ;; points between the current call and the top call to 'mm-avoid'.  It isn't
 ;; just the number of subroutine calls; there has to be a loop among choices
 ;; involved.  However, we don't insist that the loop consider > 1 choice.

 ;; The Boolean attached to an element of avoid is true if
 ;; it should be included in dif in the end, or just
 ;; checked to make sure it's still false 
 (:def mm-avoid - (Lst Maximal-match) (conj postpone dif intsect
					- (Lst Sexp)
				       avoid - (Lst (Lrcd Boolean Sexp))
				       subst - Bdgenv
				       depth - Fixnum)
    ;; AVOID is a fully varsubsted list of conjuncts that must
    ;; not match in the end
    ;; DIF is a list that we already know will not match.
;;;;    (trace-around mm-avoid
;;;;       (:> "(mm-avoid: dif = " dif ")")
    (dbg-out-indent maxmatch-dbg* 3
       (dbg-out maxmatch-dbg*
	  "Matching " (debug-varsubst conj vid subst)
	    :%  "Avoiding " (debug-varsubst avoid vid subst)
	    :% "Postponing " (debug-varsubst postpone vid subst)
	    ;; :%  "Substitution: [" vid "] " subst
	    :%)
       (cond ((null conj)
	      (let ((mml (complete-maximal-match
			      subst dif postpone intsect avoid depth)))
		 mml))
	     (t
	      (let ((c (car conj))
		    (uc unify-count*) (suc unify-success-count*)
		    nnn)
		 (multiple-value-let (many csl)
				     (deduce-check-many
					c vid subst sit th)
;;;;		    (repeat :for ((e :in csl))
;;;;		       (cond ((has-uvars c vid e)
;;;;			      (dbg-save :run-loud c vid subst e sit th)
;;;;			      (breakpoint mm-avoid
;;;;				 "Max match got answer with variables: " vid 1 c
;;;;				 :% " subst in = " subst
;;;;				 :% " subst out = " e))))
		    (let ((ud (- unify-count* uc))
			  (sud (- unify-success-count* suc)))
		       (dbg-out maxmatch-dbg*
			  "Number known: " (len csl) " Possibly more? "
			  (:q (many "yes") (t "no"))
			  :% " (Unifications: " sud "/" ud ")" :%)
		       (cond ((and (null csl)
				   (matchq (nity:is Number ?nnn)
					   c))
			      (cond ((is-Number nnn)
				     (dbg-save :run-loud sit th c vid (sub subst))
				     (breakpoint mm-avoid
					"Number anomaly: " csl 1 nnn)
				     (!= csl (list subst))))))
		       (let ((c1 (one-value
				    (safe-varsubst c vid subst)))
			     =sym)
			  (cond ((and many
				      (matchq
					 (not (?=sym d-k-lr d-k-lr))
					 c1)
				      (or (eq =sym '=)
					  (and (opt::is-Sym-with-type
						  =sym)
					       (eq (opt::Sym-with-type-actual
						      =sym)
						   '=))))
				 (dbg-save :run-loud c vid subst sit th)
				 (breakpoint mm-avoid
				    "Got many answers for " c))))
		       (nconc (include-first many csl c (cdr conj)
					     postpone dif intsect
					     avoid subst (+ depth 1))
			      (cond (consider-match-reduction*
				     (reduce-first csl c (cdr conj)
						   postpone dif intsect
						   avoid subst (+ depth 1)))
				    (t
				     !()))
			      (match-cond c
				 (:? (forall ?vars ?body)
				    (forall-exclude
				        many csl c vars body
					(cdr conj) postpone dif
					intsect avoid subst (+ depth 1)))
				 (:? (not (exists ?vars ?body))
				    (forall-exclude
				        many csl c vars `(not ,body)
					(cdr conj) postpone dif
					intsect avoid subst (+ depth 1)))
				 (t
				  (exclude-first many csl c (cdr conj)
						 postpone dif intsect
						 avoid subst (+ depth 1)))))))))))
;;;;       (:< (mml) "mm-avoid returns matches: " mml))
    )

 ;; Should be called 'include-or-postpone-first'.
 (:def include-first - (Lst Maximal-match)
		(many - Boolean csl - (Lst Bdgenv) c - Sexp
		 r postpone dif intsect - (Lst Sexp)
		 avoid - (Lst (Lrcd Boolean Sexp))
		 prev - Bdgenv depth - Fixnum)
     (cond (many
	    (dbg-out maxmatch-dbg*
	       2 "Postponed " c
	       :% 4 "because it doesn't have enough variables bound"
	       :%)
	    (mm-avoid r
		      (cons c postpone)
		      dif intsect avoid prev depth))
	   (t
	    (repeat :for (s :in csl)
		  (dbg-out maxmatch-dbg*
		      "Got: " (debug-varsubst c vid s) :%)
	       :nconc
		  (impose-new-env s prev r postpone dif
				  (cons c intsect)
				  false avoid (+ depth 1))))))

 (:def reduce-first - (Lst Maximal-match)
	       (csl - (Lst Bdgenv) c - Sexp
		r postpone dif intsect - (Lst Sexp)
		avoid - (Lst (Lrcd Boolean Sexp))
		subst - Bdgenv depth - Fixnum)
    (cond ((exists (e in csl)
	      (not (binds-any-new e vid subst)))
	   (dbg-out maxmatch-dbg*
	       2 "*Dead end reducing " c 
	       :% 4 "because it's true in the most general way it can be,"
	      :% 4 "and therefore not worth back-chaining" :%)
	   (fail depth))
	  ((plausibly-reducible (be CSymbol (car c)) th)
	   (let ((rl (find-reductions c vid subst
				      (domain-place-fact-sit th true)
				      th)))
	      (cond ((null rl)
		     (dbg-out maxmatch-dbg*
			2 "*Dead end reducing " c
			:% 4 "because it has no reductions")
		     (fail depth))
		    (t
		     (repeat :for ((red :in rl))
			(dbg-out maxmatch-dbg*
			   2 "Reducing " c " to: "
			   :% 3 (debug-varsubst
				  (car red) vid (cadr red)) :%)
		      :nconc
			(impose-new-env (cadr red) subst
					(append (car red) r)
					postpone dif intsect
					(lrecord false c)
					avoid (+ depth 1)))))))
	  (t
	   (dbg-out maxmatch-dbg*
	      2 "*Dead end reducing " c
	      :% 4 "because nothing that can imply it is achievable")
	   (fail depth))))

 ;; include and reduce have this in common:
 ;; They usually bind some of the vars, giving us a chance to
 ;; retest postponed elements and check that nothing in 'avoid'
 ;; has become unavoidable.  'impose-new-env' handles that bind/check
 ;; operation.
 ;; Note that there are no choice points in this routine.
 (:def impose-new-env - (Lst Maximal-match)
		 (e prev - Bdgenv
		  r postpone dif intsect - (Lst Sexp)
		  new-avoid-element - (~ (Lrcd Boolean Sexp))
		  avoid - (Lst (Lrcd Boolean Sexp))
		  depth - Fixnum)
		 (ignore prev)
    (multi-let (((new-r e)
		 (safe-varsubst r vid e)))
       (cond ((repeat :for ((new-g :in new-r)
			    (old-g :in r))
	       :result true
	       :until (and (not (variants new-g old-g vid vid e))
			   (not (achievable new-g th))
			   (multi-let (((many al)
					(deduce-check-many
					   new-g vid e sit th)))
			      (and (null al) (not many))))
	       :result (progn
			  (dbg-out maxmatch-dbg*
			   2 "*Dead end because progress on first"
			   " conjunct makes this"
			   :% 2 "unachievable conjunct false: "
			   new-g
			   :%)
			  (fail depth)))
	      (multi-let (((postpone e)
			   (safe-varsubst postpone vid e)))
		 (multi-let (((avoid e)
			      (safe-varsubst avoid vid e)))
		    (multi-let (((new-avoid new-dif - (Lst (Lrcd Boolean Sexp)))
				 (classify
				    (cond (new-avoid-element
					   (cons new-avoid-element
						 avoid))
					  (t avoid))
				    (\\ (p)
				       (has-qvars (cadr p)))))
				((new-postpone new-try)
				 (classify postpone #'has-qvars)))
		       (repeat :for ((a :in new-dif)
				     - (Lrcd Boolean Sexp)
				    (unwanted '()) - (Lst Bdgenv))
			:result (cond ((null unwanted)
				       (mm-avoid
					  (nconc (reverse new-try) new-r)
					  new-postpone
					  (nconc (<! (\\ (nd - (Lrcd Boolean Sexp))
							(cond ((car nd)
							       (list
								  (cadr nd)))
							      (t '())))
						    new-dif)
						 dif)
					  intsect new-avoid e depth))
				      (t
				       (signal-problem impose-new-env
					  "Can't stop! " unwanted)))
			  (!= unwanted (deduce (cadr a) vid e sit th))
			:while  (null unwanted)
			:result (progn
				  (dbg-out maxmatch-dbg*
				     2 "*Dead end because progress on first"
				     " conjunct make this"
				     :% 2 "excluded conjunct unavoidably"
				     " true: " (cadr a)
				     :%)
				  (fail depth)))))))
	     (t !()))))

 (:def exclude-first - (Lst Maximal-match) 
		(many - Boolean csl - (Lst Bdgenv) c - Sexp
		 r postpone dif intsect - (Lst Sexp)
		 avoid - (Lst (Lrcd Boolean Sexp))
		 subst - Bdgenv depth - Fixnum)
    (let ((forbidden-in-dif (memq (car c) preds-forbidden-in-dif*)))
       (cond ((achievable c th)
	      (multiple-value-let (c subst)
				  (safe-varsubst c vid subst)
		 (cond ((and (or many-allowed-in-dif* (not many))
			     (has-qvars c))
			(dbg-out maxmatch-dbg*
			   "Excluding " c :%)
			(mm-avoid r postpone dif intsect
				  (cons (tuple (not forbidden-in-dif)
					       c)
					avoid)
				  subst depth))
		       ((null csl)
			(cond (forbidden-in-dif
			       (dbg-out maxmatch-dbg*
				  2 "*Dead end avoiding " c
				  " because it can't be a difference"
				  :%)
			       (fail depth))
			      (t
			       (dbg-out maxmatch-dbg*
				  2 "Difference: "
				  (debug-varsubst c vid subst) :%)
			       (mm-avoid r postpone (cons c dif) intsect avoid
					 subst depth))))
		       (t
			(dbg-out maxmatch-dbg*
			   "*Dead end avoiding " c " because it is true"
			   " and variable-free " :%)
			(fail depth)))))
	     ((or many (not (null csl)))
	      (cond (many
		     ;; Already postponed by 'include-first'
		     !())
		    (t
		     (dbg-out maxmatch-dbg*
			 2 "*Dead end: Can't exclude "
			(debug-varsubst c vid subst)
			" [true and unachievable]" :%)
		     (fail depth))))
	     (t
	      (dbg-out maxmatch-dbg*
		 2 "*Dead end: Can't exclude "
		 (debug-varsubst c vid subst)
		 " [false and unachievable]" :%)
	      (fail depth)))))

 (:def forall-exclude - (Lst Maximal-match) 
		  (many - Boolean csl - (Lst Bdgenv)
		   c - Sexp ;; -- a 'forall' or not-'exists' goal
		   vars body r postpone dif intsect - (Lst Sexp)
		   avoid - (Lst (Lrcd Boolean Sexp))
		   subst - Bdgenv depth - Fixnum)
;;;;    (trace-around forall-exclude
;;;;       (:> "(forall-exclude: "
;;;;	   (:q (many "[many] "))
;;;;	   :% c
;;;;	   :% " Answers: " csl ")")
    (cond ((and (not many) (null csl))
	   (let ((neg-dnf (forall-refuter-dnf vars body))
		 - (Lst (Lst Sexp)))
	      ;; 'csl' is envs that make 'neg-dnf' true; each makes
	      ;; at least one conjunction in 'neg-dnf' true.
	      (let ((counterexamps
		       (counterexamp-achievables neg-dnf vid (empty-env) sit th))
		    - (Lst (Lst Sexp)))
;;;;		 (dbg-save c subst vid sit th neg-dnf)
;;;;		 (breakpoint forall-exclude
;;;;		    "counterexamps = " counterexamps)
		 (repeat :for ((c-examp :in counterexamps))
		  :nconc
		     (progn
		         (dbg-out maxmatch-dbg*
			    2 "Differences: " c-examp :%)
			 (mm-avoid r postpone (append c-examp dif) intsect avoid
				   subst (+ depth 1)))))))
	  (t
	   (cond ((not (null csl))
		  (dbg-out maxmatch-dbg*
		     "*Dead end avoiding " c " because it is true"
		     " and variable-free " :%)))
	   (fail depth)))
;;;;       (:< (val &rest _) "forall-exclude: " val))
    )

 ;;; First value is total number found so far, including the ones found
 ;;; by this call to 'complete-maximal-match'.
 (:def complete-maximal-match - (Lst Maximal-match)
				 (subst - Bdgenv
				  dif postponed intsect - (Lst Sexp)
				  avoid - (Lst (Lrcd Boolean Sexp))
				  depth - Fixnum)
;;;;    (trace-around complete-maximal-match
;;;;       (:> "(complete-maximal-match: " dif ")")
    (repeat :for ((a :in avoid) - (Lrcd Boolean Sexp)
		  al - (Lst Bdgenv) many - Boolean)
     :result (bind-remaining-variables)
        (!= < many al > (deduce-check-many (cadr a) vid subst sit th))
     :until (not (null al))
     :result (progn
	       (dbg-out maxmatch-dbg*
		  "*Dead end because of truth of " (cadr a) :%)
	       (fail depth))
	;; We can't show it's not true, so it becomes a constraint
        ;; (If (car a) is true, then it's going to be counted as a dif below,
        ;; so don't count it here.)
	(cond ((and many (not (car a)))
	       (!= postponed (cons (constraint-negation (cadr a)) *-*))))

     :where

       (:def bind-remaining-variables - (Lst Maximal-match) ()
	   (let ((alldif (nconc (<! (\\ (a) (cond ((car a) (list (cadr a)))
						   (t '())))
				    avoid)
				dif)))
	      (combine-completions alldif !() intsect postponed
				   subst 1 depth))))
;;;;       (:< (val &rest _) "complete-maximal-match: " val))
    )

  ;;; By "completion" is meant eliminating as many variables as possible
  ;;; filling in all known objects of the correct types.
  ;;; Some goals can't be completed, so their variables stick around.
  ;;; 'tree-explosion' is estimate of complexity of the the tree of
  ;;; possibilities.  If there are n variables, and each has k objects
  ;;; of that type, we can wind up with k^n fillers; if that is
  ;;; apparently happening, then we switch to considering their domains
  ;;; to be infinite.
  (:def combine-completions - (Lst Maximal-match)
			       (unk-templates dif intsect postponed - (Lst Sexp)
				subst - Bdgenv
				tree-explosion - Integer depth - Fixnum)
;;;;     (trace-around combine-completions
;;;;	(:> "(combine-completions: " unk-templates " + " dif " + " postponed ")")
     (cond ((> tree-explosion maxmatch-explosion*)
	    (cond (signal-maxmatch-excesses*
		   (signal-problem combine-completions
		      "Too many possibilities for unfilled variables in maxmatch: "
		      tree-explosion
		      (:proceed "I will treat remaining differences as constraints"))))
	    (!= postponed (append unk-templates *-*))
	    (!= unk-templates !())))
     (cond ((null unk-templates)
	    (verify-constraints-satisfiable
		      dif intsect postponed subst depth))
	   (t
	    (multi-let (((d subst)
			 (safe-varsubst (car unk-templates) vid subst)))
	       (multi-let (((polarity fmla)
			    (lit-analyze d)))
		  (let ((fty (see-function-constant (car fmla) th)))
		     (cond (fty
			    (multi-let (((found-argtypes argtypes)
					 (nisptype::simple-funtype-argtypes fty)))
			       (cond (found-argtypes
;;;;				      (dbg-save fty fmla th)
;;;;				      (breakpoint combine-completions
;;;;					 "Got argtypes " argtypes)
				      (complete-match-with-argtypes
					 argtypes polarity fmla
					 unk-templates dif intsect postponed subst
					 tree-explosion depth))
				     (t
				      (cond (signal-maxmatch-excesses*
					     (signal-problem combine-completions
						"Function too complex: " (car fmla)
						" in attempt to complete maximal match"
						(:proceed
						 "I will give up on this maximal match"))))
				      (fail depth)))))
			   (t
			    (signal-problem combine-completions
			       "Nonfunction " (car fmla)
			       " in maximal match"))))))))
;;;;	(:< (val &rest _) "combine-completions: " val))
     )

  ;; 'fmla' is the first difference being used as a template to bind
  ;; unknowns.  
 ;; If 'fmla' has unbound variables, use their expected types to generate
 ;; objects to fill them.
 ;; In addition to being used as a template, 'fmla' will wind up on the 'dif'
 ;; list, unless it has an infinite number of possible args, and must wind up
 ;; as a constraint.
 (:def complete-match-with-argtypes - (Lst Maximal-match)
			   (atypes - (Lst Opt-type)
			    polarity - Boolean fmla - Sexp
			    unk-templates dif intsect postponed - (Lst Sexp) e - Bdgenv
			    tree-explosion - Integer
			    depth - Fixnum)
;;;;    (trace-around complete-with-argtypes
;;;;       (:> "(complete-with-argtypes: " atypes
;;;;	      :% unk-templates " + " dif " + " postponed ")")
    (let ((uc unify-count*) (suc unify-success-count*)
	  (args-with-uvars (<# (\\ (a) (has-uvars a vid e))
			       (cdr fmla))))
       (cond ((exists (ty :in atypes)
		      (arg-has-uvars :in args-with-uvars)
		 (and arg-has-uvars
		      (cond ((eq ty univ-type*)
			     (domain-has-infinite-obj-terms th))
			    (t
			     (nisptype::type-feature ty 'infinite)))))
	      ;; It's numerical (presumably), or infinite for some other
	      ;; reason.
	      ;; All we can do is make it a constraint to be handled by
	      ;; Screamer at some later point
	      ;;;;(err-out "Infinite type among " atypes)
	      (combine-completions (cdr unk-templates) dif intsect
				   (cons (lit-synthesize polarity fmla)
					 postponed)
				   e tree-explosion depth))
	     ((exists (a :in args-with-uvars)
		  a)
	      (let ((filters (<! (\\ (ty a hv)
				    (cond (hv
					   (let ((p (type-predify
						       ty a
						       (nisptype::empty-vartypes th))))
					      (cond ((matchq (nity:is nity:Obj ?_)
							     p)
						     ;; Trivial
						     !())
						    (t (list p)))))
					  (t '())))
				 atypes
				 (cdr fmla)
				 args-with-uvars)))
		 (cond ((null filters)
			;; We can't enumerate all objects, so punt in this case
			;; too --
			;;;;(err-out "Can't filter args of " fmla)
			(combine-completions (cdr unk-templates) dif intsect
					     (cons (lit-synthesize polarity
								   fmla)
						   postponed)
					     e tree-explosion depth))
		       (t
			(dbg-out maxmatch-dbg*
			  "Args to difference " fmla
			  :% " must be filtered through "
			  (debug-varsubst filters vid e) :%)
			(let ((r (deduce `(and ,@filters)
					 vid e sit th)))
			  (let ((ud (- unify-count* uc))
				(sud (- unify-success-count* suc)))
			     (dbg-out maxmatch-dbg*
				      sud "/" ud " unifications" :%)
			     (cond ((null r)
				    (dbg-out maxmatch-dbg*
				       "*Dead end because nothing"
				       " satisfies type constraints "
				       :% " on args to " :% fmla :%)
				    (fail depth))
				   (t
				    (complete-given-bindings
				       r polarity fmla filters 
				       unk-templates dif intsect postponed
				       tree-explosion depth)))))))))
	     (t
	      (combine-completions
	         (cdr unk-templates)
		 (cons (lit-synthesize polarity fmla)
		       dif)
		 intsect postponed e
		 tree-explosion (+ depth 1)))))
;;;;       (:< (val &rest _) "complete-with-argtypes: " val))
    )

 (:def complete-given-bindings - (Lst Maximal-match)
			   (r - (Lst Bdgenv) polarity - Boolean fmla - Sexp
			    filters 
			    unk-templates dif intsect postponed - (Lst Sexp)
			    tree-explosion - Integer
			    depth - Fixnum)
    (let ((instl (dupfree-instances r vid filters)))
       (dbg-out maxmatch-dbg*
		"num instances: " (len instl) " [from " r "]" :%)
       (repeat :for ((c :in instl) - Bdgenv
		     :collector matches - Maximal-match)
	  (dbg-out maxmatch-dbg* 
		   "Constraints on args to " fmla
		   :%  " satisfied by "
		   (debug-varsubst filters vid c)
		   :%)
	:within
	  (dbg-out-indent maxmatch-dbg* 3
	     (:continue
	      :nconc (combine-completions
			     (cdr unk-templates)
			     (cons (lit-synthesize polarity fmla)
				   dif)
			     intsect postponed c 
			     (* tree-explosion (len instl))
			     (+ depth 1))))
	:result matches)))

 (:def verify-constraints-satisfiable - (Lst Maximal-match)
                             (dif intsect postponed - (Lst Sexp)
			      e - Bdgenv depth - Fixnum)
;;;;    (trace-around verify
;;;;       (:> "(verify: " dif ")")
    (multi-let (((d e)
		 (safe-varsubst dif vid e)))
       (multi-let (((i e)
		    (safe-varsubst intsect vid e)))
	  (multi-let (((p e)
		       (safe-varsubst postponed vid e)))
	     (multi-let (((cdifs constraints)
			  (constraints-allot-to-differences
			     (<# (\\ (e)
				    (make Constrained-goal
					  e !()))
				 d)
			     p sit th)))
		(multi-let (((consis e)
			     (constraints-satisfy
				false constraints vid e sit th)))
		   (cond (consis
			  (!= num-matches (+ *-* 1))
			  (!= depth-tot (+ *-* depth))
			  (dbg-out (or maxmatch-dbg*
				       show-matches*)
			     "[" num-matches "] **Match! -> " :%
			     2 "Dif = " cdifs :%)
			  (cond ((and (>= num-matches completion-excessive*)
				      signal-maxmatch-excesses*)
				 (signal-problem combine-completions
				    num-matches " maxmatch completions"
;;;;				    " [explosion = " tree-explosion "]"
				    :% " (Increase completion-excessive*"
				       " to avoid seeing this message again."
				    (:proceed
					    "I'll keep generating them"))))
			  ;; This is the only place we succeed --
			  (list (make Maximal-match
				      e cdifs constraints i)))
			 (t
			  ;; This is just the last place we can fail --
			  (dbg-out maxmatch-dbg*
			     "*Dead end because constraints "
			     :% constraints
			     :% " cannot be satisfied")
			  (cond (break-on-unsatisfiable-constraints*
				 (dbg-save :run-loud constraints vid e sit th)
				 (breakpoint verify-constraints-satisfiable
				    "Constraints "
				    :% constraints
				    :% " unsatisfiable"
				       " in maxmatch")))
			  (fail depth))))))))
;;;;   (:< (val &rest _) "verify: " val))
)

 (:def fail - (Lst Maximal-match) (depth - Fixnum)
    (!= num-failures (+ *-* 1))
    (!= depth-tot (+ *-* depth))
    !()))))

(deffunc counterexamp-achievables - (Lst (Lst Sexp))
	                  (disjuncts - (Lst (Lst Sexp))
			   ;; -- dnf of neg of thing to make true
			   id - Varid env - Bdgenv
			   sit - Situation dom - Domain)
   (let-fun ()
     (<# (\\ (occl - (Lst Occasion)) (<# (\\ (occ) (literal-neg (!_propo occ)))
					 (nodupq occl)))
	 (let ((cl (conjunctions-made-true disjuncts id env sit dom)))
	    (cond ((null cl)
		   (signal-problem counterexamp-achievables
		      "In " `(or ,@(<# (\\ (conjuncts) `(and ,@conjuncts))
				       disjuncts))
		      :% "  this env makes no conjunction true: "
		      :% 2 env)))
	    (one-achievable-per cl dom)))))

;;;; (:def achievables - (Lst (Lst Occasion)) (counter-envs - (Lst Bdgenv))
;;;;    (cond ((null counter-envs)
;;;;	   (list !()))
;;;;	  (t
;;;;	   (repeat :for ((env :in counter-envs))
;;;;	    :within
;;;;	      (let ((cl (conjunctions-made-true disjuncts id env sit dom)))
;;;;		 (cond ((null cl)
;;;;			(signal-problem counterexamp-achievables
;;;;			   "In " `(or ,@(<# (\\ (conjuncts) `(and ,@conjuncts))
;;;;					    disjuncts))
;;;;			   :% "  this env makes no conjunction true: "
;;;;			   :% 2 env)))
;;;;		 (let ((ach-for-rest (achievables (cdr counter-envs))))
;;;;	            (:continue
;;;;		     :append
;;;;			(<! (\\ (occl-for-env)
;;;;			       (<# (\\ (occl-from-rest)
;;;;				      (append occl-for-env occl-from-rest))
;;;;				   ach-for-rest))
;;;;			    (one-achievable-per cl dom)))))))))

;;; Each element of 'conjunctions' is a (probably variable-free) formula,
;;; already known to be made true by env.
;;; Returns a list of ways of falsifying all the conjunctions; that is,
;;; a list of one elment per conjunction whose truth value is changeable.
;;; Returns () if there is no way to make them all false;
;;; Returns (()) if 'conjunctions' is empty, so that there is only
;;; way to falsifying them -- do nothing.
(deffunc one-achievable-per - (Lst (Lst Occasion))
                         (conjunctions - (Lst (Lst Occasion)) dom - Domain)
    (cond ((null conjunctions)
	   (list !()))
	  (t
	   (let ((achievables-here
		    (repeat :for ((conjunct :in (car conjunctions)))
		     :when (achievable (!_propo conjunct) dom)
		     :collect conjunct)))
	      (cond ((null achievables-here)
		     !())
		    (t
		     (let ((one-per-rem (one-achievable-per (cdr conjunctions)
							    dom)))
			(<! (\\ (occ) (<# (\\ (occl) (cons occ occl))
					  one-per-rem))
			    achievables-here))))))))

;;; Returns dnf of occasions without variables.
(deffunc conjunctions-made-true - (Lst (Lst Occasion))
                         (disjuncts - (Lst (Lst Sexp)) id - Varid env - Bdgenv
			  sit - Situation dom - Domain)
    (repeat :for ((conjuncts :in disjuncts))
     :nconc
        (repeat :for ((ans-env :in (deduce `(and ,@conjuncts) id env sit dom)))
	 :collect (repeat :for ((c :in conjuncts))
		   :within
		      (multi-let (((c _)
				   (safe-varsubst c id ans-env)))
			 (cond ((not (has-qvars c))
				(:continue
				 :collect (uniquify-occasion c dom)))))))))

; Return true if postpone 
(deffunc many-bug - Boolean
	          (c conj - Sexp vid - Varid subst - Bdgenv)
   (cond ((memq unachievable-many-switch* '(warn-and-postpone warn-and-fail))
	  (cond ((eq unachievable-many-switch* 'warn-and-fail)
		 (signal-problem many-bug
			      "Can't tell if " (debug-varsubst c vid subst)
			      " is a difference or not in " conj :%
		    (:proceed "The maximal match will fail")))
		(t
		 (signal-problem many-bug
			      "Can't tell if " (debug-varsubst c vid subst)
			      " is a difference or not in " conj :%
                    (:proceed "The goal will be postponed"))))))
   (memq unachievable-many-switch* '(postpone warn-and-postpone)))



(deffunc constraint-negation - Sexp (c - Sexp)
   (cond ((atom c) `(not ,c))
	 ((eq (car c) 'not) (cadr c))
	 ((is Ineq-rel (car c))
	  `(,(ineq-rel-negation (be Ineq-rel (car c))) ,@(cdr c)))
	 (t `(not ,c))))

(specdecl discrim-bdgs* - Bdgenv)

(deffunc see-function-constant - (~ Opt-type) (x - Obj dom - Domain)
   (let ((ty (cond ((is Sym-with-type x)
		    (!_type x))
		   (t
		    (let ((b (find-domain-bdg x dom)))
		       (cond (b (!_type b))
			     (t
			      (signal-problem see-function-constant
				 "Undefined predicate " x
				 " in maximal match"))))))))
      (multiple-value-let (ty tenv)
			  (nisptype::follow-var-ref ty (empty-vartypes dom))
	 (cond ((nisptype::is-funtype ty)
		(nisptype::new-Type-closure ty tenv))
	       (t false)))))



;;; So far this filters only [one] zero cases...
(deffunc deduce-filter-trivial - (Mlv Boolean (Lst Bdgenv))
	                (goal - Sexp id - Varid env - Bdgenv
			 sit - Situation dom - Domain)
   (deduce-check-many goal id env sit dom))
;;;;   (multi-let (((many ansl)
;;;;		(deduce-check-many goal id env sit dom)))
;;;;      (match-cond goal
;;;;	 ?( (= ?l ?r)
;;;;	   ;; The answer "x=x" is never useful
;;;;	   (!= ansl
;;;;	       (<? (\\ (a)
;;;;		      (not (variants l r id id a)))
;;;;		   *-*))))
;;;;      (values many ansl)))

;; Returns r with duplicates removed.
(deffunc dupfree-instances - (Lst Bdgenv)
	                  (r - (Lst Bdgenv) id - Varid x - Sexp)
   (<# cadr (envs-add-to-instance-tab r x id !())))

(deffunc envs-add-to-instance-tab - (Lst (Lrcd Sexp Bdgenv))
	             (r - (Lst Bdgenv) x - Sexp id - Varid
		      tab - (Lst (Lrcd Sexp Bdgenv)))
   (repeat :for ((c :in r) - Bdgenv
		 (res tab) - (Lst (Lrcd Sexp Bdgenv)))
      (multiple-value-let (xi env)
			  (safe-varsubst x id c)
	 (cond ((not (exists (p :in res)
			(variants (car p) xi id id (cadr p))))
		(!= res (cons (lrecord xi env)
			      *-*)))))
    :result res))

(specdecl (handleable-constraint-fcns* '(+ - * / =~ /=~ > >= < =< <=
					 same-sign))
	  - (Lst Symbol))

(defproc maximal-match-constrain - (Mlv Boolean (~ Maximal-match))
                              (mm - Maximal-match constraints - (Lst Sexp)
			       mm-id cl-id - Varid env - Bdgenv
			       sit - Situation dom - Domain)
   (multi-let (((constraints env)
		(varid-convert constraints cl-id mm-id env)))
      (multi-let (((difs extra-constraints)
		   (constraints-allot-to-differences
		      (<# (\\ (d)
			     (make Constrained-goal
				   (!_goal d)
				   (list-copy (!_constraints d))))
			  (!_dif mm))
		      constraints sit dom)))
	 (multi-let (((consis env)
		      (constraints-satisfy true extra-constraints mm-id
					   env sit dom)))
	    (cond (consis
		   (values true (make Maximal-match
				      env difs !() (!_intsect mm))))
		  (t
		   (values false nil)))))))

(specdecl (numerify-early* false) - Boolean)

;;; Side-effects the dif's, which should usually be freshly copied 
;;; before 'constraints-allot-to-differences' is called.    
(defproc constraints-allot-to-differences - (Mlv (Lst Constrained-goal)
						 (Lst Sexp))
					  (dif - (Lst Constrained-goal)
					   constraints - (Lst Sexp)
					   sit - Situation dom - Domain)
   (letrec ()
      (let ((constraint-recs
	        (<# (\\ (p)
		       (tuple false
			      (cond ((or (not numerify-early*)
					 (not (numerifiable p)))
				     p)
				    (t (numerify p sit dom)))))
		    constraints))
	    - (Lst (Lrcd Boolean Sexp)))
	 (repeat :for ((d :in dif))
	    (let ((pl constraint-recs)
		  (dvl (raw-uvars (!_goal d)))
		  (cl !(Sexp)))
	       (repeat :for (sharevars disjoint
			      - (Lst (Lrcd Boolean Sexp))
			     qvl - (Lst Symboid))
		  (!= qvl (union dvl (raw-uvars cl)))
		  (!= < sharevars disjoint >
		      (classify pl
				(\\ (p)
				   (exists (v :in (raw-uvars
						     (cadr p)))
				       (memq v qvl)))))
		:until (null sharevars)
		  (repeat :for ((c :in sharevars))
		     (!= (car c) true))
		  (!= cl (nconc (<# cadr sharevars) *-*))
		  (!= pl disjoint)
		:result (!= (!_constraints d)
			    (nconc *-* cl)))))
	 (values
	    dif
	    (repeat :for ((c :in constraint-recs))
	     :when (not (car c))
	     :collect (cadr c))))
		    
;;;;	       (cond ((not (car c))
;;;;		      (dbg-save dif)
;;;;		      (signal-problem constraints-allot-to-differences
;;;;			 "Constraint " (cadr c)
;;;;			 :% " shares no variables with any difference"
;;;;			 (:continue "I will discard it")))))

))

;;;;			   (signal-problem constraints-allot-to-differences
;;;;			      "Can't find value for "
;;;;			      c)


(specdecl screamer:screamer-funs* screamer:screamer-preds*
	  - (Lst Symbol))

;;; If 'must-find', then return Bdgenv that makes them all true.
;;; Otherwise, just verify that there probably is such a Bdgenv.
(deffunc constraints-satisfy - (Mlv Boolean Bdgenv)
	              (must-find - Boolean
		       constraints - (Lst Sexp) id - Varid env - Bdgenv
		       sit - Situation dom - Domain)
   (cond ((null constraints)
	  (values true env))
	 (t
	  (!= < constraints env > (safe-varsubst constraints id env))
	  ;; First collect all the variables.
	  (repeat :for ((c :in constraints)
			(scr-vars !((Lrcd Symboid Obj (~ Symbol)))))
	     (match-cond c
		?( (opt:is ?ty ?v)
		  (cond ((is CSymbol ty)
			 (cond ((is Qvar v)
				(make-var-trip (!_sym v) ty))
			       ((and signal-maxmatch-excesses*
				     (has-qvars v))
				(signal-problem constraints-satisfy
				   "Type constraint too complex: " c
				   (:proceed "I will ignore it")))))
			(signal-maxmatch-excesses*
			 (signal-problem constraints-satisfy
			    "Unintelligible type constraint: "
			    c
			    (:continue "I will ignore it"))))))
	   :result
	      ;; Now collect the other constraints
	      (repeat :for ((c :in constraints)
			    (abort false)
			    :collectors scr-constraints ded-constraints)
	       :result (multi-let (((scr-okay scr-env)
				    (screamer-constraints-solve
				        scr-constraints)))
			  (cond (scr-okay
				 (deductive-constraints-solve
				     ded-constraints scr-env))
				(t
				 (values false nil))))
		  (cond ((and (not (matchq (is ?_ ?_) c))
			     (has-qvars c))
			(multi-let (((okay qvars-replaced)
				     (prepare-for-screamer c)))
			   (cond (okay
;;;;				  (out "New screamer constraint: " c :%)
				  (one-collect scr-constraints
					       qvars-replaced))
				 (t
;;;;				  (out "New deductive constraint: " c :%)
				  (one-collect ded-constraints c)))))
		       (t
			(signal-problem constraints-satisfy
			   "Match difference " c
			   :% "  misclassified as a constraint"
			   (:proceed "I will treat it as unsatisfiable"))
			(!= abort true)))
	       :until abort
	       :result (values false nil)

	       :where

             ;; First val is false if it finds something screamer can't
	     ;; handle
	     (:def prepare-for-screamer - (Mlv Boolean Sexp) (e - Sexp)
;;;;		(trace-around prep-scr
;;;;		   (:> " (prep-scr: " e ")")
		(cond ((is Qvar e)
		       (let ((trip (assq (!_sym e) scr-vars)))
			  (cond (trip
				 (values true (trip-scr-var trip)))
				(t
				 (dbg-save :run-loud (qv e) scr-vars)
				 (signal-problem constraints-satisfy
				    "Fumbled variable " e
				    (:continue "I will assume it's a Float"))
				 (values true
					 (trip-scr-var 
					    (make-var-trip
					       (!_sym e) 'Float)))))))
		      ((atom e) (values true e))
		      ((or (memq (car e) screamer:screamer-funs*)
			   (memq (car e) screamer:screamer-preds*))
		       (repeat :for ((x :in (cdr e))
				     sub  - Sexp okay - Boolean
				     :collectors res - Sexp)
			  (!= < okay sub > (prepare-for-screamer x))
			:result (values true `(,(car e) ,@res))
			:while okay
			:result (values false nil)
			:collect sub))
		      ((numerifiable e)
		       (let ((ne (numerify e sit dom)))
			  (cond ((equal ne e)
				 (values false nil))
				(t (prepare-for-screamer ne)))))
		      (t
		       (values false nil)))
;;;;		   (:< (okay newcon) "prep-scr: " okay 1 newcon))
		)

	     (:def trip-scr-var - Obj (trip - (Lrcd Symboid Obj Symbol))
		(cond ((not (cadr trip))
		       (!= (cadr trip)
			   (screamer-var-of-type (caddr trip) (car trip)))))
		(cadr trip))

             (:def screamer-constraints-solve - (Mlv Boolean Bdgenv)
					 (scr-constraints - (Lst Sexp))
;;;;		(trace-around screamer-constraints-solve
;;;;		   (:> "(screamer-constraints-solve: " scr-constraints ")")
		(cond ((and (null scr-constraints)
			    (not must-find))
		       (values true env))
		      (t
		       (let ((s (screamer::screamer-solve
				   scr-constraints (<# cadr scr-vars)))
			     - (Lst Number))
			  (cond ((eq s ':no-solution)
				 (values false env))
				(t
				 (values true
					 (make Bdgenv
					       (nconc (new-bdgs s)
						      (!_bdgs env)))))))))
;;;;		   (:< (val &rest _) "screamer-constraints-solve: " val))
		)

	     (:def deductive-constraints-solve - (Mlv Boolean Bdgenv)
					  (ded-constraints - (Lst Sexp)
					   envsofar - Bdgenv)
;;;;		(trace-around deductive-constraints-solve
;;;;		   (:> "(deductive-constraints-solve: " ded-constraints ")")
		(cond ((null ded-constraints)
		       (values true envsofar))
		      (t
		       (multi-let (((many al)
				    (deduce-check-many
				         `(and ,@ded-constraints)
					 id envsofar sit dom)))
			  (cond ((null al)
				 (values (and many (not must-find))
					 envsofar))
				(t
				 (values true (car al)))))))
;;;;		   (:< (val &rest _) "deductive-constraints-solve: " val))
		)

	     (:def new-bdgs - (Lst Varbdg) (s - (Lst Number))
		(repeat :for ((val :in s)
			      (trip :in scr-vars))
		 :collect
		    (make Varbdg
			  (car trip)
			  id
			  (make Expclo
				(case (caddr trip)
				   ((opt:Float opt:Number)
				    (float val))
				   (opt:Integer
				    (cond ((is-Integer val)
					   val)
					  (t
					   (cond (signal-maxmatch-excesses*
						  (signal-problem constraints-satisfy
						     "Noninteger value: "
						     val
						     (:continue "I'll round it"))))
					   (round val))))
				   (t
				    (signal-problem constraints-satisfy
				       "Fumbled constraint var " trip)))
				false)))))
	     :where
	        (:def make-var-trip - (Lrcd Symboid Obj Symbol)
			    (name - Symboid typelabel - Obj)
		   (let ((trip (assq name scr-vars)))
		      (cond ((not trip)
			     (cond ((langutils::is-Canonized-symbol typelabel)
				    (!= typelabel (langutils:Canonized-symbol-sym
						      typelabel))))
			     (cond ((not (is Symbol typelabel))
				    (signal-problem constraints-satisfy
				       "Can't handle " typelabel)))
			     (!= trip
				 (tuple name
					(screamer-var-of-type typelabel name)
					typelabel))
			     (!= scr-vars (cons trip *-*))))
		      trip))))))

(deffunc numerifiable - Boolean (p)
   (not (matchq ?(:\| (opt:is ?_ ?_)
		     (eval ?_ ?_)
		     (eval-test ?_)
		     (feasible ?@_))
		p)))

(deffunc numerify - Sexp (c - Sexp sit - Situation dom - Domain)
	  (cond ((is Qvar c) c)
		((has-qvars c)
		 (cond ((memq (car c) handleable-constraint-fcns*)
			`(,(car c)
			  ,@(<# (\\ (d) (numerify d sit dom))
				(cdr c))))
		       (t
			(signal-problem numerify
			   "Can't numerify constraint form "
			   c))))
		(t
		 (multi-let (((okay val)
			      (opt-eval c sit !() dom)))
		    (cond (okay val)
			  (t c))))))

(deffunc screamer-var-of-type (ty name)
   (case ty
      ((opt:Float opt:Number) (screamer:a-real-betweenv -10e9 10e9 name))
      (opt:Integer (screamer:an-integer-betweenv -1000000 1000000 name))
      (t
       (signal-problem screamer-var-of-type
	    "screamer-solve can't handle var ~s of type ~s"
	      name ty))))

;;; Return a new version with no env-dependent links between variables.
;;; Note: env might be the same as (!_subst mm), or it might be an extension.
(deffunc maximal-match-clean - Maximal-match (mm - Maximal-match
					      id - Varid env - Bdgenv)
   (repeat :for ((d :in (!_dif mm)) - Constrained-goal
		 e - Sexp cl - (Lst Sexp)
		 :collectors new-dif - Constrained-goal)
      (!= < e env > (safe-varsubst (!_goal d) id env))
      (!= < cl env > (safe-varsubst (!_constraints d) id env))
    :collect (make Constrained-goal e cl)
    :result
       (multi-let (((new-constraints env)
		    (safe-varsubst (!_constraints mm) id env)))
	  (multi-let (((new-intsect _)
		       (safe-varsubst (!_intsect mm) id env)))
	     (make Maximal-match
		(empty-env) new-dif new-constraints new-intsect)))))

; Return those that make X ground, ignoring duplicates
(deffunc nodup-grounders - (Lst Bdgenv) (r - (Lst Bdgenv) id - Varid x - Sexp)
   (repeat :for ((c :in r) - Bdgenv
		(res '()) - (Lst (Lrcd Sexp Bdgenv)))
      (let ((xi (unsafe-varsubst x id c)))
         (cond ((and (not (has-qvars xi))
                     (not (assoc= #'equal xi res)))
                (!= res (cons (lrecord xi c) *-*)))))
    :result (<# cadr res)))

(deffunc subst-completions - (Lst Bdgenv) (vl - (Lst Expclo) obs - (Lst Obj)
                                           subst - Bdgenv)
      (cond ((null vl) (list subst))
            (t
             (let ((r (subst-completions (cdr vl) obs subst)))
                (repeat :for (ob :in obs)
                   :nconc
                         (let ((bdg (with (car vl)
                                       (make Varbdg (!_(Qvar sym) !>skel)
                                                    !>id
                                                    (make Expclo ob false)))))
                            (repeat :for (s :in r)
                               :collect (be Bdgenv (cons bdg s)))))))))

(deffunc binds-any - Boolean (env - Bdgenv id - Varid)
   (exists (bdg in env)
      (= (!_id bdg) id)))

(deffunc binds-any-new - Boolean (env - Bdgenv id - Varid prev - Bdgenv)
   (let ((all-vars
	    `(dummy ,@(<! (\\ (bdg)
			     (cond ((= (!_id bdg) id)
				    (list (make Qvar (!_varname bdg) !())))
				   (t !())))
			  (!_bdgs env)))))
      (multi-let (((new-all-vars _)
		   (safe-varsubst all-vars id env)))
	 (not (variants new-all-vars all-vars id id prev)))))

;;;;   (exists (bdg in env)
;;;;      (and (= (!_id bdg) id)
;;;;	   (not (exists (pb in prev)
;;;;		   (and (eq (!_varname pb) (!_varname bdg))
;;;;			(= (!_id pb) (!_id bdg))))))))
;;;;

(defun lit-synthesize (polarity fmla)
       (cond (polarity fmla)
	     (t `(not ,fmla))))