;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: optexec.lisp,v 1.26 2006/01/24 13:33:46 dvm Exp $

;;; This file could be split into two by slicing out a new file
;;; expansion-cmplr.lisp to deal just with plan-expansion constructs
;;; such as 'seq' and 'with-links'.

(depends-on %module/ nisp)

(depends-on :at-run-time
	    %opt/ expansion-exc
	    %optop/ execute domain procmod maxmatch
	            discrtproj continproj planprob)

(depends-on (:at :compile-time) %optop/ next-exec-state copy-w-changes)

(specdecl ;; For keeping track of penalties during forced projection.--
          (forcings-link* (new-Plan-link '*forcings*))
	  ;; Temporary repository of values in Exec-states that have them --
	  (value-link* (new-Plan-link '*output-value*)))

(defnisptype Link-keyword (Const . #.legal-link-keywords*))

(defnisptype Lex-var-defn - (Structure subscript - Fixnum))

;;; Important: The 'plan-sit' argument is _not_ the situation in which the
;;; plan will be executed; it's there to be used only if the expansion
;;; of 'plan' depends on the value of something in the current situation.
(deffunc plan->thread-gen - Thread-gen (plan - Typed-exp
					plan-sit - Situation env - Typenv)
   (let ((h (plan-expander plan))
	 - (~ (Fun Thread-gen (Typed-exp Situation Typenv))))
      (cond (h
	     (funcall h plan plan-sit env))
	    (t
	     (act->thread-gen plan plan-sit env)))))

(datafun-table plan-expanders* plan-gen :size 100)

(specdecl plan-expanders*
          - (Htb (Fun Thread-gen (Typed-exp Situation Typenv))))

(deffunc plan-expander - (~ (Fun Thread-gen (Typed-exp Situation Typenv)))
	              (exp - Sexp)
   (let ((hd (and (consp exp) (car exp))))
      (cond ((is-symbolish hd)
	     (href plan-expanders* hd))
	    (t false))))

(datafun plan-gen wait-while
   (deffunc :^ - Thread-gen (plan - Typed-exp _ - Situation env - Typenv)
      (match-let (wait-while ?procterm)
		 plan
	 (wait-thread plan procterm false env))))

(datafun plan-gen wait-non-negative
   (deffunc :^ - Thread-gen (plan - Typed-exp _ - Situation env - Typenv)
      (match-let (wait-non-negative ?procterm ?negterm)
		  plan
	 (wait-thread plan procterm negterm env))))

(deffunc wait-thread - Thread-gen
	             (plan procterm - Sexp negterm - (~ Sexp) _ - Typenv)
   (be Thread-gen
      (\\ - Thread-cont (then - Thread-cont)
	 (\\ - TC-val (es - Exec-state stack - Thread-stack)
	    (cond ((eq (!_mode es) ':real)
		   (signal-problem wait-thread
			 "Get real"))
		  (t
		   (let ((sit (!_sit es))
			 (past-sits (<# !_sit (!_hist es))))
		      (let ((succ-sit false) - (~ Situation)
			    (failed false) (forced false) - Boolean
			    (open-subgoal false) - (~ Sexp)
			    ;;;; (steps 0) - Fixnum
			    adds dels al dl - (Lst Occasion)
			    (elapsed 0.0)  - (Alt Time-int (Const *never))
			    (prev-sit (and (not (null past-sits))
					   (car past-sits)))
			    - (~ Situation)
			    frac - Float
			    (dom (!_domain es)) - Domain)
			 (let-fun ()
			    (cond ((process-find procterm sit)
				   (!= < succ-sit adds dels elapsed frac >
				       (process-project-and-record
					  sit (!_sit es) prev-sit
					  procterm negterm (!_sit-space es)
					  dom))
				   (!= failed
				       (or (not succ-sit)
					   (let ((new-pc (!_processes succ-sit)))
					      (and new-pc
						   (not (!_kosher new-pc)))))))
				  (t
				   (!= failed true)
				   (case (!_mode es)
				      (:forced
				       (!= < succ-sit al dl >
					   (process-existence-force
						 plan procterm negterm sit past-sits
						 (!_sit-space es) dom))
				       (cond ((not (process-find procterm succ-sit))
					      (dbg-save :run-loud
						 procterm sit succ-sit dom)
					      (signal-problem wait-thread
						 "Process cannot be forced to"
						 " exist: " procterm)))
				       (!= < succ-sit adds dels elapsed frac >
					   (process-project-and-record
					      succ-sit (!_sit es) prev-sit
					      procterm negterm (!_sit-space es)
					      dom))
				       (!= < adds dels >
					   (add-del-combine al dl adds dels))
				       (!= forced true))
				      (:zero-step
				       (!= open-subgoal 
					   `(process-exists ,procterm ,negterm))))))
			    (cond ((and forced
					(not (eq elapsed '*never))
					(> frac 0.0)
					(< frac 1.0))
				   (!= elapsed (be Time-int
						   (/ (be Time-int elapsed)
						      frac)))))
			    (let ((new-es
				     (note-post-process-changes)))
			       (values
				  (be Exec-status
				      (cond ((or forced open-subgoal)
					     (cond ((null (!_active new-es))
						    (cond (open-subgoal
							   ':continuing)
							  (t ':blocked)))
						   (t ':continuing)))
					    (failed ':failed)
					    (t ':continuing)))
				  new-es))

          :where

;;;; (:def project-found-process ())
;;;;
;;;; (:def project-forced-process ()
;;;;			     )
;;;;
;;;; (:def compute-elapsed-time ()
;;;;		      )

 (:def note-post-process-changes - Exec-state ()
    (let ((elapsed-time
	     (cond ((eq elapsed '*never)
		    (be Float
			(exec-value-get max-time-link* (!_vals es))))
		   (t (be Time-int elapsed))))
	    - Float
	  (new-thread (new-Thread then `(after ,plan) stack))
	    - Thread)
       (copy-with-changes es
	    (:vals (exec-value-set
		      forcings-link*
		      !>vals
		      (be * (\\ (fc) (+ fc 1)))
		      false)
	     :active
		(cond ((or (and failed (not forced))
			   open-subgoal)
		       !>active)
		      (t
		       (cons new-thread !>active)))
	     :pp (place-Plan-prefix
		    (+ (!_length !>pp) 1)
		    (+ (!_duration !>pp) elapsed-time)
		    (or succ-sit sit)
		    (make Plan-prefix-delta
		       (uniquify-action plan dom)
		       0 elapsed-time !() !>pp)
		    (!_sit-space es)
		    (!_domain es))
	     :hist (cons es !>hist)
	     :blocked
	       (cond (open-subgoal
		      (cons (make Blockage
			       (make Gate
				  ':open-subgoals
				  (list (list open-subgoal)))
			       (list new-thread))
			    !>blocked))
		     (t !>blocked)))))))))))))))

(datafun plan-gen achieve
   (deffunc :^ - Thread-gen (plan - Sexp _ - Situation _ - Typenv)
      (match-let (achieve ?goal)
		  plan
	 (be Thread-gen
	    (\\ - Thread-cont (c - Thread-cont)
	       (\\ (e - Exec-state thst - Thread-stack)
		  (let ((e-sit (!_sit e)))
		     (cond ((eq (!_mode e) ':real)
			    (signal-problem achieve-plan-gen
			       "Can't have achieve goal " goal
			       " in :real mode"))
			   ((deduce goal ach-id* (empty-env)
				    e-sit (!_domain e))
			    (funcall c e thst))
			   ((eq (!_mode e) ':forced)
			    (multi-let (((new-occs - (Lst Occasion)
					  env - Bdgenv)
					 (find-assertable-differences
					    goal (empty-env)
					    e-sit (!_past-sits e)
					    (!_domain e))))
			       (continuing-exec
				  (copy-with-changes e
				     (:blocked 
					 (cons (make Blockage
						     (make Gate
							   ':open-subgoals
							   (tuple new-occs
								  env))
						     (list
							(make Thread
							      c thst)))
					       !>blocked)
				      :vals (exec-value-set
					       forcings-link* !>vals
					       (\\ - Number (p) (+ p 1))
					       false))))))
			   (t
			    (values ':failed e))))))))))

(specdecl (eff-id* (new-Varid)) - Varid)

(defproc process-existence-force - (Multvals new-sit - Situation
					     adds dels - (Lst Occasion))
	                         (act procterm - Sexp negterm - (~ Sexp)
				  sit - Situation past-sits - (Lst Situation)
				  sit-space - Situation-space dom - Domain)
				 (ignore negterm)
   (let ((ipdefn
	    (causation-view
	       (name-causation-defn
		   (be CSymbol (car procterm))
		   'dur-happening (cdr procterm) dom)
	       #'durative-process-view
	       dom)))
      (multi-let (((_ env)
		   (unify (!_term ipdefn) procterm eff-id* ground-id* (empty-env))))
	 (multi-let (((new-sit adds dels _)
		      (precond-force-true ipdefn act env eff-id* sit past-sits
					  sit-space dom)))
	    (values new-sit adds dels)))))

(defproc process-project-and-record - (Multvals succ - Situation
						adds dels - (Lst Occasion)
					        elapsed - (Alt Time-int
							       (Const *never))
						frac - Float)
	                            (sit start-sit - Situation
				     prev-sit - (~ Situation)
				     procterm negterm - Sexp
				     sit-space - Situation-space dom - Domain)
                            (declare (ignore start-sit prev-sit))
      (multi-let (((final-sit adds dels elapsed frac)
		   (process-project
			 sit ;;;;(list start-sit prev-sit)
			 procterm negterm sit-space dom)))
	 (!= final-sit
	     (succ-record
		false sit elapsed 0
		(uniquify-action (wait-act procterm negterm)
				 dom)
		final-sit adds dels frac !() sit-space))
	 (values final-sit adds dels elapsed frac)))

(deffunc wait-act - Sexp (procterm negterm - Sexp)
   (cond (negterm
	  `(wait-non-negative ,negterm ,procterm))
	 (t `(wait-while ,procterm))))

;;; This procedure does two rather different things, depending on whether
;;; 'negterm' is false or a fluent to be made non-negative.  
;;; In the latter case it is interested
;;; only in when that inequality becomes true.  In the former it
;;; is interested only in when the process denoted by 'procterm' finishes.
;;; So if the process ends unexpectedly, it's a failure if negterm is a
;;; term, a success if it's false.
;;; Returned value 'frac' is fraction of ground covered from starting value
;;; of negterm to 0.  It's zero if negterm=false.
(deffunc process-project - (Multvals final-sit - (~ Situation)
				     adds deletes - (Lst Occasion)
				     elapsed - (Alt Time-int
						    (Const *never))
				     frac - Float)
	       (sit - Situation 
		procterm - Sexp
		negterm - (~ Sexp)
		sit-space - Situation-space dom - Domain)
               ;; -- negterm = false means that we want to project
	       ;; until the process stops
;;;;      (dbg-out forceproj-dbg*
;;;;	       "Projecting process " procterm)
      (repeat :for ((next-sit sit) - (~ Situation)
		    ;; -- never false except when error occurs
		    (termcond
		       (and negterm (be Uterm-ineq
					`(>= ,(uniquify-continuous-fluent-subterms
					       negterm dom)
					     0.0))))
		    - (~ Uterm-ineq)
		    process - (~ Process-instance)
		    (elapsed 0.0) - Time-int
		    (numwaits 0) - Fixnum
		    (timegrain (domain-temporal-grain-size dom))
		      - Float
		    delta-exasp - (~ Time-int)
		    delta-elap - (Alt Time-int (Const *never))
		    processes - Process-complex
		    trigs - (Lst Ineq-trigger)
		    (adds !()) (dels !()) al dl - (Lst Occasion)
;;;;		    save-sit - Situation
		    :collectors
		    exasperation - Time-int
		    
		    ;; -- Forecast times until desired trigger
		    ;; point is reached, in reverse chrono order.
		    )
         ;; until we reach goal or run out of patience
	 (!= processes (!_processes next-sit))
	 (!= process (process-find procterm next-sit))
       :while process
       :result (progn
		   (cond (negterm
			  (let ((frac (compute-frac)))
			     (cond ((ground-deduce `(=~ ,frac 1.0)
						   next-sit dom)
				    (values next-sit adds dels elapsed frac))
				   (t
				    (values false !() !()
					    (cond ((=< frac 0.0) '*never)
						  (t elapsed))
					    frac)))))
			 (t
			  (values next-sit adds dels elapsed 0))))
	 (!= trigs (cond (termcond
			  (be (Lst Ineq-trigger)
			      (<? is Ineq-trigger
				  (continuous-trigs
					   (list termcond)
					   (list '(made-it))
					   (!_model processes)
					   next-sit dom))))
			 (t
			  (repeat :for ((tr :in (!_triggers processes))
					:collectors stop-trigs)
			   :when (is-stopper tr)
			   :collect tr
			   :result (cond ((null stop-trigs)
					  ;; Make one up
					  (list (make Ineq-trigger
						   `(dummy)
						   (\\ (ti - Time-int)
						      (- ti
							 way-in-the-future*))
						   `(- *T* (way-in-the-future
							    = ,way-in-the-future*))
						   (list '(made-it)))))
					 (t stop-trigs))))))
       ;; This can happen only if negterm is non-false and has become
       ;; non-negative, i.e., we made it --
       :until (null trigs)
         (!= < delta-exasp _ _ >
	     (compute-trigger-time trigs timegrain))
       :while (and delta-exasp (> delta-exasp 0.0))
       :result
          (let ((frac (compute-frac)))
;;;;	     (dbg-out forceproj-dbg*
;;;;		"Process "
;;;;		(:q ((null trigs)
;;;;		     "ends happily")
;;;;		    (t "never ends"))
;;;;		:% "  accomplished fraction " frac " of " termcond)
	     (values next-sit adds dels elapsed frac))
       :collect (:into exasperation delta-exasp)
       :until (exasperated)
       :result
           (progn
;;;;	      (dbg-out forceproj-dbg*
;;;;		 "Process ends in exasperation")
	      (values false !() !() '*never 0))
	 (!= < next-sit al dl delta-elap >
	     (wait-project true next-sit sit-space dom))
	 (cond ((not next-sit)
		(signal-problem process-project
		   "Unexpected inability to project a 'wait'")))
	 (!= adds (union al (complementq *-* dl)))
	 (!= dels (union dl (complementq *-* al)))
	 (!= numwaits (+ *-* 1))
         (!= elapsed (+ *-* (be Time-int delta-elap)))

       :where

	  (:def is-stopper - Boolean (trig - Ineq-trigger)
	     (exists (eff :in (!_effects trig))
		(cond ((is Ineq-trigger eff)
		       (is-stopper eff))
		      (t
		       (decl (eff - Sexp)
			  (matchq (process-transition
				      stopping ?,procterm)
				  eff))))))

	  (:def exasperated - Boolean ()
	     (cond ((< numwaits 10)
		    false)
		   (t
		    (let ((exlts (!_elements exasperation)))
		       (let ((delta0 (car exlts)))
			  (let ((decreases
				   (cons delta0
					 (repeat :for
						 ((delta :in (cdr exlts))
						  (prev-delta delta0)
						    - Time-int)
					  :when (< delta prev-delta)
					  :collect delta
					     (!= prev-delta delta)))))
			     (>= (* 3 (len decreases))
				 numwaits)))))))

	  (:def compute-frac - Float ()
	     (prog1
			  (cond ((and negterm (> elapsed 0))
				 (let ((initval
					  (nth-value
					     1 (opt-eval negterm sit !() dom)))
				       (finalval
					  (nth-value
					     1 (opt-eval negterm next-sit
							 !() dom)))
				       - Number)
				   (/ (- finalval initval)
				      (- (float initval)))))
				(t 0))))))

(datafun plan-gen until-non-negative
   (deffunc :^ - Thread-gen (plan - Sexp plan-sit - Situation env - Typenv)
      (match-let (until-non-negative ?quant ?eltact)
		  plan
	 (let ((eltact-tg (plan->thread-gen eltact plan-sit env)))
	    (be Thread-gen
		(\\ - Thread-cont (cont - Thread-cont)
		   (\\ - TC-val (es - Exec-state stack - Thread-stack)
		      (let-fun ()
			 (cond ((eq (!_mode es) ':real)
				(signal-problem until-non-negative
						"Get real"))
			       (t
				(until-thread-cont es quant !())))

		   :where

 (:def until-thread-cont - TC-val
                         (next-es - Exec-state
			  quant - Sexp
			  vals-so-far - (Lst Integer))
    (let ((cur-val (nth-value 1
		       (opt-eval quant (!_sit next-es) !()
				 (!_domain next-es)))))
       (cond ((not cur-val)
	      (signal-problem until-non-negative
		 "Can't evaluate " quant)))
       (decl (cur-val - Number)
	  (cond ((>= cur-val 0)
		 (continuing-exec
			 (copy-with-changes
			    next-es
			    (:active
			       (cons (new-Thread
					   cont
					   `(after ,plan) stack)
				     !>active)))))
		((nums-not-going-anywhere vals-so-far)
		 (values ':failed next-es))
		(t
		 (continuing-exec
		     (copy-with-changes next-es
			(:active (cons (make Thread
					     (funcall eltact-tg cont)
					     stack)
				       !>active)))))))))))))))))

(deffunc nums-not-going-anywhere - Boolean (vals - (Lst Number))
   (let ((delta0 (car vals)) (numvals (len vals)))
      (and (> numvals 10)
	   (let ((decreases
		    (cons delta0
			  (repeat :for
				  ((delta :in (cdr vals))
				   (prev-delta delta0)
				     - Number)
			   :when (< delta prev-delta)
			   :collect delta
			      (!= prev-delta delta)))))
	      (>= (* 3 (len decreases))
		  numvals)))))

(datafun plan-gen seq
   (deffunc :^ - Thread-gen (plan - Sexp cmplr-sit - Situation env - Typenv)
      (let ((steps-tgs (<# (\\ (step)
			      (plan->thread-gen step cmplr-sit env))
			   (cdr plan))))
	 (be Thread-gen
	    (\\ - Thread-cont (after-seq - Thread-cont)
	       (\\ - TC-val (es - Exec-state stack - Thread-stack)
		  (let-fun ((:def seq-compose - (Mlv Exec-status Exec-state)
					 (tgl - (Lst Thread-gen)
					  acts - (Lst Sexp)
					  es - Exec-state)
			       ;; 'tgl' and 'acts' are 
			       ;; lists of the same length.
			       (cond ((null tgl)
				      (funcall after-seq es stack))
				     ((null (cdr tgl))
				      (((head tgl)
					(\\ (es2 ts2)
					   (thread-enqueue
					      (make Thread after-seq ts2)
					      es2)))
				       es (make Thread-stack
						(head acts) stack)))
				     (t
				      (((head tgl)
					(\\ (es2 _)
					   (seq-compose
					      (tail tgl)
					      (tail acts)
					      es2)))
				       es stack)))))
		     (seq-compose steps-tgs (cdr plan) es))))))))


(deffunc thread-gen-seq - Thread-gen (g1 g2 - Thread-gen)
   (cond ((eq g1 no-op-thread-gen*)
	  g2)
	 ((eq g2 no-op-thread-gen*)
	  g1)
	 (t
	  (\\ - Thread-cont (c - Thread-cont)
	     (funcall g1
		(\\ - TC-val (state-after-g1 - Exec-state
			      stack-after-g1 - Thread-stack)
		   (thread-enqueue
		      (make Thread
			    (funcall g2
			       (\\ - TC-val (state-after-g2 - Exec-state
					     stack-after-g2 - Thread-stack)
				  (thread-enqueue
				     (make Thread c stack-after-g2)
				     state-after-g2)))
			    stack-after-g1)
		      state-after-g1)))))))

;;; In this version the test is just an expression to evaluate
(datafun plan-gen test
   (deffunc :^ - Thread-gen (test-plan - Sexp
			     complr-sit - Situation env - Typenv)
      (match-let (test ?exp ?iftrue ?iffalse)
		 test-plan
	 (let ((iftrue-gen (plan->thread-gen iftrue complr-sit env))
	       (iffalse-gen (plan->thread-gen iffalse complr-sit env)))
	    (\\ (after - Thread-cont)
	       (let ((iftrue-cont (funcall iftrue-gen after))
		     (iffalse-cont (funcall iffalse-gen after)))
		  (\\ (es - Exec-state st - Thread-stack)
		     (multi-let (((evaluable value)
				  (opt-eval
				      exp
				      (!_sit es) (!_bdgs es) (!_domain es))))
			(cond (evaluable
			       (cond (value (funcall iftrue-cont
					        es
						(make-Thread-stack
						       iftrue st)))
				     (t (funcall iffalse-cont
					   es
					   (make-Thread-stack
					       iffalse st)))))
			      (t
			       (signal-problem test
				  "Can't evaluate " exp)))))))))))

(specdecl symno* - Integer)

(datafun plan-gen parallel
   (deffunc :^ - Thread-gen (plan - Sexp complr-sit - Situation env - Typenv)
      (let ((sub-thread-gens
	       (<# (\\ (p) (plan->thread-gen p complr-sit env))
		   (cdr plan))))
	 (\\ (after-par - Thread-cont)
	    (\\ (es - Exec-state par-stack - Thread-stack)
		- (Mlv Exec-status Exec-state)
	       (cond ((null sub-thread-gens)
		      (funcall after-par es par-stack))
		     ((null (tail sub-thread-gens))
		      (((car sub-thread-gens)
			after-par)
		       es par-stack))

;;;;		       (\\ (es1 - Exec-state _ - Thread-stack)
;;;;			  (thread-enqueue
;;;;			      (make Thread after-par par-stack)
;;;;			      es1))
		     (t
		      (let-fun ()
			 (let* ((link (make Plan-link
					 (build-symbol (:package false)
						       par-synch- (++ symno*))
					 false nil false false))
				(dec-and-check
				   (decrement-and-proceed-when-zero
				       after-par link)))
			    (continuing-exec
				    (copy-with-changes es
				       (:active
					   (nconc (par-threads
						     sub-thread-gens
						     (cdr plan)
						     dec-and-check)
						  !>active)
					:vals
					   (cons (tuple link
							(len (tail plan)))
						 !>vals)))))

		    :where

 (:def decrement-and-proceed-when-zero - Thread-cont
                           (after-par - Thread-cont link - Plan-link)
    (\\ - TC-val (es1 - Exec-state _ - Thread-stack)
       (let ((e (be * (assq link (!_vals es))))
	     - (~ (Lrcd Plan-link Fixnum)))
	  (cond (e
		 (cond ((= (second e) 1)
			(funcall after-par
				 (copy-with-changes es1
				    (:vals
					(remove1q e
						  !>vals)))
				 par-stack))
		       (t
			(values
			   ':continuing
			   (copy-with-changes es1
			      (:vals
				  (exec-value-set
					 link !>vals
					 (be (Fun Obj Obj)
					     (\\ (v - Integer)
						(- v 1)))
					 false)))))))
		(t
		 (signal-problem
		    parallel-expander
		    "Unbound Plan-link: " link
		    :% " in " (!_env es)))))))

 (:def par-threads - (Lst Thread) (sub-gens - (Lst Thread-gen)
				   acts - (Lst Sexp)
				   dec-and-check - Thread-cont)
    (<# (\\ (sub-gen act)
	   (branch-thread sub-gen act dec-and-check))
	sub-gens
	acts))

 (:def branch-thread (tgen - Thread-gen branch - Sexp
		      dec-and-check - Thread-cont)
     (new-Thread
        (\\ (es - Exec-state stack - Thread-stack)
	   ((funcall tgen dec-and-check)
	    es stack))
	branch par-stack))))))))))

(specdecl (link-unset* (printable-as-string "<Unset Link>"))
	  - Obj)

(datafun plan-gen with-links
   (deffunc :^ - Thread-gen (exp - Sexp plan-sit - Situation env - Typenv)
      (match-let (with-links ?link-names ?do)
		 exp
	 (let ((links (<# (\\ (v) (tuple v (new-Plan-link v)))
			  link-names))
	       (exp-tg (plan->thread-gen do plan-sit env)))
	    (\\ (c - Thread-cont)
	       (let-fun ()
		  (\\ - TC-val
		       (initial-estate - Exec-state tstack - Thread-stack)
		     (continuing-exec
			(copy-with-changes initial-estate
			   (:bdgs (append links !>bdgs)
			    :vals (nconc (<# (\\ (lb)
						(tuple (second lb)
						       link-unset*))
					     links)
					 !>vals)
			    :active (cons (do-then-unlink
					      tstack
					      (!_bdgs initial-estate))
					   !>active)))))
		:where

		  (:def do-then-unlink - Thread (tstack - Thread-stack
						 initial-bdgs - Exec-bdg-tab)
		     (make Thread
			   (funcall exp-tg
			      (\\ (final-estate _)
				 (let ((es-popped
					  (unbind-links
					      (<# second links)
					      initial-bdgs
					      final-estate)))
				    (funcall c es-popped tstack))))
			   tstack))

		  (:def unbind-links - Exec-state
                               (links - (Lst Plan-link)
				initial-bdgs - Exec-bdg-tab
				final-estate - Exec-state)
		     (copy-with-changes final-estate
			(:bdgs initial-bdgs
			 :vals (remove-if (\\ (lv) (memq (car lv) links))
					       !>vals)
;;; This is wrong because side effects are cumulative --
;;;;			 (!_link-vals estate)
;;; -- If we had closures, even removing the inaccessible link values
;;; would be wrong, because there might still be closures in existence
;;; that could access or change the link values.
			 )))))))))

(datafun plan-gen link
   (deffunc :^ - Thread-gen (exp - Sexp plan-sit - Situation env - Typenv)
      (control-nest
	 (match-let (link ?step ?linkrels)
		    exp
	    ;; linkrels of form ((kind var) (kind var) ...)
	    ;; where kind = :input | :output | :wait | ...
	    (decl (linkrels - (Lst (Lrcd Link-keyword Sexp)))
	       (let  ((input-links (<? (\\ (lr) (eq (first lr) ':input))
				       linkrels))
		      (begin-blocker-links
		             (<? (\\ (lr) (memq (first lr)
						'(:input :wait)))
				 linkrels))
		      (end-blocker-links
		             (<? (\\ (lr)
				    (memq (first lr) '(:wait-end :span)))
				 linkrels))
		      (begin-output-links
		             (<? (\\ (lr)
				    (memq (first lr)
					  '(:begin-then :span)))
				 linkrels))
		      (end-output-links
			     (<? (\\ (lr) (memq (first lr)
						'(:output :then)))
				 linkrels))
		      - (Lst (Lrcd Link-keyword Sexp)))
		  :redeclare-inputs)))
       :redeclare-inputs
         (let ((inp-env
		  (cond ((null input-links) env)
			(t
			 (env-bindings-append
			    true (<# (\\ (ip)
					(let ((e (var-lookup (second ip)
							     env)))
					   (let ((lv-ty
						   (type-feature
						      (Vartype-type e)
						      'link-contents-type)))
					      (new-Vartype
						 (second ip)
						 lv-ty
						 false))))
				     input-links)
			    env)))))
	    (let ((step-tg (plan->thread-gen step plan-sit inp-env)))
	       :define-aux-fcns))
	 :define-aux-fcns
	 (let-fun ()
	    (\\ - Thread-cont (finally - Thread-cont)
	       (\\ (e-state - Exec-state tstk - Thread-stack)
		  ;; We define the continuations in backwards order because
		  ;; the ones defined earlier must be callable by the ones
		  ;; called earlier (got that?) --
  		  (let* ((transmit-then-end-cont
			    (\\ - TC-val (e-state - Exec-state
					  tstk - Thread-stack)
                               (values ':continuing
                                       (next-exec-state e-state
                                          (:vals (exec-state-vals-rebind
                                                    end-output-links
                                                    *-* !>bdgs))))))

                         ;; Deal with :wait-end and :span links
		         (wait-to-transmit-and-end-cont
                            (cond ((null end-blocker-links)
                                   transmit-then-end-cont)
                                  (t
                                   (block-on-unset-blockers
                                       end-blocker-links
                                       transmit-then-end-cont))))
		 
		         ;; Coerce :input links to their values, then do the
                         ;; action 
                         (do-the-step-cont
                            (\\ - TC-val (e-state - Exec-state
                                          tstk - Thread-stack)
                               (let ((input-alist
                                        (<# (\\ (ip)
                                               (let* ((lvar (second ip))
                                                      (d (alref
                                                           (!_bdgs e-state)
                                                           lvar
                                                           (signal-problem link-plan-gen
                                                              "Unbound link variable "
                                                              lvar))))
                                                  (tuple lvar
                                                         (alref
                                                            (!_vals e-state)
                                                            d
                                                            (signal-problem
                                                               "Link value missing " lvar)))))
                                            input-links)))
                                  ((funcall step-tg wait-to-transmit-and-end-cont)
                                   (next-exec-state e-state
                                      (:bdgs (nconc input-alist !>bdgs)))
                                   tstk))))
		
		;; Handle :wait-then and :span links
		(transmit-then-start-cont
		   (cond ((null begin-output-links)
			  do-the-step-cont)
			 (t
			  (\\ - TC-val (estate - Exec-state
					tstk - Thread-stack)
			     (transmit-to-links
				   begin-output-links !() estate
				   do-the-step-cont tstk)))))

		;; Handle :input and :wait links
		(wait-to-transmit-and-start-cont
		    (cond ((null begin-blocker-links)
			   transmit-then-start-cont)
			  (t
			   (block-on-unset-blockers
			          begin-blocker-links
				  transmit-then-start-cont))))
)
	    (\\ - Thread-cont 
wait-to-transmit-and-start-cont))))

    :where

 (:def blocked-links - (Lst Plan-link)
				  (blocker-linkrels - (Lst (Lrcd Link-keyword
								 Symbol))
				   estate - Exec-state)
		      (repeat :for ((r :in blocker-linkrels)
				      - (Lrcd Link-keyword
					      Symbol)
				    :collector blocked - Plan-link)
		       :within
			 (let* ((lnk (be Plan-link (exec-state-opt-eval
						      (second r) estate)))
				(val (exec-value-get
					lnk
					(!_vals estate))))
			    (:continue
			     :when (eq val link-unset*)
			     :collect lnk))))

 (:def block-on-unset-blockers - Thread-cont
                   (blocker-links - (Lst (Lrcd Link-keyword
                                               Symbol))
                    proceed-after-block - Thread-cont)
        (\\ - TC-val (estate - Exec-state
                      tstk - Thread-stack)
           (let ((blocked-blockers
                    (blocked-links blocker-links
                                   estate)))
              (cond ((null blocked-blockers)
                     (funcall proceed-after-block
                              estate tstk))
                    (t
                     (continuing-exec
                        (copy-with-changes estate
                           (:blocked
                              (cons (wait-for-links-block
                                       blocked-blockers
                                       proceed-after-block
                                       tstk)
                                    !>blocked)))))))))

 (:def wait-for-links-block - Blockage
                      (unset-links - (Lst Plan-link)
                      cont - Thread-cont
                      tstk - Thread-stack)
    (make Blockage
          (make Gate ':link-block unset-links)
          (list (make Thread cont tstk))))

 (:def transmit-to-links - TC-val
                    (receiving-linkrels
                        - (Lst (Lrcd Link-keyword
                                     Symbol))
                     val - (Lst Obj)
                     estate - Exec-state
                     cont - Thread-cont tstk - Thread-stack)
    (repeat :for ((linkrel :in receiving-linkrels)
                    - (Lrcd Link-keyword Symbol)
                   (vals (!_vals estate))
                     - Exec-val-tab
                  :collector links - Plan-link)
     :within
       (let ((lnk (be Plan-link
                      (exec-state-opt-eval
                         (second linkrel) estate))))
          (:continue
           :collect lnk
             (!= vals
                 (exec-value-set lnk *-* false
                    (\\ () val)))))
     :result
          (continuing-exec
             (links-unblock (<# second links)
                 (next-exec-state estate
                    (:vals vals
                     :active (cons (make Thread cont tstk)
                                   !>active)))))))

(deffunc exec-state-vals-rebind - (Lst (Lrcd Exec-bdg-id Obj))
                         (links - (Lst (Lrcd Symbol Obj))
                          vals - Exec-val-tab
                          bdgs - Exec-bdg-tab)
   (let ((bdg-ids (<# (\\ (lnk) 
                         (alref bdgs (second lnk)
                               (signal-problem exec-state-vals-rebind
                                  "Unbound link: " (second lnk))))
                      links)))
      ;; Remove all the links that will be rebound
      (!= vals (<? (\\ (b) (memq (first b) bdg-ids))
                   *-*))
      (repeat :for ((out-link :in links) - (Lrcd Symbol Obj)
                    (var-bdg-id :in bdg-ids) - Exec-bdg-id
                    (result-val (alref vals value-id*
                                       (signal-problem act-vals-rebind
                                          "No current result in Exec-state")))
                    - (Lst Obj)
                    :collector new-vals)
       :collect (tuple var-bdg-id
                       (cond ((eq (first out-link)
                                  ':output)
                              result-val)
                             (t false)))
       :result (nconc new-vals vals))))

(deffunc act->thread-gen - Thread-gen (act - Sexp
				       plan-sit - Situation
				       env - Typenv)
   (let ((dom (!_domain (find-init-situation plan-sit))))
      (multi-let (((adefn idefn env1)
		   (act-action-defn act dom)))
;;;;	 (cond ((and (is Action-defn adefn)
;;;;		     (!_has-methods (be Action-defn adefn)))
;;;;		(act-method-thread-gen
;;;;		    act (be Action-defn adefn) plan-sit env))
;;;;	       (t ... ))
;;; Whether the action is primitive or not, the compiler never expands
;;; a composite, instead relying on the add/delete list to decide what to do
		(let-fun ()
		   #'res-fcn
		 :where

Don't want args-eval below; we want args-compile above !!

 (:def res-fcn (then - Thread-cont)
    (\\ (es - Exec-state stack - Thread-stack)
       (let ((eval-act `(,(car act) ,@(args-eval (cdr act) es))))
	  (cond ((eq (!_mode es) ':real)
		 (action-really-do eval-act es adefn then stack
				   (!_sit es) env))
		(t
		 (let* ((uact (uniquify-action eval-act dom))
			(pr (prim-try-project (!_mode es) 
					     (!_sit es)
					     (!_sit-space es)
					     (<# !_sit (!_hist es))
					     uact idefn false
					     env1 eff-varid* dom)))
		    (match-cond pr
		       (:? ?(:\| :failed :ready :& ?stat)
			  (values stat es))
		       (:? (:open-subgoals ?dnf)
			  (values ':blocked
				  (copy-with-changes es
					(:blocked
					    (cons (make Blockage
							(make Gate
							   ':open-subgoals
							   dnf)
							(list (make Thread
								    then
								    stack)))
						  !>blocked)))))
		       (:? (:okay ?sit ?elapsed ?numsteps)
			  (decl (sit - Situation
				 elapsed - Float
				 numsteps - Integer)
			     (continuing-exec
				     (copy-with-changes es
					(:active
					    (cons (make Thread then stack)
						  !>active)
					 :pp (plan-prefix-extend
					        uact es numsteps elapsed
						sit)
					 :elapsed (elapsed-sum
						     !>elapsed elapsed)
					 :steps (+ !>steps numsteps))))))
		       (:? (:forced ?num-forcings ?sit ?elapsed ?numsteps)
			  (decl (sit - Situation
				 elapsed - Float
				 num-forcings numsteps - Integer)
			     (values ':forced
				     (copy-with-changes es
					(:active (cons (make Thread
							     then stack)
						       !>active)
					 :pp (plan-prefix-extend
						uact es numsteps elapsed sit)
					 :elapsed (elapsed-sum
						     !>elapsed elapsed)
					 :steps (+ !>steps numsteps)
					 :vals (exec-value-set
						   forcings-link*
						   !>vals
						   (\\ (f)
						      (+ f num-forcings))
						   false))))))
		       (t
			(signal-problem act->thread-gen
			      "Indecipherable: " pr)))))))))

 (:def plan-prefix-extend - Plan-prefix (uact - Action
					 es - Exec-state
					 numsteps - Integer
					 elapsed - Float
					 post-sit - Situation)
     (let ((pp (!_pp es)))
	(place-Plan-prefix
		    (+ (!_length pp) numsteps)
		    (+ (!_duration pp) elapsed)
		    post-sit
		    (make Plan-prefix-delta
		       uact numsteps elapsed !() pp)
		    (!_sit-space es)
		    (!_domain es))))))))

;;; The Bdgenv is the result of unifying the action of the 
;;; Internalized-action-defn with 'act'.  Varid for vars in
;;; the Internalized-action-defn is eff-varid*.
(deffunc act-action-defn - (Mlv Causation-defn Internalized-action-defn Bdgenv)
	            (act - Sexp dom - Domain)
  (let-fun ()
     (match-cond act
	?( (?(:\| dur-start dur-stop :& ?ss)
	    (?dfun ?@dargs))
	  (let ((af (find-domain-bdg-val (be CSymbol dfun) dom)))
	     (cond ((is Causation-function af)
		    (let ((ddefn (!_defn af)))
		       (cond ((is Durative-action-defn ddefn)
			      (let ((iddefn
				       (be Internalized-durative-defn
					   (causation-defn-projectable
					      ddefn dom))))
				 (do-unify
				    (cond ((eq ss 'dur-stop)
					   (!_stop-view iddefn))
					  (t (!_start-view iddefn)))
				    ddefn)))
			     (t
			      (signal-problem act-action-defn
				 "In " act ", " `(,dfun ,@dargs)
				 :% " is not a durative-action term")))))
		   (t
		    (signal-problem act-action-defn
		       "In " act ", " dfun
		       " is not a Causation-function")))))
	?( (?(:\| dur-start dur-stop)
	    ?_)
	  (signal-problem act-action-defn
	     "Mal-formed " ss " term: " act))
	?( (?afun ?@_)
	  (let ((af (find-domain-bdg-val (be CSymbol afun) dom)))
  ;;;;	      (out "af = " af " Causation-function? "
  ;;;;		   (is Causation-function af) :%)
	     (cond ((and af
			 (is Causation-function af))
		    (let ((adefn (!_(Causation-function defn) af)))
		       (cond ((is Action-defn adefn)
			      (let ((idefn (be Internalized-action-defn
					       (causation-defn-projectable
						  adefn dom))))
				 (do-unify idefn adefn)))
			     (t
			      (signal-problem act-defn-term-unify
				 "In " act ", " afun
				 :% " is not an action function")))))
		   (t
		    (signal-problem act-defn-term-unify
		       "In " act ", " afun
		       " is not a Causation-function")))))
       (t
	(signal-problem act-defn-term-unify
	   "Ill-formed action: " act)))

     :where

       (:def do-unify - (Multvals defn - Causation-defn
				  idefn - Internalized-action-defn
				  env - Bdgenv)
                (idefn - Internalized-action-defn defn - Causation-defn)
	  (multi-let (((ok e)
		       (unify (!_term idefn)
			      act
			      eff-varid* ground-id* (empty-env))))
	     (cond (ok
		    (values defn idefn e))
		   (t
		    (signal-problem act-defn-term-unify
		       "Action definition fails to unify: " act
		       :% defn)))))))

(defproc prim-try-project - Projection-result
	   (mode - Exec-mode
	    sit - Situation sit-space - Situation-space
	    past-sits - (Lst Situation)
	    uact - Action idefn - Internalized-action-defn
	    goal - (~ Sexp) env - Bdgenv id - Varid dom - Domain)
	   (ignore goal)
   (let ()
      (or (recover-cached-transition mode sit uact sit-space)
	  (let-fun ()
	     (let ((pre 
		      (cond ((!_precond idefn)
			     (bind ((optimist-fn* false))
				(deduce
				   (!_precond idefn) id env sit dom)))
			    (t
			     (list (empty-env))))))
		(cond (pre
		       (cond ((eq mode ':zero-step)
			      ':ready)
			     (t
			      (action-precond-project-result
				 mode 
				 uact idefn 0 (car pre) sit sit-space dom))))
		      ((eq mode ':forced)
		       (multi-let (((feas-sit _ _ env)
				    (precond-force-true
					idefn (!_term uact)
					env id 
					sit past-sits sit-space dom)))
			  (action-precond-project-result
			     ':forced uact idefn 1 env feas-sit
			     sit-space dom)))
		      (t
		       (tuple (atomconst :open-subgoals)
			      (disjunctive-normal-form
			          (!_precond idefn) true)))))))))

;;; This corresponds to 'make-succ-sit-with-difs' in the old regime
(defproc precond-force-true - (Multvals succ - Situation
					adds dels - (Lst Occasion)
					witness - Bdgenv)
                             (idefn - Internalized-causation-defn
			      act - Sexp
			      env - Bdgenv id - Varid 
			      sit - Situation past-sits - (Lst Situation)
			      sit-space - Situation-space dom - Domain)
   (let ((feas-act
	    (uniquify-action `(make-feasible ,act) dom)))
      (let ((sss (find-old-projection sit feas-act sit-space)))
	 (let-fun ()
	    (cond (sss
		   (with sss
		      (verify !>sit !>adds !>deletes)))
		  (t
		   (let ((new-occs !()) - (Lst Occasion))
		      (!= < new-occs env >
			  (find-assertable-differences
			     (!_condition idefn)
			     env sit past-sits dom))
		      (multi-let (((new-sit adds deletes)
				   (build-new-sit
				      true feas-act
				      new-occs !() dummy-id1* env
				      sit sit-space dom)))
			 (verify
			    (succ-record
			       true sit 0.0 1 feas-act new-sit
			       adds deletes 0.0 !() sit-space)
			    adds deletes)))))
	  :where

            (:def verify - (Multvals succ - Situation
                                     adds dels - (Lst Occasion)
                                     witness - Bdgenv)
                         (succ - Situation adds dels - (Lst Occasion))
               (let ((al (deduce (!_condition idefn) id env sit dom)))
                  (cond ((null al)
                         (signal-problem precond-force-true
                            "Precond " (!_condition idefn)
                            : % " unprovable after being forced true"))
                        (t
                         (values succ adds dels (car al))))))))))

(specdecl (difs-fill-dbg* true) - Boolean)

(deffunc find-assertable-differences - (Mlv (Lst Occasion) Bdgenv)
	                    (p - Sexp env - Bdgenv
			     sit - Situation past-sits - (Lst Situation)
			     dom - Domain)
   (multi-let (((precond env)
		(safe-varsubst p dummy-id1* env)))
      (dbg-out difs-fill-dbg*
	       "Trying to force " precond " to be true")
      (let ((dp (disjunctive-normal-form precond true))
	    - (Lst (Lst Sexp))
	    result-env - Bdgenv new-occs - (Lst Occasion))
	 (repeat :for ((conj :in dp) - (Lst Sexp)
		       (found-difs false) - Boolean)
	  :result (progn
		     (dbg-save :run-loud dp sit past-sits dom env)
		     (signal-problem find-assertable-differences
			"Unable to complete " precond))
	    (let ((mml (maximal-matches
			  conj !() dummy-id1* env sit dom)))
	       ;; Look for a maximal match whose difs are assertable.
	       (repeat :for ((mm :in mml) - Maximal-match)
		  (cond ((forall (d :in (!_dif mm))
			    (goal-has-assertable-pred (!_goal d)))
			 (let-fun ((:def elim ()
				      (!= < found-difs result-env new-occs >
					  (difs-elim-vars
						      (!_dif mm)
						      (!_constraints mm)
						      (!_subst mm)
						      sit past-sits dom))))
			    (cond (difs-fill-dbg*
				   (out (:to *error-output*)
				      "Trying difs "
				      (debug-varsubst
				          (!_dif mm) dummy-id1*(!_subst mm))
				      :% (:i> 3) :%
				      (:e (elim))))
				  (t (elim))))))
		:until found-difs))
	  :until found-difs)
	 (values new-occs result-env))))

(deffunc difs-elim-vars - (Multvals consis - Boolean
				    env - Bdgenv
				    new-occs - (Lst Occasion))
	                (difs - (Lst Constrained-goal)
			 constraints - (Lst Sexp)
			 env - Bdgenv
			 sit - Situation past-sits - (Lst Situation)
			 dom - Domain)
   (repeat :for ((dif :in difs) - Constrained-goal
		 (result-env env) - Bdgenv
		 (constraints constraints) - (Lst Sexp)
		 dif-filled - Boolean difexp goal-inst - Sexp
		 (unfilled-difs !()) (filled-difs !()) - (Lst Sexp))
      (!= < difexp result-env >
	  (safe-varsubst (!_goal dif) dummy-id1* result-env))
      (!= dif-filled (not (has-qvars difexp)))
      (cond (dif-filled
	     (!= goal-inst difexp)
	     (dbg-out difs-fill-dbg*
		      "Difference: " difexp)))
      (repeat :for ((past-sit :in past-sits) - Situation
		    (n = 1 :by 1) - Fixnum
		    al - (Lst Bdgenv))
       :until dif-filled
	 (!= al (deduce (!_goal dif) dummy-id1* result-env past-sit dom))
	 (repeat :for ((a :in al) new-env - Bdgenv
		       consis - Boolean)
	  :result nil
            (!= < goal-inst new-env >
		(safe-varsubst (!_goal dif) dummy-id1* a))
	    (!= < consis new-env >
		(constraints-satisfy
			     true (!_constraints dif) dummy-id1*
			     result-env sit dom))
	    (!= < goal-inst new-env >
		(safe-varsubst goal-inst dummy-id1* new-env))
	  :until (and consis (not (has-qvars goal-inst)))
	  :result (progn (dbg-out difs-fill-dbg*
			    "Found (" n " steps back) difference: "
			    goal-inst :%)
			 (!= dif-filled true))))
      (cond (dif-filled
	     (!= filled-difs (cons goal-inst *-*)))
	    (t
	     (!= unfilled-difs (cons (!_goal dif) *-*))
	     (!= constraints (append (!_constraints dif) *-*))))
    :result
       (multi-let (((consis result-env)
		    (constraints-satisfy
			true constraints dummy-id1*
			result-env sit dom)))
	  (cond (consis
		 (let ((new-occs-from-unfilled
			    (repeat :for ((d :in unfilled-difs) - Sexp
					  :collector instances - Occasion)
			       (!= < d result-env >
				   (safe-varsubst d dummy-id1* result-env))
			       (cond ((has-uvars d dummy-id1* result-env)
				      (signal-problem difs-elim-vars
					 "Constraining vars failed to"
					 " instantiate " d)))
			     :collect (uniquify-occasion d dom))))
		    (dbg-out difs-fill-dbg*
		       "Filled differences using constraints: "
		       (<# !_term new-occs-from-unfilled))
		    (values true
			    result-env
			    (nconc 
			       (<# (\\ (d)
				      (uniquify-occasion d dom))
				   filled-difs)
			       new-occs-from-unfilled))))
		(t
		 (dbg-out difs-fill-dbg*
		    "Attempt to fill differences founders due to"
		    " constraint inconsistency")
		 (values false nil nil))))))

(deffunc action-precond-project-result - Projection-result
	                     (mode - Exec-mode
			      uact - Action adefn - Internalized-action-defn 
			      forcings-so-far - Integer
			      bdgs - Bdgenv
			      sit - Situation sit-space - Situation-space
			      dom -  Domain)
   (multi-let (((succ _ _ _ _)
		(act-project-given-precond
			  (eq mode ':forced) uact adefn eff-varid* bdgs
			  sit sit-space dom true)))
      (cond (succ
	     (cond ((> forcings-so-far 0)
		    (make (Lstructure &flag :forced
			     num-forcings - Integer
			     sit - Situation
			     elapsed - Time-int
			     numsteps - Integer)
			  forcings-so-far succ 0.0 1))
		   (t
		    (make (Lstructure &flag :okay
				      sit - Situation
				      elapsed - Time-int
				      numsteps - Integer)
			  succ 0.0 1))))
	    ((eq mode ':forced)
	     (make (Lstructure &flag :forced
		      num-forcings - Integer
		      sit - Situation
		      elapsed - Time-int
		      numsteps - Integer)
		   (+ forcings-so-far 1)
		   sit 0.0 1))
	    (t
	     ':failed))))

(specdecl (act-id* (new-Varid)) - Varid)


Move to 'states-from-old-acts'
(deffunc act-method-thread-gen - Thread-gen
	                 (act - Sexp
			  defn - Action-defn
			  plan-sit - Situation
			  env - Typenv)
   (let ((dom (find-domain-in-vartypes env)))
      (\\ - Thread-cont (c - Thread-cont)
	 (\\ - TC-val (es - Exec-state stack - Thread-stack)
	    (let ((eval-act `(,(car act) ,@(args-eval (cdr act) es))))
	       (let-fun ()
		  (repeat :for ((bdg :in (domain-get-local-bdgs
							 dom))
				(org (own-rule-group dom))
				(found-own false))
		   :result (signal-problem act-method-thread-gen
			      "Can't find method for "
			      eval-act (:q ((not found-own)
				       " (or own rule group either)")))
		   :when (is Rule-group (!_val bdg))
		      (cond ((eq (!_val bdg) org)
			     (!= found-own true)))
		   :within
		      (multi-let (((match meth meth-id meth-env)
				   (action-meth
				      (be Rule-group (!_val bdg))
				      (!_term defn))))
			 (:continue
			  :until match
			  :result (meth-instantiate
				     meth meth-id meth-env))))

      :where

 ;; Finds the first method in 'rg' that matches the action; doesn't
 ;; even try to verify that a precondition is true.
 (:def action-meth - (Mlv Boolean
			   (~ Internalized-method-defn)
			   Varid
			   (~ Bdgenv))
                    (rg - Rule-group term - Sexp)
     (repeat :for ((rule :in (!_rules rg)))
      :result (values false nil ground-id* nil)
      :when (is Method-defn rule)
      :within
         (decl (rule - Method-defn)
            (let ((imeth (be Internalized-method-defn
			     (causation-defn-projectable rule dom)))
		  (meth-varid (new-Varid)))
	       (multi-let (((ok env)
			    (unify eval-act term act-id* meth-varid
				   (empty-env))))
		  (:continue
		   :until ok
		   :result (values true imeth meth-varid env)))))))

 (:def meth-instantiate - TC-val (imeth - Internalized-method-defn
				      meth-id - Varid meth-env - Bdgenv)
    (multi-let (((exp _)
		 (safe-varsubst (!_expansion imeth) meth-id meth-env)))
       (((plan->thread-gen exp plan-sit env) c)
	es (make Thread-stack eval-act stack))))))))))

(deffunc elapsed-sum - Time-int (elap1 elap2 - Sexp)
   (cond ((and (is Float elap1) (is Float elap2))
	  (be Time-int
	      (+ (be Float elap1) (be Float elap2))))
	 (t
	  (signal-problem elapsed-sum
	     "Attempt to find sum of " elap1 " and " elap2))))

(deffunc args-eval - (Lst Obj) (args - (Lst Sexp) estate - Exec-state)
   (bind ((plan-links* (!_vals estate)))
      (<# (\\ (arg)
	     (with estate
		(multi-let (((okay argval)
			     (opt-eval arg !>sit !>bdgs !>domain)))
		   (cond (okay argval)
			 (t arg)))))
	  args)))
;; If arg fails to evaluate, we just use it unevaluated (cf. the 
;; following function).


(deffunc exec-state-opt-eval - Obj (exp - Sexp estate - Exec-state)
   (with estate
      (multi-let (((okay val)
		   (opt-eval exp !>sit !>bdgs !>domain)))
	 (cond (okay val)
	       (t
		(signal-problem exec-state-opt-eval
		   "Problem evaluating " exp
		   :% " with respect to " estate))))))

(deffunc goal-has-assertable-pred - Boolean (goal - Sexp)
   (and (not (eq (car goal) 'not))
	(not (memq (car goal) preds-forbidden-in-dif*))))
