;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: planprob.lisp,v 1.14 2005/11/30 15:53:21 dvm Exp $

(depends-on %module/ nisp)

#+pre-chunk (depends-on %module/ nispobj)

(depends-on :at-run-time %optop/ regressmatch types rmgraph discrtproj)

;;; This file contains various routines used by the procedures in
;;; file search.  The name 'planprob' is rather arbitrary.

(defnisptype Plan-prob-soln
   (Structure plan - Plan-prefix ;;;;(Lst Action)
	      score - Number
	      ;;;id - Varid
	      match - Bdgenv
      (:handler
         (print (pps - Plan-prob-soln srm - Stream)
	    (let ((steps (reverse (!_acts (!_plan pps)))))
	       (out (:to srm)
		  (:pp-block (:pre "#<")
		     "Planning-problem-solution / length "
		     (len steps)
		     (:pp-nl :fill)
		    " / score " (!_score pps)
		     (:pp-nl :fill)
		    " [" (!_match pps) ;;;;"/" (!_id pps)
				      "] "
		      (:pp-nl :fill)
		      steps
		    (:suf ">"))))))))

(specdecl (gc-time-avail* #+allegro true #-allegro false)
	  - Boolean)

;;; !_run-time is total; !_gc-time is portion of !_run-time spent gc-ing,
;;; if available.  Units are in milliseconds.
(defnisptype Run-time-spec
    (Structure run-time gc-time - Integer
       (:handler
	   (print (rts - Run-time-spec srm - Stream)
	      (with rts
		 (out (:to srm) "#<Run-time " !>run-time " ms "
		      (:q (gc-time-avail*
			   (round (/ (* 100.0 !>gc-time) !>run-time))
			   "% gc"))
		      ">"))))))
			   
(deffunc run-time-spec-diff - Run-time-spec (rts2 rts1 - Run-time-spec)
   (make Run-time-spec
	 (- (!_run-time rts2) (!_run-time rts1))
	 (- (!_gc-time rts2) (!_gc-time rts1))))

(specdecl internal-time-units-per-second - Integer)

(deffunc get-current-run-time-spec - Run-time-spec ()
   (multi-let (#+allegro ((non-gc-user non-gc-system gc-user gc-system - Integer)
			  (excl::get-internal-run-times))
	       #-allegro ((non-gc-user non-gc-system gc-user gc-system - Integer)
			  (values (get-internal-run-time) 0 0 0)))
      (let ((non-gc-tot (+ non-gc-user non-gc-system))
	    (gc-tot (+ gc-user gc-system)))
	 (!= non-gc-tot (internal-time-in-milliseconds *-*))
	 (!= gc-tot (internal-time-in-milliseconds *-*))
	 (make Run-time-spec (+ non-gc-tot gc-tot) gc-tot))))

(deffunc internal-time-in-milliseconds - Integer (tm - Integer)
   (cond ((= internal-time-units-per-second 1000) tm)
	 (t (/ (* 1000 tm)
	       internal-time-units-per-second))))

(specdecl (optimism* 0.0) - Float)

(specdecl plan-dbg-amt* - Fixnum
;;;;          plan-dbg*
;;;;          - Boolean
	  maximal-match-count* maximal-match-tot* l-node-tot* g-node-tot*
          unify-tot* unify-success-tot*
	  sit-expansions* - Integer)

(specdecl dom* - Obj pp* - Obj)

#-(or cmu sbcl)
(deffunc plan-problem-initialize - Plan-problem (x)
   (let ((pp (coerce-to-plan-problem x))
         - Uninitialized-plan-problem)
      (let ((sp (!_sit pp)))
         (cond (sp
                (!= pp* pp)
                (let ((dom (!_domain sp)))
                   (!= dom* dom)
                   (cond ((not (is Domain dom))
                          (!= dom (domain-prepare-for-planning dom))))
                   (decl (dom - Domain)
		      (domain-make-current dom))
;;;;                   (cond ((and (!_act-dif-tab dom)
;;;;		               (exists (a in (!_ancestors dom))
;;;;			          (> (!_generation a)
;;;;			             (!_act-dif-time dom))))
;;;;                          (!= (!_achievability dom) '())
;;;;                          (!= (!_act-dif-tab dom) false)))
		   (let ((pprob (opt::problem-initialize pp)))
		      pprob)))
	       (t
		(signal-problem plan-problem-initialize
		   "Plan problem has no initial situation: "
		   pp))))))

#+(or cmu sbcl)
(defun plan-problem-initialize (x)
   (let ((pp (coerce-to-plan-problem x)))
      (let ((sp (opt::Problem-sit pp)))
         (cond (sp
                (!= pp* pp)
                (let ((dom (opt::Initial-situation-domain sp)))
                   (!= dom* dom)
                   (cond ((not (is Domain dom))
                          (!= dom (domain-prepare-for-planning dom))))
                   (domain-make-current dom)
		   (let ((pprob (opt::problem-initialize pp)))
		      pprob)))
	       (t
		(signal-problem plan-problem-initialize
		   "Plan problem has no initial situation: "
		   pp))))))

(specdecl pp* - Obj)

(deffunc coerce-to-plan-problem - (Alt Plan-problem Uninitialized-plan-problem)
                                (x)
           (cond ((is Symbol x)
                  (let ((pp (one-value (get-global-opt-sym x))))
                     (!= pp* pp)
                     (cond ((or (is Plan-problem pp)
                                (is Uninitialized-plan-problem pp))
                            pp)
                           (t
                            (error-break coerce-to-plan-problem :fatal
                               "Not the name of a plan-problem: " x)))))
	         ((or (is Plan-problem x) (is Uninitialized-plan-problem x))
                  x)
                 (t
                  (error-break coerce-to-plan-problem :fatal
                     "Not a plan-problem: " x))))

(defproc verify-domain-okay-for-optop - Void (dom - Domain)
;;;;   (err-out "Checking requirements for " dom " --> " (all-requirements dom))
   (repeat :for ((req :in (all-requirements dom)))
;;;;      (err-out "Checking " req)
      (cond ((not (memq (!_name req)
                        '(:typing :equality :strips :conditional-effects
			  :domain-axioms :processes :durative-actions
	                  :disjunctive-preconditions :existential-preconditions
	                  :expression-evaluation :fluents :numbers
			  :timed-initial-literals
			  :adl :quantified-preconditions :subgoal-through-axioms
			  :universal-preconditions :derived-predicates)))
             (out (:to *query-io*)
                  "Warning -- Optop may have trouble with domain"
		  " declaring requirement "
                  (!_name req) :%)))))             

;;;;(deffunc scores-weigh - Float (act-eff min-eff - Integer
;;;;                               planoid - (~ Action-multiset))
;;;;   (cond ((or (not planoid) (= optimism* 0))
;;;;          act-eff)
;;;;         (t
;;;;          (+ (- act-eff min-eff)
;;;;             (* optimism* (!_size planoid))
;;;;             (* (- 1 optimism*) min-eff)))))

(deffunc plan-prefix-steps - (Lst Action) (pp - Plan-prefix)
   (let-fun ()
      (collect pp '())
    :where
      (collect - (Lst Action) (l - Plan-prefix r - (Lst Action))
          (cond ((!_is-initial l) r)
                (t
                 (collect (!_prior l)
                          (cons (!_last-act l) r)))))))

(deffunc place-Plan-prefix - Plan-prefix
	               (length - Fixnum duration - Time-int
			sit - Situation
			prequel - (~ Plan-prefix-delta)
			init-sit - Situation
			sit-space - Situation-space)
   (find-or-create-Plan-prefix
      true length duration sit prequel init-sit sit-space))

(deffunc find-Plan-prefix - (~ Plan-prefix)
	               (length - Fixnum duration - Time-int
			sit - Situation
			prequel - (~ Plan-prefix-delta)
			init-sit - Situation
			sit-space - Situation-space)
   (find-or-create-Plan-prefix
      false length duration sit prequel init-sit sit-space))

(defproc find-or-create-Plan-prefix - (~ Plan-prefix)
	               (create - Boolean
			length - Fixnum duration - Time-int
			sit - Situation
			prequel - (~ Plan-prefix-delta)
			init-sit - Situation
			sit-space - Situation-space)
   (let ((tree (val-or-initialize
		  (alref (!_pp-tree sit-space) init-sit)
		  :init (make Plan-prefix-tree false !()))))
      (let-fun ((:def down-tree - (~ Plan-prefix-tree)
			   (tr - Plan-prefix-tree ppd - (~ Plan-prefix-delta))
		   (cond (ppd
			  (let ((sub-tr (down-tree tr
						   (!_prequel (!_prior ppd)))))
			     (cond (sub-tr
				    (maybe-make-a-sub-tree sub-tr ppd))
				   (t
				    ;; Subtle: If 'create' = true, then
				    ;; sub-tr will never be false.  So
				    ;; 'create' must be false.--
				    false))))
			 (t tr)))
		(:def maybe-make-a-sub-tree - (~ Plan-prefix-tree)
			   (tr - Plan-prefix-tree ppd - (~ Plan-prefix-delta))
		   (let ((e (alref (!_succs tr)
				   (!_last-act ppd))))
		      (or e
			  (cond (create
				 (let ((new-sub-tr
					  (make Plan-prefix-tree
						false !())))
				    (!= (!_succs tr)
					(cons (tuple (!_last-act ppd)
						     new-sub-tr)
					      *-*))
				    new-sub-tr))
				(t false))))))
	 (let ((sub-tr (down-tree tree prequel)))
	    (cond (sub-tr
		   (or (!_pp sub-tr)
		       (cond (create
			      (let ((new-pp (make Plan-prefix
						  length duration sit prequel)))
				 (!= (!_pp sub-tr) new-pp)
				 new-pp))
			     (t false))))
		  (t false))))))

(specdecl actspec-count* - Integer)

(defproc effort-stats-out - Void (new-goal-nodes - Integer)
   (!= maximal-match-tot* (+ *-* maximal-match-count*))
   (!= unify-tot* (+ *-* unify-count*))
   (!= unify-success-tot* (+ *-* unify-success-count*))
   (!= g-node-tot* (+ *-* new-goal-nodes))
   (cond ((> plan-dbg-amt* 1)
;;;;	  (let ((l-node-count
;;;;		   (</ (\\ (tot ei)
;;;;			  (+ tot (opt::exptree-size
;;;;				  (opt::exp_index-tree (!_l-nodes ei)))))
;;;;		       0  (!_contents last-effort-tab*)))
;;;;		(g-node-count
;;;;		   (</ (\\ (tot ei)
;;;;			  (+ tot (opt::exptree-size
;;;;				  (opt::exp_index-tree (!_effort-specs ei)))))
;;;;		       0  (!_contents last-effort-tab*))))
;;;;	  (!= l-node-tot* (+ *-* l-node-count))
	  (out (:to *error-output*)
		"Maximal matches: "
		(:f "~6D" maximal-match-count*)
		
;;;;	      "  L-nodes: "
;;;;	      (:f "~6D" l-node-count)
	      "  New goal nodes: "
	      (:f "~6D" new-goal-nodes)
	      :% "   Total so far: "
	      (:f "~6D" maximal-match-tot*)
;;;;	      "           " (:f "~6D" l-node-tot*)
	      18 (:f "~6D" g-node-tot*)
	      (:q (track-maxmatch-size*
		   :% " Biggest maximal match so far: " biggest-maxmatch*))
	      :% "Unifications: " unify-success-count* "/" unify-count*
		" in maximal matcher: " mm-unify-count*
	      :% " Effort computations:: " effort-changes*
	       "   Actspec propagations: " actspec-count*
	      :%))))

;;; This procedure relies on the fact that the initial mark on a situation
;;; is 0, and is reset only in 'next-search-state' (see file search).
(defproc evaluation-sits-flush - Void (sits - (Lst Situation)
				       sit-space - Situation-space)
   (cond ((not (null sits))
	  (let ((to-visit sits) 
		(real-visited !())
		- (Lst Situation)
		(sit-index (!_index sit-space))
		- Sit-eq-index)
	     ;; We use the terms "real" and "fantasy" to distinguish situations
	     ;; in the search space from those generated purely to evaluate
	     ;; plan prefixes.  We don't need a queue of visited fantasy situations
	     ;; because they will be marked -1 as they are visited.
	     (repeat :for (next-sit - Situation)
	      :until (null to-visit)
		 (!= next-sit (car to-visit))
		 (!= to-visit (cdr *-*))
		 (cond ((= (!_mark next-sit) 0)
			(!= (!_mark next-sit) -1)
			(opt::eq-index-remove next-sit sit-index))
		       (t
			(!= real-visited (cons next-sit *-*))))
		 (repeat :for ((succ :in (!_succs next-sit sit-space)) - Sitsuccspec
			       (flushing-succs !()) - (Lst Sitsuccspec)
			       sit - (~ Situation))
		    (!= sit (!_sit succ))
		    (cond (sit
			   (cond ((= (!_mark sit) 0)
				  (!= flushing-succs (cons succ *-*))
				  (!= to-visit (cons sit *-*)))
				 ((and (not (= (!_mark sit) -1))
				       (not (memq sit real-visited)))
				  (!= to-visit (cons sit *-*))))))
		  :result (!= (!_succs next-sit sit-space)
			      (complementq *-* flushing-succs))))))))

;;; Not currently used
(defproc must-project-fresh - (~ Situation)
                             (sit - Situation act - Action
			      sit-space - Situation-space th - Domain
			      mark - Sitmark markit - Boolean)
  (let ((succ (one-value (project false sit act sit-space th))))
     (cond (succ
            (cond ((or (not (equ (!_mark succ) mark))
	               (< (+ (len (!_path sit)) 1)
			  (len (!_path succ))))
	           (cond ((= (!_mark succ) mark)
		          (sit-reencounter-describe succ sit)))
	           (cond (markit (!= (!_mark succ) mark)))
	           (!= (!_path succ)
	               (cons act (!_path sit)))
                   succ)
                  (t false)))
           (t
	    (progn (error-break must-project-fresh
			  "Act " act " infeasible in " sit
			  (:novalue "I'll skip it"))
		   false)))))

(deffunc kosher-extension - Boolean (preserve-only - Boolean
                                     act - Feasible-action-spec
                                     pp - Plan-prefix)
   (cond (preserve-only
          (and (not (eq (!_pp act) pp))
               (repeat
                :until (!_is-initial pp)
                :result true
                :until (eq (!_prior pp) (!_pp act))
                :result (not (eq (!_last-act pp) (!_act act)))
                  (!= pp (!_prior pp)))))
         (t true)))

(specdecl succ* - Situation)

(defproc sit-reencounter-describe - Void (succ sit - Situation)
;   (!= succ* succ)
   (dbg-out (> plan-dbg-amt* 1)
      "Re-encounter " succ :%
      "old pathlen = " (len (!_path succ)) " new pathlen = "
      (+ 1 (len (!_path sit))) :%))
;   (cond ((= (len (!_path succ)) 6)
;          (error-break sit-reencounter-describe :novalue
;                       "Bingo"))))

(specdecl (plan-shower* false) - (~ (Fun Void (Plan-prefix))))

(defproc show-current-plan - Void (plan-prefix - Plan-prefix
				   score - Float
				   discrep - Integer)
   (out (:to *error-output*)
	sit-expansions* " Trying (length " (!_length plan-prefix)
	                          "/score " (:f "~,1F" score)
				  "/discrep " discrep "): ")
   (cond (plan-shower*
	  (funcall plan-shower* plan-prefix))
	 (t
	  (out (:to *error-output*)
	     :% 5
	     (:pp-block 
		(:e
		 (let-fun ((show-next - Void (pp - Plan-prefix)
			     (cond ((not (!_is-initial pp))
				    (show-next (!_prior pp))
				    (:o (!_last-act pp)
				       1 (:pp-nl :fill))))))
		    (show-next plan-prefix)))))))
   (out (:to *error-output*) :%))

(deffunc list-scramble - (Lst Obj) (l - (Lst Obj))
   (cond ((or (null l) (null (cdr l))) l)
	 (t
	  (let-fun ((scr - (Lst Obj) (l - (Lst Obj))
		       (let ((i (fxrandom (len l))))
			  (let ((tl (nthcdr i l)))
			     (cons (car tl)
				   (let ((l (nconc (ldiff l tl)
						   (cdr tl))))
				      (cond ((null (cdr l)) l)
					    (t (scr l))   )))   ))))
	     (scr (copy-list l))   ))   ))

(deffunc random-sample - (Lst Obj) (l - (Lst Obj) size - Fixnum)
   (repeat :for ((i = 1 :to size) - Fixnum
		(samp '()) 
		(rem (list-copy l)) - (Lst Obj)
                 e - (Lrcd Sexp Number))
      (!= e (randelt rem))
      (!= samp (cons e *-*))
      (!= rem (dremove1q e *-*))
     :result samp))

(deffunc randelt - Obj (l - (Lst Obj))
   (list-elt l (fxrandom (len l)))   )

(specdecl (max-incoherence* 3) - Fixnum)
          
(deffunc action-incoherence - Fixnum (a - Action rt - Reduction-tree)
      (let-fun ((walk-tree - (Multvals inco size - Fixnum)
			  (subtree - Reduction-tree left-branch - Boolean)
		  ;; 'left-branch' is true iff we're going down the leftmost
		  ;; branch of this tree, i.e., we're on the way to the
		  ;; action actually chosen.
		  ;; Two values returned are incoherence of 'a' assuming it
		  ;; occurred in the subtree; plus size of subtree.
		  ;; incoherence = max-incoherence* if it doesn't occur.
		  (cond ((and left-branch
			      (null (!_preconds subtree)))
			 ;; We've reached the chosen action
			 (values max-incoherence* 0))
			(t
			 (repeat :for ((pre :in (!_preconds subtree))
				       - Reduction-tree
				       (lb left-branch :then false)
				       - Boolean
				       sub-inco sub-size
				       (best-inco max-incoherence*)
				       (tot-size 1) ;; for this node itself
				       - Fixnum)
			    (!= < sub-inco sub-size >
				(walk-tree pre lb))
			    (!= tot-size (+ *-* sub-size))
			    (!= best-inco (min *-* sub-inco))
			  :result (values
				     (cond ((acts-match
					       (!_achiever subtree)
					       a)
					    (min best-inco tot-size))
					   (t
					    best-inco))
				     tot-size)))))
                  (acts-match - Boolean (a1 a2 - Action)
		     (or (eq a1 a2)
			 (let ((t1 (!_term a1))
			       (t2 (!_term a2)))
			    (and (consp t1)
				 (consp t2)
				 (memq (car t1)
				       '(wait-while wait-non-negative))
				 (memq (car t2)
				       '(wait-while wait-non-negative)))))))
	 (one-value (walk-tree rt true))))

;;; For debugging
(deffunc cause-int - (~ Internalized-causation-defn)
	    (sym - Symbol dom - Domain)
   (let ((v (find-domain-bdg-val sym dom)))
      (cond ((is Causation-function v)
	     (causation-defn-internalize
		(!_defn v) dom))
	    (t
	     (out (:to *error-output*)
		"Not a Causation-function: " v)
	     false))))
      