;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: procmod.lisp,v 1.8 2004/11/08 04:02:47 dvm Exp $

(depends-on %module/ nisp)

(depends-on :at-run-time %opt/ deduction index walkdom)

(depends-on :at-run-time %optop/ types domain projutil)


(deffunc construct-process-model - (Multvals model
					       - (Lst Fluent-change-model)
					     kosher - Boolean
					     triggers
					       - (Lst Triggered-effect))
                      (actives dormants - (Lst Process-instance)
		       sit - Situation dom - Domain)
   (multi-let (((influencees clean)
		(find-changers actives sit dom)))
      (letrec ()
	 (values influencees
		 clean
		 (nconc (<! (\\ (a) (find-triggered-effects
				        (!_effect a) influencees sit dom))
			    actives)
			(cont-trigs actives 'stopping)
			(cont-trigs dormants 'starting)))
       :where
	  (:def cont-trigs - (Lst Ineq-trigger)
		      (procs - (Lst Process-instance)
		       which - (Const stopping starting))
             (<! (\\ (proc)
		    (<! (\\ (c)
			   (be (Lst Ineq-trigger)
			       (<? is Ineq-trigger
				   (continuous-trigs
				      c (list `(process-transition
						  ,which
						  ,(!_term (!_uterm proc))))
				      influencees sit dom))))
			(cond ((eq which 'stopping)
			       (ineq-dnf-neg (!_continuous proc)))
			      (t (!_continuous proc)))))
		 procs)))))

(deffunc ineq-dnf-neg - (Lst (Lst Uterm-ineq)) (dl - (Lst (Lst Uterm-ineq)))
   (cond ((null dl)
	  (list '()))
	 (t
	  (let ((rn (ineq-dnf-neg (cdr dl))))
	     (repeat :for ((c :in (car dl)))
	      :nconc
		   (let ((nc (ineq-literal-neg c)))
		      (repeat :for ((n :in rn))
		       :collect (cons nc n))))))))


(deffunc ineq-literal-neg - Uterm-ineq (l - Uterm-ineq)
   `(,(be (Const < =< >= >) (ineq-rel-negation (car l))) ,@(cdr l)))

;;; Specifies that !_depquant's derivative is !_deriv, or,
;;; if !_total is false, has !_deriv as a component.
(defnisptype Deriv-spec (Lstructure () depquant - Uterm
				       total - Boolean
				       deriv - Uterm-exp))

(specdecl (forceproj-dbg* false) - Boolean)

(deffunc find-changers - (Multvals models - (Lst Fluent-change-model)
				   kosher - Boolean)
                      (actives - (Lst Process-instance)
		       sit - Situation dom - Domain)
   (let ((effs (<# !_effect actives)))
      (repeat :for ((eff :in effs) - Sexp
		    (influencers '()) - (Lst Uterm)
		    (influencees '()) - (Lst Deriv-spec))
         (let ((derivs (collect-active-derivatives
			    eff (new-Varid) (empty-env)
			    sit dom))
	       - (Lst Deriv-spec))
	    (repeat :for ((d :in derivs))
	       (!= influencees
		   (cons d *-*))
	       (!= influencers
		   (nconc (collect-uterms (!_deriv d))
			  *-*))))
       :result (progn
		  (!= influencers (nodupq *-*))
		  (cond ((exists (ftee :in influencees)
			    (memq (!_depquant ftee) influencers))
			 (signal-problem find-changers
					 :fatal
			    "Model too hairy"))
			(t
			 (repeat :for ((ftee :in influencees
					     :tail lftee)
				       - Deriv-spec
				       (nodup-ftees !()) - (Lst Deriv-spec)
				       (kosher true) - Boolean)
			     (repeat :for ((ftee2 :in (cdr lftee)))
			      :result (!= nodup-ftees
					  (cons ftee *-*))
			      :until (and (eq (!_depquant ftee2)
					      (!_depquant ftee))
					  (and (or (!_total ftee)
						   (!_total ftee2))))
			      :result
				 (cond ((not (equal (!_deriv ftee)
						    (!_deriv ftee2)))
					(!= kosher false)
					(dbg-out forceproj-dbg*
					   "Taking linear combination of"
					   " inconsistent deriv specs:"
					   :% ftee :% " vs. " ftee2)
					(!= nodup-ftees
					    (cons ftee *-*)))))
			  :result (values
				     (nconc (elapsing-models actives sit dom)
					    (build-linear-time-function
					       nodup-ftees sit dom))
				     kosher))))))))

(deffunc collect-active-derivatives - (Lst Deriv-spec)
	        (eff - Sexp aid - Varid bdgs - Bdgenv
		 sit - Situation dom - Domain)
   (multi-let (((eff bdgs)
		(safe-varsubst eff aid bdgs)))
      (cond ((atom eff)
	     (signal-problem collect-active-derivatives
		"Bogus effect: " eff))
	    (t
	     (case (car eff)
		(and
		  (<! (\\ (c) (collect-active-derivatives c aid bdgs sit dom))
		      (cdr eff)))
		(when
		 (repeat :for ((a :in (deduce (cadr eff) aid bdgs sit dom)))
		  :nconc (collect-active-derivatives
			     (caddr eff) aid a sit dom)))
		(forall
		   (deduce-forall-derivs (caddr eff) aid bdgs sit dom))
		((derivative deriv-comp)
		 (list (make Deriv-spec
			     (uniquify-term
				(varsubst-to-constant (cadr eff) aid bdgs)
				dom)
			     (eq (car eff) 'derivative)
			     (deriv-exp-utermify
				(varsubst-to-constant (caddr eff) aid bdgs)
				dom))))
		(t !()))))))

(deffunc deduce-forall-derivs - (Lst Deriv-spec)
	                     (effect - Sexp did - Varid
			      subst - Bdgenv sit - Situation
			      dom - Domain)
   (labels ((find-whens (eff)
               (cond ((eq (car eff) 'and)
		      (<! find-whens (cdr eff)))
                     ((eq (car eff) 'when)
                      (let ((ant (cadr eff)) (ceff (caddr eff)))
                         (multi-let (((many ansl)
				      (deduce-check-many
				         ant did subst sit dom)))
                             (cond (many
                                    (signal-problem deduce-forall-derivs
				       "Can't project (when " ant 1 ceff ")"
				       t "   unclear"
				       " number of objects satisfying when"
				       " clause"))
                                   (t
				    (repeat :for ((ans :in ansl))
				     :nconc (collect-active-derivatives
					       ceff did ans sit dom)))))))
                     (t
                      (signal-problem deduce-forall-derivs
			 "Can't handle forall without when: "
                             effect)))))
      (find-whens effect)))

(deffunc collect-uterms - (Lst Uterm) (ue - Uterm-exp)
   (cond ((is Uterm ue) (list ue))
	 ((atom ue) !())
	 (t
	  (<! collect-uterms
	      (be (Lst Uterm-exp) ue)))))

;;;;(specdecl (same-as* (\\ (tt) tt))
;;;;	  - (Fun Float (Time-int)))

(deffunc elapsing-models - (Lst Fluent-change-model)
	           (active-procs - (Lst Process-instance)
		    sit - Situation dom - Domain)
   (repeat :for ((ap :in active-procs))
    :collect (let ((elapexp (be Fluexp `(elapsed ,(!_term (!_uterm ap))))))
		(let ((cur-elap (fluent-current-value-float
				    elapexp sit dom)))
		   (make Fluent-change-model
			 (uniquify-term elapexp dom)
			 (\\ (tm - Time-int) (be Float (+ cur-elap tm)))
			 (be Uterm-exp
			     `(+
				(,cur-elap = ,elapexp)
				*T*)))))))

(deffunc build-linear-time-function - (Lst Fluent-change-model)
			 (influencees - (Lst Deriv-spec)
			  ;;;;influencers - (Lst Uterm)
			  sit - Situation dom - Domain)
   ;; First collect all deriv-comp's together
   (let ((deriv-tab !()) - (Lst (Lrcd Uterm (Lst Uterm-exp))))
      (repeat :for ((ee :in influencees))
	 (!= (alref deriv-tab (!_depquant ee) !())
	     (cons (!_deriv ee) *-*)))
      (letrec ((evaluate-fluents - (Mlv Float Uterm-exp) (e - Uterm-exp)
		  (cond ((is Uterm e)
			 (let ((ev (fluent-current-value-float 
				      (be Fluexp (!_term e))
				      sit dom)))
			    (values ev (be Uterm-exp `(,ev = ,e)))))
			((is CSymbol e)
			 (let ((bdg (find-domain-bdg e dom)))
			    (cond ((and bdg (is Domain-var (!_val bdg)))

				   (let ((val (!_(Domain-var val)
						 (!_val bdg))))
				      (values val (be Uterm-exp `(,val = ,e)))))
				  (t
				   (signal-problem build-linear-time-function
						   :fatal
				      "Not evaluable: " e)))))
			((is Number e)
			 (values (float e) e))
			((not (memq (strip-car (be Sexp e))
				    '(+ - * / fl-^)))
			 (decl (e - Sexp)
			    (multi-let (((okay val)
					 (opt-eval e sit !() dom)))
			       (cond ((and okay (is Number val))
				      (values val (be Uterm-exp `(,val = ,e))))
				     (t 
				      (signal-problem
				               build-linear-time-function
					 :fatal
					 "Non-numerical value in derivative: "
					 e " => " val))))))
			(t
			 (decl (e - (Lrcd Symbol . (Lst Uterm-exp)))
			    (multi-let (((argvals argexps)
					 (evaluate-fluent-args (cdr e))))
			       (let ((fn (car e)))
				  (strip-sym-with-type fn)
				  (values
				     (apply
					(case fn
					   (+ #'+)
					   (- #'-)
					   (* #'*)
					   (/ #'/)
					   (fl-^ #'identity)
					   (t
					    (signal-problem
						     build-linear-time-function
					       :fatal
					       "Can't handle function "
					       (car e))))
					argvals)
				     `(,fn ,@argexps))))))))
	       (evaluate-fluent-args - (Mlv (Lst Float) (Lst Uterm-exp))
				     (args - (Lst Uterm-exp))
		  (repeat :for ((arg :in args)
				:collectors vals exps)
		     (multi-let (((v exp)
				  (evaluate-fluents arg)))
			(one-collect vals v)
			(one-collect exps exp))
		   :result (values vals exps))))

	 (multi-let (((cur-ee-vals cur-ee-exps)
		      (evaluate-fluent-args (<# car deriv-tab)))
		     ((deriv-ee-vals deriv-ee-exps)
		      (repeat :for ((ee :in deriv-tab) - (Lrcd Uterm (Lst Uterm-exp))
				    :collectors d-vals - Float d-exps - Uterm-exp)
			 (multi-let (((infl-vals infl-exps)
				      (evaluate-fluent-args (cadr ee))))
			    (one-collect d-vals (<< + infl-vals))
			    (one-collect d-exps `(+ ,@infl-exps)))
		       :result (values d-vals d-exps))))
	    (<# (\\ (ee cur-val cur-exp deriv-val deriv-exp)
		   (make Fluent-change-model
			 (car ee)
			 (\\ (elap-time - Time-int)
			    ;; since sit's time, whatever that means
			    (+ cur-val (* deriv-val elap-time)))
			 `(+ ,cur-exp (* ,deriv-exp *T*))))
		deriv-tab
		cur-ee-vals cur-ee-exps
		deriv-ee-vals deriv-ee-exps)))))

(deffunc find-triggered-effects - (Lst (Alt Ineq-trigger Sexp))
	            (exp - Sexp 
		     changing - (Lst Fluent-change-model)
		     sit - Situation dom - Domain)
   (letrec ()
      (<# car (find-in-effect exp (new-Varid) '()))
    :where

 (:def find-in-effect - (Lst (Lrcd (Alt Ineq-trigger Sexp) Bdgenv))
          (exp - Sexp id - Varid bdgs - Bdgenv)
;;;;   (trace-around find-in-effect
;;;;      (:> "(find-in-effect: " exp "/" id 1 :% bdgs ")")
   (multi-let (((exp bdgs)
		(safe-varsubst exp id bdgs)))
      (cond ((atom exp) '())
	    (t
	     (match-cond exp
		?( (when ?pre ?eff)
		  (multi-let ((conds
			       (process-condition-classify pre dom)))
		    (repeat :for ((c :in conds)
				  - (Lstructure ()
				       discrete - (Lst Sexp)
				       continuous
					 - (Lst (Lst (Lrcd Ineq-rel Sexp Sexp))))
				  al - (Lst Bdgenv))
		       (!= al (deduce `(and ,@(!_discrete c)) id bdgs sit dom))
   ;;;;		    (dbg-save c id bdgs sit dom (d `(and ,@(!_discrete c))))
   ;;;;		    (breakpoint find-in-effect
   ;;;;		       "d = " (g d) :% "al = " al)
		     :nconc (find-in-when-effect (!_continuous c) eff id al))))
		?( (and ?@conjl)
		  (<! (\\ (d) (find-in-effect d id bdgs))
		      conjl))
		?( (?(:\| < =< = >= >) ?@_)
		  (signal-problem find-in-effect
		    "Unexpected inequality " exp
		    (:continue "I will ignore it"))
		  !())
		?( (?(:\| derivative deriv-comp) ?@_)
		  !())
		(t
		 (cond ((has-qvars exp)
			(signal-problem find-in-effect
			   "Effect " exp " has unbound variables")))
		 (list (tuple exp bdgs)))))))
;;;;      (:< (val &rest _) "find-in-effect: " val))
   )

 (:def find-in-when-effect - (Lst (Lrcd Triggered-effect Bdgenv))
		   (continuous-conds - (Lst (Lst (Lrcd Ineq-rel Sexp Sexp)))
		    eff - Sexp id - Varid al - (Lst Bdgenv))
;;;;    (trace-around find-in-when-eff
;;;;       (:> "(find-in-when-eff: " continuous-conds :% 3 eff ")")
    (<! (\\ (a)
	   (let ((subs (find-in-effect eff id a)))
	      (cond ((null subs)  ;; probably impossible
		     !())
		    (t
		     (repeat :for ((sub :in subs))
		      :nconc (multi-let (((ccl - (Lst (Lst (Lrcd Ineq-rel
								 Sexp Sexp)))
					   env - Bdgenv)
					  (safe-varsubst
					     continuous-conds id (cadr sub))))
;;;;				(dbg-save ccl continuous-conds id sub subs)
;;;;				(breakpoint find-in-when-effect
;;;;				    "ccl = " ccl)
				(repeat :for ((cc :in ccl))
				 :nconc
				    (let ((ineq-conj
					     (be (Lst (Lrcd Ineq-rel
							    Uterm-exp Uterm-exp))
						 (<# (\\ (e)
							(deriv-exp-utermify e dom))
						     (varsubst-to-constant cc id env)))))
				       (<# (\\ (te)
					      (tuple te env))
					   (continuous-trigs
					       ineq-conj
					       (list (car sub))
					       changing sit dom))))))))))
	al)
;;;;       (:< (val &rest _) "find-in-when-eff: " val))
    )))

(deffunc continuous-trigs - (Lst Triggered-effect)
			  (continuous-conds
			     - (Lst (Lrcd Ineq-rel Uterm-exp Uterm-exp))
			   effect - (Lst Triggered-effect)
			   influencees
			     - (Lst Fluent-change-model)
			   sit - Situation dom - Domain)
   (cond ((null continuous-conds)
	  effect)
	 (t
	  (multi-let (((ineq time-fcn symfcn)
		       (ineq-normalize (car continuous-conds)
				       influencees sit dom)))
	     (let ((time-fcn-val
		      (funcall time-fcn 0.0)))
		(cond ((< time-fcn-val 0)
		       ;; Not yet
		       (list (make Ineq-trigger
				   ineq
				   time-fcn
				   symfcn
				   (continuous-trigs
				      (cdr continuous-conds) effect
				      influencees sit dom))))
		      (t
		       ;; Already triggered
		       (continuous-trigs
			   (cdr continuous-conds)
			   effect influencees sit dom))))))))
				      
(deffunc deriv-exp-utermify - Uterm-exp (e - Sexp dom - Domain)
   (uniquify-continuous-fluent-subterms e dom))

(deffunc uniquify-continuous-fluent-subterms - Uterm-exp
	                        (exp - Sexp dom - Domain)
      (letrec ((harden - Uterm-exp (e - Sexp)
;;;;		  (trace-around harden
;;;;		     (:> "(harden: " e ")")
		  (cond ((atom e) e)
			(t
			 (let ((fn (strip-car e)))
			    (cond ((eq fn 'fl-v)
				   (uniquify-term (cadr e) dom))
				  (t
				   (be Uterm-exp
				       `(,fn ,@(<# harden (cdr e)))))))))
;;;;		     (:< (val &rest _) "harden: " val))
		  ))
	 (harden exp)))

(deffunc ineq-normalize - (Mlv Uterm-exp (Fun Float (Time-int)) Uterm-exp)
	               (ineq - (Lrcd Ineq-rel
				     Uterm-exp Uterm-exp)
			influencees - (Lst Fluent-change-model)
			sit - Situation dom - Domain)
   (multi-let (((vl tfl sfl)
		(evaluate-to-fcn (cadr ineq) influencees sit dom))
	       ((vr tfr sfr)
		(evaluate-to-fcn (caddr ineq) influencees sit dom)))
      (letrec ()
;;;;	 (dbg-save ineq vl tfl vr tfr influencees sit dom)
;;;;	 (breakpoint ineq-normalize
;;;;	    "tfl(0) = " (val-at-0 vl tfl)
;;;;	    :% "tfr(0) = " (val-at-0 vr tfr))
	 (cond ((or (memq (car ineq) '(> >=))
		    (and (eq (car ineq) '=~)
			 (< (val-at-0 vl tfl)
			    (val-at-0 vr tfr))))
		(values `(- ,(cadr ineq) ,(caddr ineq))
			(fcns-minus (- vl vr) tfl tfr)
			`(- ,sfl ,sfr)))
	       (t
		(values `(- ,(caddr ineq) ,(cadr ineq))
			(fcns-minus (- vr vl) tfr tfl)
			`(- ,sfr ,sfl))))
       :where
          (fcns-minus - (Fun Float (Time-int))
		      (r - Float
		       f1 f2 - (~ (Fun Float (Time-int))))
	     (cond ((and f1 f2)
		    (\\ (ti - Time-int)
		       (+ r (- (funcall f1 ti) (funcall f2 ti)))))
		   (f1
		    (\\ (ti - Time-int)
		       (be Float (+ r (funcall f1 ti)))))
		   (f2
		    (\\ (ti - Time-int)
		       (be Float (- r (funcall f2 ti)))))
		   (t
		    (\\ (_ - Time-int)
		       r)))))))

(specdecl (way-in-the-future* (expt 10 10.0))
	  - Float)

(deffunc get-trigger-time - (Multvals time - (Alt Time-int (Const *never))
				      effects 
					- (Lst Sexp)
				      new-epsilon - (~ Float))
	         (pc - Process-complex dom - Domain)
   (let ((tt (!_trigger-time pc)))
      (cond ((eq tt '*unknown)
	     (let ((grain (domain-temporal-grain-size dom)))
		(multi-let (((time - (~ Time-int)
			      effects - (Lst Sexp)
			      new-eps - (~ Float))
			     (compute-trigger-time
				(!_triggers pc)
				grain)))
		   (let ((rtime (cond (time (be Time-int time))
				      (t '*never)))
			 - (Alt Time-int (Const *never)))
		      (!= (!_trigger-time pc)
			  (cond ((eq rtime '*never)
				 '*never)
				(t
				 (tuple (be Time-int rtime)
					effects
					(be Float new-eps)))))
		      (values rtime effects new-eps)))))
	    ((eq tt '*never)
	     (values '*never !() false))
	    (t
	     (decl (tt - (Lrcd Time-int
			       (Lst Sexp)
			       Float))
		(be * (list->values tt)))))))

(deffunc domain-temporal-grain-size - Float (dom - Domain)
	     (let ((grain (find-domain-bdg 'temporal-grain-size dom)))
		(cond (grain
		       (be Float (!_(Domain-var val) (!_val grain))))
		      (t
		       (signal-problem domain-temporal-grain-size
			  "Domain " dom
			  " contains no declaration of temporal-grain-size"
			  :fatal)))))


(specdecl (trigger-time-dbg* false) - Boolean)

(deffunc compute-trigger-time - (Multvals time - (~ Time-int)
					  effects 
					    - (Lst Sexp)
					  new-epsilon - (~ Float))
	         (model - (Lst Ineq-trigger) approx-interval - Float)
   (cond ((null model)
	  (values 0.0 '() false))
	 (t
;;;;	  (dbg-save model approx-interval)
;;;;	  (breakpoint compute-trigger-time
;;;;	     "Computing trigger time for " model)
	  (repeat :for ((delta-t = 1.0 :then (* 2 delta-t)) - Float
			triggered
			   - (Lst Ineq-trigger))
	   :until (> delta-t way-in-the-future*)
	   :result (values false '() false)
	     (!= < triggered _ >
		 (find-triggered model delta-t))
	   :while (null triggered)
	   :result (binary-search-for-crossover triggered delta-t)

	   :where

 ;; Ineq-triggers in first returned list have all "sub-Ineq-triggers"
 ;; resolved away.  
 (:def find-triggered - (Multvals now later - (Lst Ineq-trigger))
	     (trigs - (Lst Ineq-trigger)
	      delta-t - Float)
;;;;    (trace-around find-triggered
;;;;       (:> " find-triggered: " delta-t 1 trigs)
    (repeat :for ((trig :in trigs) - Ineq-trigger
		  (triggered '())
		  (postponed '())
		      - (Lst Ineq-trigger))
     :result (values triggered postponed)
       (dbg-out trigger-time-dbg*
	  trig " @ " delta-t t " = " (funcall (!_timefcn trig) delta-t) :%)
       (cond ((< (funcall (!_timefcn trig) delta-t)
		 0)
	      (!= postponed (cons trig *-*)))
	     ((exists (eff in (!_effects trig))
		 (is Ineq-trigger eff))
	      (multi-let (((subtrigs - (Lst Ineq-trigger)
			    immeds - (Lst Sexp))
			   (be * (classify (!_effects trig)
				     (\\ (e)
					(is Ineq-trigger e))))))
		  (!= triggered
		      (cons (make Ineq-trigger
				  (!_fluent trig)
				  (!_timefcn trig)
				  (!_symfcn trig)
				  immeds)
			    *-*))
		  (multi-let (((now-new-trigs later-new-trigs)
			       (find-triggered subtrigs delta-t)))
		     (!= triggered
			 (nconc (trigfcns-combine
				   trig now-new-trigs)
				*-*))
		     (!= postponed
			 (nconc (trigfcns-combine
				   trig later-new-trigs)
				*-*)))))
	     (t
	      (!= triggered (cons trig *-*)))))
;;;;       (:< (now later) " find-triggered: now: " now t " later: " later))
)

 (:def trigfcns-combine - (Lst Ineq-trigger)
		   (trig - Ineq-trigger tl - (Lst Ineq-trigger))
    (repeat :for ((sub in tl))
     :collect (make Ineq-trigger	  
		    `(max ,(!_fluent trig)
			  ,(!_fluent sub))
		    (timefcns-compose
		       (be (Fun Float Float) #'max)
		       (!_timefcn trig)
		       (!_timefcn sub))
		    `(max ,(!_symfcn trig) ,(!_symfcn sub))
		    (!_effects sub))))

 (:def binary-search-for-crossover - (Multvals time - Time-int
					  effects 
					    - (Lst ;;;;(Alt
					           ;;;; (Const :process-change)
					           ;;;; ...)
						   Sexp)
					  new-epsilon - Float)
                   (triggered - (Lst Ineq-trigger) right - Float)
;;;;    (trace-around bin-search
;;;;       (:> "bin-search: " right 1 triggered)
    (repeat :for ((left 0) middle - Float)
       ;; Invariant: triggered is not ()
       (!= middle (/ (+ left right) 2.0))
     :while (> (- right left) approx-interval)
     :while (< middle right)
     :result (progn
	        (cond ((>= middle right)
		       (dbg-save :run-loud left right middle)
		       (breakpoint binary-search-for-crossover
			  "Bogosity")))
		(values (be Time-int right)
			(be * (<$ !_effects triggered))
			(epsilon-value triggered middle)))
       (dbg-out trigger-time-dbg*
	    "Bracket: [" left ", " middle ", " right "] range = "
	    (- right left) :%)
       (cond ((= middle right)
	      (dbg-save :run-loud left middle right)
	      (signal-problem binary-search
		 "oops")))
       ;; Invariant: Not triggered at 'left', but triggers by time 'right'
       (repeat :for ((trig :in triggered) - Ineq-trigger
		     (trigtoleft '()) (trigtoright '())
		     - (Lst Ineq-trigger))
          (dbg-out trigger-time-dbg*
	     trig " @ " middle :%
	       " = " (funcall (!_timefcn trig) middle) :%)
;;;;	  (cond ((= (funcall (!_timefcn trig) middle)
;;;;		    0.0)
;;;;		 (dbg-save trig middle left right)
;;;;		 (breakpoint binary-search-for-crossover
;;;;		    "Nail on head")))
	  (cond ((> (funcall (!_timefcn trig) middle)
		    0)
		 (!= trigtoleft (cons trig *-*)))
		(t
		 (!= trigtoright (cons trig *-*))))
	:result (cond ((null trigtoleft)
		       (!= left middle)
		       (!= triggered trigtoright))
		      (t
		       (!= right middle)
		       ;; We've discovered that some trigger earlier
		       ;; than the previous time we found.
		       ;; Anyone who triggers later is a loser.
		       (!= triggered trigtoleft)
		       ;;;;(!= untriggered (nconc trigtoright *-*))
		       ))))
;;;;       (:< (time effects new-eps) "bin-search: time = " time :% " effects = "
;;;;	                  effects
;;;;	   :% " new-eps = " new-eps))

    )

 (:def epsilon-value - Float (tl - (Lst Ineq-trigger) middle - Float)
    (repeat :for ((trig :in tl)
		  (eps epsilon*))
       (let ((v (funcall (!_timefcn trig) middle)))
	  (!= eps (max *-* (/ (abs v) (+ 1 (abs v))))))
     :result eps))
 ))))

(deffunc timefcns-compose - (Fun Float (Time-int))
	               (fcn - (Fun Float Float)
			&rest l - (Fun Float (Time-int)))
   (\\ (ti - Time-int)
      (apply fcn (<# (\\ (tf) (funcall tf ti))
		     l))))

(deffunc val-at-0 - Float (v - Float f - (~ (Fun Float (Time-int))))
   (cond (f (+ v (funcall f 0.0)))
	 (t v)))

;;; For efficiency (premature, no doubt), this returns 
;;; a function of time represented with a constant offset (the first
;;; component).  Second component is false if the function is a constant.
(deffunc evaluate-to-fcn - (Mlv Float
				(~ (Fun Float (Time-int)))
				Uterm-exp)
	       (exp - Uterm-exp
		changing - (Lst Fluent-change-model)
		sit - Situation dom - Domain)
   (letrec ()
      (evaluate-fluents exp)
    :where
      
   (evaluate-fluents - (Mlv Float (~ (Fun Float (Time-int))) Uterm-exp)
		     (e - Uterm-exp)
;;;;      (trace-around eval-fluents
;;;;	 (:> " (eval-fluents: " e ")")
      (cond ((is Uterm e)
	     (let ((p (lookup-fluent-change e changing)))
		(cond (p
		       (values 0.0 (!_timefcn p) (!_symfcn p)))
		      (t
		       (values (fluent-current-value-float
				  (be Fluexp (!_term e))
				  sit dom)
			       (be * false)
			       `(current-value ,e))))))
	    ((is Number e)
	     (values (float e) (be * false) e))
	    ((is CSymbol e)
	     (let ((bdg (find-domain-bdg e dom)))
		(cond ((and bdg (is Domain-var (!_val bdg)))

		       (let ((val (!_(Domain-var val)
				     (!_val bdg))))
			  (cond ((is Number val)
				 (values (float val) (be * false) e))
				(t
				 (signal-problem evaluate-to-fcn
				    :fatal
				    "Value of " e " is non-number " val)))))
		      (t
		       (signal-problem evaluate-to-fcn
				       :fatal
			  "Not evaluable: " e)))))
	    ((atom e)
	     (signal-problem evaluate-to-fcn
			     :fatal
		"Non-numerical value in numerical fluent expression: "
		e))
	    ((contains-uterms e)
	     (decl (e - (Lrcd Symbol . (Lst Uterm-exp)))
		(let ((argvals
			 (repeat :for ((a :in (cdr e)) - Uterm-exp)
			  :collect (values->list 
				      (evaluate-fluents a))))
		      - (Lst (Lrcd Float (~ (Fun Float (Time-int))) Uterm-exp)))
		   (let ((fn (car e)))
		      (strip-sym-with-type fn)
		      (case fn
			 (-
			  (cond ((= (len argvals) 1)
				 (unary-minus (car argvals)))
				(t
				 (argvals-plus
				    (list (car argvals)
					  (values->list
					     (unary-minus
						(values->list
						   (argvals-plus
						      (cdr
							argvals))))))))))
			 (+
			  (argvals-plus argvals))
			 ((* /)
			  (argvals-prod (be (Const * /) fn)
					argvals))
			 (t (signal-problem
					 evaluate-to-fcn :fatal
					 "Can't handle function "
					 fn)))))))
	    (t
	     (multi-let (((okay val) (opt-eval (be Sexp e) sit !() dom)))
		(cond (okay
		       (cond ((is Number val)
			      (values (float val) (be * false) e))
			     (t
			      (signal-problem evaluate-to-fcn :fatal
				 "Value of " e " is non-number " val))))
		      (t
		       (signal-problem evaluate-to-fcn :fatal
			  "Expression " e " cannot be evaluated"))))))
;;;;	 (:< (val fcn) "eval-fluents: " val " + " fcn))
      )

   (contains-uterms - Boolean (e - Uterm-exp)
      (cond ((is Uterm e) true)
	    ((atom e) false)
	    (t
	     (exists (x :in (be (Lst Uterm-exp) e))
		(contains-uterms x)))))

   (argvals-prod - (Mlv Float (~ (Fun Float (Time-int))) Uterm-exp)
		 (fn - (Const * /)
		  argvals
		    - (Lst (Lrcd Float (~ (Fun Float (Time-int))) Uterm-exp)))
      (let ((fcn (cond ((eq fn '/) #'/) (t #'*))))
	 (cond ((null argvals)
		(values 1.0 (be * false) 1.0))
	       ((exists (av in argvals)
		   (cadr av))
		(let ((argfcns
			 (<# (\\ (av)
				(cond ((cadr av)
				       (\\ (ti - Time-int)
					  (+ (car av)
					     (funcall (cadr av)
						      ti))))
				      (t
				       (\\ (_ - Time-int) (car av)))))
			     argvals))
		      - (Lst (Fun Float (Time-int))))
		   (values 0.0
			   (\\ (ti - Time-int)
			      (</ (\\ (prod argfcn)
				     (be Float
					 (funcall fcn
						  prod (funcall argfcn ti))))
				  (funcall (car argfcns) ti)
				  (cdr argfcns)))
			   (argvals-sym-combine fn argvals))))
	       (t
		(values (</ (\\ (prod argval)
			       (funcall fcn prod (car argval)))
			    (car (car argvals))
			    (cdr argvals))
			(be * false)
			(argvals-sym-combine fn argvals))))))

   (argvals-sym-combine - Uterm-exp (fn - Symbol
				     argvals - (Lst (Lrcd Float
							  (~ (Fun Float (Time-int)))
							  Uterm-exp)))
      `(,fn ,@(<# (\\ (av) `(+ ,(car av) ,(caddr av)))
		  argvals)))

   (argvals-plus - (Mlv Float (~ (Fun Float (Time-int))) Uterm-exp)
		(argvals
		    - (Lst (Lrcd Float (~ (Fun Float (Time-int))) Uterm-exp)))
      (cond ((null argvals)
	     (values 0.0 (be * false) 0.0))
	    (t
	     (let ((constpiece
		      (</ (\\ (tot av)
			     (+ tot (car av)))
			  (car (car argvals))
			  (cdr argvals)))
		   (argfcns
		      (repeat :for ((av :in argvals))
		       :when (cadr av)
		       :collect (cadr av))))
	       (values constpiece
		       (cond ((null argfcns) (be * false))
			     (t
			      (\\ - Float (ti - Time-int)
				 (repeat :for ((af :in argfcns)
					       (tot 0.0))
				    (!= tot (+ *-* (let ((fcn af))
						      (funcall fcn ti))))
				  :result tot))))
		       (cond ((null argfcns) constpiece)
			     (t `(+ ,@(<# caddr argvals)))))))))

   (unary-minus - (Mlv Float (~ (Fun Float (Time-int))) Uterm-exp)
		(argval
		   - (Lrcd Float (~ (Fun Float (Time-int))) Uterm-exp))
      (values (- (car argval))
	      (cond ((cadr argval)
		     (minusfcn (cadr argval)))
		    (t (be * false)))
	      `(- ,(caddr argval))))

   (minusfcn - (Fun Float (Time-int))
	     (af - (Fun Float (Time-int)))
      (\\ (ti)
	 (- (funcall af ti))))))

(deffunc fluent-current-value-float - Float
	                          (flu - Fluexp sit - Situation dom - Domain)
   (multi-let (((got-it val)
		(fluent-current-value flu sit dom)))
      (cond (got-it
	     (cond ((is Number val)
		    (float val))
		   (t
		    (signal-problem fluent-current-value-float
		       "Fluent " flu " has non-Float val " val))))
	    (t 10.0)
	    ;; -- stopgap
;;;;	     (signal-problem evaluate-fluents
;;;;		"No value for " flu " in " sit)
	     )))

(deffunc lookup-fluent-change - (~ Fluent-change-model)
	                   (fl - Uterm fcml - (Lst Fluent-change-model))
   (repeat :for ((fcm :in fcml))
    :result false
    :until (eq (!_fluent fcm) fl)
    :result fcm))

