;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: projutil.lisp,v 1.15 2005/11/30 15:53:21 dvm Exp $

(depends-on %module/ nisp)

(depends-on :at-run-time %optop/ types)
;;;;			 %opt/ deduction

(specdecl (eff-varid* (new-Varid)) - Varid)

(deffunc deduce-effects - (Multvals adds deletes - (Lst Occasion))
			(effect - Sexp did - Varid subst - Bdgenv
			 sit - Situation dom - Domain)
   (multi-let (((ok a d subst)
		(opt::expression-car-cdr effect did subst false)))
      (cond (ok
	     (op-deduce-effects a d did subst sit dom))
	    ((eq a 'true)
	     (values !() !()))
	    (t
	     (signal-problem deduce-effects
		"Weird effect: " effect 1 did 1 subst
		(:proceed "I will ignore it"))
	     (values !() !())))))

(deffunc op-deduce-effects - (Multvals adds deletes - (Lst Occasion))
	           (op - Symbol args - Sexp did - Varid subst - Bdgenv
		    sit - Situation dom - Domain)
   (letrec ()
      (case op
	 (and
	   (let ((adds '()) (deletes '()) al dl)
	     (dolist (e args)
		(multiple-value-setq
		       (al dl)
		       (deduce-effects e did subst sit dom))
		(setf adds (nconc al adds))
		(setf deletes (nconc dl deletes)))
	     (values adds deletes)))
	 (not
	  (values '()
		  (find-occasion (car args) subst)))
	 ((change assign)
	  (fluent-change-effects (car args) (cadr args)
				 did subst sit dom))
	 ((increase decrease)
	  (fluent-change-effects (car args)
				 `(,(cond ((eq op 'decrease) '-)
					  (t '+))
				   (fl-v ,(car args)) ,(cadr args))
				 did subst sit dom))
	 (when
	   (let ((adds '()) (deletes '()) al dl)
	      (dolist (a (deduce (car args) did subst sit dom))
		 (multiple-value-setq (al dl)
				      (deduce-effects (cadr args)
						      did a sit dom))
		 (setf adds (nconc al adds))
		 (setf deletes (nconc dl deletes)))
	      (values adds deletes)))
	 (forall
	    (deduce-forall-effects (cadr args) did subst sit dom))
	 (t
	  (values (find-occasion `(,op ,@args) subst)
		  '())))
   :where
     (find-occasion (a subst)
     	(let ((p (unsafe-varsubst a did subst)))
	   (cond ((has-qvars p)
		  (signal-problem op-deduce-effects
		     "Attempting to treat " p " as an occasion"
		     (:novalue "I will pretend I didn't see it"))
		  '())
		 (t
		  (list (uniquify-occasion p dom))))))))

(deffunc fluent-change-effects - (Multvals adds deletes - (Lst Occasion))
	                       (flu newflu - Sexp did - Varid subst - Bdgenv
				sit - Situation dom - Domain)
   (let ((flu (unsafe-varsubst flu did subst))
	 (newflu (unsafe-varsubst newflu did subst)))
      (cond ((or (has-qvars newflu) (has-qvars flu))
	     (signal-problem fluent-change-effects
		"Unexpected variables trying to set "
		flu 
		:% " to " newflu))
	    (t
	     (multi-let (((okay new-val)
			  (opt-eval newflu sit !() dom)))
		(cond (okay
		       (values (list (uniquify-occasion
				          `(current-value
					      ,flu ,new-val)
					  dom))
			       (nconc (assertion-find-and-uniquify
				         `(current-value ,flu ?old)
					 did subst sit)
				      (assertion-find-and-uniquify
				         `(= (fl-v ,flu) ?old)
					 did subst sit))))
		      (t
		       (dbg-save :run-loud newflu sit dom)
		       (signal-problem fluent-change-effects
			  "Can't evaluate new value of " flu
			  :% " = " newflu))))))))

(deffunc assertion-find-and-uniquify - (Lst Occasion) (pat id subst sit)
   (<# second (opt::sit-assertion-fetch pat id subst false sit)))

;;;; replaced by assertion-find-and-uniquify above --
;;;;			       (multi-let (((found1 cur-val)
;;;;					    (try-extract-one-val
;;;;					       (deduce `(current-value ,flu ?old)
;;;;						       did subst sit dom)
;;;;					       '?old did)))
;;;;				  (cond (found1
;;;;					 (list (uniquify-occasion
;;;;						  `(current-value ,flu ,cur-val)
;;;;						  dom)))
;;;;					(t !())))

(deffunc try-extract-one-val - (Mlv Boolean Obj)
	                    (ansl - (Lst Bdgenv) e - Sexp id - Varid)
   (repeat :for ((a :in ansl)
		 (answers !(Obj)))
    :result (cond ((null answers)
		   (values false nil))
		  ((null (cdr answers))
		   (values true (car answers)))
		  (t
		   (signal-problem extract-one-val
		      "Failed to extract one value for " e "/" id
		      " from"
		      :% ansl
		      :% "Got values: " answers)))
      (let ((v (varsubst-to-constant e id a)))
	 (cond ((computational v)
		(!= answers (adjoin v *-* :test #'equalp)))))))

(deffunc deduce-forall-effects - (Multvals adds deletes - (Lst Occasion))
	                     (effect - Sexp did - Varid
			      subst - Bdgenv sit - Situation
			      dom - Domain)
   (letrec ((find-whens (eff)
               (cond ((eq (car eff) 'and)
                      (let ((adds !()) (deletes !())
			    - (Lst Occasion))
                         (do ((el (cdr eff) (cdr el)))
                             ((null el)
                              (values adds deletes))
                           (multi-let (((al dl)
                                        (find-whens (car el))))
                              (setq adds (nconc al adds))
                              (setq deletes (nconc dl deletes))))))
                     ((eq (car eff) 'when)
                      (let ((ant (cadr eff)) (ceff (caddr eff)))
                         (multi-let (((many ansl)
				      (deduce-check-many
				          ant did subst sit dom)))
                             (cond (many
                                    (signal-problem deduce-forall-effects
				       "Can't project (when " ant 1 ceff ")"
				       t "   unclear"
				       " number of objects satisfying when"
				       " clause"))
                                   (t
                                    (let ((adds '()) (deletes '()))
                                      (dolist (s ansl)
                                        (multi-let (((al dl)
						     (deduce-effects
						        ceff did s sit dom)))
                                          (setq adds (nconc al adds))
                                          (setq deletes (nconc dl deletes))))
                                      (values adds deletes)))))))
                     (t
                      (signal-problem deduce-forall-effects
		         "Can't handle forall without when: "
                             effect)))))
     (find-whens effect)))

(specdecl patho-count* - Integer)

(specdecl (watch-sit* false) - (~ Situation))

(defproc delete-and-add - Void (adds deletes - (Lst Occasion)
				id - Varid bdgs - Bdgenv new-sit - Situation
				dom - Domain)
			       (ignore id bdgs)
   (letrec ()
      (repeat :for ((d :in deletes))
	 (assertion-erase (!_propo d) new-sit dom)
	 )
      (repeat :for ((a :in adds))
	 (assertion-add (!_propo a) new-sit dom)
	 )
      ;;;;(bind ((opt::exptree-dbg* true))
    :where
      ))

(deffunc new-succ-sit - Situation (sit - Situation
				   act - (Alt Action (Const wait transition)))
   (let ((ind-copy (opt::exp-index-copy (!_diff-from-init sit))))
      (make-inst Subsequent-situation
	 :frozen false
	 :diff-from-init ind-copy
	 :contents '*uncomputed
	 :path (cons act
		     (!_path sit))
	 :init (find-init-situation sit))))

(defproc succ-record - (~ Situation) ; uniquified if non-false
                 (forced - Boolean
		  sit - Situation
		  elapsed - (Alt Time-int (Const *never))
		  steps - Integer
		  act - (Alt Action (Const wait transition actual-init))
		  succ - (~ Situation) ; false if action not actually
		                       ; projectable
		  adds dels - (Lst Occasion)
		  frac - Float
		  assumptions - (Lst Sexp)
		  sit-space - Situation-space)
   (let ((sss (find-old-projection sit act sit-space)))
      (cond (sss
	     (cond ((not (eq forced (> (!_forcings sss) 0)))
		    (signal-problem succ-record
		       "Action " act " both feasible and infeasible in "
		       sit)))
	     (!_sit sss))
	    (t
	     (succ-create
	        (cond (forced 1) (t 0))
		sit elapsed steps act succ adds dels frac assumptions sit-space)))))

(deffunc find-old-projection - (~ Sitsuccspec)
	                     (sit - Situation
			      act - (Alt Action
					 (Const wait transition actual-init))
			      sit-space - Situation-space)
   (repeat :for ((sss :in (!_succs sit sit-space)))
    :result false
    :until (eq (!_act sss) act)
    :result sss))

(defproc succ-create - (~ Situation) ; uniquified if non-false
                 (forcings - Integer
		  sit - Situation
		  elapsed - (Alt Time-int (Const *never))
		  steps - Integer
		  act - (Alt Action (Const wait transition actual-init))
		  succ - (~ Situation) ; false if action not actually
		                       ; projectable
		  adds dels - (Lst Occasion)
		  frac - Float
		  assumptions - (Lst Sexp)
		  sit-space - Situation-space)
   (let (;;;;(dom (initial-situation-domain
	 ;;;;	 (find-init-situation sit)))
	 (usucc (and succ (uniquify-situation succ (!_index sit-space)))))
      (cond ((not (be Obj act))
	     (dbg-save :run-loud act succ)
	     (breakpoint succ-create
		"About to make bad succ for: " act)))
      (cond ((and (is Action act)
		  (eq (car (!_(Action term) act)) 'transition)
		  (eq usucc sit))
	     (dbg-save :run-loud act sit)
	     (breakpoint succ-create
		"About to link " sit " to itself via "
		:% act)))
      (!= (!_succs sit sit-space)
	  (cons (make-inst Sitsuccspec
		   :forcings forcings
		   :elapsed elapsed
		   :steps steps
		   :outlink false
		   :act act
		   :sit usucc
		   :adds adds
		   :deletes dels
		   :assumptions assumptions
		   :adjustments !()
		   :frac frac)
		*-*))
      usucc))

(defproc succ-erase - Void (sit - Situation act - Action sit-space - Situation-space)
   (!= (!_succs sit sit-space)
       (delete-if (\\ (s - Sitsuccspec)
		     (eq (!_act s) act))
		  *-* )))

(deffunc add-del-combine - (Multvals adds dels - (Lst Occasion))
	                 (adds-so-far deletes-so-far
			  new-adds new-deletes
			  - (Lst Occasion))
   (values (union new-adds (complementq adds-so-far new-deletes))
	   (union new-deletes (complementq deletes-so-far new-adds))))

(deffunc process-find - (~ Process-instance) (procterm - Sexp sit - Situation)
	 (repeat :for ((pinst :in (!_active (!_processes sit))))
	  :result false
	  :until (equal (!_term (!_uterm pinst))
			procterm)
	  :result pinst))

(deffunc is-wait - (Multvals procterm negterm - (~ Sexp))
	          (act - Action)
   (let ((negterm false) (procterm false)- (~ Sexp))
      (matchq ?(:\| (wait-non-negative ?negterm ?procterm)
		    (wait-while ?procterm))
	      (!_term act))
      (values procterm negterm)))

;;; For debugging
(deffunc sss-succs - (Alt (Lst Sitsuccspec)
			  (Const :nosuccessors))
                   (sl - (Lst Sitsuccspec) n - Fixnum sit-space - Situation-space)
   (let ((sit (sss-sit sl n)))
      (cond (sit
	     (!_succs sit sit-space))
	    (t
	     ':nosuccessors))))

(deffunc sss-sit - (~ Situation) (sl - (Lst Sitsuccspec) n - Fixnum)
   (cond ((and (>= n 0)
	       (< n (length sl)))
	  (!_sit (list-elt sl n)))
	 (t false)))

;;;;(deffunc sit-patho - Boolean (sit - Situation)
;;;;   (!= patho-count* (+ *-* 1))
;;;;   (cond ((signed-index-patho (!_diff-from-init sit))
;;;;	  (cond ((signed-index-patho (!_diff-from-init sit))
;;;;		 true)
;;;;		(t
;;;;		 (dbg-save sit)
;;;;		 (signal-problem sit-patho
;;;;		    "Transient pathology"))))
;;;;	 (t false)))

;;;;(defun signed-index-patho (ei)
;;;;   (bind ((opt::allow-exptree-rehash* false))
;;;;      (and (exists (socc :in (opt::exp-fetch '(current-value (traffic source-4 dest-4)
;;;;					                     1)
;;;;					     ei false))
;;;;	      (equal (caddr (opt::Occasion-prop (opt::Signed-occasion-occ socc)))
;;;;			  1))
;;;;	   (forall (socc :in (opt::exp-fetch '(current-value (traffic source-4 dest-4)
;;;;					                     ?v)
;;;;					     ei false))
;;;;	      (not (equal (caddr (opt::Occasion-prop (opt::Signed-occasion-occ socc)))
;;;;			  1))))))
