;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: regress.lisp,v 1.14 2005/12/10 03:01:27 dvm Exp $

(depends-on %module/ nisp)

(depends-on :at-run-time %optop/ types domain fluregress)

;;;;(deffunc test - (Lst Obj Obj) () (list test1* test2*))

(defproc place-regressions - (Lst Regress-method)
                           (l-node - Sexp lid - Varid th - Domain)
   (cond ((not (!_act-dif-tab th))
          (!= (!_act-dif-tab th)
              (make-act-dif-tab))))
   (let ((fetched 
	    (repeat :for ((cand :in (exp-fetch l-node (!_act-dif-tab th) true))
		       - Regress-spec)
	     :result (new-Regress-spec th l-node)
	     :until (opt::variants (!_literal cand) l-node
				  lit-id* lid '())
	     :result cand))
	 - Regress-spec)

;;;;      (multiple-value-let (ok env)
;;;;			  (unify l-node (!_literal fetched)
;;;;				 lid rs-id* '())
;;;;	      (cond ((not ok)
;;;;		     (signal-problem place-regressions
;;;;			"Variants don't unify: "
;;;;			l-node 1 (!_literal fetched))))
;;;;	      (<# (\\ (m)
;;;;		     (regress-method-subst m lid env)) ...))
      (!_methods fetched)))

(defproc new-Regress-spec - Regress-spec
                          (th - Domain eff - Sexp)
    (let ((rs (make Regress-spec eff '())))
       (!= (!_methods rs)
           (all-regression-methods eff th))
       (exp-ob-index rs (!_act-dif-tab th) true)
       rs))

(deffunc all-regression-methods - (Lst Regress-method)
                                (eff - Sexp th - Domain)
   (multi-let (((polarity p)
                (lit-analyze eff)))
      (nconc (matching-regression-methods polarity p th)
	     (fluent-regression-methods polarity p th))))

(deffunc matching-regression-methods - (Lst Regress-method)
	                     (polarity - Boolean p - Sexp th - Domain)
   (let-fun ()
      (nconc
	 (repeat :for ((cdefn :in (direct-causers p th)) - Causation-defn)
	  :nconc
	    (<! finish-it
		(causation-defn-regress
		   cdefn
		   (\\ - (Lst Embryo-regress-method)
		       (idefn - Internalized-causation-defn
			effect - Sexp)
		      (regress-unify
			 polarity p effect
			 lit-id* effect-id* (empty-env)
			 (!_condition idefn)
			 '(and) th))
		   th)))
	 (repeat :for ((rule :in (causal-impliers p th)) - Occasion)
	  :nconc
	    ;; Deductive rules don't conclude (not p), so we
	    ;; always collect those for p in this case and
	    ;; let the regressmatch code sort everything out.
	    (let ((erml
		     (rule-regress-unify
			polarity p true (!_propo rule)
			lit-id* effect-id* (empty-env) '(and) th)))
	       (repeat :for ((erm :in erml))
		   (!= (!_view erm) rule)
		   (!= (!_act-sort erm) '*watch))
	       (<! finish-it erml))))
    :where
    
       (:def finish-it (erm - Embryo-regress-method)
	  (finish-regress-method erm (cond (polarity p) (t `(not ,p)))
				 th))))

;;; For each conjunction, there's a bdgenv such that 
; if it obtains, then each element of conjunction must be true to affect
; LIT; and there's an :ADD or :DEL flag that says how it is affected.
; In the :DEL case, we obtain preservation preconditions (in
; regressmatch.nsp) by finding ways to delete every instance of this
; conjunction.

(deffunc regress-unify - (Lst Embryo-regress-method)
                       (polarity - Boolean lit - Sexp effects - Sexp
			lit-id eff-id - Varid env - Bdgenv
                        precond secondary - Sexp dom - Domain)
   (cond ((or (is Qvar effects) (atom effects))
	  (leaf-regress-unify polarity lit true effects
			      lit-id eff-id env
			      precond secondary dom))
	 (t
	  (selq (car effects)
	     (and (repeat :for (e :in (cdr effects))
		   :nconc 
		     (regress-unify polarity lit e
					      lit-id eff-id env
					      precond secondary dom)))
	     ((when if)
	      (regress-unify
	          polarity lit (caddr effects)
		  lit-id eff-id env
		  precond `(and ,(if-when-precond
				    (car effects) (cadr effects))
				,secondary)
			     dom))
	     (forall
	      (regress-unify polarity lit (caddr effects)
			     lit-id eff-id env precond secondary dom))
	     (not
	      (leaf-regress-unify polarity lit false (cadr effects)
				  lit-id eff-id env precond secondary dom))
	     (t
	      (leaf-regress-unify polarity lit true effects
				  lit-id eff-id env precond secondary dom))))))
   
(deffunc leaf-regress-unify - (Lst Embryo-regress-method)
                            (polarity - Boolean lit - Sexp
                             leaf-polarity - Boolean leaf - Sexp
			     lit-id eff-id - Varid env - Bdgenv
                             precond secondary - Sexp dom - Domain)
   (multi-let (((leaf env)
		(safe-varsubst leaf eff-id env)))
      (cond ((is Qvar leaf)
	     (cond (leaf-polarity
		    (try-reduce-precond
		       polarity lit leaf lit-id eff-id env
		       precond secondary dom))
		   (t
		    (out (:to *error-output*)
			 "Can't handle negated variable effect: " leaf :%)
		    '())))
	    ((atom leaf) '())
	    (t
	     (matching-regressions polarity lit leaf-polarity leaf
				   lit-id eff-id env
				   precond secondary)))))

;; For each conjunctive path through the preconds, 
;; find a precond that can be reduced in fact-sit of dom, thus binding
;; leaf (a Qvar).
(deffunc try-reduce-precond - (Lst Embryo-regress-method)
                       (polarity - Boolean lit - Sexp leaf - Qvar
			lit-id eff-id - Varid env - Bdgenv
                        precond secondary - Sexp dom - Domain)
   (let ((dnf-pre (disjunctive-normal-form
		     precond true))
	 (dnf-sec (disjunctive-normal-form
		     secondary true)))
      (nconc (<! (\\ (pconj)
		    (let ((rl (reduce-to-rebind
				 pconj leaf eff-id eff-id
				 env dom)))
		       (<! (\\ (r)
			      (regress-unify 
				 polarity lit (cadr r)
				 lit-id eff-id (caddr r)
				 `(and ,@(car r)) secondary dom))
			   rl)))
		 dnf-pre)
	     (<! (\\ (sconj)
		    (let ((rl (reduce-to-rebind
				 sconj leaf eff-id eff-id
				 env dom)))
		       (<! (\\ (r)
			      (regress-unify 
				 polarity lit (cadr r)
				 lit-id eff-id (caddr r)
				 precond `(and ,@(car r)) dom))
			   rl)))
		 dnf-sec))))

(deffunc rule-regress-unify - (Lst Embryo-regress-method)
	        (lit-polarity - Boolean lit rule-polarity - Boolean rule - Sexp 
		 lit-id rule-id - Varid env - Bdgenv condition - Sexp dom - Domain)
  ;; Facts never conclude (not p) directly, given that we're in a closed
  ;; world.  Instead, to make (not p) true you can collect all the ways to
  ;; infer $p$ and defeat them.  This approach presupposes that an inferrable
  ;; pred is not also an achievable one.
   (cond ((is Qvar rule)
	  (let ((dnf-pre (disjunctive-normal-form condition true)))
	     (repeat :for ((disjunct :in dnf-pre))
	      :nconc
		(let ((rl (reduce-to-rebind
				 disjunct rule rule-id rule-id env dom)))
		   (repeat :for ((trip :in rl))
		      ;; trip is (new conjuncts, new qvar val, new env)
		    :nconc (rule-regress-unify
			      lit-polarity lit rule-polarity (second trip)
			      lit-id rule-id (third trip)
			      `(and ,@(first trip) ,condition) dom))))))
	 ((atom rule) !())
	 (t
	  (let ((fctn (car rule)))
	     (opt::strip-sym-with-type fctn)
	     (let-fun ()
		(case fctn
		  ((and or)
		   (cond ((boole-eq rule-polarity (eq fctn 'and))
			  (repeat :for ((conjunct :in (cdr rule)))
			   :nconc
			      (rule-regress-unify lit-polarity lit rule-polarity conjunct
						  lit-id rule-id env
						  condition dom)))
			 (t
			  (repeat :for ((disjunct :in (cdr rule) :tail dtail))
			   :nconc
			      (rule-regress-unify
				 lit-polarity lit rule-polarity disjunct
				 lit-id rule-id env
				 `(and ,condition
				       (let ((sibs (append (ldiff (cdr rule) dtail)
							   (cdr disjunct))))
					  (cond (rule-polarity
						 (<# fmla-negate sibs))
						(t sibs))))
				 dom)))))
		   ((if -> >-)
		    (impl-regress-unify (cadr rule) (caddr rule)))
		   ((<-)
		    (impl-regress-unify (caddr rule) (cadr rule)))
		   ((not)
		    (rule-regress-unify lit-polarity lit
					(not rule-polarity) (cadr rule)
					lit-id rule-id env condition dom))
		   (t
		    (matching-regressions
		       lit-polarity lit rule-polarity rule lit-id rule-id env 
		       condition '(and))))
	      :where
	        (:def impl-regress-unify - (Lst Embryo-regress-method)
		                   (ante conse - (Lst Sexp))
		   (cond (rule-polarity
			  (rule-regress-unify lit-polarity lit true conse
					      lit-id rule-id env
					      `(and ,ante ,condition) dom))
			 (t
			  (nconc (rule-regress-unify true lit false ante
						     lit-id rule-id env
						     condition dom)
				 (rule-regress-unify false lit false conse
						     lit-id rule-id env
						     condition dom))))))))))

(deffunc matching-regressions - (Lst Embryo-regress-method)
                            (lit-polarity - Boolean lit - Sexp
                             leaf-polarity - Boolean leaf - Sexp
			     lit-id eff-id - Varid env - Bdgenv
			     precond secondary - Sexp)
   (multi-let (((ok e)
		(unify lit leaf lit-id eff-id env)))
      (cond (ok
	     (let ((which (cond ((boole-eq lit-polarity leaf-polarity)
				 ':add)
				(t
				 ':del)))
		   - (Const :add :del))
	        (list (make Embryo-regress-method
			    which false false eff-id e
			    precond secondary))))
	    (t '()))))

;;; SUP, if non-nil, is the common ancestor of all bdgenvs in R1 and R2.
;;; This is not currently used.
(deffunc regress-conjoin - (Lst (Lrcd Bdgenv . (Lst Sexp)))
                         (r1 r2 - (Lst (Lrcd Bdgenv . (Lst Sexp)))
			  sup - (~ Bdgenv))
   (repeat :for (c1 :in r1)
    :nconc
      (repeat :for (c2 :in r2)
       :nconc
	 (multiple-value-let (consis ce)
				    (bdgenvs-conjoin (car c1) (car c2) sup)
		   (cond (consis
			  ; Doesn't check for obviously inconsistent conjuncts
			  (list (cons ce (append (cdr c1) (cdr c2)))))
			 (t '()))))))


(deffunc bdgenvs-conjoin - (Mlv Boolean (~ Bdgenv))
                         (e1 e2 - Bdgenv sup - (~ Bdgenv))
   (repeat :for ((l1 e1 (cdr l1)) - (Lst Varbdg)
		 (r e2) - Bdgenv
		 b1 - Varbdg (ok true) - Boolean)
    :until (or (null l1) (eq l1 sup))
    :result (values true r)
      (!= b1 (car l1))
      (!= < ok r >
          (with (!_val b1)
             (var-unify (!_varname b1) !>skel
                        (!_id b1)      !>id
                        r)))
    :while ok
    else :result (values false r)))


;;;;(deffunc regress-method-subst - Regress-method
;;;;			      (m - Regress-method lid - Varid env - Bdgenv)
;;;;   (multiple-value-let (ok env)
;;;;		       (bdgenvs-combine env (!_bdgs m))
;;;;      (cond (ok
;;;;	     (multiple-value-let (a env)
;;;;				 (varid-convert (!_action m) (!_rid m) lid env)
;;;;		(multiple-value-let (p env)
;;;;				    (varid-convert (!_precond m) (!_rid m) lid env)
;;;;		   (make Regress-method
;;;;			 (!_change m) false false lid env a p))))
;;;;	    (t
;;;;	     (signal-problem regress-method-subst
;;;;		"Envs apparently inconsistent: "
;;;;		:% 3 env
;;;;		:% 3 (!_bdgs m))))))

(deffunc disjunctify-and-conjoin - (Lst (Lst Sexp)) (e1 e2 - Sexp)
   (dnf-conjoin (disjunctive-normal-form e1 true)
		(disjunctive-normal-form e2 true)))

;;; For debugging
(deffunc add-methods - (Lst Regress-method) (rml - (Lst Regress-method))
   (<? (\\ (rm - Regress-method) (eq (!_change rm) ':add))
       rml))