;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: regressmatch-exc.lisp,v 1.7 2006/01/24 13:33:46 dvm Exp $

(depends-on %module/ nisp)

#+pre-chunk (depends-on %module/ nispobj)

(depends-on :at-run-time %optop/ maxmatch regress types
                        queue)

(specdecl ;;;;(plan-dbg* true) - Boolean
          (infinite-effort* 1000000) - Fixnum)

(defnisptype Effort-spec :forward)

(defnisptype Reduction-step :forward)

;;; Why efforts have to be Fixnums: because in rmgraph we use them
;;; as subscripts in an array used as a queue.

; Records summary of c-nodes & cohorts below an effort-spec
; EFFECTIVE is t if it's part of the (really, an) effective subgraph
; below its parent effort-spec.
;;; Vars in !_linkenv use same Varid as 'reducee'.
;;; !_linkenv gives a consistent translation of !_reducee vars to !_precond
;;; vars.
(defnisptype Reduction
   (Structure reducee - Effort-spec
	      effective - Boolean
	      effort - Fixnum
	      instances - (Alt (Const *unknown)
			       (Lst Bdgenv))
	      linkenv - Bdgenv
	      action - (~ Sexp)  ;; NOT uniquified
		; If action = false, that means preservation
		; of LITERAL through last action in
		; plan prefix.
	      preconds - (Lst Effort-spec)
	      ;;;;constraints - (Lst Sexp)
	      step - (~ Reduction-step)
      (:handler
          (print (red - Reduction srm - Stream)
	     (with red
		(out (:to srm)
		   (:pp-block (:pre "#{")
		      "Reduction"
		      (:q (!>effective "*")) "/" !>effort
		      1 (:pp-nl :linear)
		      !>reducee
		      1 (:pp-nl :linear)
		      !>action
		      1 (:pp-nl :linear)
		      !>preconds
		      (:q ((not (eq !>instances '*unknown))
			   1 (:pp-nl :linear)
			   (len (be (Lst Bdgenv) !>instances))
			   " instances"))
		      (:q (!>step
			   1 (:pp-nl :linear)
			   "Step: " !>step))
		      (:suf "}")))))))

;; We don't need the varid because it's the same as the effort-spec this
;; reduces.
   (conser (reducee - Effort-spec
	    linkenv - Bdgenv
            action - Sexp
            preconds - (Lst Effort-spec))
	    ;;;;constraints - (Lst Sexp)
      (make ^^ reducee false
	       (cond ((null preconds) 0) (t infinite-effort*))
	       '*unknown
	       linkenv action preconds false)))

(defnisptype Red-inst-rcd (Lrcd (~ Reduction)
				(Lst (Lrcd Sexp Bdgenv))))

(defnisptype Reduction-skel (Structure act quant - Sexp
				       match - Maximal-match
				       rml - (Lst Regress-method)))
(defnisptype L-node
    (Structure occ - Occasion
	       pp - Plan-prefix
	       height - Fixnum
	       status - (Const *top *true *trivial *unchecked
			       *checked-instances *checked-all)
	       ;; -- *true means that it's ground and true in the
	       ;; current situation.  *trivial means *some instance*
	       ;; of it is ground and true, but picking one would
	       ;; be counterproductive, because the vars should
	       ;; be bound by other constraints (e.g., in
	       ;; consistent-action-combination).
	       instances - (Lst Occasion)
	       regressions - (Lst Reduction-skel)
	       effort-specs - (Lst Effort-spec)
               achiever - (~ (Lrcd Method-defn 
       (:handler
	   (print (ln - L-node srm - Stream)
	      (with ln
		 (out (:to srm)
		      (:pp-block (:pre "#<")
			 "L-node "
			 (:pp-ind :block 3)
			 (:pp-nl :linear)
			 (!_propo !>occ)
			 1 (:pp-nl :linear)
			 "@" (!_length !>pp) " ^" !>height
			 1 !>status
			 (:q ((memq !>status
				    '(*checked-instances *checked-all))
			      1 (:pp-nl :linear)
			      "Instances: "
			      (:pp-nl :linear)
			      !>instances))
			 (:q ((eq !>status '*checked-all)
			      1 (:pp-nl :linear)
			      (len !>regressions)
			      " regressions"))
			 (:suf ">")))))))
  (literal - Sexp :inline (ln - L-node) (!_propo (!_occ ln))))

;;; Effort-specs can have variables, which appear in any Reduction with the
;;; same 'varid'.
;;;  However, in the next layer down, the effort-specs made from 
;;; these preconds will have their own varid's.

(defnisptype Effort-spec
	(Structure
	   l-node - L-node
	   conjuncts - (Lrcd Symbol . (Lst Sexp))
	   mark - Fixnum
	   varid - Varid
	   effort - Fixnum
	   ;; Instances derived from all
	   ;; effective children --
	   instances - (Alt (Const *unknown) (Lst Red-inst-rcd))
	   ;; This will made more specific
	   ;; in rmgraph --
	   actions - Obj
	   parents 
	   children - (Lst Reduction)
	   ;; -- the specs derived from maximal-matches
	   ; of c-nodes derived from this through
	   ; regression past action
	   inductives - (Lst (Lrcd Uterm
				   (Lst Reduction)))
	   ;; -- Groups children derived by fluent regression
	   ;; involving the Uterm.
	(:handler
	   (print (es - Effort-spec srm - Stream)
	    (out (:to srm)
	       (:pp-block (:pre "#<")
		 "Effort-spec"
		 (:pp-ind :block 3)
		 ;;;;(:q ((!_in-progress es) "?"))
		 " [" (!_varid es) "] "
		 (:pp-nl :linear)
		 (cond ((= (len (!_conjuncts es)) 2)
			(cadr (!_conjuncts es)))
		       (t (!_conjuncts es)))
		 ;;;;"(+" (len (!_constraints es)) ")"
		 (:pp-nl :linear)
		 "/" (!_effort es)
		 1 (:pp-nl :linear)
		 (len (!_children es))
		 " children"
		 (:suf ">"))))))
   (conser (ln - L-node conj - (Lrcd Symbol . (Lst Sexp)) id - (~ Varid))
      (make-inst ^^
	 :l-node ln
	 :varid (or id (new-Varid))
	 :conjuncts conj
	 :effort infinite-effort* :instances '*unknown
	 :actions nil             
	 :parents !(Reduction)    :children !(Reduction)
	 :inductives !()))
   (literal - Sexp :inline (es - Effort-spec) (cadr (!_conjuncts es)))
   (constraints - (Lst Sexp) :inline (es - Effort-spec)
      (cddr (!_conjuncts es)))
   (pp - Plan-prefix :inline (es - Effort-spec) (!_pp (!_l-node es)))
   (height - Fixnum :inline (es - Effort-spec) (!_height (!_l-node es)))
   )

(deffunc effort-spec-conjuncts - (Lst Sexp) (es - Effort-spec)
   (!_conjuncts es))

;; These take the place of 'this-step' when required
(defnisptype Reduction-step
             (Structure r - (~ Reduction)
		(:handler
		    (print (rs - Reduction-step srm - Stream)
		       (out (:to srm)
			    (:q ((!_r rs)
				"#<the hypothetical step "
			           (!_action (!_r rs)) ">")
			       (t
				"#<a hypothetical step>"))))))
   (conser (x - (~ Reduction))
      (make ^^ x)))

(defnisptype Effort-db
    (Structure contents
	       - (Lst (Lstructure ()
			 pp - Plan-prefix
			 l-nodes effort-specs - Exp_index))))

(deffunc effort-db-specs - (Lst Effort-spec) (edb - Effort-db)
   (let-fun ()
      (let ((cl (!_contents edb)))
         (cond ((null cl) '())
               (t
                (collect (cdr cl) (car cl)))))
    :where
      (collect - (Lst Effort-spec)
	       (l - (Lst (Lstructure ()
			    pp - Plan-prefix
			    l-nodes effort-specs - Exp_index))
		main - (Lstructure ()
			  pp - Plan-prefix
			  l-nodes effort-specs - Exp_index))
	 (be *
	     (cond ((null l)
		    (opt::exp_index-contents
		        (!_effort-specs main) false))
		   (t
		    (nconc (opt::exp_index-contents
			       (!_effort-specs (car l)) true)
			   (collect (cdr l) main))))))))

(deffunc make-effort-tab - Effort-db () (make Effort-db '()))

(specdecl (l-node-id* (new-Varid))
	  (other-l-node-id* (new-Varid)) - Varid)

; First value is true if old
(defproc place-effort-spec - (Mlv Boolean Effort-spec)
                       (l-node - Sexp constraints - (Lst Sexp)
			id - (~ Varid)
			pp - Plan-prefix height - Fixnum
                        eff-db - Effort-db th - Domain)
   (let ((indx (!_effort-specs (place-pp-index pp eff-db))))
      (let ((new-conj
	       `(and ,l-node ,@constraints)))
	 (let-fun ()
	    (let ((cands (exp-fetch new-conj indx true))
		  - (Lst Effort-spec))
	       (repeat :for ((c :in cands))
		:result (new-effort-spec)
		:until (variants
			  (!_conjuncts c) new-conj
			  (!_varid c) l-node-id* (empty-env))
		:result (values true c)))
	  :where
            (new-effort-spec - (Mlv Boolean Effort-spec) ()
	       (let ((l-node
			(place-l-node l-node pp height eff-db th)))
;;;;		  (cond ((repeat :for ((c1 :in constraints :tail c1tl)
;;;;				       x)
;;;;			  :result false
;;;;			  :until (and (matchq (>= -1.0 ?x) c1)
;;;;				      (exists (c2 :in (cdr c1tl))
;;;;					 (matchq (< -1.0 ?,x) c2)))
;;;;			  :result true)
;;;;			 (dbg-save constraints l-node)
;;;;			 (breakpoint place-effort-spec
;;;;			    "Constraints = " constraints)))
		  (let ((es (make Effort-spec l-node new-conj id)))
		    (exp-ob-index es indx true)
		    (!= (!_effort-specs l-node)
			(cons es *-*))
		    (values false es))))))))

(deffunc l-node-literal - Sexp (ln - L-node)
   (!_literal ln))

(defproc place-l-node - L-node (lit - Sexp pp - Plan-prefix
				height - Fixnum
				indx - (Alt Exp_index Effort-db)
				dom - Domain)
   (let ((indx (cond ((opt::exp_index-p indx)
		      (be Exp_index indx))
		     (t
		      (!_l-nodes (place-pp-index pp (be Effort-db indx))))))
	 - Exp_index)
      (let ((cands (exp-fetch lit indx true))
	    - (Lst L-node))
	 (repeat :for ((c :in cands))
	  :result (let ((l-node (make L-node
				   (uniquify-occasion lit dom)
				   pp height
				   (cond ((eq lit 'top) '*top)
					 (t '*unchecked))
				   !() !() !(Effort-spec))))
		     (exp-ob-index l-node indx true)
		     l-node)
	  :until (variants (!_literal c) lit
			   l-node-id* other-l-node-id*
			   (empty-env))
	  :result c))))

; Same function as 'place-effort-spec', but never create
(deffunc lookup-effort-spec - (~ Effort-spec)
                       (conjuncts - (Lst Sexp) pp - Plan-prefix 
                        eff-db - Effort-db)
   (let ((p (assoc= pp (!_contents eff-db))))
      (cond ((not p) false)
	    (t
	     (let ((new-conj
		      `(and ,@conjuncts))
		   (indx (!_effort-specs p)))
		(let-fun ()
		   (let ((cands (exp-fetch new-conj indx true))
			 - (Lst Effort-spec))
		      (repeat :for ((c :in cands))
		       :result false
		       :until (variants
				 (!_conjuncts c) new-conj
				 (!_varid c) l-node-id* (empty-env))
		       :result c))))))))

(defproc lookup-l-node - (~ L-node)
	                     (lit - Sexp pp - Plan-prefix
			      indx - (Alt Exp_index Effort-db))
   (let ((indx (cond ((opt::exp_index-p indx)
		      (be Exp_index indx))
		     (t
		      (!_l-nodes (place-pp-index pp (be Effort-db indx)))))))
      (let ((cands (exp-fetch lit indx true))
	    - (Lst L-node))
	 (repeat :for ((c :in cands))
	  :result false
	  :until (variants (!_literal c) lit
			   l-node-id* other-l-node-id*
			   (empty-env))
	  :result c))))

(defproc place-pp-index - (Lstructure ()
			     pp - Plan-prefix
			     l-nodes effort-specs - Exp_index)
		     (pp - Plan-prefix eff-db - Effort-db)
   (let ((p (assoc= pp (!_contents eff-db))))
      (let-fun ()
         (cond ((not p)
		(!= p (tuple pp
			     (exp-index-init #'l-node-literal)
			     (exp-index-init #'effort-spec-conjuncts)))
		; Put first-created ones first
		(!= (!_contents eff-db)
		    (nconc *-* (list p)))))
	 p)))

;;; Initial entry point
(deffunc topnode-reduce - (Lst Effort-spec)
	               (fmla - (Lst Sexp)  ;; in dnf
			topnode - Effort-spec
                        eff-tab - Effort-db 
			pp - Plan-prefix dom - Domain)
   (let ((rskels
	    (try-causation-methods
	       'top
	       (repeat :for ((d :in (disjunctive-normal-form
				       fmla true))
			     - (Lst Sexp))
		:collect
		    (make Regress-method
		       ':add 'top false false false
		       (new-Varid) (new-Varid) (empty-env) 'done
		       d))
	       pp dom)))
      (nth-value 0 (regressions-apply rskels topnode eff-tab dom))))

(specdecl consider-preservation* - Boolean)

;;;;(specdecl es* - Effort-spec esl* - (Lst Effort-spec))

(specdecl (discard-bad-regressions* false) - Boolean)
;;; --The optimization this gates (see below) is no longer sound,
;;; now that 'maximal-matches' can exploit reductions on non-achievable
;;; goals.

(specdecl (maxmatch-trap* !()) - (Lst Sexp)
	  (reduction-dbg* false) - Boolean)

(defproc effort-spec-reduce - (Lst Effort-spec)
                       (es - Effort-spec eff-tab - Effort-db
			ss - Initial-situation dom - Domain)
   (let ((lnode (!_l-node es)))
      ;;;;(l-node-check lnode ss dom)
      (cond ((null (l-node-place-instances lnode dom))
	     (cond ((> (!_height es) 0)
		    (dbg-out reduction-dbg* 
		       "Reducing " es :%)
		    (multi-let (((new-esl immediate)
				 (regressions-apply
				    (l-node-place-regressions lnode ss dom)
				    es eff-tab dom)))
		       (dbg-out reduction-dbg*
			  (:q ((not immediate)
			       " Not achievable" :%)
			      ((not (eq immediate 'consistent))
			       " Achievable immediately by doing "
			       immediate :%))
			  "  New effort-specs: " new-esl)
		       new-esl))
		   (t
		    !(Effort-spec))))
	    (t
	     (!= (!_instances es) !())
	     (repeat :for ((inst :in (!_instances lnode))
			   (lid (new-Varid)))
		(multi-let (((_ env)
			     (unify (!_literal es) (!_propo inst)
				    (!_varid es) lid (empty-env))))
		   (multi-let (((sat con-env)
				(constraints-satisfy
				   (not (eq (!_status lnode) '*trivial))
				   (!_constraints es) (!_varid es) env
				   (!_sit (!_pp es)) dom)))
		      (cond (sat
			     ;; We use 'false' because there is no
			     ;; Reduction here
			     (!= (alist-entry false
					      (be (Lst Red-inst-rcd)
						  (!_instances es))
					      !())
				 (envs-add-to-instance-tab
				    (list (cond ((eq (!_status lnode)
						     '*trivial)
						 env)
						(t con-env)))
				    (!_conjuncts es) (!_varid es)
				    *-*))
			     (!= (!_effort es) 0))))))
	     !(Effort-spec)))))

(defproc l-node-place-instances - (Lst Occasion)
	                   (l-node - L-node dom - Domain)
   (cond ((eq (!_status l-node) '*unchecked)
	  (let ((lit (!_literal l-node))
		(pp (!_pp l-node)))
	     (cond ((is-trivial-ineq lit)
		    (!= (!_instances l-node)
			(list (!_occ l-node)))
		    (!= (!_status l-node) '*trivial))
		   (t
		    (let ((sit (!_sit pp)))
		       (multi-let (((many al)
				    (deduce-check-many
				       lit l-node-id* (empty-env) sit dom)))
			  (cond ((and (not (null al))
				      (not (has-qvars lit)))
				 (!= (!_status l-node) '*true)
				 (!= (!_instances l-node)
				     (list (!_occ l-node))))
				(t
				 (!= (!_status l-node) '*checked-instances)
				 (!= (!_instances l-node)
				     (nodupq
					(repeat :for ((a :in (cond (many !())
								   (t al))))
					 :collect
					    (uniquify-occasion
					       (one-value
						  (safe-varsubst
						     lit l-node-id* a))
					       dom)))))))))))))
   (!_instances l-node))

(defproc l-node-place-regressions - (Lst Reduction-skel)
	                    (l-node - L-node ss - Initial-situation dom - Domain)
   (cond ((memq (!_status l-node) '(*unchecked *checked-instances))
	  (l-node-place-instances l-node dom)
	  (let ((lit (!_literal l-node))
		(pp (!_pp l-node)))
	     (let ((regressions
		      (place-regressions
			 lit (new-Varid) dom)))
		(cond ((and discard-bad-regressions*
			    (not (forall (r :in regressions)
				    (and (eq (!_zone r) ss)
					 (!_good r)))))
		       (!= regressions
			   (repeat :for ((r :in regressions))
			      (cond ((not (eq (!_zone r) ss))
				     (!= (!_good r)
					 (not (regress-method-futile
					       r ss dom)))
				     (!= (!_zone r) ss)))
			    :when (!_good r)
			    :collect r))))
		(let-fun ()
		   (!= (!_regressions l-node)
		       (cond ((exists (e :in maxmatch-trap*)
				 (variants e lit dummy-id* l-node-id*
					   (empty-env)))
			      (bind ((maxmatch-dbg* true))
				 (try-methods)))
			     (t
			      (try-methods))))
		   (!= (!_status l-node) '*checked-all)

		 :where

		   (:def try-methods - (Lst Reduction-skel)
				()
		      (nconc (cond ((and consider-preservation*
					 (not (!_is-initial pp)))
				    (repeat :for ((a :in (deduce
							    lit
							    l-node-id*
							    (empty-env)
							    (!_sit (!_prior pp))
							    dom)))
				     :nconc
				       (multi-let (((lit _)
						    (safe-varsubst
						       lit l-node-id* a)))
					  (cond ((not (has-qvars lit))
						 (try-preservation-methods
						    lit regressions pp
						    dom))
						(t !())))))
				   (t !()))
			     (try-causation-methods
				lit regressions pp dom))))))))
   (!_regressions l-node))

;;; This is no good; even if the precond is unachievable, it may be
;;; reducible.
(deffunc regress-method-futile - Boolean (rm - Regress-method
                                          ss - Initial-situation
                                          th - Domain)
   (cond ((step-value-occurs-in (!_precond rm))
	  (signal-problem regress-method-futile
	     "bogus precond: " (!_precond rm))))
   ;;(out "No bogosity in " (!_precond rm) :%)

   (exists (p :in (!_precond rm))
      (and (not (has-qvars p))
           (not (ground-deduce p ss th))
           (not (achievable p th)))))

(specdecl ;;;;disjuncts* - (Lst (Lst Sexp))
	  (preservation-dbg* false)
	  maxmatch-dbg* - Boolean)

(specdecl (regress-id* (new-Varid)) - Varid)

(deffunc try-preservation-methods - (Lst Reduction-skel)
                      (lit - Sexp regressions - (Lst Regress-method)
		       pp - Plan-prefix dom - Domain)
   (let-fun ()
      (multi-let (((comb-env preservers)
		   (collect-and-combine)))
	 (repeat :for ((rm :in preservers) - Regress-method
		       (disjuncts '()) - (Lst (Lst Sexp))
		       (prior (!_prior pp)) - Plan-prefix)
	    (dbg-out preservation-dbg*
		     "Got regression method " rm)
	    (multi-let (((r-pre r-env)
			 (varid-convert
			    (!_precond rm) (!_precond-id rm)
			    regress-id* comb-env)))
	       (dbg-out preservation-dbg*
			"Looking for instances of " r-pre)
	       (let ((ansl (deduce `(and ,@r-pre)
				   regress-id* 
				   r-env
				   (!_sit prior)
				   dom)))
		  (cond ((not (null ansl))
			 (!= disjuncts
			     (nconc (instances-disjuncts
				       r-pre ansl)
				    *-*))))))
	  :nconc
	      (repeat :for ((blocker :in (dnf-neg disjuncts)))
	       :nconc
		 (cond ((null blocker)
			(dbg-save :run-loud disjuncts)
			(signal-problem try-preservation-methods
			   "Bogus inevitable nondeletion: " lit
			   :% "   after " pp
			   :% "   Regressions = " regressions
			   (:novalue "I'll ignore it"))
			!())
		       (t
			(dbg-out maxmatch-dbg*
			   "Finding maximal matches for preservation "
			   "method "
			   :% " for " lit :%)
			(match-to-derive-reduction-skeletons
			   blocker false false regress-id*
			   comb-env preservers (!_sit prior) dom))))

          :where

             (:def instances-disjuncts - (Lst Sexp)
				  (r-pre - (Lst Sexp) ansl - (Lst Bdgenv))
		(repeat :for ((a :in ansl)
			      :collectors disjuncts)
		 :result disjuncts
		 :within
		   (multi-let (((d _)
				(safe-varsubst r-pre regress-id* a)))
		      (dbg-out preservation-dbg*
			       "... killer precond instance: " d)
		      (cond ((has-qvars d)
			     (signal-problem try-preservation-methods
				"Instance of deleter precond has qvars: "
				d
				(:proceed "I will ignore this deleter")))
			    (t
			     (:continue
			      :collect d))))))))
    :where
       (collect-and-combine - (Mlv Bdgenv (Lst Regress-method))
			    ()
	  (repeat :for ((rm :in regressions) - Regress-method
			(comb-env (empty-env)) - Bdgenv
			(deleter (strip-tag (be Sexp (!_last-act pp)))) - Sexp
			ok - Boolean
			:collectors preservers - Regress-method)
	   :result (values comb-env preservers)
	   :when (and (!_preservation rm)
		      (not (!_quant rm)))
	   :collect rm
	      (dbg-out preservation-dbg*
		      "Screening regression method " rm)
	      (let ((u-env (regress-method-goal-unify rm lit (!_bdgs rm))))
		 (multi-let (((ok d-env)
			      (unify (!_action rm) deleter
				     (!_precond-id rm) ground-id*
				     u-env)))
		    (dbg-out preservation-dbg*
			   "Action " (!_action rm)
			   :% (:q (ok "  unifies")
				(t " fails to unify"))
			   " with " deleter)
		    (!= < ok comb-env >
			(bdgenvs-combine
			    comb-env
			    (ids-env-elim (list lit-id* effect-id*)
				          d-env)))
		    	      (cond ((not ok)
				     (signal-problem try-preservation-methods
					  "Preservation method envs"
					  " can't be combined")))))))))

(deffunc strip-tag - Sexp (a - Sexp)
   (repeat :while (matchq (tag ?_ ?a) a))
   a)

(deffunc try-causation-methods - (Lst Reduction-skel)
                            (lit - Sexp
			     regressions - (Lst Regress-method)
			     pp - Plan-prefix dom - Domain)
   (repeat :for ((r :in regressions) - Regress-method)
    :when (not (!_preservation r))
    :nconc
      (let ((u-env (cond ((eq lit 'top) (empty-env))
			 (t (regress-method-goal-unify r lit (!_bdgs r))))))
	 (dbg-out maxmatch-dbg*
	    "Finding maximal matches for causation method with action"
	    :% (debug-varsubst (!_action r) (!_precond-id r) u-env)
	    :% " for " lit :%)
	 (multi-let (((r-pre r-env)
		      (varid-convert
			 (!_precond r)
			 (!_precond-id r) regress-id* u-env)))
	    (multi-let (((r-quant r-env)
			 (varid-convert
			    (!_quant r)
			    (!_goal-id r) regress-id* r-env)))
	       (multi-let (((r-act r-env)
			    (varid-convert
			       (!_action r)
			       (!_precond-id r) regress-id* r-env)))
		  (!= r-env (ids-env-elim (list lit-id* effect-id*)
					  *-*))
		  (match-to-derive-reduction-skeletons
		     r-pre r-act r-quant regress-id* r-env
		     (list r) (!_sit pp) dom)))))))

;;; In the case of a causation method, 'rml' is a singleton containing
;;; that method.  For a preservation goal, 'rml' is the list of all
;;; the methods that would destroy the goal.
(deffunc match-to-derive-reduction-skeletons - (Lst Reduction-skel)
	                     (preconds - (Lst Sexp) act quant - (~ Sexp)
			      id - Varid env - Bdgenv
			      rml - (Lst Regress-method)
			      sit - Situation dom - Domain)
   (let ((mml (maximal-matches preconds !() id env sit dom)))
      (repeat :for ((mm :in mml))
       :collect 
	   (multi-let (((act env)
			(safe-varsubst act regress-id*
				       (!_subst mm))))
;;;;	      (cond ((and (eq act 'watch) (= (len rml) 1))
;;;;		     (dbg-save rml mm preconds id env sit dom)
;;;;		     (breakpoint match-to-derive-reduction-skeletons
;;;;			"Watch to achieve " (!_goal (car rml))
;;;;			:% " has preconds: " (!_dif mm)
;;;;			:% " from " preconds)))
	      (multi-let (((quant env)
			   (safe-varsubst quant regress-id* env)))
		 (make Reduction-skel
		       act quant
		       (maximal-match-clean mm regress-id* env)
		       rml))))))

(deffunc regress-method-goal-unify - Bdgenv
                      (r - Regress-method lit - Sexp env - Bdgenv)
   (let-fun ()
      (multi-let (((ok u-env)
		   (unify (normalize-if-inequality lit)
			  (!_goal r)
			  regress-id* (!_goal-id r)
			  env)))
	 (cond (ok
		u-env)
	       (t
		(signal-problem regress-method-goal-unify
		   "Effort-spec literal " lit
		   :% " doesn't unify with Regress-method goal "
			(!_goal r) "[id " (!_goal-id r) "]"
		   :% " wrt " (!_bdgs r)))))
    :where
       (normalize-if-inequality - Sexp (g - Sexp)
;;;;	  (trace-around norm-if-ineq
;;;;	     (:> "norm-if-neq: " g)
	  (repeat :for ((g1 g) g2 (polarity true))
	   :while (matchq (not ?g2) g1)
	      (!= g1 g2)
	      (!= polarity (not *-*))
	   :result (cond ((memq (car g1) ineq-rels*)
			  (multi-let (((rel lhs)
				       (inequality-normal-form polarity g1)))
			     `(,rel ,lhs 0)))
			 (t g)))
;;;;	     (:< (val &rest _) "norm-if-neq: " val))
)))

(defproc regressions-apply - (Mlv (Lst Effort-spec) (~ Sexp))
	              (regressions - (Lst Reduction-skel)
		       es - Effort-spec 
		       eff-tab - Effort-db dom - Domain)
   (repeat :for ((rs :in regressions)
		 - Reduction-skel
		 ;; If some regression makes es achievable with
		 ;; one action, 'immediate' will be that action --
		 (immediate false) - (~ Sexp)
		 ;; If no regression leads to a reduction at all
		 ;; this is false
		 (consis false) - Boolean
		 :collectors new-children - Effort-spec)
    :nconc
      (let ((lid (new-Varid)))
	 (multi-let (((_ env)
		      (unify (!_literal es) (!_literal (!_l-node es))
			     (!_varid es) lid
			     (empty-env))))
	    (multi-let (((mm-consis new-mm)
			 (maximal-match-constrain
			    (!_match rs) (!_constraints es)
			    lid (!_varid es) env (!_sit (!_pp es))
			    dom)))
	       (cond (mm-consis
		      (!= consis true)
		      (multi-let (((nc - (Lst Effort-spec)
				    act - (~ Sexp))
				   (effort-spec-link
				      es (!_act rs) (!_quant rs)
				      new-mm lid (!_rml rs)
				      eff-tab dom)))
			 (cond (act
				(cond ((or (not immediate) (eq immediate 'preserve))
				       (!= immediate act)))))
			 nc))
		     (t !())))))
    :result (values new-children
		    (or immediate
			(and consis 'consistent)))))

;;; Returns newly created children + Boolean; second val is false if some
;;; precondition of action for 'mm' is unsatisfied, otherwise either the
;;; action or the constant 'preserve' if 'act' is false.
(defproc effort-spec-link - (Mlv (Lst Effort-spec) (~ Sexp))
		   (es - Effort-spec act quant - (~ Sexp)
		    mm - Maximal-match
		    mm-id - Varid
		    rml - (Lst Regress-method)
		    eff-tab - Effort-db th - Domain)
   (let ((new-children !(Effort-spec))
	 (all-children !(Effort-spec)))
      (repeat :for ((cg :in (!_dif mm)))
	 (multi-let (((old es)
		      (place-effort-spec (!_goal cg) (!_constraints cg)
					 false
					 (cond (act (!_pp es))
					       (t (!_prior (!_pp es))))
					 (- (!_height es) 1)
					 eff-tab th)))
	    (!= all-children (cons es *-*))
	    (cond ((not old)
		   (!= new-children (cons es *-*))))))
      (let ((red (link-reduction es (!_subst mm) act
				 all-children))
	    - Reduction)
         (cond ((and (null (!_preconds red))
		     (has-qvars act))
		(dbg-save red act mm rml)
		(breakpoint effort-spec-link
		   "Action with vars: " act)))
	 (cond ((and act quant
		     (= (len rml) 1)
		     (!_inductive (car rml)))
		(let ((uquant (uniquify-term
				    (one-value
				       (safe-varsubst
					  quant
					  mm-id
					  (!_subst mm)))
				    th)))
		   (!= (alist-entry uquant (!_inductives es) !())
		       (cons red *-*)))))
	 (values new-children
		 (and (null all-children)
		      (or act 'preserve))))))

(deffunc link-reduction - Reduction (parent - Effort-spec
				     bdgs - Bdgenv
                                     act - (~ Sexp)
                                     preconds - (Lst Effort-spec))
   (let ((ersatz-step false) - (~ Reduction-step)
	 ;;(orig-bdgs bdgs)
	 )
      (let ((ersatz-definer
	       (\\ ()
		  (cond ((not ersatz-step)
			 (!= ersatz-step
			     (make Reduction-step false))))
		  ersatz-step)))
;;;;	 (dbg-save (bdgs-a bdgs))
	 (!= bdgs (hide-this-step-in-env *-* ersatz-definer))
	 (!= < act bdgs >
	     (hide-this-step act (!_varid parent) bdgs ersatz-definer))
;;;;	 (dbg-save bdgs)
;;;;	 (breakpoint link-reduction
;;;;	    "About to make a reduction of " parent
;;;;	    :% " with linkenv = " bdgs 
;;;;	    :% "      act = " act
;;;;	    :% "      preconds = " preconds)
	 (let ((re (make Reduction parent bdgs act preconds)))
;;;;	    (cond ((eq act 'watch)
;;;;		   (dbg-save re)
;;;;		   (breakpoint link-reduction
;;;;		      "Making: " re
;;;;		      :% " with preconds " preconds)))
	    (cond (ersatz-step
		   (!= (!_step re) ersatz-step)
		   (!= (!_r ersatz-step) re)))
	    (!= (!_children parent) (cons re *-*))
	    (cond ((null preconds)
		   (!= (!_effort re)
		       (cond ((eq act 'done) 0)
			     (t 1)))
		   (!= (!_effort parent) (min *-* (!_effort re)))
		   (!= (!_instances re) (list (!_linkenv re)))
		   (instances-lift re))
		  (t
		   (repeat :for ((pre :in preconds))
		      (!= (!_parents pre) (adjoin re *-*)))))
;	    (dbg-save :silent re orig-bdgs)
;	    (breakpoint link-reduction
;	       "Built reduction: " :% re)
	    re))))

;;; Lift instances from red to its parent, discarding redundant ones
;;; At this point we know that red's instances are known.
(defproc instances-lift - Void (red - Reduction)
   (let ((es (!_reducee red)))
      (cond ((eq (!_instances es) '*unknown)
	     (!= (!_instances es)
		 !(Red-inst-rcd))))
      (let ((id (!_varid es))
	    (goal (!_literal es)))
	 (!= (alist-entry red (be (Lst Red-inst-rcd)
				  (!_instances es))
			      '())
	     (envs-add-to-instance-tab
	        (red-instances-contract red)
		goal id *-*)))))

(deffunc red-instances-contract - (Lst Bdgenv) (red - Reduction)
   (<# (\\ (e) (bdgenv-contract e (+ (!_varid (!_reducee red)) 1)))
       (be (Lst Bdgenv) (!_instances red))))

(defproc hide-this-step-in-env - Bdgenv (env - Bdgenv
					 step-placer - (Fun Reduction-step
							    ()))
;;; This is very tricky.  We have to substitute out everything in each
;;; binding of env.  But this can produce a new env if done right.
;;; So we have to treat env in two different ways: As a source of
;;; bindings to clean up, and as the initial env.
;;; After we've cleaned the bindings up, we have to put them into
;;; the revised env.  The env as it undergoes revision is called 'new-env'.
   (repeat :for ((b :in env) - Varbdg
		(new-env env) - Bdgenv
		new-skel - Sexp
		(new-bdgs '()) - (Lst Varbdg))
      (with (!_val b)
	 (!= < new-skel new-env >
	     (hide-this-step !>skel !>id new-env step-placer))
	 (!= new-bdgs
	     (cons (make Varbdg (!_varname b) (!_id b)
			        (make Expclo new-skel !>id))
		   *-*)))
    :result (<# (\\ (neb)
		  (let ((nbb (opt::uvar-lookup (!_varname neb)
					       (!_id neb)
					       new-bdgs)))
		     (or nbb neb)))
	       new-env)))

(deffunc hide-this-step - (Mlv Sexp Bdgenv)
			(exp - Sexp id - Varid env - Bdgenv
			 placer - (Fun Reduction-step ()))
   (cond ((is Qvar exp)
	  (let ((b (opt::uvar-lookup (!_sym exp) id env))
		- (~ Varbdg))
	     (cond (b
		    (multiple-value-let (hide-skel hide-env)
					(hide-this-step (!_skel (!_val b))
							(!_id (!_val b))
							env placer)
		       (varid-convert hide-skel
				      (!_id (!_val b))
				      id hide-env)))
		   (t (values exp env)))))
	 ((atom exp) (values exp env))
	 ((is-step-value-occ exp)
	  (values (be Sexp `(,(car exp) ,(funcall placer)))
		  env))
	 (t
	  (let-fun ((hide-elts (el env)
		       (cond ((null el) (values '() env))
			     (t
			      (multiple-value-let
			               (aexp aenv)
				       (hide-this-step (car el)
						       id env placer)
				 (multiple-value-let
				          (dexpl denv)
					  (hide-elts (cdr el) aenv)
				    (values `(,aexp ,@dexpl)
					    denv)))))))
	     (hide-elts exp env)))))

(deffunc is-trivial-ineq - Boolean (e - Sexp)
   (or (and (is Ineq-rel (car e))
	    (letrec ((var-in-sum - Boolean (sub-e - Sexp)
			(or (is Qvar sub-e)
			    (and (not (atom sub-e))
				 (memq (car sub-e) '(+ -))
				 (exists (sub-sub :in (cdr sub-e))
				    (var-in-sum sub-sub))))))
	       (or (var-in-sum (cadr e))
		   (var-in-sum (caddr e)))))
       (and (eq (car e) 'not)
	    (is-trivial-ineq (cadr e)))))

(deffunc contradicts - Boolean (p q - Sexp)
   (cond ((eq (car p) 'not)
          (equal (cadr p) q))
         ((eq (car q) 'not)
          (equal (cadr q) p))
         (t false)))

;; These two aren't used here.
(deffunc those-true-now - (Lst Occasion) (occl - (Lst Occasion)
                                          sit - Situation th - Domain)
   (repeat :for (occ :in occl)
    :when (ground-deduce (!_propo occ) sit th)
    :collect 
      occ))

(deffunc those-false-now - (Lst Occasion) (occl - (Lst Occasion)
                                           sit - Situation th - Domain)
   (repeat :for (occ :in occl)
    :when (not (ground-deduce (!_propo occ) sit th))
    :collect 
      occ))

; Given a disjunction of conjunctions, produce its negation also
; as a disjunction of conjunctions
;;;;(deffunc dnf-neg - (Lst (Lst Sexp)) (dl - (Lst (Lst Sexp)))
;;;;   (cond ((null dl)
;;;;	  (list '()))
;;;;	 (t
;;;;	  (let ((rn (dnf-neg (cdr dl))))
;;;;	     (repeat :for (c :in (car dl))
;;;;	      :nconc
;;;;		(let ((nc (literal-neg c)))
;;;;		      (repeat :for (n :in rn)
;;;;		       :collect 
;;;;			 (cons nc n))))))))
;;;;
;;;;(deffunc literal-neg - Sexp (l - Sexp)
;;;;   (cond ((car-eq l 'not) (cadr l))
;;;;	 (t `(not ,l))))

 