;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: rmgraph-exc.lisp,v 1.11 2006/01/24 13:33:46 dvm Exp $

(depends-on %module/ nisp)

; Utilities for massaging and mining regression-match graphs.
(depends-on :at-run-time %optop/ regressmatch maxmatch discrtproj
	                queue)

;;; 0 for none, 1 for large scale, 2 for more output
(specdecl (plan-dbg-amt* 1) - Fixnum
          infinite-effort* - Fixnum)

(specdecl (check-start* false) (check-freq* false) - (~ Fixnum)
          (check-count* 0) - Integer)

;;;;(specdecl top-node* - Effort-spec
;;;;          cohorts* - (Lst (Lrcd Maximal-match
;;;;				(Lst Effort-spec))))

(specdecl last-effort-tab* - (~ Effort-db))
;;;;(large-scale-plan-dbg* true)
;;;;          - Boolean

;;; The record of an effective subgraph below an Effort-spec
;;; !_preconds corresponds to preconds that are not currently true,
;;; so !_preconds = () iff action is feasible.
(defnisptype Reduction-tree
   (Structure goal-es - Effort-spec
	      env - Bdgenv
	      ;;--The env grounds the action of the original Reduction
	      ;; to the achiever.  
	      eff - Fixnum
	      achiever - Action
	      preconds - (Lst Reduction-tree)
      (:handler
         (print (rt - Reduction-tree srm - Stream)
	    (with rt
	       (out (:to srm)
		  (:pp-block 
		     (:pp-ind :block 3)
		     "#{Reduction-tree "
		     (:pp-nl :linear)
		     " Effort: " !>eff 1
		     (:pp-nl :linear)
		     (!_literal !>goal-es)
		     (:pp-nl :linear)
		     " Achiever: " !>achiever
		     (:pp-nl :linear)
		     " Preconds: " !>preconds
			     (:suf "}"))))))))
  
;;; Property: If you follow the leftmost branch of rtree all the way to
;;; a feasible action, that feasible action will be 'act'.
(defnisptype Feasible-action-spec
    (Structure act - Action
	       pp - Plan-prefix
               ;;nodes - (Lst Effort-spec)
               rtree - Reduction-tree
	       immed-purpose - Sexp
       (:handler
           (print (feas - Feasible-action-spec srm - Stream)
              (with feas
                 (out (:to srm) "#<Feasible action " !>act
                               ": (" (!_length !>pp) ")+" !>eff ">")))))
  (eff - Fixnum (a - Feasible-action-spec)
     (!_eff (!_rtree a))))

(augtype Effort-spec
   (:type actions - (Either (Const *unknown)
		      (Lst Feasible-action-spec))))

(specdecl (reuse-rmgraphs* true) - Boolean)

(deffunc compute-effort - (Multvals topnode - Effort-spec
                                    effort - Fixnum
                                    complete-matches
			              - (Lst Maximal-match)
                                    eff-db - Effort-db
				    num-nodes - Fixnum)
                        (goal - (Lst (Lst Sexp)) ;; In dnf
                         sequel - Plan-sequel
			 pp - Plan-prefix
                         sequel - (~ Plan-sequel)
                         ;;;; feasible-final-sit - (~ Situation)
			 ;; -- situation projected after sequel,
			 ;; if sequel exists.
			 parent-top-node - (~ Effort-spec)
			 parent-effort-db - (~ Effort-db)
			 ss - Situation-space th - Domain
			 height - Fixnum)
   (let-fun ()
      (multi-let (((infeasibles top-node eff-db num-nodes)
		   (make-or-edit-rm-graph reuse-rmgraphs*)))
	 (cond ((or (not (null infeasibles))
		    (and (= (!_effort top-node) 0)
			 (not (deduce goal dummy-id* (empty-env)
				      (!_sit pp) th))))
		(cond (reuse-rmgraphs*
		       (dbg-out (> plan-dbg-amt* 1)
			  "Edited regression-match graph is unreliable")
		       (!= < infeasibles top-node eff-db num-nodes >
			   (make-or-edit-rm-graph false))))))
	 (cond ((not (null infeasibles))
		(dbg-save :run-loud top-node eff-db num-nodes infeasibles)
		(signal-problem compute-effort
		   "Infeasible \"feasible actions\" in regression-match"
		   " graph:"
		   :% infeasibles)))
	 (values top-node
		 (!_effort top-node)
		 (cond ((= (!_effort top-node) 0)
			(repeat :for ((r :in (!_children top-node)))
			 :when (= (!_effort r) 0)
			 :collect
			    (make Maximal-match
				  (!_linkenv r) !() !() !())))
		       (t !()))
		 eff-db
		 num-nodes))
    :where

 (:def make-or-edit-rm-graph - (Multvals
				    infeas-acts
					 - (Lst Feasible-action-spec)
				    top-node - Effort-spec
				    eff-db - Effort-db
				    num-nodes - Fixnum)
                        (copy-vs-redo - Boolean)
   (multi-let (((top-node open-nodes eff-db)
		(eff-db-move-forward
		    copy-vs-redo parent-top-node parent-effort-db
		    goal pp feasible-final-sit height th))
	       (num-nodes 0))
;;;;      (out "Setting last-effort-tab* to " eff-db)
      (!= last-effort-tab* eff-db)
      (!= check-count* 0)
      (repeat :for ((l-node-queue (make Queue open-nodes)))
       :until (!_empty l-node-queue)
	 (let ((new (effort-spec-reduce (dequeue l-node-queue)
				   last-effort-tab* (!_initsit ss) th)))
	    (!= num-nodes (+ *-* (len new)))
	    (list-enqueue new l-node-queue)))
      (let ((all-l-nodes
	       (effort-db-specs last-effort-tab*)))
	 (tally-efforts all-l-nodes)
	 (dbg-out (> plan-dbg-amt* 1)
	    "Graph effort: " (!_effort top-node) :%)
;;;;	 (cond ((> (!_effort top-node) 1000)
;;;;		(dbg-save all-l-nodes top-node)
;;;;		(breakpoint compute-effort
;;;;		   "Ooomph")))
	 (!= all-l-nodes
	     (sort *-* #'es-easier))
	 (repeat :for ((ln :in all-l-nodes))
	    (effort-spec-sort ln))
	 (tally-actspecs all-l-nodes th)
	 (values (repeat :for ((feas-act :in (be (Lst Feasible-action-spec)
						 (!_actions top-node)))
			       - Feasible-action-spec
			       :collector infeasibles - Feasible-action-spec)
		  :within
		    (let ((act (!_act presumed-feas-act)))
		       (cond ((eq (!_term act) 'watch)
			      (:continue
			       :collect presumed-feas-act))
			     ((not (eq (!_term act)
				       'done))
			      (multi-let (((succ _ _ _ _ _ _)
					   (contin-discrete-project
					      (!_sit (!_pp presumed-feas-act))
					      act ss th)))
				 (:continue
				  :when (not succ)
				  :collect presumed-feas-act))))))
		 top-node
		 eff-db
		 num-nodes)))
   )))

(deffunc eff-db-move-forward - (Multvals top-node - Effort-spec
					 open-nodes - (Lst Effort-spec)
					 eff-db - Effort-db)
                         (copy-vs-redo - Boolean
			  old-top-node - (~ Effort-spec)
			  old-eff-db - (~ Effort-db)
			  goal - (Lst (Lst Sexp))
                          sequel - (~ Plan-sequel)
			  pp - Plan-prefix
			  feasible-final-sit - (~ Situation)
			  height - Fixnum dom - Domain)
   (let ((reduction-skels
	    (cond (feasible-final-sit
		   (repeat :for ((d :in goal) - (Lst Sexp))
		    :nconc
		      (let ((mml (maximal-matches preconds !() regress-id* env
						  feasible-final-sit dom)))
			 (repeat :for ((mm :in mml))
			  :collect
			    (multi-let ((clean-mm
					   (maximal-match-clean mm regress-id* env)))
			       (make Reduction-skel
				     'done false mm !())))))
		   
)
		  (t
		   (repeat :for ((d :in goal) - (Lst Sexp))
		    :nconc
		      (match-to-derive-reduction-skeletons
			 goal 'done false regress-id* (empty-env)
			 !() (!_sit pp) dom))))))
				     

				     

	       (repeat :for ((d :in (disjunctive-normal-form
				       fmla true))
			     - (Lst Sexp))
		:collect
		    (make Regress-method
		       ':add 'top false false false
		       (new-Varid) (new-Varid) (empty-env) 'done
		       d))


	          (t		       
		   (try-causation-methods
		      'top
		      (repeat :for ((d :in goal) - (Lst Sexp))
		       :collect
			   (make Regress-method
			      ':add 'top false false false
			      (new-Varid) (new-Varid) (empty-env) 'done
			      d))
		      pp dom))


;;; Scheme:
;;; Call regressions-apply, after producing regressions from maximal matches.

   (cond ((and copy-vs-redo old-top-node)
	  (dbg-out (> plan-dbg-amt* 1)
	     "Editing old regression-match graph" :%)
	  (multi-let (((top-node open-nodes new-eff-db)
		       (rmgraph-copy-and-trim old-top-node old-eff-db
					      pp dom)))
	     (cond ((memq top-node open-nodes)
		    ;; We must be getting close!
		    (multi-let ((new-children
				   (topnode-reduce goal top-node
						   new-eff-db pp dom)))
		       (values top-node new-children new-eff-db)))
		   (t
		    (values top-node open-nodes new-eff-db)))))
	 (t
	  (dbg-out (> plan-dbg-amt* 1)
	     "Creating new regression-match graph" :%)
	  (let ((new-eff-db (make-effort-tab)))
	     (multi-let (((_ top-node)
			  (place-effort-spec 'top !() false pp
					     (+ height 1)
					     new-eff-db dom)))
		(multi-let ((new-children
			       (topnode-reduce goal top-node
					       new-eff-db
					       pp dom)))
		   (values top-node new-children new-eff-db)))))))

(deffunc rmgraph-copy-and-trim - (Multvals newtop - Effort-spec
					   open-effort-specs
					      - (Lst Effort-spec)
					   new-eff-db - Effort-db)
	         (oldtop - Effort-spec old-effort-db - Effort-db
		  new-pp - Plan-prefix dom - Domain)
		 (ignore old-effort-db)
   (let ((open-effort-specs !(Effort-spec))
	 (new-eff-db (make-effort-tab)))
      (let-fun ((:def walk-and-copy - Effort-spec (old-es - Effort-spec)
		  (multi-let (((already-constructed es)
			       (place-effort-spec
				   (!_literal (!_l-node old-es))
				   (!_constraints old-es)
				   (!_varid old-es)
				   (plan-prefix-forward (!_pp old-es) new-pp)
				   (!_height old-es)
				   new-eff-db dom)))
		     (cond ((not already-constructed)
			    (cond ((or (=< (!_effort old-es) 1)
				       (exists (red :in (!_children old-es))
					  (not (!_action red))))
				   ;; Have to redo preservations
				   (!= open-effort-specs
				       (cons es *-*)))
				  (t
				   (repeat :for ((c :in (!_children old-es)))
				      (link-reduction
					 es (!_linkenv c) (!_action c)
					 (<# walk-and-copy
					     (!_preconds c)))
;;;;				      (dbg-save oldtop es c open-effort-specs)
;;;;				      (breakpoint rmgraph-copy-and-trim
;;;;					 "Linked " es " :to " c)
				      )))))
		     es)))
	 (let ((new-top (walk-and-copy oldtop)))
	    (values new-top open-effort-specs new-eff-db))
       :where
          (:def plan-prefix-forward - Plan-prefix (old new - Plan-prefix)
	     (cond ((!_is-initial new)
		    (signal-problem rmgraph-copy
				"Can't find "
				old :% " not prior of " new))
		   ((eq old (!_prior new))
		    new)
		   (t old))))))

;;;;	     (repeat :for ((res = new :then (!_prior new))
;;;;			   (old-find = old :then (!_prior old-find)))
;;;;	      :until (and (not (!_is-initial new))
;;;;			  (eq old-find old))
;;;;	      :result (cond ((!_is-initial new)
;;;;			     (signal-problem rmgraph-copy
;;;;				"Can't find 
;;;;				old :% " not prior of " new)
;;;;		    new)
;;;;		   (t))))))

(specdecl effort-changes* - Integer)

(defproc tally-efforts - Void (leaves - (Lst Effort-spec))
   (let ((level (adjustable-array 10 '()))
         - Obj   ;(ARY (LST effort-spec) 1) declares it as simple-vector
         (size 10) - Fixnum
         (esl (<? (\\ (es - Effort-spec) (= (!_effort es) 0))
                  leaves))
         - (Lst Effort-spec))
      (!= (aref level 1)
          (<? (\\ (es - Effort-spec) (= (!_effort es) 1))
	      leaves))
      (!= effort-changes* 0)
      (repeat :for ((lev 0) - Fixnum)
         (repeat 
          ; If no preservation methods, this loop iterates no more than
          ; once
	    (repeat :for ((es :in esl))
	       (cond ((= (!_effort es) lev)
		      (repeat :for ((par :in (!_parents es))
				   - Reduction)
			 (cond ((or (< (!_effort par) infinite-effort*)
				    (forall (sib :in (!_preconds par))
				       (< (!_effort sib)
					  infinite-effort*)))
				(let ((neweff (propagate-effort-up par)))
				   (cond (neweff
					  (cond ((>= neweff size)
                                                (!= size (* 2 neweff))
                                                (!= level
                                                    (adjust-array
                                                       *-*
                                                       (list size)
                                                       :initial-element '()))))
					  (!= (aref level neweff)
					      (cons (!_reducee par)
						    *-*)))))))))))
            (!= esl (aref level lev))
          :until (null esl)
            (!= (aref level lev) '()))
         (!= lev (+ *-* 1))
       :until (= lev size)
         (!= esl (aref level lev))
         (!= (aref level lev) '()))))

(defun adjustable-array (dims init)
   (make-array dims
               :initial-element init
               :adjustable true))

;; Returns new effort of reducee of red, or false if the new would be
;; no improvement on the old.
(defproc propagate-effort-up - (~ Fixnum) (red - Reduction)
   (let ((up (!_reducee red))
	 (neweff (reduction-effort red)))
      (cond ((and (< neweff (!_effort red))
		  (< neweff infinite-effort*))
	     (!= (!_effort red) neweff)
	     (!= (!_instances red) '*unknown)
	     (cond ((=< neweff (!_effort up))
		    (let ((ral (place-reduction-instances red)))
		       (cond ((null ral)
			      ;; This shouldn't happen very often
			      (err-out "No instances for reduction " red)
			      (dbg-save :run-loud red)
			      (breakpoint propagate-effort-up
				 "No instances")
			      false)
			     ((< neweff (!_effort up))
			      (!= effort-changes* (+ *-* 1))
			      (!= (!_effort up) neweff)
			      (!= (!_instances up) !())
			      (instances-lift red)
			      neweff)
			     (t
			      (instances-lift red)
			      false)
			     )))
		   (t false))) 
	    (t false))))

(deffunc reduction-effort - Fixnum (reduc - Reduction)
   (</ (\\ (te pre)
	  (+ te (!_effort pre)))
       (reduction-delta-effort reduc)
       (!_preconds reduc)))

;; Bdgenv is in terms of reducee's varid.
(defproc place-reduction-instances - (Lst Bdgenv) (r - Reduction)
   (cond ((eq (!_instances r) '*unknown)
	  (cond ((forall (es in (!_preconds r))
		    (not (eq (!_instances es) '*unknown)))
		 (let ((envs (collect-effort-spec-instances (!_preconds r)
							    (!_linkenv r))))
		    (!= (!_instances r) envs)
		    envs))))
	 (t
	  (be (Lst Bdgenv) (!_instances r)))))

;;; Called only when we know that every element of pl's instances are known
(deffunc collect-effort-spec-instances - (Lst Bdgenv)
	                          (pl - (Lst Effort-spec) env - Bdgenv)
   (cond ((null pl)
	  (list env))
	 (t
	  (let ((cdr-instances (collect-effort-spec-instances (cdr pl) env)))
	     (repeat :for ((a-inst
			    :in
			    (<$ cadr
				(be (Lst Red-inst-rcd)
				    (!_instances (car pl))))))
	      :nconc (progn
			(repeat :for ((d-inst :in cdr-instances)
				      combined - Bdgenv
				      ok - Boolean)
			   (!= < ok combined >
			       (bdgenvs-combine
				  (cadr a-inst)
				  d-inst))
			 :when ok
			 :collect combined)))))))

; Tag reductions as effective or ineffective.  
; Make sure effective children come first; this used to be necessary for
; various depth-first-sweep algorithms; now done purely for superstitious 
; reasons. 
(defproc effort-spec-sort - Void (es - Effort-spec)
      (repeat :for ((sub :in (!_children es))
		   (eff (!_effort es)))
	 (!= (!_effective sub) (= (!_effort sub) eff)))
      (!= (!_children es)
	  (nconc (<? !_effective *-*)
		 (<? neg !_effective *-*))))

(specdecl actspec-count* actspec-size* actspec-and-iter* - Integer)

(specdecl redo-count* - Integer
          (collect-all-actions* false) - Boolean)

; Effort-specs are sorted by effort
(defproc tally-actspecs - Void (l-nodes - (Lst Effort-spec) dom - Domain)
   (dbg-out (> plan-dbg-amt* 1) "Extracting actspecs" :%)
   (!= actspec-count* 0)
   (!= actspec-size* 0)
   (!= actspec-and-iter* 0)
   (!= redo-count* 0)
   (repeat :for ((esl l-nodes) 
		redo - (Lst Effort-spec)
		; Nodes at lower effort level that are
		; parents and need to be redone
		es - Effort-spec)
    :until (null esl)
      (!= es (car esl))
      (!= esl (cdr esl))
      ;(out "Trying " es t)
      (let ((old (!_actions es))
	    (new (select-reasonable-actions
                    (combine-subacts (!_children es) es dom))))
	 (cond ((or (eq old '*unknown)
                    (not (actspecs-subset
			    new (be (Lst Feasible-action-spec) old))))
  		(!= (!_actions es) new)
                (!= redo (done-parents es esl))
                      )
               (t (!= redo '())))
      (cond (collect-all-actions*
             (!= esl (nconc (sort redo #'es-easier)
		            *-*))))
      )
;   (OUT "Actspec count = " ACTSPEC-COUNT* " ... size = " ACTSPEC-SIZE* " ["
;	(F "~3,1F" (/ ACTSPEC-SIZE* (->FLOAT ACTSPEC-COUNT*))) "] "
;	" ... iters = " ACTSPEC-AND-ITER*
;	" [" (F "~3,1F" (/ ACTSPEC-AND-ITER* (->FLOAT ACTSPEC-COUNT*)))"] "
;   )
	;" Redo count = " REDO-COUNT* T
   ))

(specdecl (flush-infinite-actions* true) - Boolean)

(deffunc select-reasonable-actions - (Lst Feasible-action-spec)
                                   (al - (Lst Feasible-action-spec))
          ;;(<? (\\ (a) (not (has-qvars (!_act a)))) ...)
       (cond (flush-infinite-actions*
	      (<? (\\ (a) (< (!_eff a) infinite-effort*))
		  al))
	     (t al)))

(deffunc combine-subacts - (Lst Feasible-action-spec)
                         (chl - (Lst Reduction)
                          par - Effort-spec
			  dom - Domain)
   (!= actspec-count* (+ *-* 1))
   (repeat :for ((sub :in chl) - Reduction
		 (specs !()) - (Lst Feasible-action-spec)
		 delta
		 - Fixnum)
    :result specs
        (!= delta (reduction-delta-effort sub))
        (!= specs
	   (actspecs-merge-or
	      *-*
	      (cond ((forall (pc :in (!_preconds sub))
			 (= (!_effort pc) 0))
		     (cond ((eq (!_action sub) 'watch)
			    (dbg-save :run-loud par sub dom)
			    (signal-problem combine-subacts
			       "Feasible 'watch' action to achieve"
			       " goal that is already true: "
			       :% 3 (!_literal (!_reducee sub))))
			   ((and (!_action sub)
				 (exists (pc :in (!_preconds sub))
				    (not (eq (!_pp par) (!_pp pc)))))
			    (dbg-save :run-loud sub par)
			    (signal-problem combine-subacts
			       "Goal with all true preconds: " par
			       :% " wrt plan-prefix " (!_pp par)
			       :% " has precond with different prefix: "
			       :% 3 (<? (\\ (pc)
					   (not (eq (!_pp par) (!_pp pc))))
					(!_preconds sub)))))
		     (let ((instances
			      (uniquify-action-instances sub dom))
			   - (Lst (Lrcd Action Sexp Bdgenv)))
			(cond ((null instances)
			       ;; Anomaly: the action is easy
			       ;; but has free variables
			       (signal-problem combine-subacts
				  "Action has variables after combining"
				  " reductions:"
				  :% par)))
			(<# (\\ (ai)
			       (cond ((act-qvar-check (car ai))
				      (dbg-save sub ai)
				      (breakpoint combine-subacts
					 "ai = " ai)))
			       (let (proc)
				  (cond ((matchq (wait-non-negative
						    (- (fl-v (elapsed ?_)) ?_)
						    ?proc)
						 (!_term (car ai)))
					 (dbg-save :run-loud ai par proc)
					 (breakpoint combine-subacts
					    "Got: " ai))))
;;;;			       (cond ((eq (!_term (car ai)) 'watch)
;;;;				      (dbg-save ai dom)
;;;;				      (breakpoint combine-subacts
;;;;					 "Got watch: " ai)))
;;;;			       (cond ((not (!_act (car ai)))
;;;;				      (signal-problem combine-subacts
;;;;					 "Null action " ai)))
			       (make Feasible-action-spec
				  (first ai)
				  (!_pp par)
				  (make Reduction-tree
				     par
				     (third ai)
				     delta
				     (first ai)
				     !())
				  (second ai)))
			    instances)))
		    (t
		     (reduction-actspecs-merge sub dom)))))))

(defproc act-qvar-check - Boolean (uact - Action)
	 (cond ((has-qvars (!_term uact))
		(err-out
		   "Action with vars: " uact)
		true)
	       (t false)))

;;;;		   (:novalue "I'll go ahead; who knows what will happen?")))))

(deffunc es-easier - Boolean (es1 es2 - Effort-spec)
   (let ((ef1 (!_effort es1)) (ef2 (!_effort es2)))
      (or (< ef1 ef2)
	  (and (= ef1 ef2)
               (let ((l1 (!_length (!_pp es1)))
                     (l2 (!_length (!_pp es2))))
                  (or (< l1 l2)
                      (and (= l1 l2)
                           (< (!_height es1) (!_height es2)))))))))

;;; Given a set of envs derived from a Reduction, derive the 
;;; possible achievers.  We need only one env per achiever.
;;; The second element of each triple returned is the literal of the
;;; reducee, with variables replaced by values that satisfy the
;;; constraints. 

(deffunc uniquify-action-instances - (Lst (Lrcd Action
						Sexp
						Bdgenv))
	                          (red - Reduction dom - Domain)
   (let ((envs (place-reduction-instances red))
	 (act (!_action red))
	 (es (!_reducee red)))
      (repeat :for ((env :in envs)
		    (res !((Lrcd Action Sexp Bdgenv))))
	 (multi-let (((consis act-inst purp env)
		      (->action act env)))
	    (cond (consis
		   (let ((trip (assq act-inst res)))
		      (cond ((not trip)
			     (!= trip (tuple act-inst purp env))
			     (!= res (cons trip *-*))))))))
       :result res
       :where
          (:def ->action - (Mlv Boolean (~ Action) (~ Sexp) (~ Bdgenv))
		   (act - (Alt Sexp Action) env - Bdgenv)
	     (cond ((is Uterm act)
		    (signal-problem uniquify-action-instances
		       "Unexpected uniquified action in Reduction"
		       :proceed)
		    (values true (be Action act) '(true) env))
		   (t
		    (multi-let (((consis env)
				 (constraints-satisfy
				    true
				    (!_constraints es) (!_varid es)
				    env
				    (cond ((!_action red)
					   (!_sit (!_pp es)))
					  (t (!_sit (!_prior (!_pp es)))))
				    dom)))
		       (cond (consis
			      (multi-let (((okay act env)
					   (reduction-act-inst-uniquify
					      (be Sexp act)
					      (!_varid es) env dom)))
				 (multi-let (((purp env)
					      (safe-varsubst
					         (!_literal es)
						 (!_varid es)
						 env)))
				    (values okay act purp env))))
			     (t
			      (dbg-save es red dom)
			      (breakpoint uniquify-action-instances
				 "Inconsistent constraints: "
				 (!_constraints es))
			      (values false nil nil nil))))))))))

(deffunc reduction-act-inst-uniquify - (Mlv Boolean (~ Action) (~ Bdgenv))
                            (act - Sexp id - Varid env - Bdgenv dom - Domain)
   (multi-let (((act-inst - Sexp env - Bdgenv)
		(safe-varsubst act id env)))
     (cond ((and (or (atom act-inst)
		     (not (memq (car act-inst)
				'(wait-while
				  wait-non-negative))))
		 (has-qvars act-inst))
;;;;	    (signal-problem reduction-act-inst-uniquify
;;;;	       "Instance vars fail to instantiate"
;;;;	       " reduction action: "
;;;;	       act-inst
;;;;	       (:continue "I will discard the reduction action"))
	    (values false nil nil))
	   (t
	    (values true
		    (uniquify-action act-inst dom)
		    env)))))

(deffunc actspecs-merge-or - (Lst Feasible-action-spec)
                       (al1 al2 - (Lst Feasible-action-spec))
   (let ((res
   (nconc (actspecs-symmetric-dif al1 al2)
          (repeat :for ((p1 :in al1))
	   :nconc
                (let ((p2 (lookup-feasible-action p1 al2)))
                   (cond (p2
			  (act-qvar-check (!_act p1))
			  (multi-let (((rt purp)
				       (cond ((< (!_eff p2) (!_eff p1))
					      (values (!_rtree p2)
						      (!_immed-purpose p2)))
					     (t
					      (values (!_rtree p1)
						      (!_immed-purpose p1))))))
			     (list (make Feasible-action-spec
				      (!_act p1) (!_pp p1) rt purp))))
                         (t '())))))))
;;;;   (cond ((> (len res) 1)
;;;;	  (dbg-save al1 al2 res)
;;;;	  (breakpoint actspecs-merge-or
;;;;	     "res = " res)))
   res)
)
                                
(deffunc actspecs-symmetric-dif - (Lst Feasible-action-spec)
                            (al1 al2 - (Lst Feasible-action-spec))
   (nconc (repeat :for (p1 :in al1)
           :when (not (lookup-feasible-action p1 al2))
           :collect p1)
          (repeat :for (p2 :in al2)
	   :when (not (lookup-feasible-action p2 al1))
           :collect p2)))

(deffunc lookup-feasible-action - (~ Feasible-action-spec)
	                    (a - Feasible-action-spec
			     al - (Lst Feasible-action-spec))
   (repeat :for ((a1 :in al))
    :result false
    :until (same-feasible-action a a1)
    :result a1))

(deffunc reduction-actspecs-merge - (Lst Feasible-action-spec)
                         (red - Reduction dom - Domain)
   (cond ((forall (p :in (!_preconds red))
	     (or (= (!_effort p) 0)
		 (and (not (eq (!_actions p) '*unknown))
		      (not (null (!_actions p))))))
	  (repeat :for ((es :in (!_preconds red)) - Effort-spec
			(res '())
			  - (Lst Feasible-action-spec))
	   :result (trim-actspec-list res)
	     (repeat :for ((cand-act :in (be (Lst Feasible-action-spec)
					     (!_actions es)))
			   (sibs (remove1q es (!_preconds red))))
		(let ((rt (consistent-action-combination
			       cand-act sibs red dom)))
		   (cond (rt
			  (repeat :for ((prev-act :in res))
			   :result (!= res (cons (new-act) *-*))
			   :until (same-feasible-action prev-act cand-act)
			   :result (cond ((< (!_eff rt)
					     (!_eff (!_rtree prev-act)))
					  (!= res (cons (new-act)
							(dremove1q prev-act
								   *-*)))))
			   :where
			      (new-act - Feasible-action-spec () 
;;;;				 (dbg-save cand-act rt res)
;;;;				 (breakpoint new-act
;;;;				    "cand-act = " cand-act)
				 (with cand-act
				    (act-qvar-check !>act)
				    (make Feasible-action-spec
				       !>act !>pp rt
				       (!_immed-purpose cand-act)))))))))))
	 (t
;;;;	  (dbg-save red)
;;;;	  (breakpoint reduction-actspecs-merge
;;;;	     "Preconds don't work: " red)
	  !())))

(deffunc consistent-action-combination - (~ Reduction-tree)
	                      (act - Feasible-action-spec
			       sibs - (Lst Effort-spec)
			       red - Reduction dom - Domain)
   (let-fun ((:def combine - (Multvals consis - Boolean
				 env - Bdgenv
				 rtl - (Lst Reduction-tree))
		   (sl - (Lst Effort-spec)) ; envsofar - Bdgenv)
	       (cond ((null sl)
		      (multi-let (((consis env)
				   (bdgenvs-combine (!_env (!_rtree act))
						    (!_linkenv red))))
			 (cond ((not consis)
				(out "Ooops! ")
;;;;				(dbg-save act red)
;;;;				(breakpoint consistent-action-combination
;;;;				   "Oops")
				))
			 ;; (!_rtree act) put on front later
			 (values consis env !())))
		     (t
		      (multi-let (((d-consis d-inst d-rtl)
				   (combine (cdr sl))))
			 (cond (d-consis
				(cond ((= (!_effort (car sl)) 0)
				       (values true d-inst d-rtl))
				      (t
				       (actions-combine
					  (be (Lst Feasible-action-spec)
					      (!_actions (car sl)))
					  d-inst d-rtl))))
			       (t
				(values false nil nil)))))))
	    (:def actions-combine - (Multvals consis - Boolean
					      env - Bdgenv
					      rtl - (Lst Reduction-tree))
	              (actions - (Lst Feasible-action-spec)
		       d-inst - Bdgenv d-rtl - (Lst Reduction-tree))
	       (cond ((not (shorter actions 2))
		      (!= actions
			  (sort (list-copy *-*)
				(\\ (a1 a2
				     - Feasible-action-spec)
				   (< (!_eff a1)
				      (!_eff a2)))))))
	       (repeat :for ((a :in actions)
			     - Feasible-action-spec
			     consis inst)
		:result (values false nil nil)
		  (!= < consis inst >
		      (bdgenvs-combine 
			  (!_env (!_rtree a))
			  d-inst))
		:until consis
		:result (values true inst
				(cons (!_rtree a)
				      d-rtl)))))
      (multi-let (((consis rtl-env rtl)
		   (combine sibs)))
	 (cond (consis
		(let ((reducee (!_reducee red)))
		   (!= < consis rtl-env >
		       (constraints-satisfy
			  true
			  (!_constraints reducee) (!_varid reducee)
			  rtl-env
			  (cond ((!_action red)
				 (!_sit (!_pp reducee)))
				(t (!_sit (!_prior (!_pp reducee)))))
			  dom))
		   (cond (consis
			  (!= rtl (cons (!_rtree act) *-*))
			  (let ((eff (</ (\\ (tot rtree)
					    (+ tot (!_eff rtree)))
					 (reduction-delta-effort red)
					 rtl)))
			     (multi-let (((okay act env)
					  (reduction-act-inst-uniquify
					     (!_action red)
					     (!_varid reducee)
					     (bdgenv-contract
					         rtl-env (+ (!_varid reducee)
							    1))
					     dom)))
				(cond (okay
				       (make Reduction-tree
					  reducee env eff act rtl))
				      (t false)))))
			 (t false))))
	       (t false)))))

(deffunc reduction-delta-effort - Fixnum (reduc - Reduction)
   (cond ((and (!_action reduc)
	       (not (memq (!_action reduc) '(done watch))))
	  1)
	 (t 0)))

;;;;(deffunc l-node-reduction-instances - (Lst Sexp)
;;;;	              (es - Effort-spec red - Reduction env - Bdgenv)
;;;;   (let ((p (assq red (!_instances es))))
;;;;      (cond (p
;;;;	     (repeat :for ((r :in (cadr p)) - (Lrcd Sexp Bdgenv))
;;;;	      :nconc
;;;;		(multi-let (((instance env)
;;;;			     (safe-varsubst (car p) (!_varid es) env)))
;;;;		   (cond ((has-qvars instance)
;;;;			  (signal-problem instances-from-reduction
;;;;			     "Effort-spec goal still has vars: "
;;;;			     instance
;;;;			     (:continue "I will ignore it"))
;;;;			  !())
;;;;			 (t (list instance))))))
;;;;	    (t
;;;;	     (signal-problem instances-from-reduction
;;;;		"Reduction " red
;;;;		:% " has no instances in " es)))))

;;;;(deffunc red-tree-instances-combine - (Lst (Lrcd Sexp Bdgenv))
;;;;	                          (a-insts d-insts  - (Lst (Lrcd Sexp Bdgenv)))
;;;;   (repeat :for ((a :in a-insts))
;;;;    :nconc
;;;;      (repeat :for ((d :in d-insts)
;;;;		    ok env)
;;;;         (!= < ok env > (bdgenvs-combine (cadr a) (cadr d)))
;;;;       :when ok
;;;;       :collect (tuple (car a) env))))

(deffunc done-parents - (Lst Effort-spec) (es - Effort-spec
                                           rem - (Lst Effort-spec))
		(repeat :for ((par :in (!_parents es)) - Reduction
			     (eseff (!_effort es)) - Integer
                             (redo '()) - (Lst Effort-spec))
		   (let ((par-es (!_reducee par)))
		      (let ((pareff (!_effort par-es)))
			 (cond ((or (< pareff eseff)
				    (and (= pareff eseff)
					 (repeat :for ((sib :in rem))
					  :result true
					  :until (eq sib par-es)
					  :result false
					  :until (> (!_effort sib)
						   pareff)
					  :result false)))
				(!= redo (cons par-es *-*))))))
                 :result redo))

; Is AS1 a subset of AS2? (Also check to see that new efforts are no lower
; than old.)
(deffunc actspecs-subset - Boolean (as1 as2 - (Lst Feasible-action-spec))
   (forall (a1 in as1)
      (exists (a2 in as2)
         (and (eq (!_act a2) (!_act a1))
              (=< (!_eff a2) (!_eff a1))))))

(specdecl (act-comparator*
             (\\ (f1 f2 - Feasible-action-spec)
                (< (!_eff f1) (!_eff f2))))
          - (Fun Boolean (Feasible-action-spec Feasible-action-spec) nil))

; Discard actions to avoid combinatoric explosion.
(specdecl (max-merged-out* 20) - Fixnum)

(deffunc trim-actspec-list - (Lst Feasible-action-spec)
                           (r - (Lst Feasible-action-spec))
   (cond ((> (len r) max-merged-out*)
	  (!= r (sort *-* act-comparator*))
	  (let ((rem (take max-merged-out* r)))
             rem))
	 (t r)))

(deffunc same-feasible-action - Boolean (a1 a2 - Feasible-action-spec)
   (and (eq (!_act a1) (!_act a2))
	(or (not consider-preservation*)
	    (eq (!_pp a1) (!_pp a2)))))

; for debugging
(defproc rmgraph-clear - Void (n - Effort-spec)
   (let-fun ()
      (clear n)
    :where
      (clear - Void (n - Effort-spec)
         (cond ((not (eq (!_actions n) '*unknown))
		(!= (!_actions n) '*unknown)
                (repeat :for ((ch :in (!_children n)))
                   (repeat :for ((p :in (!_preconds ch)))
                       (clear p))))))))