;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)

;;; Code for generating answers according to IPC 2004 conventions

(depends-on %module/ nisp)

(depends-on (:at :run-time) %optop/ search)

(specdecl (problem-dir* '("home" "ipc4" "DOMAINS"))
	  (soln-dir* '("home" "optop" "optop-results"))
	  - (Lst String)
	  domain-pn* domains-dir* - Pathname)

#|

Values at home --

(!=/ problem-dir* '("Docume~1" "dvm2" "workingdir" "domains"))
;;;;		    "airport" "temporal-timewindows"
;;;;		    "adl-temporal-timedliterals"))

(!=/ soln-dir* '("Docume~1" "dvm2" "optop-results"))


|#

(!= domains-dir*
    (make-Pathname :directory `(:absolute ,@problem-dir*)
		   :type "pddl"))

(!= domain-pn* (make-Pathname :name "domain"))

(defun domain-pathname () (make-Pathname :name "domain"))

#+clisp (specdecl ext:*args* - (Lst String))

#+clisp 
(defproc try-one-prob-and-exit - Void ()
   (out (:to *query-io*) "Problem file: ")
   (let ((fn (string-trim " \t" (in (:from *query-io*) :linestring))))
      (cond ((char= (elt fn 0) #\()
	     (!= fn (read-from-string fn))))
      (read-prob-and-solve fn))
   (ext:exit))

(defproc read-probs-and-solve - Symbol (dir - Sexp prob-files - (Lst String)
				      max-time - Float)
   (repeat :for ((prob :in prob-files))
    :result 'Nirvana
    :until (> (read-prob-and-solve `(,dir ,prob))
	      max-time)
    :result 'Too-much-time))

(defproc read-prob-and-solve - Float (in-file)
   (cond ((fboundp 'local-reset)
	  (local-reset)))
   (let ((start-run-time (get-internal-run-time)) - Integer)
      (let ((in-pn (coerce-to-pathname in-file))
	    domain-pn - Pathname)
;;;;	 (!= in-pn
;;;;	     (merge-pathnames *-* domains-dir*))
	 (!= domain-pn (merge-pathnames domain-pn* in-pn))
	 (let ((out-pn (soln-pathname in-pn)))
	    (bind ((when-to-dump* ':never))
	       (let-fun ()
	          (no-synerrs-allowed (opt-file-syncheck domain-pn))
	          (no-synerrs-allowed (opt-file-syncheck in-pn))
		:where
		  (:def no-synerrs-allowed (r)
		      (cond ((not (equal r '(flagged 0 expressions)))
			     (signal-problem read-prob-and-solve
				"Syntax problems: " r))))))
	    (match-cond last-entity-defined*
	       (:? (problem ?probname)
		  (let ((prob (plan-problem-initialize probname))
			  - Plan-problem
			(parse-run-time
			    (- (be Integer (get-internal-run-time))
			       start-run-time))
			  - Integer)
		     (with-open-file (outsrm out-pn
					:direction ':output
				        :if-does-not-exist ':create
					:if-exists ':overwrite)

			(prob-try-and-output-catching-errors
				  prob
				  start-run-time parse-run-time
				  outsrm))))
	       (t
		(signal-problem read-prob-and-solve
		   "Reading file " in-pn " failed to define a problem")))))))

(deffunc coerce-to-pathname - Pathname (x)
   (cond ((consp x)
	  (car (filespecs->pathnames x)))
	 (t (->pathname x))))

;;; The top of dir
(defun soln-pathname (prob-pathname)
   (let ((dir (Pathname-directory prob-pathname))
	 (name (Pathname-name prob-pathname)))
      (let ((hypos (or (position #\- name)
		       (position #\. name))))
	 (cond ((and hypos
		     (>= (length dir) (+ (length problem-dir*) 1))
		     (eq (car dir) ':absolute)
		     (<& #'equal (cdr dir) problem-dir*))
		(let ((subdirs (nthcdr (+ (length problem-dir*) 1)
				       dir)))
		   (ensure-directories-exist
		      (make-Pathname
			 :host (Pathname-host prob-pathname)
			 :device (Pathname-device prob-pathname)
			 :directory `(:absolute ,@soln-dir* ,@subdirs)
			 :name (subseq name 0 hypos)
			 :type "soln"))))
	       (t
		(dbg-save dir)
		(signal-problem dir-soln-pathname
		   "Can't derive solution pathname from " prob-pathname
		   :% " hypos = " hypos " dir = " dir))))))

(defun prob-try-and-output-catching-errors
               (prob start-run-time parse-run-time outsrm)
   (handler-case
            (prob-try-and-output prob start-run-time parse-run-time
				 outsrm)
      (error (var)
	 (bind ((*print-pretty* true))
	    (out (:to *error-output*)
	       "Planner aborted with error: " var :%)))))

;;; Example call: (prob-try-and-output '(%airport/ p01-airport1-p1-w1))

(defproc prob-try-and-output - Float (prob - Plan-problem
				     start-run-time parse-run-time - Integer
				     outsrm - Stream)
      (let ((metric (prob-figure-out-metric prob)))
	 ;; If 'metric' is false, that means use built-in metric.
	 (bind ((reduction-dbg* false)
		(plan-dbg-amt* 0)
;;;;		(plan-dbg* false)
;;;;		(large-scale-plan-dbg* false)
		(warn-about-opt-eval-problems* false)
		(signal-maxmatch-excesses* false)
;;;;		(max-hyb-queue-length* 10)
;;;;		(max-hyb-branching* 5)
		(maxmatch-dbg* false))
	    (multi-let (((solns pps-tried pps-produced mms _ _ _)
			 (search-for-plan prob
					  :metric metric
					  :require-quiescence true)))
	       (let ((final-time (get-internal-run-time))
		     (count-steps (equal metric '(minimize (total-steps)))))
		  (let ((tot-run-time
			   (/ (float (- final-time start-run-time))
			      internal-time-units-per-second))
			(parse-run-time
			   (/ (float parse-run-time)
			      internal-time-units-per-second)))
		     (cond ((null solns)
			    (out (:to *error-output*)
			       "Failed to solve " (!_name prob) :%))
			   (t
			    (let ((soln (car solns)))
			       (out (:to *error-output*)
				  "Solved problem " (!_name prob) :%)
			       (out (:to outsrm)
				  "; Time " (:f "~,2F" tot-run-time)
				  :% "; ParsingTime " (:f "~,2F" parse-run-time)
				  :% "; NrActions" (:q (count-steps
							(!_length (!_plan soln))))
				  :% "; MakeSpan"
				  :% "; MetricValue " (:q ((not count-steps)
							   (!_duration (!_plan soln))))
				  :%)
			       (ipc-format-print 
				  (optop->ipc (!_plan soln))
				  true outsrm))))
		     (out (:to *error-output*)
			:% " Total run time (ms): " (:f "~,2F" tot-run-time)
			:% " Number of plan prefixes produced: " pps-produced
			:% " Number explored: " pps-tried
			:% " Number of maximal matches: " mms :%)
		     tot-run-time))))))

(deffunc prob-figure-out-metric - (~ Sexp) (prob - Plan-problem)
   (cond ((!_metric prob) false)
	 (t
	  (let ((dom (domain-prepare-for-planning (!_domain prob))))
	     (repeat :for ((vb :in (domain-get-local-bdgs dom)))
	      :result '(minimize (total-steps))
	      :within
		(cond ((is Domain-bdg vb)
		       (let ((v (!_val vb)))
			  (cond ((is Causation-function v)
				 (:continue
				  :until (let ((d (!_defn v)))
					    (or (is Durative-action-defn d)
						(is Process-defn d)))
				  :result '(minimize (total-time)))))))))))))

;;; Records a real time plus some infinitesimal skips
(defnisptype Time-plus (Structure time - Float skips - Integer))

;;; !_duration = 0 means that it is a classical step.
(defnisptype Act-occurrence
    (Structure time duration - Time-plus act - Sexp))

;;; Plan with duratives made explicit
(defnisptype IPC-plan
    (Structure dur-mode - (Const :durative :point :mixed)
	       steps - (Lst Act-occurrence)
	       makespan - Float))
;;; -- Note that the !_makespan can be greater than the endpoint
;;; of the last action if we have to wait for some external event
;;; to make a goal true
			  
;;; Transform from Optop format to IPC format
(deffunc optop->ipc - IPC-plan (pp - Plan-prefix)
   (let ((last-time (!_duration pp))
	 (dom (!_domain pp))
	 (has-duratives false)
	 (has-points false)
	 (has-waits false))
      (let-fun ((:def transform-prequel - (Lst Act-occurrence)
				   (pre-pp - Plan-prefix
				    enders - (Lst (Lrcd Sexp Float)))
		  ;; enders is a list of duratives that are going
		  ;; to end in the "future" of 'pre-pp'.
;;;;		  (trace-around transform-prequel (:if (not (!_is-initial pre-pp)))
;;;;		     (:> "(transform-prequel: " (!_last-act pre-pp)
;;;;			 :% 3 enders")")
		  (cond ((!_is-initial pre-pp)
			 !())
			(t
			 (let ((act (!_term (!_last-act pre-pp))) - Sexp)
			    (let ((ipc-before
				     (transform-prequel
				        (!_prior pre-pp)
					(enders-cull-stoppers enders pre-pp)))
				  (time-before (!_duration (!_prior pre-pp)))
				  (skip (step-count-just-before pre-pp)))
			       (cond ((eq (car act) 'dur-start)
				      (cons (durative-act (cadr act)
							  time-before skip enders)
					    ipc-before))
				     ((memq (car act)
					    '(wait wait-while wait-non-negative))
				      (!= has-waits true)
				      ipc-before)
				     (t
				      (!= has-points true)
				      (let ((steps-before
					      (cond ((null ipc-before) 0)
						    (t
						     (!_skips
						       (!_time (car ipc-before)))))))
					 (cons (make Act-occurrence
						  (make Time-plus
						      time-before steps-before)
						  (make Time-plus
						      0.0
						      (+ steps-before 1))
						  act)
					       ipc-before))))))))
;;;;		     (:< (val &rest _) "transform-prequel: " val))
		  )

		(:def step-count-just-before - Integer (pre-pp - Plan-prefix)
		    (let ((prior (!_prior pre-pp)))
		       (cond ((!_is-initial prior) 0)
			     ((matchq (dur-start ?@_)
				      (!_term (!_last-act prior)))
			      0)
			     (t
			      (- (!_length prior)
				 (!_length (!_prior prior)))))))

		(:def enders-cull-stoppers - (Lst (Lrcd Sexp Float))
				   (enders - (Lst (Lrcd Sexp Float))
				    pre-pp - Plan-prefix)
;;;;		   (trace-around enders-cull
;;;;		      (:> "(enders-cull: " enders ")")
		   (let ((result (!_sit pre-pp)) - Situation)
		      (!= enders (list-copy *-*))
		      (repeat :for ((a :in (deduce
					      '(process-transition ?w (dur-happening ?d))
					      dummy-id* (empty-env)
					      result dom))
				    - Bdgenv
				    durative - Sexp)
			 (!= durative
			     (one-value (safe-varsubst '?d dummy-id* a)))
			 (cond ((eq (get-var-val-1 'w a)
				   'starting)
				(!= enders
				   (delete-if
				      (\\ (e) (equal (car e) durative))
				      *-*)))
			       (t
				(!= enders
				   (cons (tuple durative
						(!_duration pre-pp))
					 *-*)))))
		      enders)
;;;;		      (:< (val &rest _) "enders-cull: " val))
		   )

		(:def durative-act - Act-occurrence
		     (dur - Sexp pre-time - Float
		      skip - Integer enders - (Lst (Lrcd Sexp Float)))
		   (!= has-duratives true)
		   (let ((e (assoc= #'equal dur enders)))
		      (let ((end-time
			       (cond (e (cadr e))
				     (t
				      (dbg-save :run-loud dur enders)
				      (signal-problem durative-act
					 dur " never ends"
					 (:proceed
					    !"I'll assume it ends when the plan does"))
				      last-time))))
			 (make Act-occurrence
			    (make Time-plus pre-time skip)
			    (make Time-plus (- end-time pre-time) 0)
			    dur)))))

	 (let ((steps (transform-prequel
			 pp
			 (repeat :for ((pin :in (!_active (!_processes (!_sit pp))))
				       - Process-instance
				       term - Sexp)
			    (!= term (!_term (!_uterm pin)))
			  :when (eq (car term) 'dur-happening)
			  :collect
			     (tuple (cadr term)
				    last-time)))))
	    (make IPC-plan
		(cond ((or has-duratives has-waits)
		       (cond (has-points ':mixed)
			     (t ':durative)))
		      (t ':point))
		(reverse steps)
		(!_length pp))))))
		      
(specdecl (time-list* !()) - (Lst Time-plus))

;;; Print in competition format.  If second arg is true, suppress zero
;;; durations for :point plans.
(defproc ipc-format-print - Void (plan - IPC-plan
				   points-fixed - Boolean
				   outsrm - Stream)
    (cond ((not (eq (!_dur-mode plan) ':point))
	   (!= points-fixed false)))
    (repeat :for ((step :in (!_steps plan)))
       (let ((start-time (!_time step))
	     (dur (!_duration step)))
	  (cond ((not (is Time-plus dur))
		 (signal-problem ipc-format-print
		    "dur = " dur)))
	  (out (:to outsrm)
	       (:a (time-out start-time outsrm))
	       ": " (!_act step)
	       " [" (:a (time-out dur outsrm)) "]" :%))

	    :where
	    
	      (:def time-out - Void (tp - Time-plus osrm - Stream)
		 ;;;;(on-list tp time-list*)
		  (out (:to :string)
		     (:q ((and points-fixed
			       (infinitesimal-time tp))
			  (!_skips tp))
			 (t
			  (:f "~,2f" (!_time tp))
			  (:q ((not (= (!_skips tp) 0))
			       "," (!_skips tp)))))))

	      (:def infinitesimal-time - Boolean (tp - Time-plus)
		 (and (= (!_time tp) 0)
		      (not (= (!_skips tp) 0))))))

(defvar host-specific* false)


   (cond ((not host-specific*)
	  (out (:to *query-io*) "Type name of machine-specific run file: ")
	  (let ((r (in (:from *query-io*) :linestring)))
	     (load r))
	  (!= host-specific* true)))
