;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;;$Id: search-exc.lisp,v 1.19 2006/01/24 13:33:47 dvm Exp $

(depends-on %module/ nisp)

#+pre-chunk
(depends-on %module/ nispobj)

(depends-on :at-run-time %optop/ rmgraph regressmatch planprob execute
			         optexec rtree-exec types continproj)

 ;;; The !_mode of !_estate is always :forced.
 ;;; The !_sit of !_estate is a plausible final situation for the sequel.
 (defnisptype Plan-sequel
     (Lstructure () rtree - Reduction-tree
 		   estate - Exec-state
                   in-progress - (Lst (Lrcd L-node Plan-prefix Thread-stack))))
;;; The rtree could be made a thread, and incorporated into the estate

cccc(defnisptype Sit-search-state
    (Structure score effort - Float
	       discrepancy incoherence - Fixnum
	       preserve-only - Boolean ; if true, only consider steps that
	                               ; preserve goals
	       plan-prefix - Plan-prefix

	       sequel - (~ Plan-sequel)
               parent - (~ Sit-search-state)
               children - (Lst Sit-search-state)
               children-found - Boolean
	       ;; These are false if reuse-rmgraphs* = false --
	       parent-top-node - (~ Effort-spec)
	       parent-eff-db - (~ Effort-db)

       (:handler
           (print (sss - Sit-search-state srm - Stream)
              (out (:to srm) "#<Sit-search-state/"
		   (:q ((!_preserve-only sss) "?/"))
                   (:f "~,1F" (!_score sss))
                   (:q ((not (= (!_score sss)
                               (!_effort sss)))
                       "[" (:f "~,1F" (- (!_effort sss) (!_score sss))) "]"))
                   ":" (!_incoherence sss)
                   1 (!_acts (!_plan-prefix sss)) ">"))))
  (situation - Situation *integrable (st - Sit-search-state)
     (!_sit (!_plan-prefix st))))

(deffunc init-sit-search-state - Sit-search-state (pp - Plan-prefix)
   (make-inst Sit-search-state
      :score 0 :effort 0 :discrepancy 0 :incoherence 0
      :preserve-only false
      :plan-prefix pp 
      :sequel false
      :in-progress !()
      :parent false :children !() :children-found false
      :parent-top-node false
      :parent-eff-db false))

(deffunc sprout-child-state - Sit-search-state
                       (effort - Float incoherence - Fixnum
                        preserve-only - Boolean
                        plan-prefix - Plan-prefix
                        sequel - Plan-sequel
                        parent - (~ Sit-search-state)
			parent-top-node - (~ Effort-spec)
			parent-eff-db - (~ Effort-db)
                        th - Domain)
                       (ignore th)
   (let ((new (make-inst Sit-search-state
		 :score effort
		 :effort effort
		 :discrepancy (cond (parent (!_discrepancy parent))
				    (t 0))
		 ;; -- Rationale: If the parent diverged from
		 ;; optimality at some point, this search state
		 ;; starts with the same degree of divergence.
		 ;; (Which can increase after its score is computed.)
		 :incoherence incoherence
		 :preserve-only preserve-only
		 :plan-prefix plan-prefix
		 :sequel sequel
		 :parent parent
		 :children !()
		 :children-found false
		 :parent-top-node parent-top-node
		 :parent-eff-db parent-eff-db)))
      (cond (parent
             (!= (!_children parent) (cons new *-*))))
      new))

(specdecl maximal-match-count* maximal-match-tot* g-node-tot*
	  sit-expansions* mm-unify-count* 
          infinite-effort* - Integer
	  (max-hyb-branching* 20) (max-hyb-queue-length* 100) - Integer)

;;;;(specdecl loser* queue* alt-queue* - Obj)

(defun load-domain-files (&rest filenames)
   (repeat :for ((fn :in (filespecs->pathnames `(domains/ ,@filenames))))
      (format t "Loading ~a~%" (namestring fn))
      (format t "~s~%"
	      (opt-file-syncheck fn))))

(specdecl (save-last-effort-tab* false)
	  (save-last-sit-space* false) - Boolean
	  (last-sit-space* false) - (~ Situation-space))

(specdecl (default-require-quiescence* false) - Boolean)

(specdecl sit-init-completer* - (Fun Void (Situation) t))

;;; 'require-quiescence' means that, in a domain with durative actions,
;;; none are in progress when the plan is at an end.
(defun search-for-plan (pp &key (metric false)
				(maxspan false)
				(maxlength 1000)
				(maxplans 10000)
				(maxtime 10)   ; search time, in minutes
				(initial false)
				(sit-space false)
				(require-quiescence
				 default-require-quiescence*))
   (cond (maxspan (!= maxspan (float *-*))))
   (!= pp (plan-problem-initialize *-*))
   (!= maxtime (* 60 maxtime))
   (decl (pp - Plan-problem metric - (~ Sexp) maxspan - (~ Float)
	  maxlength maxplans maxtime - Fixnum initial - (~ Situation)
	  sit-space - (~ Situation-space))
      (cond (sit-space
	     (cond (save-last-sit-space*
		    (!= last-sit-space* sit-space))))
	    (t
	     (cond ((and save-last-sit-space* last-sit-space*)
		    (!= sit-space last-sit-space*))
		   (t
		    (!= sit-space (make Situation-space (!_sit pp)))
		    (cond (save-last-sit-space*
			   (!= last-sit-space* sit-space)))))))
      (bind ((sit-init-completer* (sit-start-processes sit-space)))
	 (!= (!_sit pp)
	     (be Initial-situation
                 (uniquify-situation *-* (!_index sit-space))))
	 (cond (initial
		(!= initial (uniquify-situation *-* (!_index sit-space)))))
	 (multi-let (((solns numtried goal-nodes-num maxmatches
		       run-time real-time 
		       mode eff-db)
		      (hyb-solve pp initial metric maxspan maxlength
				 maxplans maxtime require-quiescence
				 sit-space)))
	    (cond (save-last-effort-tab*
		   (!= last-effort-tab* eff-db)))
;;;;	    (out (:to *query-io*) "goal-nodes-num = " goal-nodes-num :%)
	    (values solns
		    numtried goal-nodes-num maxmatches run-time real-time 
		    mode)))))

(specdecl search-for-plan
	  - (Fun (Multvals solns - (Lst Plan-prob-soln)
			   numtried goal-nodes-num maxmatches - Integer
			   run-time - Run-time-spec
			   real-time - Integer
			   mode - (Const :best-first
					 :limited-discrepancy))
		 (Obj . Obj)))			   

(specdecl (fluenv* (domain-place-env
		    (try-domain-with-name 'fluents false)))
          - Typenv)

(deffunc hyb-solve - (Multvals solns - (Lst Plan-prob-soln)
                              numtried goal-nodes-num maxmatches - Integer
			      run-time - Run-time-spec
			      real-time - Integer
	                      mode - (Const :best-first
					    :limited-discrepancy)
			      eff-db - (~ Effort-db))
	         (pp - Plan-problem initsit - (~ Situation) metric - (~ Sexp)
		  maxspan - (~ Float) maxlength maxplans maxtime - Fixnum
		  require-quiescence - Boolean sit-space - Situation-space)
   (!= pp (plan-problem-initialize pp))
   (verify-domain-okay-for-optop (!_domain pp))
   (cond ((not maxspan)
	  (!= maxspan (domain-temporal-scope (!_domain pp)))))
   (multi-let (((metric-te - (~ Typed-exp)
		 max-or-min - (~ (Const minimize maximize)))
		(cond ((and metric
			    (not (eq metric ':effort)))
		       (problem-metric-parse metric (!_domain pp)))
		      (t
		       (values false false))))
	       (start-run (get-current-run-time-spec))
	       (start-real (get-internal-real-time)))
      (cond ((and metric-te
		  (> (!_ totbugs metric-te)
		     0))
	     (signal-problem hyb-solve
		"Ill-defined metric: " metric)))
      (cond ((and (not metric-te)
		  (not (eq metric ':effort))
		  (!_metric pp))
	     (!= max-or-min (car (!_metric pp)))
	     (!= metric-te (cadr (!_metric pp)))))
      (let ((metric
               (cond ((eq metric ':effort)
		      false)
		     (metric-te
		      (let ((m (nth-value 0
				  (internalize
				     metric-te !()
				     ':antecedent))))
			 (cond ((eq max-or-min 'maximize)
				(!= m `(- ,m))))
			 m))
		     (t 'total-steps))))
	 ;; -- from here on, metric = false means "use estimated effort"
         (bind ((last-effort-tab* false) - (~ Effort-db))
	    (multi-let (((solns numtried goal-node-num bf-mode)
			 (with pp
			    (hyb-search !>goal initsit !>sit !>domain
					metric maxspan maxlength
					maxplans maxtime require-quiescence
					sit-space))))
	       (values solns
		       numtried
		       goal-node-num
		       maximal-match-tot* 
		       (run-time-spec-diff
			   (get-current-run-time-spec)
			   start-run)
		       (internal-time-in-milliseconds
			  (- (get-internal-real-time) start-real))
		       (cond (bf-mode ':best-first)
			     (t ':limited-discrepancy))
		       last-effort-tab*))))))

(specdecl ;;;;last-state* next* - (~ Sit-search-state)
          ;;;;new* - (Lst Sit-search-state)
	  )

(specdecl unify-count* unify-success-count*
          unify-tot* unify-success-tot* - Integer)

(specdecl (fat-thresh* 9) - Fixnum)
; If there are N states on the queue that are all the same, and N >= fat-thresh*,
; switch to limited-discrepancy mode

; If t, successors are in random order (modulo sorting) --
(specdecl (scramble-successors* false)
  ; If nil, when backtracking in HC-mode pick from arbitrary point in queue.
	  (hc-backtrack-depth-first* true) - Boolean)
	  
(specdecl (ying-or-yang ':yang) - (Const :ying :yang))

(deffunc hyb-search - (Multvals solns - (Lst Plan-prob-soln)
				numtried num-goal-nodes - Integer
				bf-mode - Boolean)
                   (ur-goal - Sexp s - (~ Situation)
		    ss - Initial-situation
		    th - Domain metric - Sexp
		    maxspan - Float maxlength maxplans maxtime - Fixnum
		    require-quiescence - Boolean sit-space - Situation-space)
   (!= th (domain-prepare-for-planning th))
   (domain-make-current th)
   (cond ((and require-quiescence
	       (domain-declares-requirement th ':durative-actions)
	       (domain-declares-requirement th ':universal-preconditions))
	  (!= ur-goal `(and ,ur-goal
			    (forall (p D= D> D<)
			       (not (dur-in-progress p D= D> D<)))))))
   (cond ((not (eq ss (!_initsit sit-space)))
	  (signal-problem hyb-search
	     "Situation-space " sit-space
	     :% " does not have initial-situation " ss)))
   (let ((init-sit (get-actual-init-sit s ss sit-space))
	 - Situation)
      (let ((bf-mode true) - Boolean
	    (bf-mark (make Sitmark))
	    (hc-mark (make Sitmark))
	    mark - Sitmark
	    (init-state (init-sit-search-state 
			   (place-Plan-prefix 0 0.0 init-sit false
					      init-sit sit-space)))
	    - Sit-search-state
	    (init-run-time (get-internal-run-time))
	    (goal (disjunctive-normal-form ur-goal)))
	 (!= mark bf-mark)
	 (!= (!_mark init-sit) mark)
	 (!= (!_mark (!_plan-prefix init-state)) mark)
	 (!= maximal-match-tot* 0)
	 (!= biggest-maxmatch* 0)
	 ;;;;(!= l-node-tot* 0)
	 (!= g-node-tot* 0)
	 (!= unify-tot* 0)
	 (!= unify-success-tot* 0)
	 (!= sit-expansions* 0)
	 (!= maxtime (* *-* internal-time-units-per-second))
	 (multiple-value-let (solns - (Lst Plan-prob-soln)
			      new - (Lst Sit-search-state)
			      gn-num - Fixnum)
			     (sit-search-expand 
				 init-state goal metric
				 (be Time-int maxspan)
				 maxlength
				 init-sit sit-space th mark false)
	    ;;;;(!= last-state* false)
	    (repeat :for ((queue '()) - (Lst Sit-search-state)
			  (i = 1 :by 1) - Fixnum ;CURLEN
			  (comparator #'score-smaller)
			  - (Fun Boolean (Sit-search-state Sit-search-state) nil)
			  (prevlength 1000) - Integer prevscore - Float
			  ; Keep track of how much estimate rises as plan
			  ; gets longer
			  (tot-lengthening-lossage 0.0) av-lengthen-loss - Float
			  (times-lengthened 0)
			  - Integer
			  next - Sit-search-state scores - (Lst Float)
			  trimmed - Boolean)
		(effort-stats-out gn-num)
	     :while (null solns)
	     else :result (values solns i g-node-tot* bf-mode)
	        (cond ((eq ying-or-yang ':ying)
		       (!= new (ying-set-discrepancies new))))
		(!= new (stable-sort (cond (scramble-successors*
					    (list-scramble new))
					   (t new))
				     comparator))
		(repeat :for ((s :in new))
		 :result nil
		 :until (> (!_length (!_plan-prefix s))
			  prevlength)
		 :result (progn (!= tot-lengthening-lossage
				   (+ *-* (/ (- (!_score s)
						prevscore)
					     (- (!_length (!_plan-prefix s))
						prevlength))))
			       (!= times-lengthened (+ *-* 1))))
		(!= trimmed (> (length new) max-hyb-branching*))
		(cond (trimmed
		       (!= new (dtake max-hyb-branching* new))))
		(show-new-succs new trimmed)
		(!= queue
		    (bf-queue-merge new queue comparator))
	     :until (or (null queue)
			(> sit-expansions* maxplans)
			(> (- (get-internal-run-time)
			      init-run-time)
			   maxtime))
	     :result (progn (cond ((null queue)
				   (dbg-out (> plan-dbg-amt* 0)
				      "No more plans to try" :%))
				  ((> sit-expansions* maxplans)
				   (dbg-out (> plan-dbg-amt* 0)
				      "Tried too many plans" :%))
				  (t
				   (dbg-out (> plan-dbg-amt* 0)
				      "Search took too long")))
			   (values !() i g-node-tot* bf-mode))
   ;;;;	     (!= queue* queue)
		(cond ((> times-lengthened 0)
		       (!= av-lengthen-loss (/ tot-lengthening-lossage
					       times-lengthened))))
		(dbg-out (> plan-dbg-amt* 0)
		      "Mode = "
		     (:q (bf-mode "BF") (t "HC"))
		     (:q ((> times-lengthened 0)
			 " Av eff gain/step = " (:f "~,1F" av-lengthen-loss)))
		     " Obesity = " (obesity queue) "/" (len queue) 
   ;;;;		  " [" (<# (\\ (s) (!_length (!_plan-prefix s)))
   ;;;;			   queue) "]"
		     :%)
		(cond ((and bf-mode
			    (> (obesity queue) fat-thresh*))
		       (dbg-out (> plan-dbg-amt* 1) "Switching to hill-climbing mode" :%)
		       (!= queue
			   (stable-sort (copy-list queue)
					(correct-for-length-compare
					    av-lengthen-loss)))
		       (!= next (car queue))
		       (state-re-mark next hc-mark)
		       (!= bf-mode false)
		       (!= mark hc-mark)
		       (!= comparator #'score-less-discrepant))
		      (t
		       (!= next
			   (cond ((or hc-backtrack-depth-first*
				      (not (null new)))
				  (car queue))
				 (t
				  (one-of-best queue comparator))))))

		;;;;(!= last-state* next)
		(cond ((and (> plan-dbg-amt* 0) (not (memq next new)))
		       (out (:to *error-output*) "Backtracking" :%)))
		(cond ((> (length queue) max-hyb-queue-length*)
		       (!= queue (dtake max-hyb-queue-length* *-*))))
		(cond ((> plan-dbg-amt* 1)
		       (!= scores (<# !_score queue))
		       (out (:to *error-output*)
			    i " Queue length: " (len queue) " / Scores: "
			    scores :%
			    )))
		(cond ((> plan-dbg-amt* 0)
		       (show-current-plan (!_plan-prefix next)
					  (!_score next)
					  (!_discrepancy next))
   ;;;;		    (cond ((= (!_length (!_plan-prefix next)) 9)
   ;;;;			   (!= forceproj-dbg* true)))
		       ))
		;(error-break bf-search :novalue "ready to expand")
		(!= prevlength (!_length (!_plan-prefix next)))
		(!= prevscore (!_score next))
		(!= < solns new gn-num >
		    (sit-search-expand next goal metric
				       (be Time-int maxspan)
				       maxlength
				       init-sit sit-space th mark
				       (cond (bf-mode true)
					     (t bf-mark))))
   ;;;;	     (!= loser* next)
		(!= queue (dremove1q next *-*)))))))

;;; Find situation with process complex in motion, either s, ss,
;;; or an "actual init" that is a successor of one of them.
(deffunc get-actual-init-sit - Situation (s - (~ Situation)
					  ss - Initial-situation
					  sit-space - Situation-space)
   (repeat :for ((i = 1 :to 2))
    :result
      (signal-problem hyb-search
	 "Can't find initial situation with process complex")
    :within
      (let ((isit (or s (let ((actual-ss
				    (find-old-projection
				       ss 'actual-init sit-space)))
			   (cond (actual-ss
				  (!_sit actual-ss))
				 (t ss)))))
		    - Situation)
	 (:continue
	  :until (!_processes isit)
	     (funcall sit-init-completer*
		      (or s ss))
	  :result isit))))

(specdecl (count-incoherence* true) - Boolean)

(deffunc score-smaller - Boolean (s1 s2 - Sit-search-state)
   (cond ((and count-incoherence* (= (!_score s1) (!_score s2)))
          (< (!_incoherence s1) (!_incoherence s2)))
         (t
          (< (!_score s1) (!_score s2)))))

(deffunc score-less-discrepant - Boolean (s1 s2 - Sit-search-state)
   (< (!_discrepancy s1) (!_discrepancy s2)))
	  
;         ((= (!_DISCREPANCY S1) (!_DISCREPANCY S2))
;          (SCORE-SMALLER S1 S2))


(specdecl (max-discrepancy* 1) - Integer)

;;; Set discrepancies, and 
;;; discard all states with discrepancy > max-discrepancy*
(defproc ying-set-discrepancies - (Lst Sit-search-state)
	              (search-states - (Lst Sit-search-state))
   (cond ((not (null search-states))
	  (!= search-states
	      (stable-sort *-* #'score-smaller))
;;; Unnecessary given that this was set in 'sprout-child-state' --
;;;;	  (!= (!_discrepancy (head search-states))
;;;;	      0)
	  (repeat :for ((ss :in (tail search-states)) - Sit-search-state
			(current-score (!_score (head search-states)))
			   - Number
			(discrep 0) - Fixnum)
	     (cond ((not (= (!_score ss)
			    current-score))
		    (!= current-score (!_score ss))
		    (!= discrep (+ *-* 1))))
	     (!= (!_discrepancy ss) (+ *-* discrep))
	   :until (> discrep max-discrepancy*)
	   :result (cond ((> discrep max-discrepancy*)
			  (repeat :for ((ssl = search-states :then (tail ssl))
					- (Lst Sit-search-state))
			   :until (= (!_discrepancy (cadr ssl))
				     discrep)
			   :result (!= (tail ssl) !())))))))
   search-states)

(deffunc one-of-best - Sit-search-state
                           (queue - (Lst Sit-search-state)
			    comparator - (Fun Boolean (Sit-search-state
						       Sit-search-state)
					              nil))
   (cond ((or (null (cdr queue))
	      (funcall comparator (car queue) (cadr queue)))
	  (car queue))
	 (t
	  (let ((competitors (list (car queue))))
	     (repeat :for ((s :in (cdr queue)))
	      :until (funcall comparator (car queue) s)
	        (!= competitors (cons s *-*)))
	     (randelt competitors)))))

(deffunc correct-for-length-compare
                      - (Fun Boolean (Sit-search-state Sit-search-state))
	      (av-lengthen-loss - Float)
              (ignore av-lengthen-loss)
   (\\ (s1 s2 - Sit-search-state)
      (< (!_discrepancy s1)
	 (!_discrepancy s2))))
	     
;;;;	    (t
;;;;	     (< (- (!_score s1)
;;;;		   (* (!_length (!_plan-prefix s1))
;;;;		      av-lengthen-loss))
;;;;		(- (!_score s2)
;;;;		   (* (!_length (!_plan-prefix s2))
;;;;		      av-lengthen-loss)))))))

;;; Q is non-().  How many elements at the front have the same score,
;;; within 10% ?
(deffunc obesity - Fixnum (q - (Lst Sit-search-state))
   (let ((e (!_score (car q)))
         (n 1))
      (repeat :for ((s :in (cdr q)))
       :until (> (!_score s) (* 1.1 e))
         (!= n (+ n 1))
       :result n)))

;;;; Alternative defn. of obesity: How many elements in queue are almost as 
;;;; good as first element.  (This is a stopgap, since we really care about depth,
;;;; not length, but they're the same if we neglect preservation.)
;;;;(deffunc obesity - Fixnum (q - (Lst Sit-search-state))
;;;;   (let ((e (- (!_length (!_plan-prefix (car q)))
;;;;               1))
;;;;	 (e (!_score (car q)))
;;;;         (n 1))
;;;;      (repeat for ((s in (cdr q)))
;;;;         (cond ((>= (!_length (!_plan-prefix s))
;;;;                    e)
;;;;                (!= n (+ n 1)))))
;;;;      n))

;;;;(specdecl sit* - Situation act* - Action topnode* - Effort-spec)

(deffunc show-new-succs - Void (new - (Lst Sit-search-state) trimmed - Boolean)
             (dbg-out (> plan-dbg-amt* 0)
                      "Considering "
		      (:q (trimmed "[trimmed] "))
                      (repeat :for (ss :in new)
		       :collect 
			 (lrecord ;(!_SCORE SS)
                                          (!_effort ss)
                                          (!_discrepancy ss)
                                          (!_incoherence ss)
					  (!_last-act (!_plan-prefix ss))))
                      :%)
;;;;    (cond ((null new)
;;;;	   (breakpoint show-new-succs
;;;;	      "Dead end")))
    )

(specdecl (stop-after-expansion* false) - Boolean)

; Assign a new sitmark to S and its ancestors
(defproc state-re-mark - Void (s - Sit-search-state hc-mark - Sitmark)
   (let ((pr (!_plan-prefix s)))
      (repeat
         (!= (!_mark (!_sit pr)) hc-mark)
       :until (!_is-initial pr)
         (!= pr (!_prior pr)))))

(specdecl act-comparator*
          - (Fun Boolean (Feasible-action-spec Feasible-action-spec) nil))

(deffunc sit-search-expand - (Mlv (Lst Plan-prob-soln)
				  (Lst Sit-search-state)
				  Fixnum)
                    (next - Sit-search-state goal - (Lst Sexp)
		     metric - Sexp maxspan - Time-int maxlength - Integer
                     init-sit - Situation sit-space - Situation-space
		     th - Domain mark - Sitmark
                     allow-reencounter - (Either Sitmark (Const t nil)))
   (!= sit-expansions* (+ *-* 1))
   (!= maximal-match-count* 0)
   (!= mm-unify-count* 0)
   (!= unify-count* 0)
   (!= unify-success-count* 0)
   (bind ((act-comparator*
	     (construct-act-comparator next))
	  - (Fun Boolean (Feasible-action-spec Feasible-action-spec) nil))
      (let ((pp (!_plan-prefix next))
	    (blocking-subgoals
	        (</ (\\ (bg - (Lst (Lst Sexp)) ; blocking goal in dnf
			 block - Blockage)
		       (match-cond (!_gate block)
			  (:? (:open-subgoals ?dnf)
			     (dnf-conjoin bg dnf))
			  (t bg)))
		    (disjunctive-normal-form 'true)
		    (!_blocked (!_estate (!_sequel next))))))
         (let ((pathlen (!_length pp))
	       (timespent (!_duration pp))
	       (estate (exec-state-normalize estate)))
	    (multi-let (((topnode eff mml eff-db num-goal-nodes)
			 (compute-effort goal (!_sequel next)
                                         pp estate
                                         (!_parent-top-node next)
                                         (!_parent-eff-db next)
					       sit-space th
					       (+ (- maxlength pathlen)
						  2))))
               (values (repeat :for (mm :in mml)
                        :collect
                          (let ((score (plan-solution-score
                                          pp metric th)))
                             (make Plan-prob-soln
                                       pp  
                                       score
                                       (!_subst mm))))
                       (cond ((and (null mml)
                                   (< eff infinite-effort*)
                                   (< pathlen maxlength)
                                   (< timespent maxspan))
                              (let ((acts
                                       (be (Lst Feasible-action-spec)
                                           (!_actions topnode)))
                                    - (Lst Feasible-action-spec))
                                 (dbg-out (> plan-dbg-amt* 1)
                                    "Feasible actions to try [pathlen = "
                                                               pathlen "]: "
                                    (:e (repeat :for ((a :in acts))
                                           (:o :%(!_act a)
                                               (:q ((not (= (!_length (!_pp a))
                                                            pathlen))
                                                    " [" (!_length (!_pp a))
                                                     "]"))))))
				 (successor-states
				    (!_actions topnode)
				    next pp maxspan maxlength dom)))
                             (t
                              '()))
                       num-goal-nodes))))))

(deffunc construct-act-comparator - (Fun Boolean
					 (Feasible-action-spec
					  Feasible-action-spec))
                         (next - Sit-search-state)
   (cond ((and count-incoherence*
	       (> (!_length (!_plan-prefix next))
		  0))
	  (\\ (a1 a2 - Feasible-action-spec)
	     (let ((e1 (!_eff a1)) (e2 (!_eff a2)))
		(or (< e1 e2)
		    (and (= e1 e2)
			 (< (action-incoherence
				(!_act a1)
				(car (!_sequel next)))
			    (action-incoherence
				(!_act a2)
				(car (!_sequel next)))))))))
	 (t
	  (\\ (a1 a2 - Feasible-action-spec)
	     (< (!_eff a1) (!_eff a2))))))

(specdecl ;;;;new-state* - Sit-search-state
          deletes* - (Lst Occasion)
          (include-likelies* false) - Boolean)

;;; This takes the place of 'state-after-act' --
(deffunc successor-states - (Lst Sit-search-state)
	              (acts - (Lst Feasible-action-spec)
		       current - Sit-search-state
		       maxspan - Time-int maxlength - Integer dom - Domain)
   (let ((pp (!_plan-prefix current))
	 (sequel (!_sequel current)))
      (cond (stop-after-expansion*
	     (dbg-save :run-loud topnode)
	     (breakpoint successor-states
		"Expanded")))
      (!= (!_children-found current) true)
      (let-fun ()
	 (nconc (states-from-old-acts)
		(states-from-new-acts))
       :where

         (:def states-from-old-acts - (Lst Sit-search-state)
            (repeat :for ((th :in (!_active (!_estate sequel))))
             :within
                (cond ((not (!_is-empty (!_stack th)))
                       (match-cond (!_this-frame (!_stack th))
                          (:? (:act ?a)
                             (multi-let (((defn int-defn env)
                                          (act-action-defn a dom)))
                                (


                             (let ((act (uniquify-action a dom)))
                                (


	     (let ((act-specs
		      (execute-one-step (!_estate sequel)))
		   - (Lst (Lrcd (~ Action) Exec-status Exec-state)))
	        (cond ((null act-specs)
		       ;; No feasible acts proposed by current Exec-state
		       !())
		      (t
		       (repeat :for ((trip :in act-specs)
				       - (Lrcd (~ Action)
					       Exec-status
					       Exec-state)
				     (best-rtree-threads
					 (cond ((null acts) !(Thread))
					       (t
						(threads-for-best-rtree
						    acts))))
				      - (Lst Thread))
		       :within
			  (let ((augmented-exec-state
				   (next-exec-state (third trip)
				       (:active (append best-rtree-threads
							!>active)))))
			     (:continue
			      :nconc
			         (state-from-forced-projection
				    augmented-exec-state)))
		       :where
		          (:def threads-for-best-rtree - (Lst Thread)
				          (acts - (Lst Feasible-action-spec))
			     (let* ((best-act
				       (</ (\\ (best-so-far
						next-act)
					      (cond ((< (!_eff next-act)
							(!_eff best-so-far))
						     next-act)
						    (t best-so-far)))
					   (first acts)
					   (rest acts)))
				     ;; 'best-act' carries most
				     ;; promising r-tree for the new
				     ;; acts.  We use it when
				     ;; projecting all the old acts.
				    (new-act-threads
				       (list (make Thread
						(funcall 
						   (rtree-thread-gen 
						      (!_rtree best-act)
						      false (!_sit pp) dom)
						   no-op-thread-cont*)
						(make Thread-stack
						   ':rtree-thread
						   empty-thread-stack*)))))
			       new-act-threads)))))))

         (:def states-from-new-acts - (Lst Sit-search-state)
	                            ()
            This should now never expand -- On next cycle 'states-from-old-acts'
            will do that.  So every act is projected at least once at abstract
            level.

	    (repeat :for ((a :in acts))
	     :nconc
               (multi-let (((adefn idefn env1)
                            (act-action-defn (!_act a) dom)))
	          (cond ((and (is Action-defn adefn)
                              (!_has-methods (be Action-defn adefn)))
                         (act-method-thread-gen
                             act (be Action-defn adefn) plan-sit env))
	       (t ... ))



	       (let ((augmented-exec-state
			(exec-state+feasible-act
			   (!_estate sequel)
                           (!_rtree a)
                           true pp)))
                  (multi-let (((status-for-a new-state cycles)
                               (continue-execution 
                                  augmented-exec-state
                                  (- maxspan (!_elapsed augmented-exec-state))
                                  (- maxlength (!_steps augmented-exec-state))
                                  projection-cycles*)))
                     (cond ((memq status-for-a '(:continuing :finished))
                            (state-from-forced-projection
                               augmented-exec-state))
                           (t
                            (signal-problem states-from-new-acts
                               "Nominally feasible act " a
                               " could not be projected"))))))

         (:def state-from-forced-projection - (Lst Sit-search-state)
                                        (e-state - Exec-state)
            (multi-let (((status post-projection cyc)
                         (continue-execution
                             (next-exec-state e-state
                                (:mode ':forced
                                 :vals (exec-value-set
                                          forcings-link* *-*
                                          false (\\ () 0))))
                             (- maxspan (!_elapsed e-state))
                             (- maxlength (!_length e-state))
                             projection-cycles*)))
               (list
                  (sprout-child-state
                     (
              (effort - Float incoherence - Fixnum
                        preserve-only - Boolean
                        plan-prefix - Plan-prefix
                        sequel - Plan-sequel
                        parent - (~ Sit-search-state)
			parent-top-node - (~ Effort-spec)
			parent-eff-db - (~ Effort-db)
                        th - Domain)

               (cond ((eq status ':stalled)
                      ;; Possible infinite loop
                      )
                     ((

		 (let ((best-act
		  (</ (\\ (best-so-far
			   next-act)
			 (cond ((< (!_eff next-act)
				   (!_eff best-so-far))
				next-act)
			       (t best-so-far)))
		      (first acts)
		      (rest acts))))
	    (list (make Thread
		     (funcall 
			(rtree-thread-gen 
			   (!_rtree best-act)
			   false (!_sit pp) dom)
			(make Thread-cont
			   (\\ (after-state - Exec-state
				_ - Thread-stack)
			      (values ':continuing
				      after-state))))
		     (make Thread-stack
			   ':rtree-thread
			   empty-thread-stack*))))


(exec-state+feasible-act
   (!_estate sequel)
   a)


1 0.0 feas-act
			  adds deletes prev
			  metric maxspan
			  old-top-node old-eff-db
			  allow-reencounter linkid assumptions mark
			  dom)				


(deffunc projection-score - Number (final - Situation
                                    metric - Sexp
                                    time-est - Number step-est - Integer
                                    dom - Domain)
	  (multi-let (((evaluable basic-score)
		       (opt-eval metric final
				 (list (tuple 'total-steps
					      step-est)
				       (tuple 'total-time
					      time-est))
				 dom)
		       ))
	     (cond (evaluable
		    (let ((res (* basic-score
				  (expt penalty-fac* penalty))))
		       (dbg-out (> plan-dbg-amt* 0)
			  "Plan prefix score = " res)
		       res))
		   (t
		    (dbg-save :run-loud
			      metric final pp dom step-est time-est)
		    (signal-problem plan-prefix-score
		       "Can't evaluate metric " metric)))))



		(cond ((kosher-extension
			  (!_preserve-only next)
			  a pp)
		      (prog1
			  (state-after-act 
			     next a metric maxspan
			     topnode eff-db
			     mark allow-reencounter
			     th)
			))
		     (t '())))))
      (set-discrepancies new-states)
      (evaluation-sits-flush
	 (<# !_situation new-states))
      new-states))
	        


)

;;; Returns 0 or 1 Sit-search-states, depending on whether the new situation
;;; has been seen before or not.
(deffunc next-search-state - (Lst Sit-search-state)
	                (next-sit - Situation
			 pp - Plan-prefix
			 delta-steps - Fixnum elapsed - Time-int
			 feas-act - Feasible-action-spec
			 adds deletes - (Lst Occasion)
			 prev - Sit-search-state
			 metric - Sexp maxspan - Time-int
			 old-top-node - Effort-spec
			 old-eff-db - Effort-db
			 init-sit - Situation sit-space - Situation-space
			 allow-reencounter - (Either Sitmark (Const t nil))
			 linkid - (~ Symbol) assumptions - (Lst Sexp)
			 mark - Sitmark dom - Domain)
   (!= deletes
       (nconc (repeat :for ((a :in adds))
	       :collect (uniquify-occasion `(not ,(!_propo a))
					   dom))
	      *-*))
   (let ((succ-pp
	  (plan-prefix-grow
	     pp delta-steps elapsed (!_act feas-act) linkid
	     assumptions next-sit init-sit sit-space dom))
	 - Plan-prefix
	 (sit-fresh (or (not (equ (!_mark next-sit) mark))
			(and (eq allow-reencounter true)
			     (< (+ (len (!_path (!_sit pp))) 1)
				(len (!_path next-sit))))))
	 (seen-bf (and (is Sitmark allow-reencounter)
		       (equ (!_mark next-sit)
			    allow-reencounter)))
	 - Boolean)
      (let ((pp-fresh (not (equ (!_mark succ-pp) mark))))
	 (cond (pp-fresh
		(!= (!_mark succ-pp) mark)
		(cond ((= (!_mark next-sit) mark)
		       (sit-reencounter-describe next-sit (!_sit pp)))
		      (t
		       (!= (!_mark next-sit) mark)))
		(cond ((or sit-fresh
			   (not (!_preserve-only prev)))
		       (cond ((and seen-bf
				   (< (length (!_path next-sit))
				      (+ (!_length pp)
					 1)))
			      (dbg-out (> plan-dbg-amt* 1)
				       "State seen again with longer path"
				       :%)))
		(prog1
		   (list
		      (sprout-child-state
			 (prog1
			    (plan-prefix-score
			       succ-pp feas-act metric maxspan
			       dom)
			    )
			 (cond ((!_sequel prev)
				(action-incoherence
				   (!_act feas-act)
				   (car (!_sequel prev))))
			       (t 0))
			 (not fresh)
			 succ-pp 
			 (make Plan-sequel 
			       (!_rtree feas-act)
			       
			 prev
			 (and reuse-rmgraphs* old-top-node)
			 (and reuse-rmgraphs* old-eff-db)
			 dom))
;;;;		   (cond ((or (sit-patho (!_sit pp))
;;;;			      (sit-patho (!_sit succ-pp)))
;;;;			  (dbg-save next-sit pp succ-pp dom)
;;;;			  (signal-problem next-search-state
;;;;			     "Pathology leaving next-search-state"
;;;;			     :% " pred -- " (sit-patho (!_sit pp))
;;;;			     :% " succ -- " (sit-patho (!_sit succ-pp)))))
		   )))
	    (t
	     (dbg-out (> plan-dbg-amt* 1) "State seen before -- discarding")
;;;;	     (dbg-save :run-loud prev next-sit feas-act)
;;;;	     (breakpoint next-search-state
;;;;		next-sit " old hat")
	     !()))))

(deffunc plan-prefix-grow - Plan-prefix
	                  (pp - Plan-prefix 
			   delta-step - (Const 0 1) elapsed - Time-int
			   act - Action
			   linkid - (~ Symbol)
			   assumptions - (Lst Sexp)
			   succ init-sit - Situation
			   sit-space - Situation-space dom - Domain)
   (let ((act-term (act-term-reword (!_term act) elapsed succ dom)))
      (letrec ()
	 (let ((act-seq
		  (<# (\\ (a) (uniquify-action a dom))
		      (nconc (cond (linkid
				    (list `(link ,act-term
						 :output ,linkid)))
				   (t (list act-term)))
			     (<# verification-step assumptions))))
	       - (Lst Action))
	    (repeat :for ((uact :in (cdr act-seq)) - Action
			  (res-pp
			     (place-Plan-prefix
				   (+ (!_length pp) delta-step)
				   (+ (!_duration pp) elapsed)
				   succ
				   (make Plan-prefix-delta
					 (car act-seq) delta-step elapsed
					 (and linkid (list linkid))
					 pp)
				   init-sit sit-space)
			   :then
			      (place-Plan-prefix
				    (!_length res-pp)
				    (!_duration res-pp)
				    succ
				    (make Plan-prefix-delta
					  uact 0 0.0 false res-pp)
				    init-sit sit-space))
			  - Plan-prefix)
	     :result res-pp))
      :where
	 (:def verification-step - Sexp (a - Sexp)
	    `(verify
		,(cond ((car-eq a 'step-value-is)
			(cond (linkid
			       `(= (lv ,linkid) ,(cadr a)))
			      (t
			       (signal-problem plan-prefix-grow
				  "No link for value of step " act-term))))
		       (t a)))))))

;;; If 'act' is a wait for a specific condition that didn't occur,
;;; replace it with a simple 'wait'.   
(deffunc act-term-reword - Sexp (act - Sexp elapsed - Time-int
			       sit - Situation dom - Domain)
   (let-fun ()
      (match-cond act
	 (:? (wait-while ?proc)
	    (let ((pmod (!_processes sit)))
	       (cond ((and pmod
			   (exists (pci :in (!_active pmod))
			      (equal (!_term (!_uterm pci))
				     proc)))
		      (reword-as-wait))
		     (t act))))
	 (:? (wait-non-negative ?term ?proc)
	    (multi-let (((ok val)
			 (opt-eval term sit !() dom)))
	       (cond (ok
		      (cond ((>= val 0)
			     act)
			    (t
			     (reword-as-wait))))
		     (t
		      (out (:to *error-output*)
			 "Unable to evaluate " term " in " sit :%)
		      (reword-as-wait)))))
	 (t
	  act))
    :where
       (:def reword-as-wait - Sexp ()
	   `(wait ,elapsed))))

(specdecl (penalty-fac* 1.1) - Float)

(deffunc plan-prefix-score - Number (pp - Plan-prefix
				     act - Feasible-action-spec
				     metric - Sexp maxspan - Float
				     sit-space - Situation-space dom - Domain)
   (cond ((not metric) 
	  (+ (!_length pp)
	     -1
	     (!_eff act)))
	 (t
	  (dbg-out (> plan-dbg-amt* 1)
	     "Scoring new plan prefix " pp
	     " [duration = " (!_duration pp) "]" :%)	 
	  (multi-let (((step-est time-est final penalty)
		       (reduction-tree-project
			   (!_rtree act) true
			   (!_length pp)
			   (!_duration pp)
			   (- maxspan (!_duration pp))
			   (!_sits pp)
			   sit-space dom)))
	     (cond ((and (!_prequel pp)
			 (not (eq (cadr (!_sits pp))
				  (!_sit (!_prior pp)))))
		    (signal-problem plan-prefix-score
		       "You don't understand plan prefixes!")))
	     (multi-let (((evaluable basic-score)
			  (opt-eval metric final
				    (list (tuple 'total-steps
						 step-est)
					  (tuple 'total-time
						 time-est))
				    dom)
			  ))
		(cond (evaluable
		       (let ((res (* basic-score
				     (expt penalty-fac* penalty))))
			  (dbg-out (> plan-dbg-amt* 1)
			     "Plan prefix score = " res)
			  res))
		      (t
		       (dbg-save :run-loud
				 metric final pp dom step-est time-est)
		       (signal-problem plan-prefix-score
			  "Can't evaluate metric " metric))))))))

(deffunc plan-solution-score - Number (pp - Plan-prefix
				       metric - Sexp dom - Domain)
   (cond ((not metric) 
	  (!_length pp))
	 (t
	  (multi-let (((evaluable score)
		       (opt-eval metric (!_sit pp) 
				 (list (tuple 'total-steps
					      (!_length pp))
				       (tuple 'total-time
					      (!_duration pp)))
				 dom)))
	     (cond (evaluable
		    score)
		   (t
		    (signal-problem plan-soloution-score
		       "Can't evaluate metric " metric)))))))

(deffunc optimistically-project - (Multvals sit - (~ Situation)
					    adds deletes - (Lst Occasion)
					    linkid - (~ Symbol)
					    assumptions - (Lst Sexp))
                    (allow-anomalies - Boolean
		     uact - Action idefn - Internalized-action-defn
		     aid - Varid env - Bdgenv
		     sit - Situation
		     sit-space - Situation-space dom - Domain
		     alter - Boolean)
      (cond ((step-value-occurs-in
		 (!_effect idefn))
	     (let ((nid (new-Varid)))
		(let ((norm (deduce `(normal-step-value ,(!_term uact) ?norm)
				    nid '() sit dom)))
		   (let ((normval
			    (cond ((null norm)
				   (dbg-out optimism-dbg*
					    "No normal step value for "
					    uact :%)
				   '*nonorm)
				  (t
				   (unsafe-varsubst '?norm nid (car norm))))))
		      (cond ((has-uvars normval nid env)
			     (signal-problem norm-replace
				"Normal step value exists but contains"
				" variables: " normval)))
		      (norm-project normval allow-anomalies uact idefn aid env
				    sit sit-space dom alter)))))
	    (t
;;;;	     (cond ((sit-patho sit)
;;;;		    (dbg-save sit uact)
;;;;		    (signal-problem optimistically-project
;;;;		       "Pathology before effect-project")))
	     (multi-let (((succ adds dels assumptions)
			  (effect-project
			     allow-anomalies uact idefn aid env
			     sit sit-space dom alter)))
		(values succ adds dels false assumptions)))))

;;;;(decl () (!= optimistic-projector* #'optimistically-project))

(defvar step-count* 0)

(deffunc norm-project - (Multvals sit - (~ Situation)
				adds deletes - (Lst Occasion)
				linkid - (~ Symbol)
				assumptions - (Lst Sexp))
                    (normval - Sexp allow-anomalies - Boolean act - Action
		     idefn - Internalized-action-defn aid - Varid env - Bdgenv
		     sit - Situation
		     sit-space - Situation-space dom - Domain
		     alter - Boolean)
   (multi-let (((normeffect assumptions linkid)
		(cond ((eq normval '*nonorm)
		       (let ((linkid (new-link-id)))
			  (values
			     (replace-step-value (!_effect idefn)
						 `(lv ,linkid))
			     '()
			     linkid)))
		      (t
		       (values
			  (replace-step-value (!_effect idefn)
					      normval)
			  `((step-value-is ,normval))
			  false)))))
	 (multi-let (((succ adds dels)
		      (do-effect allow-anomalies act normeffect
				 aid env sit sit-space dom)))
	    (cond (linkid
		   (let ((domenv (empty-vartypes dom)))
		      (let ((act-type
			       (recover-type
				  (!_term act)
				  (designated-type '(Step Obj)
						   true domenv)
				  domenv)))
			 (let ((new-link
				  (make Plan-link
				     linkid
				     (not (eq normval '*nonorm))
				     normval
				     sit act))
			       (domenv (opt::domain-place-env dom)))
			    (symbol-declare-locally
			       linkid
			       (opt::make-link-type
				  (opt::steptype-restype
				     act-type domenv)
				  domenv)
			       new-link succ dom))))))
	    (cond (alter
		   (!= succ
		       (succ-record
			  false sit 0.0 1 act succ
			  adds dels 0.0 assumptions sit-space))))
	    (values succ adds dels linkid assumptions))))

(specdecl (linkno* 0) - Fixnum)

(deffunc new-link-id - Symbol ()
   (build-symbol link- (++ linkno*)))

(defproc yang-set-discrepancies - (Lst Sit-search-state)
                         (sl - (Lst Sit-search-state))
   (cond ((not (or (null sl) (null (cdr sl))))
          (let ((best (</ (\\ (sofar next - Sit-search-state)
                             (cond ((score-smaller next sofar)
                                    next)
                                   (t sofar)))
                          (car sl)
                          (cdr sl))))
             (repeat :for ((st :in sl) - Sit-search-state)
                (cond ((score-smaller best st)
                       (!= (!_discrepancy st)
                           (+ *-* 1))))))))
   sl)

(specdecl (incoherence-anomaly* 0)
          - Fixnum)

;;;;(specdecl act* - Action l* - (Lst (Lrcd Action Fixnum)))


;;;;(deffunc compute-score - Float (act-eff min-eff - Integer
;;;;                                planoid - (~ Action-multiset))
;;;;   (scores-weigh act-eff min-eff planoid))

(specdecl (default-goal-penalty* 1.0) - Float)

(specdecl (reckless-deletion-speculation* false) - Boolean)

(specdecl (down-decay* 1.0) - Float
          (down-rule* 'average) - (Const average max))

(deffunc bf-queue-merge - (Lst Sit-search-state)
                        (new old - (Lst Sit-search-state)
                         comparator 
                         - (Fun Boolean (Sit-search-state Sit-search-state) nil))
   (cond ((null new) old)
         ((null old) new)
         ((funcall comparator (car old) (car new))
          (cons (car old) (bf-queue-merge new (cdr old) comparator)))
         (t
          (cons (car new) (bf-queue-merge (cdr new) old comparator)))))

;;;;(deffunc arbitrary-acts - (Lst Action) (pp - Plan-prefix
;;;;                                        th - Domain
;;;;                                        acts - (Lst Feasible-action-spec))
;;;;   (let ((id (new-Varid)))
;;;;      (let ((ml (all-regression-methods '?any th)))
;;;;	 (repeat :for (m :in ml)
;;;;	    (splice
;;;;	       (for (e in (deduce `(and ,@(!_precond m)) id '() (!_sit pp) th))
;;;;                  (splice
;;;;                     (let ((a (uniquify-action (unsafe-varsubst (!_action m) id e)
;;;;                                               th)))
;;;;                        (cond ((not
;;;;				 (exists (fas in acts)
;;;;				    (and (eq (!_act fas) a)
;;;;					 (eq (!_pp fas) pp))))
;;;;                               (list a))
;;;;                              (t '()))))))))))

(defun new-step-id ()
  (build-symbol step- (++ step-count*)))

(deffunc init-pp - Plan-prefix (ss - Situation sit-space - Situation-space)
   (place-Plan-prefix 0 0.0 ss false ss sit-space))

(deffunc prob-space - (Mlv Situation-space Plan-problem)
		    (pp - Obj)
   (let ((pp (coerce-to-plan-problem pp)))
      (let ((sit-space (make Situation-space (!_sit pp))))
	 (bind ((sit-init-completer* (sit-start-processes sit-space)))
	    (situation-initialize (!_sit pp))
	    (!= (!_sit pp)
		(be Initial-situation
		    (uniquify-situation
		       *-* (!_index sit-space))))
	    (values sit-space pp)))))

(deffunc seq-pp - Plan-prefix (seq - (Lst Sexp)
			       init-sit - (~ Situation)
			       sit-space - Situation-space)
   (let* ((init-sit (or init-sit (!_initsit sit-space)))
	  - Situation
	  (pp0 (init-pp init-sit sit-space))
	  - Plan-prefix)
      (bind ((sit-init-completer* (sit-start-processes sit-space)))
	 (!= (!_sit pp0)
	     (be Initial-situation
		 (uniquify-situation *-* (!_index sit-space))))
	 (let ((dom (!_domain init-sit)))
	    (</ (\\ (sofar - Plan-prefix a - Sexp)
		   (let ((act (uniquify-action a dom)))
		      (multiple-value-let (succ adds deletes delta elap
					   linkid assumptions)
					  (contin-discrete-project
					      (!_sit sofar) act
					      sit-space dom)
					  (ignore adds deletes stepid
						  assumptions)
			 (cond ((not succ)
				(dbg-save :run-loud
					  (sit (!_sit sofar)) act
					  sit-space dom)
				(error-break seq-pp :fatal
				   act " is unfeasible in " (!_sit sofar))))
			 (decl (elap - Time-int)
			    (place-Plan-prefix
				  (+ (!_length sofar) delta)
				  (+ (!_duration sofar) elap)
				  succ
				  (make Plan-prefix-delta
					act delta elap
					(and linkid (list linkid))
					sofar)
				  init-sit sit-space)))))
		pp0
		seq)))))

;; Destructively remove all but the first n elements of l.
(deffunc dtake - (Lst Obj) (n - Integer l - (Lst Obj))
   (cond ((< n 1) '())
	 (t
	  (repeat :for ((tl l (cdr tl)))
	   :until (= n 1)
	   :result (!= (cdr tl) '())
	     (!= n (- n 1)))
	  l)))
	      
(defproc plan-print-procs - Void (pp - Plan-prefix)
   (let ((dom (!_domain pp)))
      (letrec ((print-prequel - Float (pre-pp - Plan-prefix act - (~ Action))
		  (let ((printed-time
			   (cond ((not (!_is-initial pre-pp))
				  (print-prequel (!_prior pre-pp)
						 (!_last-act pre-pp)))
				 (t false)))
			- (~ Float)
			(time-here (!_duration pre-pp)))
		     (cond ((or (not printed-time)
				(not (= printed-time time-here)))
			    (out "Time " (:f "~,1f" time-here) :%)))
		     (let ((sit (!_sit pre-pp)))
;;;;		        (out "Traffic between mid-s & mid-d: "
;;;;			     (nth-value 1
;;;;				(opt-eval '(fl-v (traffic mid-s mid-d))
;;;;					  sit !() dom))
;;;;			     :%)
			(let ((stoppers (deduce
					   '(process-transition stopping ?proc)
					   dummy-id* (empty-env) sit dom))
			      (starters (deduce
					   '(process-transition starting ?proc)
					   dummy-id* (empty-env) sit dom)))
			   (cond ((not (null stoppers))
				  (out 3 "Stop: "
				       (<# (\\ (a) (get-var-val-1 'proc a))
					   stoppers)
				       :%)))
			   (cond ((not (null starters))
				  (out 3 "Start: " 
				        (<# (\\ (a) (get-var-val-1 'proc a))
					   starters)
				       :%)))))
		     (cond (act (out 3 "Action: " (!_term act) :%)))
		     time-here)))
	 (print-prequel pp false))))

(defproc plan-print-2 - Plan-prefix (pp - Plan-prefix)
   (repeat :for ((dom (!_domain pp)) - Domain
		 prev-pp - Plan-prefix)
    :until (!_is-initial pp)
    :result prev-pp
       (let ((sit (!_sit pp)))
	  (let ((enders-here
		   (repeat
		      :for ((a :in (deduce
				      '(process-transition
					   stopping
					   (dur-happening ?d))
				      dummy-id* (empty-env)
				      sit dom))
			    - Bdgenv)
		    :collect (debug-varsubst '?d dummy-id* a))))
	    (out (:f "~,2f" (!_duration pp))
		 1 (!_last-act pp)
		 :% 2 enders-here :% :%)))
       (!= prev-pp pp)
       (!= pp (!_prior pp))))

;;; Print in 2002 competition format:
(defproc plan-print - Void (pp - Plan-prefix)
   (let ((last-time (!_duration pp))
	 (dom (!_domain (find-init-situation (!_sit pp)))))
      (letrec ((print-prequel - Void (pre-pp - Plan-prefix
				      enders - (Lst (Lrcd Sexp Float)))
		  ;; enders is a list of duratives that are going
		  ;; to end in the "future" of 'pre-pp'.
;;;;		  (trace-around print-prequel (:if (not (!_is-initial pre-pp)))
;;;;		     (:> "(print-prequel: " (!_last-act pre-pp)
;;;;			 :% 3 enders ")")
		  (cond ((not (!_is-initial pre-pp))
			 (let ((result (!_sit pre-pp)) - Situation
			       (act (!_term (!_last-act pre-pp))) - Sexp
			       (enders-here enders)
			       - (Lst (Lrcd Sexp Float)))
			    (!= enders (list-copy *-*))
			    (repeat
			       :for ((a :in (deduce
					       '(process-transition
						    ?w
						    (dur-happening ?d))
					       dummy-id* (empty-env)
					       result dom))
				     - Bdgenv
				     durative - Sexp)
			       (!= durative
				   (one-value (safe-varsubst '?d dummy-id* a)))
;;;;			       (out "Transition: "
;;;;				    (debug-varsubst
;;;;				       '(process-transition
;;;;						    ?w
;;;;						    (dur-happening ?d))
;;;;				       dummy-id* a) :%)
			       (cond ((eq (get-var-val-1 'w a)
					  'starting)
				      (!= enders
					  (delete-if
					     (\\ (e) (equal (car e) durative))
					     *-*)))
				     (t
				      (!= enders
					  (cons (tuple durative
						       (!_duration pre-pp))
						*-*)))))
			    (print-prequel (!_prior pre-pp) enders)
			    (cond ((eq (car act) 'dur-start)
				   (let ((e (assoc= #'equal
						    (cadr act) enders-here)))
				      (let ((end-time
					       (cond (e (cadr e))
						     (t
						      (dbg-save :run-loud
						         act enders)
						      (signal-problem
							 (cadr act)
							 " never ends"
							 (:proceed
							    !"I'll assume it ~
							      ends when the ~
							      plan does"))
						      last-time))))
					 (out
					    (!_duration pre-pp)
					    " : "
					    (cadr act)
					    " ["
					    (- end-time
					       (!_duration pre-pp))
					    "]"
;;;;					    1 end-time 1 (!_duration pre-pp)
;;;;					    1 e
					    :%)
;;;;					 (dbg-save enders act e)
;;;;					 (breakpoint print-prequel)
					 )))
				  ((not (memq (car act)
					      '(wait-while wait-non-negative)))
				   (out (!_duration pre-pp)
					" : " act :%))))))
;;;;		     (:< (&rest _) "print-prequel exit"))
		  ))
	 (print-prequel
	    pp
	    (repeat :for ((pin :in (!_active (!_processes (!_sit pp))))
			  - Process-instance
			  term - Sexp)
	       (!= term (!_term (!_uterm pin)))
	     :when (eq (car term) 'dur-happening)
	     :collect
	        (tuple (cadr term)
		       last-time))))))

(defproc plan-print-procs - Void (pp - Plan-prefix)
   (let ((dom (!_domain (!_space (!_sit pp)))))
      (letrec ((print-prequel - Float (pre-pp - Plan-prefix act - (~ Action))
		  (let ((printed-time
			   (cond ((not (!_is-initial pre-pp))
				  (print-prequel (!_prior pre-pp)
						 (!_last-act pre-pp)))
				 (t false)))
			- (~ Float)
			(time-here (!_duration pre-pp)))
		     (cond ((or (not printed-time)
				(not (= printed-time time-here)))
			    (out "Time " (:f "~,1f" time-here) :%)))
		     (let ((sit (!_sit pre-pp)))
;;;;		        (out "Traffic between mid-s & mid-d: "
;;;;			     (nth-value 1
;;;;				(opt-eval '(fl-v (traffic mid-s mid-d))
;;;;					  sit !() dom))
;;;;			     :%)
			(let ((stoppers (deduce
					   '(process-transition stopping ?proc)
					   dummy-id* (empty-env) sit dom))
			      (starters (deduce
					   '(process-transition starting ?proc)
					   dummy-id* (empty-env) sit dom)))
			   (cond ((not (null stoppers))
				  (out 3 "Stop: "
				       (<# (\\ (a) (get-var-val-1 'proc a))
					   stoppers)
				       :%)))
			   (cond ((not (null starters))
				  (out 3 "Start: " 
				        (<# (\\ (a) (get-var-val-1 'proc a))
					   starters)
				       :%)))))
		     (cond (act (out 3 "Action: " (!_term act) :%)))
		     time-here)))
	 (print-prequel pp false))))

(defproc plan-print-2 - Plan-prefix (pp - Plan-prefix)
   (repeat :for ((dom (!_domain (!_space (!_sit pp)))) - Domain
		 prev-pp - Plan-prefix)
    :until (!_is-initial pp)
    :result prev-pp
       (let ((sit (!_sit pp)))
	  (let ((enders-here
		   (repeat
		      :for ((a :in (deduce
				      '(process-transition
					   stopping
					   (dur-happening ?d))
				      dummy-id* (empty-env)
				      sit dom))
			    - Bdgenv)
		    :collect (debug-varsubst '?d dummy-id* a))))
	    (out (:f "~,2f" (!_duration pp))
		 1 (!_last-act pp)
		 :% 2 enders-here :% :%)))
       (!= prev-pp pp)
       (!= pp (!_prior pp))))

(deffunc domain-temporal-scope - Float (dom - Domain)
	     (let ((grain (find-domain-bdg 'temporal-scope dom)))
		(cond (grain
		       (be Float (!_(Domain-var val) (!_val grain))))
		      (t
		       (signal-problem domain-temporal-scope
			  "Domain " dom
			  " contains no declaration of temporal-scope"
			  :fatal)))))
