;-*- Mode: Common-lisp; Package: optop; Readtable: ytools; -*-
(in-package :optop)
;;; $Id: types.lisp,v 1.51 2005/12/12 19:09:31 dvm Exp $

(depends-on %module/ nisp opt)

;;;;(depends-on :at-compile-time %ydecl/ objtype objdcl)

;;;;(depends-on %module/ nispobj)

(depends-on :at-run-time %langutils/ symboid)

;;;;  ALSO DEPENDS ON ...? .%ydecl/ compnisp

;;; Most of the stuff in here depends on OPT types
;;;  -- which should already have been loaded in optop.lsy

(eval-when (:slurp-toplevel :compile-toplevel :execute)
   (!= lisp-class-package* (find-package :opt)))

(nisptype::syn-type-former Alt opt::Alt (:system opt))

(lisp-class-as-nisp-type Opt-type nisptype::Type
   (:package :nisptype))

;; Just need this to define Name-repository
(lisp-class-as-nisp-type Type-system
   (:package :nisptype))

(specdecl opt-type-sys* - Type-system)

(defnisptype Typenv :forward)

(lisp-class-as-nisp-type Typed-exp lisplang::Typed-exp
   (:package :lisplang)
   (:access env - Typenv)
   (:access type - Opt-type)
   (:access totbugs - Integer)
   (:initexp (make-inst lisplang::Typed-exp
                :flag 'null
                :type univ-type*)))

(lisp-class-as-nisp-type Wheresym
   (:package :ytools)
   (:expect-nisp-class-descriptor false)
   (:conser *integrable (name - String) (make-Wheresym name))
   (name - String))

(defnisptype Symboid (Alt Symbol Wheresym)
   (conser :inline (name - String) (make-Wheresym name))
   (name - String :inline (b - Symboid) (Symboid-name b)))

(specdecl symboid-get - (nisp::Fun Obj (Symboid Symbol)))

(defnisptype Qvar Obj
  (conser *integrable (sym - Symboid notes - (Lst Obj))
     (nisp::make-Qvar sym notes))
  (sym - Symboid *integrable (qv) (nisp::Qvar-sym qv))
  (= *integrable (x y) (eq (nisp::Qvar-sym x) (nisp::Qvar-sym y)))
  (is *integrable (x) (nisp::is-Qvar x)))

(eval-when (:compile-toplevel :load-toplevel :slurp-toplevel)
   #-pre-chunk (:slurp-filter :slurp-nisp-types)
   (nisp::declare-type-acceptable 'Qvar 'Sexp))

;;; Canonical symbol: either a real one or a Canonized-symbol --
(defnisptype CSymbol Obj
   (is - Boolean *integrable (x)
       (or (is-Symbol x)
           (langutils::is-Canonized-symbol x)))
   (sym - Symbol (cs)
      (cond ((langutils::is-Canonized-symbol cs)
             (langutils::Canonized-symbol-sym cs))
            (t cs))))

(eval-when (:compile-toplevel :load-toplevel :slurp-toplevel)
   #-pre-chunk (:slurp-filter :slurp-nisp-types)
   (nisp::declare-type-acceptable 'Symbol 'CSymbol)
   (nisp::declare-type-acceptable 'CSymbol 'Sexp))

(lisp-class-as-nisp-type Ref
   (:both contents - Obj))

(lisp-class-as-nisp-type Sym-with-type
   (actual - CSymbol)
   (:access type - Opt-type))

(deffunc strip-car - (~ CSymbol) (l - Sexp)
   (cond ((atom l) false)
         (t
          (let ((a (car l)))
             (opt::strip-sym-with-type a)
             (cond ((is CSymbol a) a)
                   (t false))))))

(defnisptype Timestamp Integer)

;;; An S-expression whose type (in Opt) is (Fluent ...).
(defnisptype Fluexp Sexp)

(lisp-class-as-nisp-type Uterm 
   (:conser (x - Sexp) (opt::make-Uterm x))
   (:access term - Sexp)
   (= - Boolean *integrable (occ1 occ2 - Uterm) (eq occ1 occ2)))

(lisp-class-as-nisp-type Occasion
   (:conser (x - Sexp a d - Boolean) (opt::make-Occasion x a d))
   (:include Uterm)
   (:access propo - Sexp *integrable (c - Occasion) (!_term (be Uterm c)))
   (:initexp (opt::make-Occasion '(_) true false))
)

(lisp-class-as-nisp-type Signed-occasion
   (:conser (s - Boolean x - Occasion)
      (opt::make-Signed-occasion s x))
   (:access sign - Boolean)
   (:access occ - Occasion)
:augtype
   (= - Boolean *integrable (occ1 occ2 - Signed-occasion) (eq occ1 occ2))
   (propo - Sexp (x - Signed-occasion)
      (!_propo (!_occ x))))

;;; We need only one kind, so we hide the 'piecefn' arg.
(lisp-class-as-nisp-type Sit-eq-index opt:Eq_setindex
   (:conser ()
      (opt:new-Eq_setindex #'opt:get-situation-contents)))

(lisp-class-as-nisp-type Fact 
   (:access statement - Typed-exp)
   (internalized - Occasion))

(lisp-class-as-nisp-type Requirement
   (:access name - Symbol)
   (:initexp (opt::try-requirement-with-name ':strips false)))

;;;;(cond ((not strips-req*)
;;;;       (signal-problem Requirement
;;;;          ":strips requirement not defined")))
;;;;
;;;;(!= (ydecl::type-feature 'Requirement 'ydecl::initexp)
;;;;    'strips-reqz*)

(defnisptype Domain :forward)

(defnisptype Domain-bdg (Structure (:already-defined nisptype::Vartype))
   (conser (n - CSymbol ty - Opt-type val - Obj dom - Domain)
      (opt::make-Domain-bdg
         :sym n :type ty :val val :domain dom))
   (is - Boolean (x) (nisptype::is-Vartype x))
   (varname - CSymbol (vb - Domain-bdg) (nisptype::Vartype-var vb))
   (:all val - Obj (vb - Domain-bdg) (nisptype::Vartype-val vb))
   (:all type - Opt-type (vb - Domain-bdg) (nisptype::Vartype-type vb)))

;;; Local and global bindings are of this type.  (See "Clarification"
;;; in langutils/namespace.lisp.)
(defnisptype Sym-bdg Domain-bdg
   (:type varname - Symbol))

;;; Confusion Inc.: The underlying classes opt::Domain and Optop-domain
;;; correspond to types Opt-domain and Domain respectively.
;;; Note especially that objects of class opt::Domain are
;;; not (necessarily) of type Domain.

(lisp-class-as-nisp-type Opt-domain opt::Domain 
   (:access name - Sexp)
   (:all generation - Timestamp)
   (:access ancestors - (Lst Opt-domain))
   (parents - (Lst Opt-domain))
   (:access requirements - (Lst Requirement))
   ;;(access facts - (lst Typed-exp))
   (:access local-bdgs - Ref)
                                     ; Obj that can respond to name-lookup
   (:access inherited-bdgs - (~ (Htb Domain-bdg))))

(defnisptype Name-repository (Alt Domain Type-system))

;; This is meaningful only if the domain is local to an Initial-situation
(augtype Opt-domain
   (:access main-parent - Opt-domain (odom - Opt-domain)
      (car (!_parents odom))))

(deffunc all-requirements - (Lst Requirement) (dom - Opt-domain)
   (nodupq (repeat :for (a :in (!_ancestors dom))
              :append (!_requirements a))))

 (lisp-class-as-nisp-type Exptree opt::exptree
   (:expect-nisp-class-descriptor false)
   (:initexp (opt::new-exptree nil nil !() !() nil))
;;;;   (:conser :inline (cell - Obj superior - (~ Exptree)
;;;;                     objects tables - (Lst Obj)
;;;;                     rehash-prospects - Obj)
;;;;      (opt::new-exptree cell superior objects tables rehash-prospects))
)

(lisp-class-as-nisp-type Exp_index opt::exp_index
   (tree - Exptree)
   (:expect-nisp-class-descriptor false)
   (:conser :inline (tree - Exptree rehash - Obj)
      (opt::make-exp_index :tree tree :rehash rehash)))

(defnisptype Indexed-domain
             (nisp::Object (:already-defined opt::Indexed-domain)
                           (:include Opt-domain))
   (is - Boolean (x) (opt::is-Indexed-domain x))
   (:access fill-time - Timestamp *integrable (d - Indexed-domain)
       (opt::Indexed-domain-fill-time d)) 
;   (ACCESS AXIOMS - (~ Exp_index) *INTEGRABLE (D - indexed-domain)
;       (opt::INDEXED-DOMAIN-AXIOMS D))
   (:all actions - (~ Exp_index) *integrable (d - Indexed-domain)
       (opt::Indexed-domain-actions d))
   (:access occasions - (~ Exp_index) *integrable (d - Indexed-domain)
       (opt::Indexed-domain-occasions d))
   (:access namespace - Obj *integrable (d - Indexed-domain)
      (opt::Indexed-domain-namespace d))
   (:access namespace-tab - Obj *integrable (d - Indexed-domain)
      (opt::Indexed-domain-namespace-tab d))
)

;;;(defnisptype Action-defn :forward)

(eval-when (:slurp-toplevel :compile-toplevel :load-toplevel)
    (def-class Optop-domain
               (:options (:kind :object)
                         (:include opt::Indexed-domain))
               (:handler
                  (print (d s)
                     (format s "#<Optop domain ~s>"
                             (opt::Domain-name d))))
	    
           (act-dif-time 0 :type integer)
           (act-dif-tab false)
	   (pred-causers false)
;;;;	   (direct-causers false)
;;;;	   (causal-inferrers false)
;;;;           (achievability false) 
           (causation-effects false)
	   ;; -- Previous slots are for linking predicates to actions
	   ;; and methods for changing its truth values.
	   ;; This one is a Cache linking act names to
	   ;; methods for achieving it --
	   (act-meth-tab false)
           (fact-sit false)
	   (infinite-objects ':compute)
	   ;; -- when computed (by method of 'domain-make-current'
	   ;; in domain.lisp), should be true iff it's possible
	   ;; that there are an infinite number of terms of type
	   ;; Obj.
    )) 

; (defnisptype Achievability-rcd
;     (Structure causal - Boolean ;; if true, achieve-method is an Action-defn
;                achieve-method   ;; if false, it is a way of inferring a 
;                - (Either Action-defn ;; pattern starting with the symbol in
;                          Axiom       ;; question.
;                          Occasion)))

(lisp-class-as-nisp-type Typenv opt::Env
   (:all bdgs - (Lst (Alt Domain-bdg Name-repository)) opt::Env-vartypes))

(defnisptype Internalized-causation-defn
    (Structure function - CSymbol
               term condition effect - Sexp
               type-filter - (Lst Sexp)))

(lisp-class-as-nisp-type Causation-defn
   (:all internalized - (~ Internalized-causation-defn))
   (function - CSymbol)
   (term - Sexp)
   (varbdgs - (Lst Sym-bdg))
;;;;   (param-constraints - (Lst Vartype))
;;;;   (var-constraints - (Lst Vartype))
   (condition - Typed-exp)
   (effect - Typed-exp))

(defnisptype Internalized-process-defn
       (Structure (:include Internalized-causation-defn)
                  sorted-condition   
                    - (Lst (Lstructure () discrete - (Lst Sexp)
                                          continuous - (Lst (Lst Sexp))))
                  ;; For each !_discrete conjunction, !_continuous
                  ;; is a dnf statement 
                  start-effect stop-effect - Sexp))

(lisp-class-as-nisp-type Process-defn
   (:include Causation-defn)
   (start-effect - (~ Typed-exp))
   (stop-effect - (~ Typed-exp))
   (:initexp (opt::make-Process-defn
                 :start-effect null-Typed-exp*
                 :stop-effect null-Typed-exp*
                 :term '(null)
                 :varbdgs !()))
 :augtype
   (:type internalized - (~ Internalized-process-defn)))

;;;;(lisp-class-as-nisp-type Process-defn
;;;;   (function - CSymbol)
;;;;   (internalized - (~ Internalized-process-defn))
;;;;   (term - Sexp)
;;;;   (varbdgs - (Lst nisptype::Vartype))
;;;;   (condition - Typed-exp)
;;;;   (effect - Typed-exp))

(defnisptype Internalized-action-defn Internalized-causation-defn
   (precond - Sexp *integrable (iad - Internalized-action-defn)
      (!_condition iad)))

(lisp-class-as-nisp-type Action-defn
   (:include Causation-defn)
   (:all precond - Typed-exp opt::Action-defn-condition)
   (:access has-methods - Boolean opt::Action-defn-has-methods)
 :augtype
   (:type internalized - (~ Internalized-action-defn)))
   
(defnisptype Internalized-method-defn
               (Structure (:include Internalized-causation-defn)
                          maint - Sexp
			  expansion - Sexp))

(lisp-class-as-nisp-type Method-defn
   (:include Causation-defn)
   (:access name - (~ Symbol))
   (:access maint - Typed-exp)
   (:access expansion - Typed-exp)
 :augtype
   (:type internalized - (~ Internalized-method-defn)))

(defnisptype Internalized-durative-defn
      (Structure    ;;(:include Internalized-causation-defn)
          duration - Sexp
          process-view - Internalized-process-defn
          start-view - Internalized-action-defn
          stop-view - (~ Internalized-action-defn)))

;;;;(lisp-clss-as-nisp-type Duration-constraint
;;;;   (:all op - (const = <= >=))
;;;;   (:all term - Typed-exp))

(lisp-class-as-nisp-type Durative-action-defn
   (:include Causation-defn)
   (duration - Typed-exp)
 :augtype
   (:type internalized - (~ Internalized-durative-defn)))

(lisp-class-as-nisp-type Rule-group
   (rules - (Lst (Alt Causation-defn Fact)))) ; Occasion? Method-defn?

(defnisptype Domain (nisp::Object (:already-defined Optop-domain)
                                  (:include Indexed-domain))
   (is - Boolean (x) (is-Optop-domain x))
   ; The time when act-dif-tab and achievability were computed
   ; (actually, when they began to be computed):
   (:all act-dif-time - Integer *integrable (d - Domain)
        (Optop-domain-act-dif-time d))
   (:all act-dif-tab - (~ Exp_index) *integrable (d - Domain)
        (Optop-domain-act-dif-tab d))
   ;; (href (!_causers d) p):
   ;; First element is Causations that can change truth value of
   ;; atomic formulas with pred 'p'; second is rules that can
   ;; be used to infer that an atomic formula with pred 'p' is true.
   (:all pred-causers - (~ (Htb (Lrcd (Alt (Lst Causation-defn)
					  (Const :unknown))
				     (Alt (Lst Occasion)
					  (Const :unknown)))))
      :integrable (d - Domain)
          (Optop-domain-pred-causers d))
	
;;;;   (:all direct-causers - (Lst (Lrcd CSymbol (Lst Causation-defn)))
;;;;        :integrable (d - Domain)
;;;;           (Optop-domain-direct-causers d))
;;;;   (:all causal-inferrers - (Lst (Lrcd CSymbol Occasion))
;;;;	:integrable (d - Domain)
;;;;           (Optop-domain-causal-infferers d))
;;;;   (:all achievability - (Lst (Lrcd Symbol (Lst (Alt Causation-defn))))
;;;;        *integrable (d - Domain)
;;;;           (Optop-domain-immed-achievability d))

   ;; Used to record various things.  For a Causation-function,
   ;; entry is all preds that can be affected by it.
   ;; For a Back-chainer, entry is either :non-causal, if the rule
   ;; can't be used in an inference chain from a Causation-function
   ;; to any conclusion; or a list of all the preds it can be used
   ;; to infer.
   (:all causation-effects - (~ (Htb (Lst CSymbol)))
        *integrable (d - Domain) (Optop-domain-causation-effects d))
   (:all env - (~ Typenv) *integrable
        (dom - Domain) (Optop-domain-env dom))
   (:all infinite-objects - (Alt (Const :compute) Boolean)
	 *integrable (dom - Domain) (Optop-domain-infinite-objects dom))
)
   
(lisp-class-as-nisp-type Opt-constant opt::Constant
   (:access name - CSymbol)
   (:access type - Opt-type))

(lisp-class-as-nisp-type Functional-constant
   (:include Opt-constant)
   (:access callable - Boolean)
   (:access type-inverter - Obj))

(lisp-class-as-nisp-type Predicate 
   (:include Functional-constant))

(lisp-class-as-nisp-type Causation-function-no-arg-types
                         opt::Causation-function
   (:include Functional-constant)
   ;;(ACCESS NAME - symbol)
   (:access defn - Causation-defn))

(defnisptype Causation-function Causation-function-no-arg-types
   (argtypes - Opt-type (af - Causation-function)
      (must-have-argtype (!_type af))))

(deffunc must-have-argtype - Opt-type (afty - Opt-type)
         (cond ((nisptype::is-funtype afty)
                (nisptype::type-find-feature
                   afty 'nisptype::argtype
                   (nisptype::empty-vartypes opt-type-sys*)))
               (t
                (signal-problem must-have-argtype
                   "Expected a function constant, got " afty
                   :fatal))))
                         
;(lisp-class-as-nisp-type Axiom
;   (vars - (Lst symbol))
;   (antecedent - Typed-exp)
;   (consequent - Typed-exp))

(lisp-class-as-nisp-type Back-chainer opt::Back-chainer
   (:package :opt)
   (:include Occasion)
   (consequent - Sexp)
   (antecedent - Sexp)
   (stratified - Boolean))
;   (generation - Timestamp)

(lisp-class-as-nisp-type Prelink-list
   (:all rules - (Lst Back-chainer)))

(defnisptype Linkspot (Lstructure &flag prelinked
                                  prop - Sexp
                                  rules - Prelink-list))

(lisp-class-as-nisp-type Domain-var 
   (name - CSymbol)
   (val - Obj)
   (:access type - Opt-type))

(defnisptype Varid Integer)

(defnisptype Expclo (Structure () 
                       skel - Sexp
                       id - (~ Varid))
   (equ - Boolean *integrable (e1 e2 - Expclo)
      (and (equal (!_skel e1) (!_skel e2))
           (eql (!_id e1) (!_id e2)))))

(defnisptype Varbdg
    (Lstructure () varname - Symboid
                   id - Varid
                   val - Expclo))

; (lisp-class-as-nisp-type Varbdg
;    (:expect-nisp-class-descriptor false)
;    (CONSER (N I V) (opt::MAKE-VARBDG N I V))
;    (VARNAME - symbol)
;    (ID - varid)
;    (VAL - expclo))

(defnisptype Bdgenv (Lst Varbdg)
   (conser *integrable (l - (Lst Varbdg)) l)
   (bdgs - (Lst Varbdg) *integrable (e - Bdgenv) e))

(specdecl optimist-fn* - (~ Symbol)
          (optimism-dbg* true) - Boolean)

(!= optimist-fn* false)
;;;;(!= optimist-fn* 'step-value)

(lisp-class-as-nisp-type Optimist-var)
   
(lisp-class-as-nisp-type Optimism-rcd
   (polarity - Boolean)
   (opt-pat - Sexp)
   (other-pat - Sexp)
   (id1 - Varid)
   (id2 - Varid))

(declare-infinite-types opt
                        opt:Number opt:Float opt:Rational opt:Integer
                        opt:Ratio opt:Char opt:Symbol opt:String opt:Obj)

(opt::def-opt-form-handler declare-infinite-types opt::top-level (dcl-form)
   (eval dcl-form)
   dcl-form)

(defnisptype Action Uterm
   ;;(conser - Action *integrable (e - Sexp) (be Action e))
   ;;(= - Boolean *integrable (ac1 ac2 - Action) (eq ac1 ac2))
   (act - Sexp *integrable (ac - Action) (!_term ac)))

(specdecl (dummy-id1* (new-Varid))
          (dummy-id2* (new-Varid))
          - Varid)

(deffunc uniquify-action - Action (a - Sexp context - Domain)
   (let ((index (domain-place-action-index context)))
      (repeat :for ((cand :in (exp-fetch a index true))
                    - Action
                    prev - Sexp)
       :result (let ((new (make Uterm a)))
                  (exp-ob-index new index t)
                  (be Action new))
         (!= prev (!_term cand))
       :until (variants prev a dummy-id1* dummy-id2* '())
       :result cand)))

(defproc domain-place-action-index - Exp_index (idom - Domain)
    (let ((ai (!_actions idom)))
       (cond ((not ai)
              (!= ai (exp-index-init #'opt::Occasion-prop))
              (!= (!_actions idom) ai)))
       ai))

(specdecl (sitmark* 0) - Fixnum)

(defconstant +null-sitmark+ 0)
#-(or cmu sbcl)
(specdecl +null-sitmark+ - Fixnum)
#+(or cmu sbcl)
(nisp::typedecl '+null-sitmark+ 'Fixnum '*noinit)

; Mark situations with small numbers
(defnisptype Sitmark Integer
   (conser () (!= sitmark* (+ *-* 1)) sitmark*)
   (equ - Boolean *integrable (m1 m2 - Sitmark)
      (= m1 m2)))

;;; An S-expression, but with some of the leaves Uterms and CSymbols
(defnisptype Uterm-exp
    (Alt CSymbol Number String Uterm
         (Lrcd CSymbol . (Lst Uterm-exp))))

(defnisptype Time-int Float)

(eval-when (:compile-toplevel :load-toplevel :slurp-toplevel)
   #-pre-chunk (:slurp-filter :slurp-nisp-types)
   (nisp::declare-type-acceptable 'Time-int 'Float))

(deffunc float->time-int - Time-int (r - Float) (be * r))

(defnisptype Uterm-ineq (Lrcd (Const < =< >= >) Uterm-exp Uterm-exp))

;;; Nisp is not smart enough to figure out that 
(eval-when (:compile-toplevel :load-toplevel :slurp-toplevel)
   #-pre-chunk (:slurp-filter :slurp-nisp-types)
   (nisp::declare-type-acceptable 'Uterm-ineq 'Uterm-exp))

;;; Records a process term plus one disjunct of its definition's
;;; condition, divided into discrete and continuous conjuncts, and
;;; the corresponding effect.
(defnisptype Process-instance
       (Structure
          uterm - Uterm
          discrete - (Lst Sexp)
          continuous - (Lst (Lst Uterm-ineq)) ; in dnf
          start-effect effect stop-effect - Sexp
        (:handler
            (print (pri - Process-instance srm - Stream)
               (out (:to srm) "#{Process-instance " (!_uterm pri) "}"))))
  (term - Sexp (pin - Process-instance) (!_term (!_uterm pin))))

(defnisptype Triggered-effect :forward)

;;; The !_timefcn gives the predicted time evolution of !_fluent.
;;; The Time-int is the time from the situation this model occurs in.  
;;; That is, 0 = the time of that situation.
(defnisptype Fluent-change-model
   (Structure fluent - Uterm-exp
              timefcn - (Fun Float (Time-int))
              symfcn - Uterm-exp ;; symbolic version of the timefcn
      (:handler
         (print (fcm - Fluent-change-model srm - Stream)
            (with fcm
               (fluent-model-print srm !>fluent !>symfcn !()))))))

;;; When the 'fluent' becomes non-negative, the effects occur (as if
;;; they were effects of an ordinary action); a new Situation may have to be
;;; created.
(defnisptype Ineq-trigger
    (Structure (:include Fluent-change-model)
               effects - (Lst Triggered-effect)
       (:handler
           (print (trig - Ineq-trigger srm - Stream)
              (with trig
                 (fluent-model-print srm !>fluent !>symfcn !>effects))))))

(defproc fluent-model-print - Void (srm - Stream flu symfcn - Uterm-exp
                                    effects - (Lst Triggered-effect))
   (out (:to srm)
      (:pp-block (:pre "#{")
         (:q ((null effects)
              "Fluent-model ")
             (t
              "Fluent-model-trigger "))
         flu 1
         (:pp-nl :linear)
         " = "  symfcn 1
         (:q ((not (null effects))
              1 (:pp-nl :linear) " >=0 -> "
              effects))
         (:suf "}"))))

;;; If effect is an Sexp, it's a proposition
;;; that will becomes true.
;;; Special case (process-transition [starting | stopping] <name>).
;;; If effect is an Ineq-trigger, then we wait for *its* fluent to 
;;; become non-negative (which may already have happened).
(defnisptype Triggered-effect (Alt ;;;;(Const :process-change)
                                   Sexp
                                   Ineq-trigger))

;;;;(defnisptype Continuous-change-model (Lst Ineq-trigger)
;;;;   (conser (l - (Lst Ineq-trigger)) (be * l))
;;;;   (triggers - (Lst Ineq-trigger) *integrable (ccm - Continuous-change-model)
;;;;      ccm))
;;;;
;;; !_triggers has the same length as the list of Floats returned
;;; by the !_fcn.
;;;;(defnisptype Continuous-change-model
;;;;     (Structure triggers - (Lst Ineq-trigger)
;;;;                fcn - (Fun (Lst Float) (Time-int))))

(defnisptype Process-complex
   (Structure active dormant - (Lst Process-instance)
              trigger-time - (Alt (Const *unknown *never)
                                  (Lrcd Time-int   
                                        (Lst Sexp) ; effects
                                        Float))    ; new epsilon
              ;;; !_model explains how all the atomic fluents
              ;;; change.
              ;;; !_triggers summarizes the inequalities
              ;;; such that when they become true the process
              ;;; complex will (probably) change.
              model - (Lst Fluent-change-model)
              triggers - (Lst Ineq-trigger)
              kosher - Boolean
              ;; -- if false, the complex is not well defined, but this is
              ;; a half-assed version of reality.
      (:handler
          (print (pc - Process-complex srm - Stream)
             (out (:to srm)
                (:pp-block (:pre "#{")
                   "Process-complex [-> " (!_trigger-time pc) "] "
                   (:pp-nl :linear)
                   (:pp-ind :block 3)
                   "Active: " (!_active pc)
                   1 (:pp-nl :linear)
                   "[Dormant: " (!_dormant pc) "] "
                   (:pp-nl :linear)
                   "Changing: " (<# !_fluent (!_model pc))
                   1 (:pp-nl :linear)
                   "Triggers: " (!_triggers pc)
                   (:suf "}")))))))

(defnisptype Sitsuccspec :forward)

;;; This is the time "up to" a Situation, i.e., the time it 
;;; occurs in some frame of reference.  The !_elapsed is
;;; the time that passes until it occurs.
(defnisptype Sit-time
   (Structure elapsed - Time-int ;; in seconds since ...
              nudge - Integer  ;; in dt's (possibly negative)
      (:handler
          (print (st - Sit-time srm - Stream)
             (let ((nudge (!_nudge st)))
                (out (:to srm)
                     "#<" (:f "~6,3F" (!_elapsed st))
                          (:q ((not (= nudge 0))
                               (:q ((> nudge 0) "+" ))
                               nudge "dt"))
                          ">"))))))
                                    
(deffunc sit-times-compare - Integer (st1 st2 - Sit-time)
   (cond ((= (!_elapsed st1) (!_elapsed st2))
          (cond ((= (!_nudge st1) (!_nudge st2))
                 0)
                ((< (!_nudge st1) (!_nudge st2))
                 -1)
                (t 1)))
         ((< (!_elapsed st1) (!_elapsed st2))
          -1)
         (t 1)))

(defnisptype Sitmore
   (Structure ()
;;;;      date - Sit-time NO! time depends on plan prefix *to* this sit
      processes - (~ Process-complex)
      succs - (Lst Sitsuccspec)
      process-defns - (Lst (Alt Process-defn Durative-action-defn))
;;;;      reality - (Const :unknown :search-space :evaluation-space)
      )
  (conser () (make ^^ false
                      !(Sitsuccspec)
                      !((Alt Process-defn Durative-action-defn))
;;;;                      ':unknown
                      )))

;;;;(defnisptype Init-sitmore
;;;;    (Structure ()
;;;;       processes - (~ Process-complex)
;;;;       succs - (Lst Sitsuccspec)
;;;;       process-defns - (Lst (Alt Process-defn Durative-action-defn)))
;;;;  (conser () (make ^^ false !(Sitsuccspec) !(Process-defn))))

(lisp-class-as-nisp-type Situation
   (:all path - (Lst Action))
   (contents - (Either (Const opt::*uncomputed) (Lst Occasion)))
   (:all more - (~ Sitmore))
   (:all mark - Sitmark)
   (:both frozen - Boolean)
   (:all diff-from-init - Exp_index)
 :augtype
;;;;   (date - Sit-time (sit - Situation)
;;;;      (!_date (ensure-sitmore sit)))
   (:all processes - (~ Process-complex) (sit - Situation)
      (!_processes (ensure-sitmore sit)))
;;;;   (:all reality - (Const :unknown :search-space :evaluation-space)
;;;;               (sit - Situation)
;;;;      (!_reality (ensure-sitmore sit)))
   )

(lisp-class-as-nisp-type Initial-situation 
   (:include Situation)
   (:access name - Symbol)
   (:access domain - Domain)
   (:access is - Boolean opt::is-Initial-situation)
   (:access fact-index - Obj)
   (:access init-index - Obj)
;;;;   (:access sit-index - Obj)
 :augtype
   (:all process-defns - (Lst (Alt Process-defn Durative-action-defn))
            (s - Initial-situation)
      (!_process-defns (ensure-sitmore s))))

(lisp-class-as-nisp-type Subsequent-situation 
   (:include Situation)
   (:access init - Initial-situation))

(deffunc sit-contents - (Lst Occasion) (s - Situation)
   (let ((cl (exp_index-contents
                (!_init-index (cond ((is Initial-situation s) s)
                                    (t (!_(Subsequent-situation init) s))))
                false))
         - (Lst Occasion))
      (nconc
         (repeat :for ((occ :in cl))
          :when (not (variant-fetch (!_propo occ) ':deleted (!_diff-from-init s)))
          :collect occ)
         (repeat :for ((socc :in (be (Lst Signed-occasion)
                                     (exp_index-contents (!_diff-from-init s)
                                                         false))))
          :when (!_sign socc)
          :collect (!_occ socc)))))

(deffunc ensure-sitmore - Sitmore (sit - Situation)
   (or (!_more sit)
       (let ((sm (make Sitmore)))
          (!= (!_more sit) sm)
;;;;          (cond ((is Initial-situation sit)
;;;;                 (!= (!_reality sm) ':search-space)))
          sm)))

(defnisptype Sitsuccspec
           (Structure 
              forcings - Integer ;; > 0  if act is not actually feasible
              elapsed - (Alt Time-int (Const *never))
              steps   - Integer
              act - (Alt Action (Const wait transition actual-init))
              ;; If false, then act is not feasible
              ;; But for forced acts, can be true  
              ;; even though act is infeasible. --
              sit - (~ Situation)
              outlink - (~ Symbol)
                      ;; -- The Symbol that names the link receiving the
                      ;; output of this step
                      ;; if it gets incorporated
                      ;; into a plan.
                      ;; May be false if link to be named later, or if
                      ;; output is not used by anyone
              adds deletes - (Lst Occasion)
              values  - (Lst Obj)
              ;; -- Empty list if value Void; singleton if one value.
              open-subgoals - (Lst (Lst Sexp))
              ;; -- dnf of subgoal that makes !_act infeasible (when it is)
              frac - Float
              ;; -- If act involves a term becoming non-negative, this
              ;; is what fraction of the ground was covered.  0.0 if
              ;; no term involved.
                    ;;;;opt - Boolean
              assumptions - (Lst Sexp)
              adjustments - (Lst (Lrcd Occasion Situation)))

  (access forced - Boolean *integrable (ss - Sitsuccspec) (> (!_forcings ss) 0))
  )

;;;;(defnisptype Situation-space Initial-situation
;;;;   (:access init - Initial-situation *integrable (ss - Situation-space)
;;;;        ss))

(lisp-class-as-nisp-type Uninitialized-plan-problem opt::Problem
   (:access name - Symbol)
;;;;   (:all space - (~ Initial-situation) opt::Problem-sit)
   (:all sit - (~ Initial-situation))
   (:all goal - Sexp)
   (:all expansion - Sexp)
   (:all length - (~ Integer))) 

(lisp-class-as-nisp-type Plan-problem opt::Initialized-problem
   (:include Uninitialized-plan-problem)
   (:access goal - Sexp opt::Initialized-problem-expanded-goal)
   (:access metric - (~ (Lrcd (Const minimize maximize) Typed-exp))))

(augtype Plan-problem
   (domain - Domain (pp - Plan-problem)
      (!_domain (!_sit pp))))

(defnisptype Regress-method (Structure ()
                                       zone - (~ Initial-situation)
                                       good - Boolean
                                       change - (Const :add :del)
                                       ;; free vars wrt goal-id --
                                       goal - Sexp
                                       causer - (~ (Alt Occasion
                                                        Internalized-causation-defn))
                                       inductive - Boolean
                                       ;; See fluregress for what these are --
                                       quant - (~ Sexp) 
                                       goal-id precond-id - Varid
                                       bdgs - Bdgenv
;;; bdgs specify how variables in Regress-spec relate to variables
;;; in !_precond.                                       
                                       ;; free vars wrt precond-id --
                                       action - Sexp
                                       precond - (Lst Sexp)
                                (:handler
                                   (print (rm - Regress-method
                                           srm - Stream)
                                      (with rm
                                         (let ((b (cond ((not !>good)
                                                         "Bad ")
                                                        (t ""))))
                                            (out (:to srm)
                                               (:pp-block (:pre "#<")
                                                 (:a b)
                                                 "Regression method "
                                                 !>change 1
                                                 !>goal
                                                 (:q (!>quant
                                                      1
                                                      (:pp-nl :linear)
                                                      !>quant)
                                                     (!>inductive
                                                      1 '~ "? "))
                                                 1
                                                 (:pp-nl :linear)
                                                 !>action
                                                 (:suf ">"))))))))
  (conser (change - (Const :add :del)
           goal - Sexp
	   causer - (~ (Alt Occasion
			    Internalized-causation-defn))
           inductive - Boolean quant - (~ Sexp)
           goal-id precond-id - Varid bdgs - Bdgenv
           action - Sexp precond - (Lst Sexp))
     (make ^^ false true change goal causer inductive quant 
              goal-id precond-id bdgs action precond))
  (causal - Boolean :integrable (rm - Regress-method)
     (not (is Occasion (!_causer rm))))
  (preservation - Boolean :integrable (rm - Regress-method)
     (eq (!_change rm) ':del)))

(defnisptype Regress-spec (Structure literal - Sexp
                                     methods - (Lst Regress-method)))

(deffunc regress-spec-literal - Sexp (rs - Regress-spec)
   (!_literal rs))

(specdecl uvar-topcoord* - Obj) ;; ?? Is this still used?

(deffunc make-act-dif-tab - Exp_index ()
   (exp-index-init #'regress-spec-literal))

;;;;(deffunc make-act-eff-tab - Exp_index ()
;;;;   (exp-index-init #'Progress-spec-action))

;;; A Link(-name) is just a uniquified term.
;;; Chosen so that execution of same plan in same state will always
;;; create the same Link-name for a link.
(defnisptype Link-name Uterm)
;;; Eventually will be changed to
;;;;(defnisptype Link-name Uterm
;;;;   (conser (type - Opt-type id - Symbol args - (Lst Uterm-exp) env - Typenv)
;;;;      (opt::make-Skolem-term type id args argtypes env))
;;;;   (:access id - Symbol *integrable (ln - Link-name) (opt::Skolem-term-id ln))
;;;;   (:access num - Integer *integrable (ln - Link-name) (opt::Skolem-term-num ln))
;;;;   (:access args - (Lst Uterm-exp) *integrable
;;;;          (ln - Link-name) (opt::Skolem-term-argsum ln)))

(defnisptype Plan-link (Structure linkid - Symbol
                                  has-expected-val - Boolean
                                  expected-val - Obj
                                  sit - (~ Situation)
                                  setter - (~ Action)
         ;; Most of those fields are obscure and mostly unused
			  (:handler
			      (print (lnk - Plan-link srm - Stream)
				 (out (:to srm) "#-<" (!_linkid lnk) ">-")))))

(deffunc new-Plan-link - Plan-link (name - Symbol)
   (make Plan-link name false nil false false))

(deffunc gen-Plan-link - Plan-link (name - Symbol)
   (make Plan-link (gen-var name) false nil false false))

;;;;(defnisptype Plan-prefix-delta :forward)
(defnisptype Plan-prefix :forward)

(defnisptype Plan-prefix-delta
   (Structure last-act - Action
              delta-steps - (Const 0 1)
              delta-time - Time-int
              links - (Lst Symbol)
              prior - Plan-prefix))

(defnisptype Plan-prefix
     (Structure 
         length - Fixnum
         duration - Time-int
         sit - Situation
	 mark - Sitmark
         prequel - (~ Plan-prefix-delta)
       (:handler
           (print (pp - Plan-prefix srm - Stream)
              (out (:to srm)
                   "#<Plan prefix [" (!_length pp) ", " (!_duration pp) "] "
;;;;                   (:q ((not (null (!_links pp)))
;;;;                       "[Links] "))
                   (:q ((not (!_prequel pp)) "initial")
                      (t "... " (!_last-act pp)))
                   ">"))))
                      
  (conser (length - Fixnum duration - Time-int sit - Situation
	   prequel - (~ Plan-prefix-delta))
     (make ^^ length duration sit +null-sitmark+ prequel))

  (is-initial - Boolean *integrable (pp - Plan-prefix)
      (not (!_prequel pp)))

  (last-act - Action *integrable (pp - Plan-prefix)
     (let ((preq (!_prequel pp)))
        (!_last-act preq)))

  (domain - Domain (pp - Plan-prefix)
     (!_domain (!_sit pp)))

  (prior - Plan-prefix *integrable (pp - Plan-prefix)
     (!_prior (!_prequel pp)))

  ; These things are consed just once, then re-encountered through
  ; CDRing, so they're always EQ.
  (= - Boolean *integrable (pp1 pp2 - Plan-prefix)
     (eq pp1 pp2))
  (acts - (Lst Action) (pp - Plan-prefix)
     ; in reverse order
     (cond ((!_is-initial pp) '())
           (t (cons (!_last-act pp) (!_acts (!_prior pp))))))

  (sits - (Lst Situation) (pp - Plan-prefix)
     (cond ((!_is-initial pp) (list (!_sit pp)))
           (t (cons (!_sit pp) (!_sits (!_prior pp)))))))

;;; Indexing structure for uniquifying Plan-prefixes.
(defnisptype Plan-prefix-tree 
      (Structure pp - (~ Plan-prefix)
		 ;; -- The one that ends here, if any
		 succs - (Lst (Lrcd Action Plan-prefix-tree))
		 ;; -- Tree containing those for which this one
		 ;; is a prefix
		 ))

(defnisptype Situation-space
    (Structure initsit - Initial-situation
	       succ-tab - (Htb (Lst Sitsuccspec))
	       index - Sit-eq-index
	       ;; For a given initial situation, this stores
	       ;; all the Plan-prefixes reachable from it, for
	       ;; uniquification. --
	       pp-tree - (Lst (Lrcd Situation Plan-prefix-tree))
       (:handler
	(print (ss - Situation-space srm - Stream)
	   (out (:to srm) "#<Situation space for " (!_initsit ss) ">"))))
   (conser (initsit - Initial-situation)
      (make ^^ initsit
	       (make-eq-hash-table)
	       (make Sit-eq-index)
	       !())))

(augtype Situation
   (:access succs - (Lst Sitsuccspec) :integrable
	    (sit - Situation space - Situation-space)
      (href (!_succ-tab space) sit !()))

   (:set succs - (Lst Sitsuccspec) :integrable
	 (sit - Situation space - Situation-space sl - (Lst Sitsuccspec))
      (!= (href (!_succ-tab space) sit)
	  sl))

   (domain - Domain (s - Situation)
      (!_domain (find-init-situation s))))

;;;;   (space - Situation-space (s - Situation)
;;;;      (be Situation-space (find-init-situation s))))


(specdecl dummy-id* - Varid)

(specdecl qvar-topcoord* - Obj)

(yt::subr-synonym uniquify-term uniquify-occasion)

(eval-when (:compile-toplevel :load-toplevel :slurp-toplevel)
 (specdecl (ineq-rels* '(=< < > >= =~ /=~)) - (Lst Symbol))
)

;;; (=~ x y) means x and y are equal to within some floating-point
;;; approximation.
(defnisptype Ineq-rel #.`(Const ,@ineq-rels*))

; Type declarations for various builtin functions
(specdecl opt::Type-desig - (Fun Sexp Opt-type)
          univ-type* num-type* prop-type* - Opt-type
;;;;          (float-type* (designated-type 'Float true opt::global-opt-env*))
;;;;                - Opt-type
          global-opt-env* - Typenv
          nisptype::simple-funtype-argtypes
              - (Fun (Mlv Boolean (Lst Opt-type))
                     (Opt-type) false)
          term-check-and-tidy
          term-check - (Fun Typed-exp (Sexp Opt-type Obj Typenv))
          recover-type - (Fun Opt-type (Sexp Opt-type Typenv))
          unify - (Fun (Mlv Boolean (~ Bdgenv))
                       (Sexp Sexp Varid Varid Bdgenv)
                       false)
          var-unify - (Fun (Mlv Boolean (~ Bdgenv))
                           (Symboid Obj Varid Varid Bdgenv) false)
          uvar-lookup - (Fun (~ Varbdg) (Symboid Varid Bdgenv))
          env-elim safe-varsubst
            - (Fun (Mlv Sexp Bdgenv) (Sexp Varid Bdgenv) false)
          unsafe-varsubst - (Fun Sexp (Sexp Varid Bdgenv) false)
          varsubst-to-constant - (Fun Sexp (Sexp Varid Bdgenv) false)
          varid-convert - (Fun (Mlv Sexp Bdgenv) (Sexp Varid Varid Bdgenv))
          empty-env - (Fun Bdgenv () false)
          new-Varid - (Fun Varid ())
          conn-flatten - (Fun (Lst Sexp) (Symbol Sexp) false)
          conn-join - (Fun Sexp (Symbol Sexp Sexp) false)
          n-conn-join - (Fun Sexp (Symbol . Sexp) false)
          opt::expression-car-cdr - (Fun (Mlv Boolean Sexp Sexp Bdgenv)
                                         (Sexp Varid Bdgenv Boolean))
          ;EXPTREE-CONTENTS - (FUN (LST obj) (Exptree Boolean))
	  find-domain-in-vartypes - (Fun (~ Domain) (Typenv) false)
          place-domain-bdg - (Fun Domain-bdg (CSymbol Domain) false)
          find-domain-bdg - (Fun (~ Domain-bdg) (CSymbol Domain) false)
          find-domain-bdg-val - (Fun (~ Obj) (CSymbol Domain) false)
          var-lookup - (Fun (~ Domain-bdg) (CSymbol Typenv))
          get-global-opt-sym - (Fun (Mlv Obj Boolean Opt-type)
                                    (Symbol) false)
	  domain-declares-requirement - (Fun Boolean (Domain Symbol) false)
          set-global-opt-sym - (Fun Void (Symbol Obj Opt-type) true)
          domain-get-local-bdgs - (Fun (Lst (Alt Domain-bdg Name-repository))
                                       (Opt-domain))
          all-rule-groups - (Fun (Lst Rule-group) (Domain))
          next-generation - (Fun Integer () true)
          opt::domain-place-env - (Fun Typenv (Domain))
          reset-ancestors - (Fun Void (Domain) true)
          raw-uvars - (Fun (Lst Symboid) (Sexp) false)
          domain-place-situation - (Fun Initial-situation (Domain Symbol))
          domain-place-fact-sit -  (Fun Initial-situation (Domain Boolean))
          type-predify - (Fun Sexp (Opt-type Sexp Obj) false)
          internalize - (Fun (Mlv Sexp (Lst Linkspot))
                             (Typed-exp
                              (Lst Symbol)
                              (Const :antecedent :consequent nil)))
          opt::add-type-filter-conj - (Fun Typed-exp
                                           (Typed-exp (Lst Sym-bdg)))
          uniquify-term - (Fun Uterm (Sexp Domain))
          uniquify-occasion - (Fun Occasion (Sexp Domain))
	  uniquify-occasion-with-vars - (Fun Occasion (Sexp Domain))
          exp-index-init - (Fun Exp_index ((Fun Sexp (Obj) false)))
          exp-fetch - (Fun Obj (Sexp Exp_index Boolean))
          exp-ob-index - (Fun Void (Obj Exp_index Boolean) true)
          exp_index-contents - (Fun (Lst Obj) (Exp_index Boolean) false)
          find-init-situation - (Fun Initial-situation (Situation) false)
	  opt:get-situation-contents - (Fun (Lst Signed-occasion) (Situation))
          opt:uniquify-situation - (Fun Situation (Situation Sit-eq-index))
          find-reductions - (Fun (Lst (Lrcd (Lst Sexp) Bdgenv))
                                 (Sexp Varid Bdgenv Situation Domain))
          backchainer-is-stratified -
             (Fun Boolean (Back-chainer Situation))
	  forall-refuter-dnf - (Fun (Lst (Lst Sexp)) ((Lst Symbol) Sexp))
          disjunctive-normal-form - (Fun (Lst (Lst Sexp))
                                         (Sexp Boolean))
          dnf-neg - (Fun (Lst (Lst Sexp)) ((Lst (Lst Sexp))))
          literal-neg - (Fun Sexp (Sexp))
          lit-analyze - (Fun (Mlv Boolean Sexp) (Sexp))
          
          assertion-add 
            - (Fun (Lst Occasion) (Sexp Situation Domain))
          assertion-erase 
            - (Fun (Lst Occasion) (Sexp Situation Domain))
          assertion-find-and-erase
            - (Fun (Lst Occasion) (Sexp Varid Bdgenv Situation))
          variant-fetch 
            - (Fun (~ (Either Occasion Signed-occasion))
                   (Sexp (~ (Const :neutral :added :deleted)) Exp_index))
          deducible - (Fun Boolean
                           (Sexp Varid Bdgenv Situation Domain)
                           false)
          deduce - (Fun (Lst Bdgenv) 
                        (Sexp Varid Bdgenv Situation Domain)
                        false)
          deduce-check-many - (Fun (Mlv Boolean (Lst Bdgenv)) 
                                   (Sexp Varid Bdgenv Situation Domain)
                                   false)
          ground-deduce - (Fun Boolean (Sexp Situation Domain) false)
          loc-deduce - (Fun (Lst Bdgenv) (Sexp Situation) false)
          conj-deduce - (Fun (Lst Bdgenv)
                             ((Lst Sexp) Varid Bdgenv Situation Domain)
                             false)
          
          opt-eval - (Fun (Mlv Boolean Obj)
                          (Sexp Situation (Lst (Lrcd CSymbol Obj)) Domain))
	  try-postpone-negs - (Fun (Mlv Boolean (Lst Sexp))
				   ((Lst Sexp) (Lst Symboid)) false)
          get-var-val - (Fun Obj (CSymbol (Lst Bdgenv)) false)
          has-qvars - (Fun Boolean (Sexp) false)
          has-uvars - (Fun Boolean (Sexp Varid Bdgenv) false)
          all-uvars - (Fun (Lst Expclo) (Sexp Varid Bdgenv) false)
          fluent-current-value - (Fun Obj (Fluexp Situation Domain))
          nisp::is-Qvar - (Fun Boolean (Obj) False)
          ;;;;uniquify-action - (Fun Action (Sexp Domain))
          try-domain-with-name - (Fun (~ Domain) (Symbol Boolean))
	  domain-make-current - (Fun Void (Opt-domain))
          problem-initialize - (Fun Void (Uninitialized-plan-problem)
                                    true)
          situation-initialize - (Fun Void (Situation) true)
          walk-domain-bdgs
             - (Fun Void (Domain (Fun Void (Domain-bdg Opt-domain)
                                      true))
                    true)
          sit-ini - (Fun Initial-situation (Symbol Domain))
          own-rule-group - (Fun Rule-group (Domain))
          fact-internalize - (Fun Sexp (Fact))
          ground-id* - Varid unify-count* unify-success-count* - Integer
          opt::exp-index-copy - (Fun Exp_index (Exp_index))
          opt::try-rehash* - Boolean
          opt::rehash-attempts* opt::rehash-failures* - Integer
          opt::break-on-rehash-failure* - Boolean
          opt::uvar-topcoord* - Obj
          opt::indexed-domain-place-occasion-index - Domain
          opt::make-link-type 
          opt::steptype-restype - (Fun Opt-type (Opt-type Typenv))
          epsilon* - Float
          num-fluent-type* - Opt-type
          bad-reduction-preds* - (Lst CSymbol)
	  last-entity-defined* - Obj)

(eval-when (:compile-toplevel :slurp-toplevel :load-toplevel)
   (cond (+opt-uses-ytools+
          (nisp::typedecl 'yt::true 'Boolean t)
          (nisp::typedecl 'yt::false 'Boolean nil)
          ;; Double colon ('::') required here because the form
          ;; is read even if +opt-uses-ytools+ is false.
;;; In cmucl, this proclaims true and false special, which
;;; is not allowed for constants --          
;;;          (specdecl ytools::true ytools::false - Boolean)
          )))


(deffunc loc-contents - (Lst Signed-occasion) (sit - Situation)
   (exp_index-contents (!_diff-from-init sit) false))